"""
(c) 2023 «Dacom M» Ltd
"""

import json
import os
import tempfile
import time
import subprocess
import shutil

from veilrpc.utils.bazalt import states_response, CommandExecutionError, \
    delete_file_with_owner, get_output, usr_cmd, ionice_cmd, ZPOOL_STORAGES_PATH, PATH_ISO_STORAGE, \
    PATH_FILE_STORAGE, PATH_TMP, STORAGE_ROOT_DIR
from veilrpc.utils.file import chown_directory

# Слишком частая отправка данных в event bus чревата ощутимым
# замедлением выполнения основной задачи.
SEND_PROGRESS_DELAY = 2
HW_ASSISTED_CIPHER = 'aes128-gcm@openssh.com'
SOFT_CIPHER = 'chacha20-poly1305@openssh.com'


class RsyncError(Exception):
    pass


class DeleteError(Exception):
    pass


def rsync(is_local, original_source, destination_source, ssh_options,
          owner, progress_data, redis, rsync_options=None):
    assert bool(ssh_options) if not is_local else True

    retcode = 0
    error = ''
    same_fs = False  # not use, because only for move, it will be finalize
    tmp_destination_source = None
    split_list_dest = destination_source.split('/')
    if STORAGE_ROOT_DIR in destination_source and \
            not ('zfs' in split_list_dest and len(split_list_dest) == 6 and split_list_dest.index('zfs') == 2) and \
            not (PATH_TMP in split_list_dest):
        if PATH_FILE_STORAGE in split_list_dest:
            index = split_list_dest.index(PATH_FILE_STORAGE)
            split_list_dest[index] = PATH_TMP
        elif PATH_ISO_STORAGE in split_list_dest:
            index = split_list_dest.index(PATH_ISO_STORAGE)
            split_list_dest[index] = PATH_TMP
        else:
            split_list_dest.insert(-1, PATH_TMP)
        tmp_destination_source = '/'.join(split_list_dest)
    tmp_ssh_key_path = None
    rsh_option = None
    rsync_cmd = None
    if is_local:
        s_dir = os.path.dirname(original_source)
        d_dir = os.path.dirname(destination_source)
        check_fs_cmd = f'test `df --output=source {s_dir} | tail -n1` == ' \
                       f'`df --output=source {d_dir} | tail -n1` && echo true || echo false'
        cmd_ret = get_output(check_fs_cmd)
        if not cmd_ret['result']:
            raise CommandExecutionError(
                f'An error occurred while checking for a filesystem match between {s_dir} and '
                f'{d_dir}: {cmd_ret["comment"]}')
        if cmd_ret['comment'].strip() == 'false':
            same_fs = False
        if not same_fs and tmp_destination_source:
            rsync_cmd = f'rsync "{original_source}" "{tmp_destination_source}"'
        elif not same_fs:
            rsync_cmd = f'rsync {original_source} {destination_source}'
    else:
        tmp_fh, tmp_ssh_key_path = tempfile.mkstemp()
        with os.fdopen(tmp_fh, 'w') as tmp_file:
            tmp_file.write(ssh_options['private_ssh_key'])
        chown_directory(tmp_ssh_key_path, owner)
        cipher = HW_ASSISTED_CIPHER if ssh_options.get('aes-ni') else SOFT_CIPHER
        rsh_option = "ssh -o \\\"StrictHostKeyChecking no\\\" -c {} -i {}".format(
            cipher, tmp_ssh_key_path)
        remote_destination_source = '{}@{}:"{}"'.format(
            ssh_options['user'], ssh_options['destination_ip'],
            tmp_destination_source if tmp_destination_source else destination_source
        )
        rsync_cmd = f'rsync --protect-args --rsh "{rsh_option}" "{original_source}" {remote_destination_source}'
    if is_local and same_fs:
        try:
            shutil.move(original_source, destination_source)
        except Exception as e:
            retcode = 1
            error += f'\nFailed to move {original_source} to {destination_source}. Error: {e}'
    else:
        rsync_options = rsync_options or "--whole-file"
        user_rsync_cmd = f"su {owner} -c 'ionice -c 3 {rsync_cmd} {rsync_options} --human-readable --info=progress2'"
        try:
            if 'steps' in progress_data and 'step' in progress_data['steps']:
                progress_data['steps']['step'] += 1
            p = popen_with_progress(user_rsync_cmd, "rsync", progress_data, redis)
            if p.returncode:
                retcode = p.returncode
                error += f'\n{p.stderr.read()}'
            if not retcode and tmp_destination_source:
                if is_local:
                    try:
                        shutil.move(tmp_destination_source, destination_source)
                    except Exception as e:
                        retcode = 1
                        error += f'\nFailed to move to {destination_source} after rsync. Error: {e}'
                else:
                    ssh_rsh_option = rsh_option.replace('\\', '')
                    ssh_cmd = f"mv {tmp_destination_source} {destination_source}"
                    ssh_cmd = f"{ssh_rsh_option} {ssh_options['user']}@{ssh_options['destination_ip']} {ssh_cmd}"
                    ret_ssh_move_cmd = get_output(ssh_cmd, timeout=300)
                    if not ret_ssh_move_cmd['result']:
                        retcode = 1
                        error += f'\nAn error occurred while remote move to {destination_source}'\
                                 f' after rsync. Error: {ret_ssh_move_cmd["comment"]}.'
        finally:
            if tmp_ssh_key_path:
                os.remove(tmp_ssh_key_path)
    return {'retcode': retcode, 'error': error}


def rsync_multi_move(is_local, original_sources, destination_datapool_path, destination_datapool_uuid, ssh_options,
                     owner, progress_data, redis, rsync_options=None, **kwargs):
    assert bool(ssh_options) if not is_local else True
    comment = ''
    retcode = False
    matched_rebase = kwargs.get('matched_rebase', [])
    qemu_rebase_cmd = "qemu-img rebase -u -f qcow2 -b {backing_file} -F qcow2 {snapshot}"
    rebase_command = ":"  # doing nothing
    with_snapshots = False
    tmp_destination_source = None
    split_list_dest = destination_datapool_path.split('/')
    if STORAGE_ROOT_DIR in destination_datapool_path and 'zfs' not in split_list_dest \
            and PATH_TMP not in split_list_dest:
        if PATH_FILE_STORAGE in split_list_dest:
            index = split_list_dest.index(PATH_FILE_STORAGE)
            split_list_dest[index] = PATH_TMP
        elif PATH_ISO_STORAGE in split_list_dest:
            index = split_list_dest.index(PATH_ISO_STORAGE)
            split_list_dest[index] = PATH_TMP
        else:
            split_list_dest.append(PATH_TMP)
        tmp_destination_source = '/'.join(split_list_dest)
    for s in original_sources:
        if s.split('.')[-1] == 'snapshot':
            with_snapshots = True
            if matched_rebase:
                break
            else:
                ret_info = get_output(f'qemu-img info --output=json "{s}"')
                if not ret_info['result']:
                    raise CommandExecutionError(f'Failed to get info about "{s}". Error: {ret_info["comment"]}')
                ret_dict = json.loads(ret_info["comment"])
                try:
                    matched_rebase.append({
                        'snapshot': f"{destination_datapool_path.rstrip('/')}/{s.split('/')[-1]}",
                        'backing_file': f"{destination_datapool_path.rstrip('/')}/"
                                        f"{ret_dict['backing-filename'].split('/')[-1]}"})
                except Exception as e:
                    raise CommandExecutionError(f'Filed get backing filename from json dict {ret_dict}. Error: {e}.')

    meta_original_sources = []
    for s in original_sources:
        if s.split('.')[-1] == 'qcow2':
            meta_source = s.replace('qcow2', 'meta')
            try:
                os.path.isfile(meta_source)
            except Exception as e:
                raise CommandExecutionError(
                    f'An error occurred while checking path meta file {meta_source}: {e}')
            else:
                meta_original_sources.append(meta_source)
    original_sources = original_sources + meta_original_sources
    str_original_sources = ' '.join(original_sources)
    # replace uuid datapool into meta file of vdisk
    meta_cmd = f"sed -i -r '/.* \"datapool\": \"{destination_datapool_uuid}\".*/! s/(.* \"datapool\": \")" \
               f"[^\"]*(\".*)/\\1{destination_datapool_uuid}\\2/' {destination_datapool_path.rstrip('/')}/*.meta"
    tmp_ssh_key_path = None
    ssh_connect = ''
    same_fs = is_local
    if is_local:
        for s in original_sources:
            s_dir = os.path.dirname(s)
            check_fs_cmd = f'test `df --output=source {s_dir} | tail -n1` == ' \
                           f'`df --output=source {destination_datapool_path} | tail -n1` && echo true || echo false'
            cmd_ret = get_output(check_fs_cmd)
            if not cmd_ret['result']:
                raise CommandExecutionError(
                    f'An error occurred while checking for a filesystem match between {s} and '
                    f'{destination_datapool_path}: {cmd_ret["comment"]}')
            if cmd_ret['comment'].strip() == 'false':
                same_fs = False
                break
        if same_fs:
            rsync_cmd = f'mv {str_original_sources} {destination_datapool_path}'
        else:
            if tmp_destination_source:
                rsync_cmd = f'rsync {str_original_sources} {tmp_destination_source}'
            else:
                rsync_cmd = f'rsync {str_original_sources} {destination_datapool_path}'
        if with_snapshots:
            rebase_command = ";".join(qemu_rebase_cmd.format(
                backing_file=vs['backing_file'], snapshot=vs['snapshot']) for vs in matched_rebase)
    else:
        tmp_fh, tmp_ssh_key_path = tempfile.mkstemp()
        with os.fdopen(tmp_fh, 'w') as tmp_file:
            tmp_file.write(ssh_options['private_ssh_key'])
        chown_directory(tmp_ssh_key_path, owner)
        cipher = HW_ASSISTED_CIPHER if ssh_options.get('aes-ni') else SOFT_CIPHER
        rsh_option = "ssh -o \\\"StrictHostKeyChecking no\\\" -c {} -i {}".format(
            cipher, tmp_ssh_key_path)
        remote_destination_source = '{}@{}:"{}"'.format(
            ssh_options['user'], ssh_options['destination_ip'],
            tmp_destination_source if tmp_destination_source else destination_datapool_path
        )
        ssh_rsh_option = rsh_option.replace('\\', '')
        ssh_connect = f"{ssh_rsh_option} {ssh_options['user']}@{ssh_options['destination_ip']}"
        ssh_meta_cmd = meta_cmd.replace('(', '\\(').replace(')', '\\)').replace('"', '\\"').replace(' \\', '\\ \\')\
            .replace(' s', '\\ s').replace('*.meta', '\\*.meta').replace('\\1', '\\\\1').replace('\\2', '\\\\2')
        meta_cmd = f"{ssh_connect} {ssh_meta_cmd}"
        if with_snapshots:
            rebase_cmd = ';'.join(qemu_rebase_cmd.format(
                backing_file=vs['backing_file'], snapshot=vs['snapshot']) for vs in matched_rebase)
            rebase_command = f"{ssh_rsh_option} {ssh_options['user']}@{ssh_options['destination_ip']} '{rebase_cmd}'"
        rsync_cmd = f'rsync --protect-args --rsh "{rsh_option}" {str_original_sources} {remote_destination_source}'
    if is_local and same_fs:
        user_rsync_cmd = f"su {owner} -c 'ionice -c 3 {rsync_cmd}'"
    else:
        rsync_options = rsync_options or "--whole-file"
        user_rsync_cmd = f"su {owner} -c 'ionice -c 3 {rsync_cmd} {rsync_options} --human-readable --info=progress2'"
    try:
        if 'steps' in progress_data and 'step' in progress_data['steps']:
            progress_data['steps']['step'] += 1
        p = popen_with_progress(user_rsync_cmd, "rsync", progress_data, redis)
        if p.returncode:
            raise CommandExecutionError(f'An error occurred while rsync or mv files. '
                                        f'Error: {str(p.stderr.read())}.')
        if tmp_destination_source and not same_fs:
            tmp_destination_source_files = [f'{tmp_destination_source}/{os.path.basename(f)}' for f in original_sources]
            str_tmp_des_files = ' '.join(tmp_destination_source_files)
            mv_cmd = f'mv {str_tmp_des_files} {destination_datapool_path}'
            if is_local:
                mv_ret = get_output(mv_cmd, timeout=300)
                if not mv_ret['result']:
                    raise CommandExecutionError(f'An error occurred while move files from {PATH_TMP} to '
                                                f'{destination_datapool_path}. Error: {mv_ret["comment"]}.')
            else:
                mv_ret = get_output(f"{ssh_connect} {mv_cmd}", timeout=300)
                if not mv_ret['result']:
                    raise CommandExecutionError(f'An error occurred while remote move files from {PATH_TMP} to '
                                                f'{destination_datapool_path}. Error: {mv_ret["comment"]}.')
        if meta_original_sources:
            meta_ret = get_output(meta_cmd, timeout=300)
            if not meta_ret['result']:
                raise CommandExecutionError(f'An error occurred while fix meta files. '
                                            f'Error: {meta_ret["comment"]}.')
        if with_snapshots:
            cmd_ret = get_output(rebase_command, timeout=300)
            if not cmd_ret['result']:
                raise CommandExecutionError(f'An error occurred while rebase snapshots. '
                                            f'Error: {cmd_ret["comment"]}.')
    except Exception as e:
        comment += "\n" + str(e)
        retcode = True
        if is_local and same_fs:
            for s in original_sources:
                if not os.path.isfile(s):
                    destination_file = f"{destination_datapool_path.rstrip('/')}/{s.split('/')[-1]}"
                    if os.path.isfile(destination_file):
                        mv_back = get_output(f'mv "{destination_file}" "{s}"')
                        if not mv_back['result']:
                            comment += "\n" + mv_back['comment']
                    else:
                        comment += "\n" + f"File {s} lost."
    else:
        if not (is_local and same_fs):
            delete_any_type(str_original_sources, owner=owner, quiet=True)
            if meta_original_sources:
                delete_any_type(meta_original_sources, owner=owner, quiet=True)
    finally:
        if tmp_ssh_key_path:
            os.remove(tmp_ssh_key_path)
    return {'retcode': True if p.returncode or retcode else False,
            'error': comment}


@states_response
def write_vdisk_meta(is_local, path, ssh_options, owner, metadata):
    meta_path = path.replace('.qcow2', '.meta')
    if metadata:
        if is_local:
            from .directory_datapool import write_meta
            write_meta(meta_path, metadata, owner)
        else:
            metadata = '\n'.join(metadata)
            tmp_ssh_key_path = None
            try:
                tmp_fh, tmp_ssh_key_path = tempfile.mkstemp()
                with os.fdopen(tmp_fh, 'w') as tmp_file:
                    tmp_file.write(str(ssh_options['private_ssh_key'].encode('utf-8').decode('unicode_escape')))
                chown_directory(tmp_ssh_key_path, owner)
                ssh_cmd = "ssh -o \"StrictHostKeyChecking no\" -i {} {}@{} ".format(
                    tmp_ssh_key_path, ssh_options['user'], ssh_options['destination_ip'])
                cmd = f"echo '{metadata}' | {ssh_cmd} 'cat > {meta_path}'"
                cmd_ret = get_output(cmd)
                if not cmd_ret['result']:
                    raise CommandExecutionError(f'An error occurred while creating remote meta {meta_path}: {cmd_ret["comment"]}')
            except Exception as e:
                comment = f"Unknown error: {e}"
                raise CommandExecutionError(comment) from None
            finally:
                if tmp_ssh_key_path:
                    os.remove(tmp_ssh_key_path)
        return f"Meta with path {meta_path} was successfully created."


def convert(original_source, destination_source, destination_format, owner, progress_data, redis, preallocate=False,
            compress=False, **kwargs):
    tmp_destination_source = None
    split_list_dest = destination_source.split('/')
    if STORAGE_ROOT_DIR in destination_source and \
            not ('zfs' in split_list_dest and len(split_list_dest) == 6 and split_list_dest.index('zfs') == 2) and \
            not (PATH_TMP in split_list_dest):
        if PATH_FILE_STORAGE in split_list_dest:
            index = split_list_dest.index(PATH_FILE_STORAGE)
            split_list_dest[index] = PATH_TMP
        elif PATH_ISO_STORAGE in split_list_dest:
            index = split_list_dest.index(PATH_ISO_STORAGE)
            split_list_dest[index] = PATH_TMP
        else:
            split_list_dest.insert(-1, PATH_TMP)
        tmp_destination_source = '/'.join(split_list_dest)
    convert_cmd = 'qemu-img convert {} -U "{}" "{}" {} -O {} -p {}'.format(
        '-c' if compress else '', original_source,
        tmp_destination_source if tmp_destination_source else destination_source,
        '-S 0' if preallocate else '',
        destination_format,
        kwargs.get('additional_options') or '')
    convert_cmd = ionice_cmd(convert_cmd)
    user_convert_cmd = usr_cmd(convert_cmd, owner)
    if 'steps' in progress_data and 'step' in progress_data['steps']:
        progress_data['steps']['step'] += 1
    p = popen_with_progress(user_convert_cmd, "convert", progress_data, redis)
    retcode = p.returncode
    error = f'{p.stderr.read()}'
    if tmp_destination_source:
        try:
            shutil.move(tmp_destination_source, destination_source)
        except Exception as e:
            retcode = 1
            error += f'\nFailed to move to {destination_source} after convert. Error: {e}'
    return {'retcode': retcode, 'error': error}


def popen_with_progress(command, cmd_type, progress_data, red, flag=None):
    """
    Выполнение команды, и отправка в redis прогресса выполнения этой команды.
    flag -- опциональный путь к файлу, содержащему PID, cmdline и command выполняемого процесса. Создаётся на время
     работы задачи, может быть использован для проверки и снятия к.-л. блокировок при ненормальном завершении или
     снятии задачи. пример -- zerofill блочных устройств, флаг проверяется storage_transport_helper
    """

    p = subprocess.Popen(command, shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE, executable='/bin/bash')
    if isinstance(flag, str):
        with open(flag, 'w') as flg:
            with open(f'/proc/{p.pid}/cmdline', 'r') as cmdline:
                with open(f'/proc/{p.pid}/comm', 'r') as comm:
                    flg.write(f'{p.pid}\n{cmdline.read()}\n{comm.read()}')
    current_progress = []
    send_ts = 0
    if cmd_type in ['rsync', 'blockcommit', 'qemu_commit']:
        end_data = '\r'
    elif cmd_type == 'pv':
        end_data = '\n'
    else:
        end_data = '\r'
    steps = progress_data.get('steps', {'total_steps': 1, 'step': 1})
    progress_data = {k: v for k, v in progress_data.items() if k != 'steps'}
    accumulate_stdout = ''
    while p.poll() is None:
        stdout = p.stdout.read(1).decode()
        accumulate_stdout += stdout
        if stdout == end_data:
            if time.time() - send_ts >= SEND_PROGRESS_DELAY and len(current_progress) > 0:
                data = {'progress': ''.join(current_progress),
                        'parser': cmd_type,
                        'total_steps': steps['total_steps'],
                        'step': steps['step']}
                progress_data['comment'] = json.dumps(data)
                if red:
                    red.lpush('node_results', json.dumps(progress_data))
                send_ts = time.time()
            current_progress = []
        elif not stdout == '':
            current_progress.append(stdout)
    if isinstance(flag, str) and os.path.exists(flag):
        # если popen() завершается мгновенно, в флаг м.б. записаны неверные значения, и при везении
        # его может удалить storage_transport_helper
        os.remove(flag)
    p.comment = accumulate_stdout[-512:]
    return p


def python_function_with_progress(function, progress_data, red, *args, stdout_pipe=None, **kwargs):
    """
    Выполнение python функции с отправкой в redis прогресса выполнения этой команды.
    """
    import sys
    import io
    import threading

    send_ts = 0
    steps = progress_data.get('steps', {'total_steps': 1, 'step': 1})
    progress_data = {k: v for k, v in progress_data.items() if k != 'steps'}

    stdout = None

    if not stdout_pipe:
        stdout = sys.stdout
        sys.stdout = stdout_pipe = io.StringIO()

    thread = threading.Thread(target=function, args=args, kwargs=kwargs, daemon=True)
    thread.start()
    cur_pos = stdout_pipe.tell()
    data_out = ''
    while thread.is_alive():
        stdout_pipe.seek(cur_pos)
        out = stdout_pipe.readline()
        cur_pos = stdout_pipe.tell()
        if out:
            data_out = out
        if time.time() - send_ts >= SEND_PROGRESS_DELAY and data_out:
            data = {'progress': data_out,
                    'parser': "pv",
                    'total_steps': steps['total_steps'],
                    'step': steps['step']}
            progress_data['comment'] = json.dumps(data)
            if red:
                red.lpush('node_results', json.dumps(progress_data))
            send_ts = time.time()
    if not stdout_pipe:
        sys.stdout = stdout
    return thread


@states_response
def create_local_zfs_dataset(zpool_name, destination_id, destination_source, owner):
    dataset_name = f'{zpool_name}/{destination_id}'
    dataset_exists = False
    list_dataset = get_output(f'zfs list {zpool_name} -rH -o name')
    if not list_dataset['result']:
        raise CommandExecutionError(f'An error occurred while getting list of dataset from zpool {zpool_name}. '
                                    f'Error: {list_dataset["comment"]}')
    for line in list_dataset['comment'].split():
        if f'{zpool_name}/{destination_id}' in line:
            dataset_exists = True
            check_ret = get_output(f'test -f {destination_source}/vdisk.qcow2')
            if check_ret['result']:
                raise CommandExecutionError(f'An error occurred while creating dataset {dataset_name}. '
                                            f'Error: {destination_source}/vdisk.qcow2 is already exists.')
    if not dataset_exists:
        zfs_cmd = f'zfs create {dataset_name}'
        cmd_ret = get_output(zfs_cmd)
        if not cmd_ret['result']:
            if os.path.exists(destination_source):
                delete_file_with_owner(destination_source, owner)
            raise CommandExecutionError(f'An error occurred while creating dataset {dataset_name}. '
                                        f'Error: {cmd_ret["comment"]}')
    try:
        chown_directory(destination_source, owner)
    except Exception as e:
        raise CommandExecutionError(str(e)) from None
    return f"ZFS dataset {dataset_name} was successfully created."


@states_response
def create_delete_remote_zfs_dataset(create_delete, zpool_name, destination_id, ssh_options, owner):
    dataset_name = '{}/{}'.format(zpool_name, destination_id)
    if create_delete:
        # TODO наверно нужно по другому давать права на создание dataset'а пользователю veil-node
        zfs_cmd = 'sudo zfs allow {0} create,mount,canmount {1}; /sbin/zfs create -o canmount=noauto {2};'\
                  ' sudo zfs unallow {0} create,mount,canmount {1}; sudo zfs mount {2}; sudo zfs set canmount=on {2}'\
            .format(owner, dataset_name.split('/')[0], dataset_name)
    else:
        # -rR удаляет все снапшоты и клоны от удаляемого датасета
        zfs_cmd = f'sudo zfs destroy {dataset_name} -rR;'
    tmp_ssh_key_path = None
    try:
        tmp_fh, tmp_ssh_key_path = tempfile.mkstemp()
        with os.fdopen(tmp_fh, 'w') as tmp_file:
            tmp_file.write(str(ssh_options['private_ssh_key'].encode('utf-8').decode('unicode_escape')))
        chown_directory(tmp_ssh_key_path, owner)
        ssh_cmd = "ssh -o \"StrictHostKeyChecking no\" -i {} {}@{} ".format(
            tmp_ssh_key_path, ssh_options['user'], ssh_options['destination_ip'])
        cmd_ret = get_output(ssh_cmd + "'" + zfs_cmd + "'")
        if not cmd_ret['result']:
            raise CommandExecutionError(f'An error occurred while creating remote dataset {dataset_name}:'
                                        f' {cmd_ret["comment"]}')
    except Exception as e:
        comment = f"Unknown error: {e}"
        raise CommandExecutionError(comment) from None
    finally:
        if tmp_ssh_key_path:
            os.remove(tmp_ssh_key_path)

    if not cmd_ret['result']:
        comment = 'An error occurred while {} dataset {}: {}'.format(
            'creating' if create_delete else 'deleting', dataset_name, cmd_ret['comment'])
        raise CommandExecutionError(comment)
    return "ZFS dataset {} was successfully {}.".format(dataset_name, 'created' if create_delete else 'deleted')


@states_response
def copy_zfs_to_not_zfs(is_local, src_vdisk_uuid, src_pool_name, destination_id, destination_source, ssh_options, owner,
                        task=None, redis=None, progress_data=None, rsync_options=None):
    """
    Копировании из ZFS на не ZFS датапул при наличии снапшотов. Копируется певое сохраненное состояние.
    """
    get_snapshots_disk = get_output(usr_cmd(
        f'sudo zfs list -Ht snapshot -o name -s creation -r {src_pool_name}/{src_vdisk_uuid}', owner))
    if not get_snapshots_disk['result'] or not get_snapshots_disk['comment'].strip().split():
        raise CommandExecutionError(
            f'Could not get list snapshots from dataset {src_pool_name}/{src_vdisk_uuid}!. '
            f'Error: {get_snapshots_disk["comment"]}')
    else:
        fist_snapshot = get_snapshots_disk['comment'].strip().split()[0]
    # делаем временный клон от первого снапшота
    cmd_ret = get_output(usr_cmd(f'sudo zfs clone {fist_snapshot} {src_pool_name}/{destination_id}_tmp_clone', owner))
    if not cmd_ret['result']:
        raise CommandExecutionError(
            f'An error occurred while creating clone {src_pool_name}/{destination_id}_tmp_clone: {cmd_ret["comment"]}')
    if not progress_data:
        progress_data = dict(type='progress', jid=task, id='', name='dir_to_dir_copy', result=True)
    rsync_ret = rsync(is_local, f'{ZPOOL_STORAGES_PATH}/{src_pool_name}/{destination_id}_tmp_clone/vdisk.qcow2',
                      destination_source, ssh_options, owner, progress_data, redis, rsync_options)
    error_rsync = ''
    if rsync_ret['retcode']:
        error_rsync = f'An error occurred while rsync: {rsync_ret["error"]}\n'
    # удаляем временный клон
    cmd_ret = get_output(usr_cmd(f'sudo zfs destroy -rR {src_pool_name}/{destination_id}_tmp_clone', owner))
    if not cmd_ret['result']:
        raise CommandExecutionError(
            f'{error_rsync}An error occurred while deleting dataset {src_pool_name}/{destination_id}_tmp_clone: '
            f'{cmd_ret["comment"]}')
    elif error_rsync:
        raise CommandExecutionError(error_rsync[:-1])
    return f"File successfully {'locally' if is_local else 'remote'} copied from dataset {src_pool_name}/{src_vdisk_uuid} to {destination_source}."


@states_response
def delete_local_zfs_dataset(zpool_name, destination_id, only_clone=False):
    comment = ''

    dataset_name = f'{zpool_name}/{destination_id}'
    # Проверяем, является ли dataset клоном, если да, то получаем снимок от которого он создан
    cmd_snapshot = f'zfs list -Ho origin -r {dataset_name}'
    get_snapshot = get_output(cmd_snapshot)
    if not get_snapshot['result']:
        comment = f"An error occurred get original snapshot from dataset {dataset_name}: {get_snapshot['comment']}"
        raise CommandExecutionError(comment)
    # -rR удаляет все снапшоты и клоны от датасета
    zfs_cmd = f'sudo zfs destroy -rR {dataset_name}'
    cmd_ret = get_output(zfs_cmd)
    if not cmd_ret['result']:
        comment = f"An error occurred while deleting dataset {dataset_name}: {cmd_ret['comment']}"
        raise CommandExecutionError(comment)
    # only_clone удаляем только клон датасета, без удаления снапшота от которого он сделан
    if get_snapshot['comment'].strip() != '-' and not only_clone:
        cmd_delete_snapshot = f"zfs destroy {get_snapshot['comment'].strip()}"
        ret_delete_snapshot = get_output(cmd_delete_snapshot)
        if not ret_delete_snapshot['result']:
            comment += f"\nAn error occurred while deleting snapshot {get_snapshot['comment'].strip()}: " \
                       f"{ret_delete_snapshot['comment']}"

    return f"ZFS dataset {dataset_name} was successfully deleted.{comment}"


@states_response
def dir_to_dir_copy(is_local, original_source, destination_source,
                    task, ssh_options, owner, redis=None, progress_data=None,
                    write_meta=True, skip_not_exists=False, **kwargs):

    if kwargs.get('destination_type') == 'zfs' and not kwargs.get('is_clone') and \
            (kwargs.get('original_type') != 'zfs' or not kwargs.get('snapshots_exists')):
        if is_local:
            cmd_ret = create_local_zfs_dataset(kwargs['destination_zpool_name'], kwargs['destination_id'],
                                               os.path.dirname(destination_source), owner)
            if not cmd_ret['result']:
                raise CommandExecutionError(cmd_ret['comment'])
        else:
            cmd_ret = create_delete_remote_zfs_dataset(True, kwargs['destination_zpool_name'], kwargs['destination_id'],
                                                       ssh_options, owner)
            if not cmd_ret['result']:
                raise CommandExecutionError(cmd_ret['comment'])
    elif kwargs.get('destination_type') == 'zfs' and kwargs.get('original_type') == 'zfs' \
            and not kwargs.get('is_clone'):
        from .snapshot import zfs_to_zfs_copy
        ret = zfs_to_zfs_copy(is_local, kwargs['original_id'], kwargs['original_zpool_name'], kwargs['destination_id'],
                              kwargs['destination_zpool_name'], {}, ssh_options, owner, kwargs['with_snapshots'], task,
                              redis=redis, progress_data=progress_data)
        if not ret['result']:
            raise CommandExecutionError(ret['comment'])
        else:
            return "Dataset successfully copied from {} to {}.".format(os.path.dirname(original_source),
                                                                       os.path.dirname(destination_source))
    elif kwargs.get('destination_type') == 'zfs' and kwargs.get('original_type') == 'zfs' and kwargs.get('is_clone'):
        from .snapshot import zfs_create_clone
        ret = zfs_create_clone(kwargs['destination_id'], kwargs['original_id'],
                               kwargs['original_zpool_name'], owner, snapshot=kwargs.get('snapshot'))
        if not ret['result']:
            raise CommandExecutionError(ret['comment'])
        else:
            return f"Dataset {os.path.dirname(original_source)} successfully clone to " \
                   f"{os.path.dirname(destination_source)}."
    # при простом копировании с zfs на не zfs датапул,
    # будем копировать первое(?) сохраненное состояние, оно должно быть
    elif kwargs.get('destination_type') != 'zfs' and kwargs.get('original_type') == 'zfs' \
            and kwargs.get('snapshots_exists'):
        ret = copy_zfs_to_not_zfs(is_local, kwargs['original_id'], kwargs['original_zpool_name'],
                                  kwargs['destination_id'], destination_source, ssh_options, owner, task,
                                  redis=redis, progress_data=progress_data,
                                  rsync_options=kwargs.get('rsync_options', None))
        if not ret['result']:
            raise CommandExecutionError(ret['comment'])
        else:
            return ret['comment']

    if write_meta:
        meta_ret = write_vdisk_meta(is_local, destination_source, ssh_options, owner, kwargs.get('destination_metadata'))
        if not meta_ret['result']:
            raise CommandExecutionError(meta_ret['comment'])
    if skip_not_exists and not os.path.exists(original_source):
        return f"File {original_source} does not exists."
    if not progress_data:
        progress_data = dict(type='progress', jid=task, id='', name='dir_to_dir_copy', result=True)
    from .vdisk import is_need_preallocation
    if kwargs.get('preallocation') and is_local and is_need_preallocation:
        convert_ret = convert(original_source, destination_source, 'qcow2', owner, progress_data, redis,
                              preallocate=True)
        if convert_ret['retcode']:
            comment = f"An error occurred while rsync: {convert_ret['error']}"
            raise CommandExecutionError(comment)
    else:
        rsync_ret = rsync(is_local, original_source, destination_source,
                          ssh_options, owner, progress_data, redis, kwargs.get('rsync_options', None))
        if rsync_ret['retcode']:
            comment = f"An error occurred while rsync: {rsync_ret['error']}"
            raise CommandExecutionError(comment)
    return f"File successfully {'locally' if is_local else 'remote'} copied from {original_source} " \
           f"to {destination_source}."


@states_response
def move(**kwargs):
    ret = dir_to_dir_copy(**kwargs)
    if not ret['result']:
        raise CommandExecutionError(ret['comment'])
    comment = f"File successfully {'locally' if kwargs.get('is_local') else 'remote'} moved from {kwargs['original_source']} to {kwargs['destination_source']}."
    try:
        delete_any_type(kwargs.get('original_source'),
                        owner=kwargs.get('owner'))
    except DeleteError as e:
        comment += "\nOriginal removal has failed. Error: {}".format(str(e))
    return comment


@states_response
def dir_to_dir_move(is_local, original_source, destination_source,
                    task, ssh_options, owner, **kwargs):
    from .vdisk import is_need_preallocation, get_vdisk_virtual_size
    if kwargs.get('is_powered') and is_local:
        from .snapshot import create_directory_snapshot, create_snapshot_of_powered_domain
        comment = ''
        size = get_vdisk_virtual_size(original_source)
        if kwargs.get('destination_type') == 'zfs':
            cmd_ret = create_local_zfs_dataset(kwargs['destination_zpool_name'], kwargs['original_id'],
                                               os.path.dirname(destination_source), owner)
            comment += cmd_ret['comment']
            if not cmd_ret['result']:
                raise CommandExecutionError(comment)
        meta_ret = write_vdisk_meta(is_local, destination_source, ssh_options, owner,
                                    kwargs.get('destination_metadata'))
        comment += meta_ret['comment']
        if not meta_ret['result']:
            raise CommandExecutionError(comment)
        cmd_ret = create_directory_snapshot(destination_source, original_source,
                                            kwargs.get('backing_file_format', 'qcow2'), size, owner)
        comment += cmd_ret['comment']
        if not cmd_ret['result']:
            raise CommandExecutionError(comment)
        # if need source
        # blk_domain = get_output(f"virsh domblklist {kwargs['domain']} | awk 'NR>2&&NF>0' | grep -v {kwargs['dev']}")
        blk_domain = get_output(f"virsh domblklist {kwargs['domain']} "
                                f"| awk 'NR>2&&NF>0{{print $1}}' | grep -v {kwargs['dev']}")
        if not blk_domain['result']:
            raise CommandExecutionError(comment + f"\nAn error occurred while get domblklist {blk_domain['comment']}")
        split_path = destination_source.split('/')
        file_name = split_path[-1]
        split_file_name = file_name.split('.')
        snapshot_name = split_file_name[0]
        cmd_ret = create_snapshot_of_powered_domain(
            kwargs['domain'],
            snapshot_name,
            [{'source': destination_source,
              'dev': kwargs['dev'],
              'type_datapool': 'local' if kwargs['destination_type'] == 'zfs' else kwargs['destination_type']}],
            # if need source
            # [(lambda dev, source: {'dev': dev, 'source': source}
            #  )(*x.split()) for x in blk_domain['comment'].splitlines() if x[-1] != '-'],
            [{'dev': dev} for dev in blk_domain['comment'].splitlines()],
            None)
        if not cmd_ret['result']:
            raise CommandExecutionError(cmd_ret['comment'])
        comment += cmd_ret['comment']['comment']
        progress_data = dict(type='progress', jid=task, id='', name='dir_to_dir_move', result=True)
        cmd = f"virsh blockpull {kwargs['domain']} {kwargs['dev']} --verbose --wait 2>&1"
        p = popen_with_progress(cmd, "blockcommit", progress_data, kwargs.get('redis', None))
        if p.returncode:
            raise CommandExecutionError(
                comment +
                f'An error occurred while COMMIT snapshot (stage 1): '
                f'{str(p.stderr.read()) + str(p.stdout.read()) + str(p.comment if hasattr(p, "comment") else "")}')
        blk_domain = get_output(f"virsh domblklist {kwargs['domain']} | awk 'NR>2&&NF>0{{print $2}}'")
        if blk_domain['result']:
            list_blk = blk_domain['comment'].split()
        else:
            list_blk = []
        if 'Pull complete' in p.comment or (destination_source in list_blk and original_source not in list_blk):
            comment += 'Pull complete.' if 'Pull complete' in p.comment else ''
            try:
                delete_any_type(original_source, owner=owner)
                delete_any_type(original_source.replace('.qcow2', '.meta'), owner=owner)
            except DeleteError as e:
                comment += f"\nRemoval failed. Error: {str(e)}."
            if kwargs.get('original_type') == 'zfs':
                cmd_ret = delete_local_zfs_dataset(kwargs['original_zpool_name'], kwargs['original_id'])
                if not cmd_ret['result']:
                    comment += f"\nDelete dataset failed. Error: {str(cmd_ret['comment'])}."
            comment = f"Vdisk has been successfully locally moved from {original_source} to {destination_source}."
        return comment
    if kwargs.get('destination_type') == 'zfs':
        if is_local:
            cmd_ret = create_local_zfs_dataset(kwargs['destination_zpool_name'], kwargs['original_id'],
                                               os.path.dirname(destination_source), owner)
            if not cmd_ret['result']:
                raise CommandExecutionError(cmd_ret['comment'])
        else:
            cmd_ret = create_delete_remote_zfs_dataset(True, kwargs['destination_zpool_name'], kwargs['original_id'],
                                                       ssh_options, owner)
            if not cmd_ret['result']:
                raise CommandExecutionError(cmd_ret['comment'])

    meta_ret = write_vdisk_meta(is_local, destination_source, ssh_options, owner, kwargs.get('destination_metadata'))
    if not meta_ret['result']:
        raise CommandExecutionError(meta_ret['comment'])

    progress_data = dict(type='progress', jid=task, id='', name='dir_to_dir_move', result=True)
    if kwargs.get('preallocation') and is_local and is_need_preallocation(original_source, owner=owner):
        convert_ret = convert(original_source, destination_source, 'qcow2', owner, progress_data,
                              kwargs.get('redis', None), preallocate=True)
        if convert_ret['retcode']:
            comment = f"An error occurred while rsync: {convert_ret['error']}"
            raise CommandExecutionError(comment)
    else:
        rsync_ret = rsync(is_local, original_source, destination_source,
                          ssh_options, owner, progress_data,
                          kwargs.get('redis', None), kwargs.get('rsync_options', None))
        if rsync_ret['retcode']:
            comment = f"An error occurred while rsync: {rsync_ret['error']}"
            raise CommandExecutionError(comment)
    matched_rebase = kwargs.get('snapshots')
    if is_local and matched_rebase:
        from veilrpc.states.snapshot import rebase_unsafe
        rebase_unsafe_ret = rebase_unsafe(matched_rebase, is_local, ssh_options, owner)
        if not rebase_unsafe_ret['result']:
            raise CommandExecutionError(rebase_unsafe_ret['comment'])
    comment = f"Vdisk has been successfully {'locally' if is_local else 'remote'} moved from " \
              f"{original_source} to {destination_source}."
    try:
        delete_any_type(original_source, owner=owner)
        delete_any_type(original_source.replace('.qcow2', '.meta'), owner=owner)
    except DeleteError as e:
        comment += "\nRemoval failed. Error: {}".format(str(e))
    if kwargs.get('original_type') == 'zfs':
        cmd_ret = delete_local_zfs_dataset(kwargs['original_zpool_name'], kwargs['original_id'])
        if not cmd_ret['result']:
            comment += "\nDelete dataset failed. Error: {}".format(str(cmd_ret['comment']))
    return comment


@states_response
def general_copy(original, destination, task, owner, **kwargs):
    progress_data = dict(type='progress', jid=task, id='', name='general_move', result=True)
    convert_ret = convert(original['source'],
                          destination['source'],
                          destination['format'],
                          owner, progress_data, kwargs.get('redis', None))
    if convert_ret['retcode']:
        comment = f"An error occurred while convert: {convert_ret['error']}"
        try:
            delete_any_type(destination['source'],
                            datapool_type=destination['type'],
                            owner=owner)
            comment += "\nCopy has been removed."
        except DeleteError as e:
            comment += "\nCopy removal has failed. Error: {}".format(str(e))
        raise CommandExecutionError(comment)
    else:
        return "Vdisk successfully copied from {}({}) to {}({}).".format(
            original['source'], original['type'], destination['source'], destination['type'])


@states_response
def general_move(original, destination, task, owner, **kwargs):
    progress_data = dict(type='progress', jid=task, id='', name='general_move', result=True)
    convert_ret = convert(original['source'],
                          destination['source'],
                          destination['format'],
                          owner, progress_data, kwargs.get('redis', None))
    if convert_ret['retcode']:
        comment = f"An error occurred while convert: {convert_ret['error']}"
        try:
            delete_any_type(destination['source'],
                            datapool_type=destination['type'],
                            owner=owner)
            comment += "\nCopy has been removed."
        except DeleteError as e:
            comment += "\nCopy removal has failed. Error: {}".format(str(e))
        raise CommandExecutionError(comment)
    else:
        comment = "Vdisk successfully moved from {}({}) to {}({}).".format(
            original['source'], original['type'], destination['source'], destination['type'])
        try:
            delete_any_type(original['source'],
                            datapool_type=original['type'],
                            owner=owner)
            comment += "\nOriginal has been removed."
        except DeleteError as e:
            comment += "\nOriginal removal has failed. Error: {}".format(str(e))
        return comment


def delete_any_type(source, owner, datapool_type='directory', quiet=False):
    if datapool_type == 'directory':
        cmd_ret = delete_file_with_owner(source, owner, multi_path=True)
    else:
        cmd = "lvremove -f {}".format(source)
        cmd_ret = get_output(cmd)
    if not cmd_ret['result'] and not quiet:
        raise DeleteError(cmd_ret['comment'])


@states_response
def dir_to_dir_move_with_snapshots(is_local, original_sources, destination_datapool_path, destination_datapool_uuid,
                                   task, ssh_options, owner, **kwargs):

    progress_data = dict(type='progress', jid=task, id='', name='dir_to_dir_move_with_snapshot', result=True)
    rsync_ret = rsync_multi_move(is_local, original_sources, destination_datapool_path, destination_datapool_uuid,
                                 ssh_options, owner, progress_data, kwargs.get('redis', None),
                                 kwargs.get('rsync_options', None), matched_rebase=kwargs.get('matched_rebase'))
    if rsync_ret['retcode']:
        comment = f"An error occurred while rsync_multi_move: {rsync_ret['error']}"
        raise CommandExecutionError(comment)
    else:
        return f"Vdisks has been successfully move to {destination_datapool_path}."
