"""
(c) 2023 «Dacom M» Ltd
"""

import os
import psutil
import time
import json
import tempfile

import shutil
from veilrpc.utils.bazalt import states_response, get_output, CommandExecutionError,\
    get_result, action_service, states_result, is_service_active, is_service_failed, WORKER_USER_GID, WORKER_USER_UID
from veilrpc.utils.grpc import RpcCommandCrt
from veilrpc.utils import mount
from glustercli.cli.bricks import add, remove_start, remove_stop, remove_commit, remove_status, replace_commit
from glustercli.cli.heal import enable, disable, full, statistics, info as _heal_info
from glustercli.cli.peer import probe, detach, pool, status
from glustercli.cli.rebalance import start as r_start, stop as r_stop, fix_layout_start
from glustercli.cli.volume import create, start, stop, delete, optset, info, status_detail, vollist
from glustercli.cli.utils import GlusterCmdException, volume_execute
import secrets
import uuid
import zlib

BRICK_DOWN_TIMEOUT = 60
NEW_BRICK_APPEAR_TIMEOUT = 10
TMPDIR_PREFIX = 'gluster_replace_'


# Peer
@states_response
def remove_peers():
    real_peers = status()
    disconnected_peers = []
    for peer in real_peers:
        if peer.get('connected') == 'Disconnected':
            disconnected_peers.append(peer.get('uuid'))
    if disconnected_peers:
        for disc_peer in disconnected_peers:
            try:
                os.remove(f'/var/lib/glusterd/peers/{disc_peer}')
            except OSError:
                pass
        action, service = 'restart', 'glusterd'
        cmd_ret = action_service(service, action)
        if not cmd_ret['result']:
            raise CommandExecutionError(f"Error in {action} {service}: {cmd_ret['comment']}")
        return f"Remove disconnected nodes {', '.join(disconnected_peers)}"
    else:
        return "No disconnected nodes"


@states_response
def create_cluster(nodes, certs=None):
    with open('/etc/hostname', 'r') as hostname:
        this_node = hostname.read().strip()
    for node in nodes:
        if certs and (cert := certs.get(node)):
            if node == this_node:
                delete_cluster()
            else:
                RpcCommandCrt().apply(target=node, crt=cert, module='gluster', function='delete_cluster')
        probe(node)
    return pool()


@states_response
def delete_cluster():
    """
    Removes All Hosts Visible From Current Node
    """
    action_service("glusterd", "stop")
    time_0 = time.monotonic()
    while is_service_active("glusterd"):
        time.sleep(1)
        if time.monotonic() - time_0 > 60:
            raise CommandExecutionError("glusterd stop timeout")

    for proc in psutil.process_iter():  # после остановки сервиса иногда остаются процессы и не дают рестартовать
        if proc.name() in ['glusterfs', 'glustershd', 'glusterd']:
            proc.kill()

    for file in os.listdir("/var/lib/glusterd/peers"):
        os.remove(os.path.join("/var/lib/glusterd/peers", file))
    voldir = '/var/lib/glusterd/vols'
    for vol in os.listdir(voldir):
        shutil.rmtree(os.path.join(voldir, vol))

    with open('/var/lib/glusterd/glusterd.info', 'w+') as gl_info:
        cfg = {(spl := line.split('=', 1))[0]: spl[1] for line in gl_info.readlines() if '=' in line}
        gl_info.seek(0)
        for k, v in cfg.items():
            if k == 'UUID':
                gl_info.write(f'UUID={uuid.uuid4()}')
            else:
                gl_info.write(f'{k}={v}')

    action_service("glusterd", "start")
    time_0 = time.monotonic()
    while not is_service_active("glusterd"):
        time.sleep(1)
        if is_service_failed("glusterd"):
            raise CommandExecutionError("glusterd failed to start")
        if time.monotonic() - time_0 > 60:
            raise CommandExecutionError("glusterd start timeout")
    time.sleep(3)
    return status()


@states_response
def status_cluster():
    return status()


@states_response
def pool_list():
    cmd = 'gluster pool list'
    cmd_ret = get_output(cmd)
    if not cmd_ret['result']:
        raise CommandExecutionError(cmd_ret['comment'])
    return cmd_ret['comment']


@states_response
def status_pool():
    return pool()


@states_response
def detach_nodes_from_cluster(nodes):
    # local hostname will be shown as localhost so we should never detach a node on itself
    peers_list = [peer['hostname'] for peer in pool()]

    for node in nodes:
        try:
            peer_name = [peer for peer in peers_list if peer.startswith(node)][0]  # if ext.network name is uuid-extintf
        except IndexError:
            raise CommandExecutionError(f'No {node} in storage!') from None
        detach(peer_name)
    return pool()


# Volume
@states_response
def create_volume(volname, volbricks, replica=0, stripe=0, arbiter=0, disperse=0,
                  disperse_data=0, redundancy=0, transport='tcp', force=True, sharding=64):
    create(volname, volbricks, replica, stripe, arbiter, disperse, disperse_data, redundancy, transport, force)
    # TODO: следить за починкой шардинга и cluster.lookup-optimize
    vol_info = info(volname)
    if vol_info[0]['type'] == 'DISTRIBUTE':
        optset(volname, {"group": "distributed-virt"})
    elif 'DISPERSE' in vol_info[0]['type']:
        # num threads for SHD is for replicated only, removed it here
        optset(volname, {"group": "dispersed-virt"})
    else:
        optset(volname, {"group": "virt"})
    optset(volname, {'group': 'veil'})

    if isinstance(sharding, int) and sharding in range(2, 2049) and bin(sharding).count('1') == 1:
        optset(volname, {"features.shard": "on", "features.shard-block-size": f"{sharding}MB"})
    else:
        optset(volname, {"features.shard": "off"})

    if [br for br in vol_info[0]['bricks'] if br['type'] == 'Arbiter']:
        optset(volname, {'server-quorum-type': "server"})
    optset(volname, {'storage.owner-gid': str(WORKER_USER_GID)})
    optset(volname, {'storage.owner-uid': str(WORKER_USER_UID)})
    time_0 = time.monotonic()
    while time.monotonic() - time_0 < 50:
        try:
            start(volname, force)
        except GlusterCmdException as e:
            if "Another transaction is in progress" in str(e):
                time.sleep(2)
                continue
            else:
                raise CommandExecutionError(str(e)) from None
        else:
            break

    return status_detail(volname)


@states_response
def delete_volume(volname, force=True):
    if volname in vollist():
        stop(volname, force)
        return delete(volname)
    else:
        return "No such volume."


@states_response
def stop_volume(volname, force=True):
    stop(volname, force)
    return info(volname)


@states_response
def start_volume(volname, force=True):
    start(volname, force)
    return status_detail(volname)


@states_response
def set_volume_options(volname, options=None):
    if options:
        return optset(volname, options)
    return 'No options'


@states_response
def get_volume_options(volname, options='all'):
    cmd = f'gluster volume get {volname} {options}'
    cmd_ret = get_output(cmd)
    if not cmd_ret['result']:
        raise CommandExecutionError(cmd_ret['comment'])
    return cmd_ret['comment']


@states_response
def info_volume(volname):
    return info(volname)


@states_response
def status_detail_volume(volname=None, group_subvols=False):
    return status_detail(volname, group_subvols=group_subvols)


@states_response
def volume_list():
    return vollist()


@states_response
def rebalance_start(volname):
    return r_start(volname)


@states_response
def rebalance_stop(volname):
    return r_stop(volname)


@states_response
def rebalance_status(volname):
    cmd = f'gluster volume rebalance {volname} status'
    cmd_ret = get_output(cmd)
    if not cmd_ret['result']:
        raise CommandExecutionError(cmd_ret['comment'])
    return [line for line in cmd_ret['comment'].splitlines()]


@states_response
def heal_enable(volname):
    return enable(volname)


@states_response
def heal_disable(volname):
    return disable(volname)


@states_response
def heal_full(volname):
    return full(volname)


@states_response
def heal_statistics(volname):
    return statistics(volname)


@states_response
def heal_info(volname):
    return _heal_info(volname)


# Brick
@states_response
def get_bricks(volname):
    try:
        return [{"name": br['name'].rsplit('/', 1)[1] if '/' in br['name'] else br['name'],
                 'online': br['online'],
                 'path': br['name']}
                for br in status_detail(volname)[0]['bricks']]
    except (IndexError, KeyError):
        return []


@states_response
def brick_add(volname, bricks, stripe=None, replica=None, arbiter=None, force=True):
    old_info = info(volname)[0]
    had_arbiter = [br for br in old_info['bricks'] if br['type'] == 'Arbiter']
    add(volname, bricks, stripe, replica, arbiter, force)
    new_info = info(volname)[0]
    have_arbiter = [br for br in new_info['bricks'] if br['type'] == 'Arbiter']
    if old_info['type'] == 'DISTRIBUTE' and new_info['type'] != 'DISTRIBUTE':
        optset(volname, {'group': 'virt'})
        optset(volname, {'group': 'veil'})

    if have_arbiter and not had_arbiter:
        optset(volname, {'server-quorum-type': "server"})
    return status_detail(volname) + apply_fix_layout(volname, old_info, new_info)


@states_response
def brick_remove(volname, bricks, replica=None, force=True, task=None, redis=None):
    try:
        stage = 'pre_start'
        vol_detail = status_detail(volname)[0]

        # если том с репликацией и мы убираем только реплики, не требуя миграции данных, возникает ошибка,
        # что нужно использовать force commit; включаем это сами, если надо
        if ('REPLICATE' in vol_detail['type']
                and replica not in (None, vol_detail['replica'])
                and (vol_detail['replica'] - replica) * vol_detail['distribute'] == len(bricks)):

            stage = 'commit_force'
            out_test = remove_force(volname, bricks, replica)
        else:
            size_to_migrate = sum([brk['size_used'] for brk in vol_detail['bricks'] if brk['name'] in bricks]) or 1
            # 1 -- чтоб на ноль случайно не поделить
            stage = 'start'
            remove_start(volname, bricks, replica=replica)
            stage = 'progress'
            while True:
                try:
                    aggregated_stat = remove_status(volname, bricks, replica=replica)['aggregate']
                    stat = aggregated_stat['status']
                    size_migrated = int(aggregated_stat['size'])
                    progress_data = dict(type='progress', jid=task, id='', name='brick_remove', result=True,
                                         comment=json.dumps(
                                             {'parser': 'pv',
                                              'progress': str(min(100, int(size_migrated / size_to_migrate * 100))),
                                              'total_steps': 1,
                                              'step': 1}))
                    if redis:
                        redis.lpush('node_results', json.dumps(progress_data))
                except KeyError as err:
                    raise CommandExecutionError("Error occurred during bricks removal status acquiring") from err
                if stat not in ['in progress', 'completed']:
                    remove_stop(volname, bricks, replica=replica)
                    raise CommandExecutionError(f'Abnormal status "{stat}" of brick(s) removal on volume {volname}')
                if stat == 'completed':
                    out_test = f'*** Stats: "{json.dumps(aggregated_stat)}" *** New volume config: '
                    break
                time.sleep(2)

            if aggregated_stat['failures'] == '0':
                stage = 'commit'
                remove_commit(volname, bricks, replica=replica)
            else:
                # если ненулевое число ошибок при миграции, коммит не проходит
                stage = 'stop'
                remove_stop(volname, bricks, replica=replica)
                raise CommandExecutionError(f'Some errors occurred during data migration before bricks removal. '
                                            f'Remove cancelled (failures={aggregated_stat["failures"]}).')
        stage = 'info'
        dict_info = info(volname)[0]
        if [br for br in vol_detail['bricks'] if br['type'] == 'Arbiter'] and\
                not [br for br in dict_info['bricks'] if br['type'] == 'Arbiter']:
            optset(volname, {'server-quorum-type': "none"})

    except GlusterCmdException as e:
        # pylint: disable=used-before-assignment
        raise CommandExecutionError(f"Stage: {stage} Error: {e.args[0][2]}") from None
    return [out_test, ' *** New volume config: ',
            json.dumps({k: dict_info[k] for k in dict_info if k in ['name', 'type', 'num_bricks', 'replica',
                                                                    'distribute', 'disperse', 'disperse_redundancy']})
            ] + apply_fix_layout(volname, vol_detail, dict_info)


def remove_force(volname, bricks, replica=None):
    """
    Remove Bricks start

    :param volname: Volume Name
    :param bricks: List of Bricks
    :param replica: Replica Count
    :returns: Output of remove-brick force command, raises
     GlusterCmdException((rc, out, err)) on error
    """
    cmd = ["remove-brick", volname]
    if replica is not None:
        cmd += ["replica", "{0}".format(replica)]
    cmd += bricks
    cmd += ["force"]
    return volume_execute(cmd)


@states_response
def brick_remove_start(volname, bricks, replica=None, force=True):
    remove_start(volname, bricks, replica, force)
    return status_detail(volname)


@states_response
def brick_remove_stop(volname, bricks, replica=None, force=True):
    remove_stop(volname, bricks, replica, force)
    return status_detail(volname)


@states_response
def brick_remove_commit(volname, bricks, replica=None, force=True):
    remove_commit(volname, bricks, replica, force)
    return status_detail(volname)


@states_response
def brick_remove_status(volname, bricks, replica=None, force=True):
    return remove_status(volname, bricks, replica, force)


@states_response
def brick_reset(volname, brick):
    volume_execute(['reset-brick', volname, brick, 'start'])
    return status_detail(volname)


@states_response
def brick_replace(volname, old_brick, new_brick):
    peers = pool()
    # bringing old brick down if it is still online
    resetted = False
    rst_time = time.monotonic()
    while True:
        if not [br for br in status_detail(volname)[0]['bricks'] if br['name'] == old_brick][0]['online']:
            break
        if time.monotonic() - rst_time > BRICK_DOWN_TIMEOUT:
            raise CommandExecutionError(f'Timeout {BRICK_DOWN_TIMEOUT} s expired while bringing down old brick. '
                                        f'Operation canceled.')
        if not resetted:
            volume_execute(['reset-brick', volname, old_brick, 'start'])
            resetted = True
        time.sleep(1)
    # now we need have volume mounted
    if check_volume_mounted(':' + volname):
        cmd_ret = get_volume_mount_point(':' + volname)
        if cmd_ret['result']:
            mntpoint = cmd_ret['comment'].strip()
        else:
            raise CommandExecutionError(f'Unable to get mountpoint for mounted volume {volname}')
        need_umount = False
    else:
        need_umount = True
        mntpoint = tempfile.mkdtemp(dir='/tmp', prefix=TMPDIR_PREFIX)
        if 'localhost' in (connected_hosts := [h['hostname'] for h in peers if h['connected'] == 'Connected']):
            act_host = 'localhost'
        else:
            if not connected_hosts:
                raise CommandExecutionError('No connected peers seen.')
            else:
                act_host = connected_hosts[0]
        mount.mount(name=mntpoint, device=f'{act_host}:{volname}', fstype='glusterfs')
    # triggering AFR:
    # 1. create and remove directory that did not exist earlier
    try:
        with tempfile.TemporaryDirectory(dir=mntpoint):
            pass
    except OSError:  # on dispersed volumes remove directory sometimes fails with I/O err but directory disappears
        pass
    # 2. create and remove ext.attr that did not exist earlier too; the same about i/o error so no check for 'result'
    get_output(f'setfattr -n trusted.non-existent-key -v abc {mntpoint}')
    #  if not cmd_ret['result']:
    #    raise CommandExecutionError(f'Error setting ext.attrs: e={cmd_ret["comment"]}')
    get_output(f'setfattr -x trusted.non-existent-key {mntpoint}')
    #  if not cmd_ret['result']:
    #    raise CommandExecutionError(f'Error removing ext.attrs: e={cmd_ret["comment"]}')
    if need_umount:
        mount.umount(mntpoint)
    # switching in new brick
    volume_execute(['replace-brick', volname, old_brick, new_brick, 'commit', 'force'])
    repl_start = time.monotonic()
    while True:
        try:
            if [br['online'] for br in status_detail(volname)[0]['bricks'] if br['name'] == new_brick][0]:
                break
        except IndexError:  # if brick not yet appeared
            pass
        if time.monotonic() - repl_start > NEW_BRICK_APPEAR_TIMEOUT:
            raise CommandExecutionError(f'Timeout {NEW_BRICK_APPEAR_TIMEOUT} s exceeded '
                                        f'while waiting new brick to appear and become online.')
        time.sleep(1)


@states_response
def brick_replace_commit(volname, old_brick, new_brick, force=False):
    return replace_commit(volname, old_brick, new_brick, force)


def check_volume_mounted(volname, **kwargs):
    cmd = "mount -l | awk '{{print $1}}' | grep -e '{}$'".format(volname)
    return get_result(cmd, **kwargs)


@states_response
def get_volume_mount_point(volname, **kwargs):
    return states_result(get_output,
                         cmd="mount -l | grep -e '{}' | awk '{{print $3}}'".format(volname),
                         **kwargs)


def test_rw(mountpoint=None, volume=None, addr='localhost', duration=600, fsize_mb=100):

    from gluster import gfapi

    class FuseIo:

        def __init__(self, mnt):
            self.mountpoint = mnt

        def create_write(self, name, data):
            with open(os.path.join(self.mountpoint, name), 'w') as rand_f:
                return rand_f.write(data)

        def read(self, name):
            with open(os.path.join(self.mountpoint, name), 'r') as rand_f:
                return rand_f.read()

        def remove(self, name):
            if os.path.isfile(pf := os.path.join(self.mountpoint, name)):
                os.remove(pf)

        @property
        def free_size(self):
            return (stat := os.statvfs(self.mountpoint)).f_bfree * stat.f_bsize

    class GfapiIo:

        def __init__(self, addr, volume):
            vol = gfapi.Volume(hosts=[addr], volname=volume)
            vol.mount()
            self.vol = vol

        def create_write(self, name, data):
            with self.vol.fopen(name, 'w') as rand_f:
                return rand_f.write(data)

        def read(self, name):
            with self.vol.fopen(name, 'r') as rand_f:
                return rand_f.read()

        def remove(self, name):
            if self.vol.isfile(name):
                self.vol.unlink(name)

        @property
        def free_size(self):
            return (stat := self.vol.statvfs('.')).f_bfree * stat.f_bsize

    if bool(mountpoint) == bool(volume):
        print('Either "mountpoint" for FUSE-based access or "volume" and optionally "addr" ("localhost" by default) parameters must be present. Exiting...')
        return

    if mountpoint:
        io = FuseIo(mountpoint)
    else:
        io = GfapiIo(addr=addr, volume=volume)

    files_hashes = dict()
    files, failures, wrong_sizes = 0, 0, 0
    starttime = time.monotonic()
    logname = f"/tmp/test_rw_{time.strftime('%Y%m%d_%H%M%S')}.log"
    open(logname, 'w').close()
    hours = 0

    def rm_files(full=False):  # pylint: disable=redefined-outer-name
        fi, h_err, s_err = 0, 0, 0
        for _ in range(len(files_hashes) if full else secrets.SystemRandom().randrange(0, len(files_hashes))):
            fn, stored_crc = secrets.choice(list(files_hashes.items()))

            try:
                contents = io.read(fn)
                crc = hex(zlib.crc32(contents))
                print(lll := f'CRC32 {" OK " if crc == stored_crc else "FAIL"} {fn} {len(contents) - fsize_mb * 1048576}')
                with open(logname, 'a') as f_log:
                    f_log.write(lll + '\n')
            except OSError as err:
                crc = contents = ''
                print(f"OSError during read: e={err}")
                with open(logname, 'a') as f_log:
                    f_log.write(f'{time.strftime("%D %T")} {fn} read error, e={err}\n')
            try:
                io.remove(fn)
            except OSError as err:
                print(f"OSError during remove: e={err}")
                with open(logname, 'a') as f_log:
                    f_log.write(f'{time.strftime("%D %T")} {fn} remove error, e={err}\n')
            fi += 1
            files_hashes.pop(fn)
            if crc != stored_crc:
                h_err += 1
            if len(contents) != fsize_mb * 1048576:
                s_err += 1
        return fi, h_err, s_err

    try:
        while (time.monotonic() - starttime < duration) if duration > 0 else True:
            if (time.monotonic() - starttime) / 3600 > hours:
                hours += 1
                with open(logname, 'a') as log:
                    log.write(f'{time.strftime("%D %T")} test_rw: {files} files; {failures} bad CRCs; {wrong_sizes} size errors\n')
            free = io.free_size
            if free < 10 * 1048576 * fsize_mb:
                if not files_hashes:
                    return
                f, h, s = rm_files()
                files += f
                failures += h
                wrong_sizes += s

            rand_file = open('/dev/urandom', 'rb').read(1024) * 1024 * fsize_mb
            rand_name = str(uuid.uuid4())
            try:
                written = io.create_write(name=rand_name, data=rand_file)
            except OSError as e:
                print(f"OSError during write: e={e}")
                with open(logname, 'a') as log:
                    log.write(f'{time.strftime("%D %T")} write error, e={e}\n')

                io.remove(rand_name)
                f, h, s = rm_files()
                files += f
                failures += h
                wrong_sizes += s
            else:
                print(lll := f'{round((time.monotonic() - starttime)/duration * 100, 1)}% W '
                      f'{" OK " if written == len(rand_file) else "FAIL"} {rand_name}')
                with open(logname, 'a') as log:
                    log.write(lll + '\n')

                calc_crc = hex(zlib.crc32(rand_file))
                files_hashes[rand_name] = calc_crc
    except KeyboardInterrupt:
        print("Ctrl-C pressed! aborting.")

    if files_hashes:
        f, h, s = rm_files(full=True)
        files += f
        failures += h
        wrong_sizes += s
    print(f"Time elapsed: {int(time.monotonic() - starttime)} sec Files generated: {files} Hash errors: {failures} "
          f"Size errors: {wrong_sizes}")
    with open(logname, 'a') as log:
        log.write(f'{time.strftime("%D %T")} END test_rw: {files} files; {failures} bad CRCs; {wrong_sizes} size errors\n')


def apply_fix_layout(volname, old_info, new_info):
    if old_info['distribute'] != new_info['distribute']:
        for rty in range(10):
            try:
                fix_layout_start(volname)
            except GlusterCmdException as e:
                if "transaction in progress" in e[0][2]:
                    time.sleep(6)
                else:
                    raise CommandExecutionError(e) from None
            else:
                break
        return [f'Fix layout completed in {rty*10} secs.']
    return []


@states_response
def gluster_info(**kwargs):
    ret = {}
    ret['pool_list'] = pool_list()
    ret['status_detail_volume'] = status_detail_volume()
    return ret
