"""
(c) 2023 «Dacom M» Ltd
"""
import hashlib
import os

from veilrpc.utils.bazalt import states_response, CommandExecutionError, get_output, \
    delete_file_with_owner, ListVars, WORKER_USER
from veilrpc.utils.file import chown_directory, chown_files_in_dir


class LIBRARY_TYPES(ListVars):
    """
    Класс, описывающий доступные для использования типы файлов.
    """

    vmware = 'VMware4 disk image'
    hyperv = 'Microsoft Disk Image eXtended'
    qcow = 'QEMU QCOW Image'
    qcow2 = 'QEMU QCOW2 Image'
    ascii = 'ASCII text'
    init = 'Generic INItialization configuration'
    tar_gnu = 'POSIX tar archive (GNU)'
    tar = 'POSIX tar archive'
    zip = 'Zip archive data'
    data = 'data'
    raw = 'raw image'
    mbr = 'DOS/MBR boot sector'
    run = 'shell script executable (binary data)'
    elf = 'ELF 64-bit LSB executable'
    xml = 'XML document'
    xml_10 = 'XML 1.0 document'
    xml_11 = 'XML 1.1 document'
    unicode = 'UTF-8 Unicode text'
    unicode_16 = 'UTF-16 Unicode text'
    zstd = 'Zstandard compressed data'
    zlib = 'zlib compressed data'
    deb = 'Debian binary package'
    rpm = 'RPM'
    pub = 'OpenSSH'


class EXTENSION_TYPES(ListVars):
    """
    Класс, описывающий доступные для использования расширения файлов.
    """

    img = '.img'
    raw = '.raw'
    bin = '.bin'
    vmdk = '.vmdk'
    zip = '.zip'
    json = '.json'
    run = '.run'
    ovf = '.ovf'
    xml = '.xml'
    tar = '.tar'
    qcow = '.qcow'
    qcow2 = '.qcow2'
    vhdx = '.vhdx'
    ova = '.ova'
    iso = '.iso'
    zst = '.zst'
    vfd = '.vfd'
    key = '.key'
    rdp = '.rdp'
    dmg = '.dmg'
    fd = '.fd'
    deb = '.deb'
    rpm = '.rpm'
    yml = '.yml'
    yaml = '.yaml'
    zip = '.zip'
    pub = '.pub'


class EXTENSION_TEXT_TYPES(ListVars):
    """
    Класс, описывающий доступные ASCII для использования расширения файлов.
    """

    json = '.json'
    run = '.run'
    ovf = '.ovf'
    xml = '.xml'
    yml = '.yml'
    yaml = '.yaml'


class EXTENSION_ZIP_TYPES(ListVars):
    """
    Класс, описывающий доступное сжатие для использования расширения файлов.
    """

    zst = '.zst'


class LIBRARY_ZIP_TYPES(ListVars):
    """
    Класс, описывающий доступное сжатие для использования типы файлов.
    """

    zstd = 'Zstandard compressed data'


class ADDITIONAL_FLAGS:
    """
    Класс, описывающий доступные дополнительные флаги файлов.
    """
    convert_import = {'can_be_converted': True, 'can_be_imported': True}
    imported = {'can_be_imported': True}
    domained = {'can_be_domained': True}
    unpacked = {'can_be_unpacked': True}
    node_profile = {'node_profile': True}
    node_backup_os = {'node_backup_os': True}
    backup = {'can_be_unpacked': True, 'can_be_domained': True}
    backup_incremental = {'can_be_unpacked': True, 'can_be_domained': True, 'can_be_incremental': True}
    backup_ova = {'can_be_unpacked': True, 'can_be_converted': True}
    ssh = {'can_be_injected': True}


class ASSIGNMENT_TYPE:
    backup = 'backup'
    backup_incremental = 'backup_incremental'
    backup_os = 'backup_os'
    backup_ova = 'backup_ova'
    domain_xml = 'domain_xml'
    domain_ovf = 'domain_ovf'
    node_backup_os = 'node_backup_os'
    node_profile = 'node_profile'
    image_qcow2 = 'image_qcow2'
    fd = 'fd'
    image_vmdk = 'image_vmdk'
    image_vhdx = 'image_vhdx'
    image_dmg = 'image_dmg'
    image_raw = 'image_raw'
    archive = 'archive'
    xml_unsupported = 'xml_unsupported'
    ovf_unsupported = 'ovf_unsupported'
    floppy = 'floppy'
    key = 'key'
    rdp = 'rdp'
    deb = 'deb'
    rpm = 'rpm'
    yaml = 'yaml'
    zip = 'zip'
    pub = 'pub'


class CONVERT_VDISK_EXTENSTIONS(ListVars):
    ESXI = 'vmdk'
    HyperV = 'vhd'
    VirtualBox = 'vdi'

    @classmethod
    def with_options(cls, item: str) -> bool:
        return item == cls.ESXI

    @classmethod
    def check_ext(cls, ext: str) -> str:
        """
        В определенных случаях (например для HyperV) формат файла не совпадает с переменной ключа -О
        """
        return 'vpc' if ext == cls.HyperV else ext

    @classmethod
    def choice_format(cls) -> list:
        return [(value, f"{value} ({key})") for key, value in cls.__dict__.items() if not key.startswith('__') and
                not callable(getattr(cls, key))]


def additional_flags(filename, **kwargs) -> dict:
    from .directory_datapool import get_file_info
    info_ret = get_file_info(filename, **kwargs)
    if not info_ret['result']:
        return {}
    out = info_ret['comment']
    file_extension = os.path.splitext(filename)[1]
    if file_extension in (EXTENSION_TYPES.img, EXTENSION_TYPES.raw, EXTENSION_TYPES.bin) or \
            ((os.path.splitext(filename)[1] in [EXTENSION_TYPES.vmdk, EXTENSION_TYPES.vhdx, EXTENSION_TYPES.dmg]) and
             (LIBRARY_TYPES.data in out or LIBRARY_TYPES.mbr in out)):
        return ADDITIONAL_FLAGS.convert_import
    elif file_extension == EXTENSION_TYPES.run:
        return ADDITIONAL_FLAGS.node_backup_os
    elif file_extension == EXTENSION_TYPES.pub:
        return ADDITIONAL_FLAGS.ssh
    else:
        if LIBRARY_TYPES.vmware in out:
            return ADDITIONAL_FLAGS.convert_import
        elif LIBRARY_TYPES.hyperv in out:
            return ADDITIONAL_FLAGS.convert_import
        elif LIBRARY_TYPES.qcow in out or LIBRARY_TYPES.qcow2 in out:
            return ADDITIONAL_FLAGS.imported
        elif any(header in out for header in [LIBRARY_TYPES.tar_gnu, LIBRARY_TYPES.tar] + LIBRARY_ZIP_TYPES.list()):
            return ADDITIONAL_FLAGS.unpacked
        elif LIBRARY_TYPES.ascii in out or LIBRARY_TYPES.unicode in out:
            if file_extension == EXTENSION_TYPES.json:
                return ADDITIONAL_FLAGS.node_profile
            elif file_extension in [EXTENSION_TYPES.yml, EXTENSION_TYPES.yaml]:
                return {}
            else:
                return ADDITIONAL_FLAGS.domained
        elif (LIBRARY_TYPES.xml_10 in out or LIBRARY_TYPES.xml_11 in out) and file_extension in EXTENSION_TYPES.ovf:
            return ADDITIONAL_FLAGS.domained
        elif LIBRARY_TYPES.zip in out:
            return {}
        else:
            return {}


def assignment_type(flags: dict, extension: str) -> str:
    extension = extension.lower()
    if dict(ADDITIONAL_FLAGS.backup_incremental.items() & flags.items()) == ADDITIONAL_FLAGS.backup_incremental:
        return ASSIGNMENT_TYPE.backup_incremental
    elif dict(ADDITIONAL_FLAGS.backup_ova.items() & flags.items()) == ADDITIONAL_FLAGS.backup_ova:
        return ASSIGNMENT_TYPE.backup_ova
    elif dict(ADDITIONAL_FLAGS.backup.items() & flags.items()) == ADDITIONAL_FLAGS.backup:
        return ASSIGNMENT_TYPE.backup
    elif list(ADDITIONAL_FLAGS.node_backup_os.keys())[0] in flags:
        return ASSIGNMENT_TYPE.node_backup_os
    elif list(ADDITIONAL_FLAGS.node_profile.keys())[0] in flags:
        return ASSIGNMENT_TYPE.node_profile
    elif list(ADDITIONAL_FLAGS.unpacked.keys())[0] in flags:
        return ASSIGNMENT_TYPE.archive
    elif list(ADDITIONAL_FLAGS.domained.keys())[0] in flags and extension == EXTENSION_TYPES.ovf:
        return ASSIGNMENT_TYPE.domain_ovf
    elif list(ADDITIONAL_FLAGS.domained.keys())[0] in flags:
        return ASSIGNMENT_TYPE.domain_xml
    elif extension == EXTENSION_TYPES.xml:
        return ASSIGNMENT_TYPE.xml_unsupported
    elif extension in [EXTENSION_TYPES.yml, EXTENSION_TYPES.yaml]:
        return ASSIGNMENT_TYPE.yaml
    elif extension == EXTENSION_TYPES.vfd:
        return ASSIGNMENT_TYPE.floppy
    elif extension == EXTENSION_TYPES.key:
        return ASSIGNMENT_TYPE.key
    elif extension == EXTENSION_TYPES.pub:
        return ASSIGNMENT_TYPE.pub
    elif extension == EXTENSION_TYPES.rdp:
        return ASSIGNMENT_TYPE.rdp
    elif extension == EXTENSION_TYPES.deb:
        return ASSIGNMENT_TYPE.deb
    elif extension == EXTENSION_TYPES.rpm:
        return ASSIGNMENT_TYPE.rpm
    elif extension == EXTENSION_TYPES.fd:
        return ASSIGNMENT_TYPE.fd
    elif extension == EXTENSION_TYPES.ovf:
        return ASSIGNMENT_TYPE.ovf_unsupported
    elif extension == EXTENSION_TYPES.zip:
        return ASSIGNMENT_TYPE.zip
    elif dict(ADDITIONAL_FLAGS.convert_import.items() & flags.items()) == ADDITIONAL_FLAGS.convert_import \
            and extension == EXTENSION_TYPES.vmdk:
        return ASSIGNMENT_TYPE.image_vmdk
    elif dict(ADDITIONAL_FLAGS.convert_import.items() & flags.items()) == ADDITIONAL_FLAGS.convert_import \
            and extension == EXTENSION_TYPES.vhdx:
        return ASSIGNMENT_TYPE.image_vhdx
    elif dict(ADDITIONAL_FLAGS.convert_import.items() & flags.items()) == ADDITIONAL_FLAGS.convert_import \
            and extension == EXTENSION_TYPES.dmg:
        return ASSIGNMENT_TYPE.image_dmg
    elif dict(ADDITIONAL_FLAGS.convert_import.items() & flags.items()) == ADDITIONAL_FLAGS.convert_import:
        return ASSIGNMENT_TYPE.image_raw
    elif dict(ADDITIONAL_FLAGS.imported.items() & flags.items()) == ADDITIONAL_FLAGS.imported:
        return ASSIGNMENT_TYPE.image_qcow2
    else:
        return 'unknown'


def _extension(path) -> str:
    """
    Получает расширение файла
    :return: Пример: .ova
    """
    try:
        _, file_extension = os.path.splitext(path)
        return file_extension
    except TypeError:
        return ''


@states_response
def library_discovery(datapool_path, library_storage_path, with_ass_type=False, chown_before=False,
                      with_filename=False, remove_unknown=False, **kwargs):
    from .directory_datapool import get_file_info, find_with_size_cmd, _create_nested_directory
    from .iso import ISO_TYPES
    path = os.path.join(datapool_path, library_storage_path)
    try:
        _create_nested_directory(path, WORKER_USER, True)
    except Exception as e:
        raise CommandExecutionError(str(e)) from None
    if chown_before:
        try:
            chown_files_in_dir(path)
        except OSError:
            pass
    results, unknowns = [], []
    ret = find_with_size_cmd(path, **kwargs)
    if not ret['result']:
        raise CommandExecutionError(f"Can not discover LIBRARY images in datapool {datapool_path}: {ret['comment']}")
    for line in ret['comment']:
        try:
            size, filename = line.split()
            info_ret = get_file_info(filename, **kwargs)
            if not info_ret['result']:
                if remove_unknown:
                    unknowns.append(filename)
                continue
            if not [_ for _ in LIBRARY_TYPES.list() if _ in info_ret['comment']]:
                if remove_unknown:
                    unknowns.append(filename)
                continue
            elif (LIBRARY_TYPES.ascii in info_ret['comment'] or LIBRARY_TYPES.unicode in info_ret['comment']) \
                    and f'.{filename.split(".")[-1].lower()}' not in EXTENSION_TEXT_TYPES.list():
                if remove_unknown:
                    unknowns.append(filename)
                continue
            elif ISO_TYPES[0] in info_ret['comment']:
                if remove_unknown:
                    unknowns.append(filename)
                continue
            flags = additional_flags(filename, **kwargs)
            info = {'path': filename, 'size': int(size.strip()), 'additional_flags': flags}
            if with_ass_type:
                info['assignment_type'] = assignment_type(flags, _extension(filename))
            if with_filename:
                info['filename'] = os.path.basename(filename)
            results.append(info)
        except (KeyError, ValueError):
            continue
    if remove_unknown and unknowns:
        cmd_ret = multiple_delete(unknowns, WORKER_USER)
        if not cmd_ret['result']:
            raise CommandExecutionError(cmd_ret['comment'])
    return results


@states_response
def delete(file_path, owner):
    cmd_ret = delete_file_with_owner(file_path, owner)
    if not cmd_ret['result']:
        raise CommandExecutionError(cmd_ret['comment'])
    return f"File with path {file_path} successfully deleted."


@states_response
def multiple_delete(file_paths, owner):
    for file_path in file_paths:
        cmd_ret = delete_file_with_owner(file_path, owner)
        if not cmd_ret['result']:
            raise CommandExecutionError(cmd_ret['comment'])
    return f"Files with paths {', '.join(file_paths)} successfully deleted."


@states_response
def get_file_mtime(path, create_as_root, owner, **kwargs):
    if create_as_root:
        cmd = "stat -c %Y {}".format(path)
    else:
        cmd = "su {} -c \'stat -c %Y {}\'".format(owner, path)

    cmd_ret = get_output(cmd, **kwargs)
    if not cmd_ret['result']:
        raise CommandExecutionError(cmd_ret['comment'])
    return cmd_ret['comment']


def check_file_backup_os(path):
    cmd = f'file "{path}" |sed "s/.*: //"'
    ret = get_output(cmd)
    if not ret['result']:
        raise CommandExecutionError(ret['comment'])
    if LIBRARY_TYPES.run in ret['comment'].strip():
        cmd = f'dd if="{path}" count=1 bs=30 2>/dev/null'
        ret = get_output(cmd)
        if not ret['result']:
            raise CommandExecutionError(ret['comment'])
        if ret['comment'].strip() == '#!/bin/sh\n#\n# \"Dacom M\" Ltd\n#':
            return True
        else:
            return False
    else:
        return False


@states_response
def get_description_backup_os(path):
    if not check_file_backup_os(path):
        return ''
    cmd = f'dd if="{path}" count=1 bs=8k 2>/dev/null | sed -n "/DESCRIPTION=/ s/DESCRIPTION=//p" | xargs printf'
    ret = get_output(cmd)
    if not ret['result']:
        raise CommandExecutionError(ret['comment'])
    return ret['comment']


@states_response
def compress_backup(backup, compress_type, user, task, **kwargs):
    from .copy import popen_with_progress

    redis = kwargs.get('redis', None)
    progress_data = dict(type='progress', jid=task, id='', name='compress_backup', result=True)
    new_backup = ''

    if not os.path.isfile(backup):
        raise CommandExecutionError(f'File {backup} not found.')
    cmd = f'set -o pipefail; exec 3>&1; pv -n {backup} 2>&3 | '
    if compress_type == LIBRARY_ZIP_TYPES.zstd.split()[0]:
        new_backup = f'{backup}{EXTENSION_ZIP_TYPES.zst}'
        cmd += f'/usr/local/sbin/veil-zstd > "{new_backup}"'
    else:
        raise CommandExecutionError(f'Unsupported format {compress_type}.')
    p = popen_with_progress(cmd, "pv", progress_data, redis)
    if p.returncode:
        if os.path.exists(new_backup):
            try:
                os.remove(new_backup)
            except OSError:
                pass
        raise CommandExecutionError(f'Failed to compress backup to {new_backup}. Error: {p.stderr.read()}')
    if os.path.exists(backup):
        try:
            os.remove(backup)
        except OSError:
            pass
    try:
        chown_directory(new_backup, user)
    except Exception as e:
        raise CommandExecutionError(str(e)) from None
    return {'filename': os.path.basename(new_backup), 'path': new_backup, 'size': os.path.getsize(new_backup)}


@states_response
def decompress_backup(backup, user, task, **kwargs):
    from .copy import popen_with_progress

    redis = kwargs.get('redis', None)
    progress_data = dict(type='progress', jid=task, id='', name='decompress_backup', result=True)
    new_backup = ''

    if not os.path.isfile(backup):
        raise CommandExecutionError(f'File {backup} not found.')
    cmd = f'set -o pipefail; exec 3>&1; pv -n {backup} 2>&3 | '
    ret_type = get_output(f'file {backup}')
    if not ret_type['result']:
        raise CommandExecutionError(f'Failed to get type from backup "{backup}". Error: {ret_type["comment"]}')
    if LIBRARY_ZIP_TYPES.zstd in ret_type['comment']:
        new_backup = '.'.join(backup.split('.')[:-1])
        cmd += f'zstd --rm -d -c > "{new_backup}"'
    else:
        raise CommandExecutionError(f'Failed to recognize format of compress of "{backup}". '
                                    f'Got type: {ret_type["comment"]}')
    p = popen_with_progress(cmd, "pv", progress_data, redis)
    if p.returncode:
        if os.path.exists(new_backup):
            try:
                os.remove(new_backup)
            except OSError:
                pass
        raise CommandExecutionError(
            f'Failed to decompress backup from {backup}. Error: {p.stderr.read()}')
    if os.path.exists(backup):
        try:
            os.remove(backup)
        except OSError:
            pass
    try:
        chown_directory(new_backup, user)
    except Exception as e:
        raise CommandExecutionError(str(e)) from None
    return {'filename': os.path.basename(new_backup), 'path': new_backup, 'size': os.path.getsize(new_backup)}


@states_response
def generate_integrity_hash(path, **kwargs):
    import io
    from .copy import python_function_with_progress

    redis = kwargs.get('redis', None)
    task = kwargs.get('task', None)
    progress_data = kwargs.get('progress_data',
                               dict(type='progress', jid=task, id='', name='generate_hash_progress', result=True))
    stdout_pipe = io.StringIO()
    result = {}

    try:
        python_function_with_progress(_generate_hash, progress_data, redis, path, result, stdout_pipe,
                                      stdout_pipe=stdout_pipe)
    except Exception as e:
        raise CommandExecutionError(
            f'An error occurred while getting an integrity hash for the file. Error: {e}') from None
    return result


class IntegrityHashCompromisedError(Exception):
    pass


@states_response
def check_integrity_hash(path, **kwargs):
    import io
    from .copy import python_function_with_progress

    redis = kwargs.get('redis', None)
    task = kwargs.get('task', None)
    current_hash = kwargs.get('current_hash', None)
    progress_data = kwargs.get('progress_data',
                               dict(type='progress', jid=task, id='', name='generate_hash_progress', result=True))
    stdout_pipe = io.StringIO()
    result = {}

    try:
        python_function_with_progress(_generate_hash, progress_data, redis, path, result, stdout_pipe,
                                      stdout_pipe=stdout_pipe)
    except Exception as e:
        raise CommandExecutionError(
            f'An error occurred while getting an integrity hash for the file. Error: {e}') from None

    # проверка хеша (кидаем ошибку, чтобы статус таски стал failed)
    if current_hash and result.get('integrity_hash') != current_hash:
        raise IntegrityHashCompromisedError("File integrity is compromised.")  # не менять текст ошибки!

    return result


def _generate_hash(path, result, stdout_pipe):
    file_hash = hashlib.md5()  # nosec
    size = os.path.getsize(path)
    piece = 8192 * 1024
    accumulator = 0
    with open(path, "rb") as f:
        chunk = f.read(piece)
        while chunk:
            file_hash.update(chunk)
            accumulator += piece
            print(int((accumulator * 100) / size), file=stdout_pipe)
            chunk = f.read(piece)
    result.update({'integrity_hash': file_hash.hexdigest()})


def gen_hash(path):
    file_hash = hashlib.md5()  # nosec
    piece = 8192 * 1024
    accumulator = 0
    with open(path, "rb") as f:
        chunk = f.read(piece)
        while chunk:
            file_hash.update(chunk)
            accumulator += piece
            chunk = f.read(piece)
    return file_hash.hexdigest()
