"""
(c) 2023 «Dacom M» Ltd

(U)Mounting of filesystems
=======================
"""

import os.path
import pwd
import re
import six
from six import string_types


from veilrpc.utils.bazalt import states_response, CommandExecutionError, get_output
from veilrpc.utils.mount import active, fstab, rm_fstab, umount, user_info, group_info, \
    set_fstab, mount as storage_mount, remount
from veilrpc.utils.file import mkdir


@states_response
def mounted(name,
            device,
            fstype,
            skip_permissions_check=False,
            mkmnt=False,
            opts='defaults',
            owner=None,
            dump=0,
            pass_num=0,
            config='/etc/fstab',
            persist=True,
            mount=True,
            user=None,
            match_on='auto',
            device_name_regex=None,
            extra_mount_invisible_options=None,
            extra_mount_invisible_keys=None,
            extra_mount_ignore_fs_keys=None,
            extra_mount_translate_options=None,
            hidden_opts=None,
            systemd=False):
    """
    Verify that a device is mounted

    name
        The path to the location where the device is to be mounted

    device
        The device name, typically the device node, such as ``/dev/sdb1``
        or ``UUID=066e0200-2867-4ebe-b9e6-f30026ca2314`` or ``LABEL=DATA``

    fstype
        The filesystem type, this will be ``xfs``, ``ext2/3/4`` in the case of classic
        filesystems, and ``fuse`` in the case of fuse mounts

    mkmnt
        If the mount point is not present then the state will fail, set ``mkmnt: True``
        to create the mount point if it is otherwise not present

    opts
        A list object of options or a comma delimited list

    dump
        The dump value to be passed into the fstab, Default is ``0``

    pass_num
        The pass value to be passed into the fstab, Default is ``0``

    config
        Set an alternative location for the fstab, Default is ``/etc/fstab``

    persist
        Set if the mount should be saved in the fstab, Default is ``True``

    mount
        Set if the mount should be mounted immediately, Default is ``True``

    user
        The user to own the mount

    match_on
        A name or list of fstab properties on which this state should be applied.
        Default is ``auto``, a special value indicating to guess based on fstype.
        In general, ``auto`` matches on name for recognized special devices and
        device otherwise.

    device_name_regex
        A list of device exact names or regular expressions which should
        not force a remount. For example, glusterfs may be mounted with a
        comma-separated list of servers in fstab, but the /proc/self/mountinfo
        will show only the first available server.

        .. code-block:: jinja

            {% set glusterfs_ip_list = ['10.0.0.1', '10.0.0.2', '10.0.0.3'] %}

            mount glusterfs volume:
              mount.mounted:
                - name: /mnt/glusterfs_mount_point
                - device: {{ glusterfs_ip_list|join(',') }}:/volume_name
                - fstype: glusterfs
                - opts: _netdev,rw,defaults,direct-io-mode=disable
                - mkmnt: True
                - persist: True
                - dump: 0
                - pass_num: 0
                - device_name_regex:
                  - ({{ glusterfs_ip_list|join('|') }}):/volume_name

    extra_mount_invisible_options
        A list of extra options that are not visible through the
        ``/proc/self/mountinfo`` interface.

        If a option is not visible through this interface it will always remount
        the device. This option extends the builtin ``mount_invisible_options``
        list.

    extra_mount_invisible_keys
        A list of extra key options that are not visible through the
        ``/proc/self/mountinfo`` interface.

        If a key option is not visible through this interface it will always
        remount the device. This option extends the builtin
        ``mount_invisible_keys`` list.

        A good example for a key option is the password option::

            password=badsecret

    extra_ignore_fs_keys
        A dict of filesystem options which should not force a remount. This will update
        the internal dictionary. The dict should look like this::

            {
                'ramfs': ['size']
            }

    extra_mount_translate_options
        A dict of mount options that gets translated when mounted. To prevent a remount
        add additional options to the default dictionary. This will update the internal
        dictionary. The dictionary should look like this::

            {
                'tcp': 'proto=tcp',
                'udp': 'proto=udp'
            }

    hidden_opts
        A list of mount options that will be ignored when considering a remount
        as part of the state application
    """
    result = True
    comment = ''

    if device_name_regex is None:
        device_name_regex = []

    # Make sure that opts is correct, it can be a list or a comma delimited
    # string
    if isinstance(opts, string_types):
        opts = opts.split(',')

    if isinstance(hidden_opts, string_types):
        hidden_opts = hidden_opts.split(',')

    # remove possible trailing slash
    if not name == '/':
        name = name.rstrip('/')

    device_list = []
    # Get the active data
    active_mount = active(extended=True)
    real_name = os.path.realpath(name)

    if fstype == 'nfs' and not skip_permissions_check and owner:
        opts.append(f'user={owner}')

    def _check_mountpoint(device, mpoint):
        """
        Check mount point and device, if mount point with another device exist in /etc/fstab
        we must stop task.
        return: True if mount point in fstab with another device.
        """
        with open('/etc/fstab', 'r') as fstab_f:
            for record in fstab_f.readlines():
                record.strip(' \t\n\r')
                if record and record.startswith('#'):
                    continue
                else:
                    elements = record.split()
                    if device not in elements and mpoint in elements:
                        return True
        return False

    def _check_perms(owner, name):
        try:
            owner_uid, owner_gid = pwd.getpwnam(owner).pw_uid, pwd.getpwnam(owner).pw_gid
        except KeyError:
            return False
        mpoint_uid, mpoint_gid = os.stat(name).st_uid, os.stat(name).st_gid
        if owner_uid != mpoint_uid or owner_gid != mpoint_gid:
            return False
        return True

    def _chown_directory(owner, name):
        try:
            owner_uid, owner_gid = pwd.getpwnam(owner).pw_uid, pwd.getpwnam(owner).pw_gid
        except KeyError:
            return False
        try:
            os.chown(name, owner_uid, owner_gid)
        except Exception:
            return False
        return True

    if device.startswith('/'):
        if 'bind' in opts and real_name in active_mount:
            _device = device
            if active_mount[real_name]['device'].startswith('/'):
                # Find the device that the bind really points at.
                while True:
                    if _device in active_mount:
                        _real_device = active_mount[_device]['device']
                        opts = list(set(opts + active_mount[_device]['opts'] + active_mount[_device]['superopts']))
                        active_mount[real_name]['opts'].append('bind')
                        break
                    _device = os.path.dirname(_device)
                real_device = _real_device
            else:
                # Remote file systems act differently.
                if _device in active_mount:
                    opts = list(set(opts + active_mount[_device]['opts'] + active_mount[_device]['superopts']))
                    active_mount[real_name]['opts'].append('bind')
                real_device = active_mount[real_name]['device']
        else:
            real_device = os.path.realpath(device)
    elif device.upper().startswith('UUID='):
        real_device = device.split('=')[1].strip('"').lower()
    elif device.upper().startswith('LABEL='):
        _label = device.split('=')[1]
        cmd = f'blkid -t LABEL={_label}'
        res = get_output(cmd)
        if not res['result']:
            comment = f'Unable to find device with label {_label}.'
            raise CommandExecutionError(comment)
        else:
            # output is a list of entries like this:
            # /dev/sda: LABEL="<label>" UUID="<uuid>" UUID_SUB="<uuid>" TYPE="btrfs"
            # exact list of properties varies between filesystems, but we're
            # only interested in the device in the first column
            for line in res['stdout']:
                dev_with_label = line.split(':')[0]
                device_list.append(dev_with_label)
            real_device = device_list[0]
    else:
        real_device = device

    # LVS devices have 2 names under /dev:
    # /dev/mapper/vg--name-lv--name and /dev/vg-name/lv-name
    # No matter what name is used for mounting,
    # mount always displays the device as /dev/mapper/vg--name-lv--name
    # Note the double-dash escaping.
    # So, let's call that the canonical device name
    # We should normalize names of the /dev/vg-name/lv-name type to the canonical name
    lvs_match = re.match(r'^/dev/(?P<vg_name>[^/]+)/(?P<lv_name>[^/]+$)', device)
    if lvs_match:
        double_dash_escaped = dict((k, re.sub(r'-', '--', v)) for k, v in six.iteritems(lvs_match.groupdict()))
        mapper_device = '/dev/mapper/{vg_name}-{lv_name}'.format(**double_dash_escaped)
        if os.path.exists(mapper_device):
            real_device = mapper_device

    # When included, FUSE devices are prefaced by the
    # filesystem type and a hash, e.g. sshfs.  In the mount list only the
    # hostname is included.  So if we detect that the device is a FUSE device
    # then we remove the prefaced string so that the device in state matches
    # the device in the mount list.
    fuse_match = re.match(r'^\w+\#(?P<device_name>.+)', device)
    if fuse_match and 'device_name' in fuse_match.groupdict():
        real_device = fuse_match.group('device_name')

    if real_name in active_mount:
        if 'superopts' not in active_mount[real_name]:
            active_mount[real_name]['superopts'] = []
        if mount:
            device_list.append(active_mount[real_name]['device'])
            device_list.append(os.path.realpath(device_list[0]))
            alt_device = active_mount[real_name]['alt_device'] if 'alt_device' in active_mount[real_name] else None
            uuid_device = active_mount[real_name]['device_uuid'] if 'device_uuid' in active_mount[real_name] else None
            label_device = active_mount[real_name]['device_label'] if 'device_label' in active_mount[real_name] else None
            if alt_device and alt_device not in device_list:
                device_list.append(alt_device)
            if uuid_device and uuid_device not in device_list:
                device_list.append(uuid_device)
            if label_device and label_device not in device_list:
                device_list.append(label_device)
            if opts:
                mount_invisible_options = [
                    'x-systemd.automount',
                    'noauto',
                    '_netdev',
                    'actimeo',
                    'bg',
                    'comment',
                    'defaults',
                    'delay_connect',
                    'direct-io-mode',
                    'intr',
                    'loop',
                    'nointr',
                    'nobootwait',
                    'nofail',
                    'password',
                    'reconnect',
                    'retry',
                    'soft',
                    'auto',
                    'users',
                    'bind',
                    'nonempty',
                    'transform_symlinks',
                    'port',
                    'backup-volfile-servers',
                ]

                if extra_mount_invisible_options:
                    mount_invisible_options.extend(extra_mount_invisible_options)

                if hidden_opts:
                    mount_invisible_options = list(set(mount_invisible_options) | set(hidden_opts))

                # options which are provided as key=value (e.g. password=Zohp5ohb)
                mount_invisible_keys = [
                    'actimeo',
                    'comment',
                    'direct-io-mode',
                    'password',
                    'retry',
                    'port',
                ]

                if extra_mount_invisible_keys:
                    mount_invisible_keys.extend(extra_mount_invisible_keys)

                # Some filesystems have options which should not force a remount.
                mount_ignore_fs_keys = {
                    'ramfs': ['size']
                }

                if extra_mount_ignore_fs_keys:
                    mount_ignore_fs_keys.update(extra_mount_ignore_fs_keys)

                # Some options are translated once mounted
                mount_translate_options = {
                    'tcp': 'proto=tcp',
                    'udp': 'proto=udp',
                }

                if extra_mount_translate_options:
                    mount_translate_options.update(extra_mount_translate_options)

                for opt in opts:
                    if opt in mount_translate_options:
                        opt = mount_translate_options[opt]

                    keyval_option = opt.split('=')[0]
                    if keyval_option in mount_invisible_keys:
                        opt = keyval_option

                    size_match = re.match(r'size=(?P<size_value>[0-9]+)(?P<size_unit>k|m|g)', opt)
                    if size_match:
                        converted_size = int(size_match.group('size_value'))
                        if size_match.group('size_unit') == 'm':
                            converted_size = int(size_match.group('size_value')) * 1024
                        if size_match.group('size_unit') == 'g':
                            converted_size = int(size_match.group('size_value')) * 1024 * 1024
                        opt = f"size={converted_size}k"
                    # make cifs option user synonym for option username which is reported by /proc/mounts
                    if fstype in ['cifs'] and opt.split('=')[0] == 'user':
                        opt = "username={0}".format(opt.split('=')[1])

                    # convert uid/gid to numeric value from user/group name
                    name_id_opts = {'uid': user_info,
                                    'gid': group_info}
                    if opt.split('=')[0] in name_id_opts and len(opt.split('=')) > 1:
                        _givenid = opt.split('=')[1]
                        _param = opt.split('=')[0]
                        _id = _givenid
                        if not re.match('[0-9]+$', _givenid):
                            _info = name_id_opts[_param](_givenid)
                            if _info and _param in _info:
                                _id = _info[_param]
                        opt = _param + '=' + str(_id)

                    if opt not in active_mount[real_name]['opts'] \
                            and opt not in active_mount[real_name].get('superopts', []) \
                            and opt not in mount_invisible_options \
                            and opt not in mount_ignore_fs_keys.get(fstype, []) \
                            and opt not in mount_invisible_keys:
                        # Some file systems require umounting and mounting if options change
                        # add others to list that require similiar functionality
                        if fstype in ['nfs', 'cvfs'] or fstype.startswith('fuse'):
                            unmount_result = unmounted(real_name)
                            if unmount_result['result']:
                                result = storage_mount(real_name, device, mkmnt=mkmnt, fstype=fstype, opts=opts,
                                                       systemd=systemd)
                            else:
                                comment = f"Unable to unmount {real_name}: {unmount_result['comment']}."
                                raise CommandExecutionError(comment)
                        else:
                            result = remount(real_name, device, mkmnt=mkmnt, fstype=fstype, opts=opts)
                            # Cleanup after the remount, so we
                            # don't write remount into fstab
                            if 'remount' in opts:
                                opts.remove('remount')

            if real_device not in device_list:
                # we mast fail state, because something mounted to this name
                comment = f'Unable to mount, real device {real_device} with name {name} is mounted'
                raise CommandExecutionError(comment)
            else:
                comment = f'Target {real_device} was already mounted'
    # using a duplicate check, so I can catch the results of an umount
    if real_name not in active_mount:
        if mount:
            # The mount is not present! Mount it
            # We can't mount already mounted device
            for active_device in active_mount:
                if real_device == active_mount[active_device]['device']:
                    comment = f'Unable to mount, this device {device} is mounted. Try to scan node.'
                    raise CommandExecutionError(comment)
            # And we can't mount device with such record in fstab.
            if _check_mountpoint(device, real_name):
                comment = 'This mount point already in fstab'
                raise CommandExecutionError(comment)

            if not os.path.exists(name) and not mkmnt:
                comment = f'Mount directory {name} is not present'
                raise CommandExecutionError(comment)

            out = storage_mount(name, device, mkmnt, fstype, opts, user=user, systemd=systemd)
            active_mount = active(extended=True)
            if isinstance(out, string_types):
                # Failed to (re)mount, the state has failed!
                comment = out
                if os.path.exists(name):
                    try:
                        os.rmdir(name)
                    except OSError:
                        comment += ' Mount point not empty, cann\'t delete.'
                raise CommandExecutionError(comment)
            elif real_name in active_mount:
                if owner:
                    if fstype == 'nfs':
                        if not skip_permissions_check and not _check_perms(owner, name):
                            comment = 'Mount point owner is not equal to the expected owner.'
                            unmounted(name, device, user=user)
                            raise CommandExecutionError(comment)
                    elif fstype in ['glusterfs', 'ocfs2', 'gfs2']:
                        if not (_check_perms(owner, name) or _chown_directory(owner, name)):
                            comment = 'Mount point owner is not equal to the expected owner.'
                            unmounted(name, device, user=user)
                            raise CommandExecutionError(comment)
                # (Re)mount worked!
                comment = f'Target {real_device} was successfully mounted'
        elif not os.path.exists(name):
            if mkmnt:
                mkdir(name, user=user)
                comment = f'{name} was created, not mounted'
            else:
                comment = f'{name} not present and not mounted'
        else:
            comment = f'{name} not mounted'

    if persist:
        out = set_fstab(name, device, fstype, opts, dump, pass_num, config, match_on=match_on)

        if out == 'present':
            comment += '. Entry already exists in the fstab.'
            return comment
        if out == 'new':
            comment += '. Added new entry to the fstab.'
            return comment
        if out == 'change':
            comment += '. Updated the entry in the fstab.'
            return comment
        if out == 'bad config':
            comment += '. However, the fstab was not found.'
            raise CommandExecutionError(comment)

    if result:
        return comment
    raise CommandExecutionError(comment)


@states_response
def unmounted(name,
              device=None,
              config='/etc/fstab',
              persist=False,
              user=None,
              lazy=False):
    """
    Verify that a device is not mounted

    name
        The path to the location where the device is to be unmounted from

    device
        The device to be unmounted.  This is optional because the device could
        be mounted in multiple places.

    config
        Set an alternative location for the fstab, Default is ``/etc/fstab``

    persist
        Set if the mount should be purged from the fstab, Default is ``False``

    user
        The user to own the mount
    """
    result = True

    # Get the active data
    active_mounts = active(extended=True)
    if name not in active_mounts:
        # Nothing to unmount
        comment = f'Target {name} was already unmounted.'
        for k, v in active_mounts.items():
            if device and v.get('device') == device:
                out = umount(k, device, lazy)
                if not out['result']:
                    # Failed to umount, the state has failed!
                    comment = out['comment']
                    result = False
                else:
                    # umount worked!
                    comment += f' Target {k} was also unmounted.'
    else:
        # The mount is present! Unmount it
        if device:
            out = umount(name, device, lazy)
        else:
            out = umount(name, lazy)

        if not out['result']:
            # Failed to umount, the state has failed!
            comment = out['comment']
            result = False
        else:
            # umount worked!
            comment = f'Target {name} was successfully unmounted.'

    if persist:
        fstab_data = fstab(config)

        if name not in fstab_data:
            comment += f'. fstab entry for name {name} not found'
        else:
            if device and fstab_data[name]['device'] != device:
                comment += f'. fstab entry for device {device} not found.'
                return comment

            out = rm_fstab(name, device, config)
            if out is not True:
                result = False
                comment += f'. Failed to persist purge: {out}.'
            else:
                comment += '. Removed target from fstab.'

        for k, v in fstab_data.items():
            if device and v.get('device') == device:
                out = rm_fstab(k, device, config)
                if out is not True:
                    comment += f'. Also removed target {k} from fstab.'
    if not result:
        raise CommandExecutionError(comment)
    return comment
