"""
(c) 2023 «Dacom M» Ltd
"""

from itertools import zip_longest, groupby
from copy import deepcopy
from collections import defaultdict, OrderedDict
import json
import os
import time
import re
from re import findall, split as rsplit
from subprocess import getstatusoutput
import xmltodict
import uuid
import shutil
import stat

from veilrpc.utils.bazalt import states_response, CommandExecutionError, ROLE_SCRIPT_PATH, PATH_NODES_HEARTBEAT, \
    get_output, LOCAL_MODULES_PATH, TransportProtocols, thread_pool, FENCE_PATH, states_result, \
    HEARTBEAT_PATH, INSTALL_TESTS_PATH, INSTALL_NODE_PASS, BASE_SCRIPT_PATH, RUN_PATH, distribution, \
    VEIL_NODE_LOCAL_PATH, get_config, VEIL_ROLE_FILE, ZPOOL_STORAGES_PATH, \
    OCFS2_PATH, GFS2_PATH, GLUSTER_PATH, LOCAL_STORAGES_PATH, get_result, CLUSTER_PATH, DOMAINS_FOLDER, get_oem, \
    PATH_FILE_STORAGE
from veilrpc.states.puppet import puppet_modify_param
from veilrpc.utils.virt import list_domains, node_info, cpu_baseline, create_node_device, \
    virt_destroy_by_id, virt_shutdown_by_id, virt_get_xml, list_all_devices, \
    list_all_storage_pools, list_all_vols
from veilrpc.utils.smbios import records
from veilrpc.utils.file import modify_file, chown_directory
from veilrpc.states.domain import get_shutoff_domains, CommonDataOperator
from veilrpc.utils.size import memory_units_convertor

PARSER = re.compile(r"\d+")
NTP_PATH = f'{LOCAL_MODULES_PATH}/ntp.yaml'
SYSCTL_PATH = f'{LOCAL_MODULES_PATH}/sysctl.yaml'
MDEV_DB_DIR = '/etc/mdev-instances.d'
MDEV_BASE = '/sys/class/mdev_bus'
MDEV_DEVICES = '/sys/bus/mdev/devices'
GRUB_FILE = '/etc/default/grub'
GRID_CFG = '/etc/modprobe.d/nvidia.conf'
SRIOV_MGR_LIST = ['/usr/lib/nvidia/sriov-mgr', '/usr/lib/nvidia/sriov-manage']  # different names in different versions
GRID_SRIOV_ENABLE_LIST = '/etc/veil-grid-sriov'
PCI_PATTERN = "veil_node::grub::iommu_mode: '{}'"
PCIE_ACS_OVERRIDE_PATTERN = "veil_node::grub::pcie_acs_override: '{}'"
DATAPOOLS_CMD = "find {storages}/*/ -maxdepth 1 -type f -name datapool*.meta 2>&-"
SHARED_STORAGES_CMD = "find {storages}/ -maxdepth 1 -type f -name shared_storage*.meta 2>&-"
ZFS_FIND_META_CMD = "find {storages} -maxdepth 1 -type f -name datapool*.meta 2>&-"
BMC_ACTIONS = {'no_action': 0,
               'hard_reset': 1,
               'power_off': 2,
               'power_cycle': 3}
PUBLIC_KEY_IDENTIFIER = ' veil_public_key'
ROOT_AUTHORIZED_KEYS = '/root/.ssh/authorized_keys'
USER_HOME = '/home/{username}'
USER_SSH_FOLDER = f'{USER_HOME}/.ssh'
USER_AUTHORIZED_KEYS = f'{USER_SSH_FOLDER}/authorized_keys'


def _zerofill_flag(devname):
    """
    флаг создается, когда устройство затирается нулями
    Решетка в начале значит, что флаг будет проверяться storage_transport_helper на наличие процесса с PID, cmdline,
    comm, указанными в флаге (см. popen_with_progress), когда таковой не найден, флаг удаляется
    """
    return os.path.join(RUN_PATH, '#zerofill_' + devname.replace('/', '_'))


def get_cpu_topology():
    """
    Собирает информацию о топологии узла.
    """

    def _find_numbers(line):
        return re.findall(PARSER, line)

    topology = {"cpu_map": {"all": {}}, 'cpu_sockets': 0, 'cpu_cores': 0, 'cpu_threads': 0}

    # Get numa topology
    numa_cmd = "numactl -H"
    cmd_ret = get_output(numa_cmd)
    if not cmd_ret['result']:
        comment = f"An error occurred while getting node cpu topology: {cmd_ret['comment']}"
        raise CommandExecutionError(comment)
    else:
        lines = cmd_ret['comment'].splitlines()
        for i, line in enumerate(lines):
            matches = _find_numbers(line)
            if line.startswith('available'):
                topology['cpu_map'].update(
                    {str(node): {'cpus': [0], 'distances': "10"} for node in range(int(matches[0]))})
            elif 'cpus' in line:
                topology['cpu_map'][matches[0]]['cpus'] = list(map(int, matches[1:]))
                for cpu_num in matches[1:]:
                    topology['cpu_map']['all'][cpu_num] = []
            elif 'size' in line:
                topology['cpu_map'][matches[0]]['size'] = matches[1]
            elif 'free' in line:
                topology['cpu_map'][matches[0]]['free'] = matches[1]
            elif 'distances' in line:
                for node in range(len(topology['cpu_map'].keys()) - 1):
                    distances = _find_numbers(lines[i + 2 + node])
                    topology['cpu_map'][distances[0]]['distances'] = ','.join(distances[1:])
        return topology


@states_response
def get_meta(storages_dir, **kwargs):
    cmd_ret = get_output(SHARED_STORAGES_CMD.format(storages=storages_dir), **kwargs)
    if not cmd_ret['result']:
        return {}
    out = [x for x in cmd_ret['comment'].splitlines() if x]
    if out:
        try:
            with open(out[0]) as f:
                return json.load(f)
        except (OSError, json.JSONDecodeError):
            pass
    return {}


@states_response
def discovery_datapool(storages_dirs, check_lvm=True, check_head=True, node_id=None, check_space=False, **kwargs):
    from .directory_datapool import check_files_owner, get_free_space_shell
    results = []
    datapools_1 = []

    # find meta
    for path in storages_dirs:
        cmd = ZFS_FIND_META_CMD if ZPOOL_STORAGES_PATH in path else DATAPOOLS_CMD
        cmd_ret = get_output(cmd.format(storages=path), **kwargs)
        if not cmd_ret['result']:
            continue
        out = [x for x in cmd_ret['comment'].splitlines() if x]
        if out:
            datapools_1.extend(out)

    # check owners
    res = check_files_owner(datapools_1, **kwargs)
    datapools_2 = res['comment'] if res['result'] else []

    # check heartbeat
    if node_id:
        datapools_3 = []
        for path in datapools_2:
            file_dir = os.path.join(os.path.dirname(path), PATH_NODES_HEARTBEAT)
            if os.path.isdir(file_dir):
                file_path = os.path.join(file_dir, node_id)
                if modify_file(file_path):
                    datapools_3.append(path)
            else:
                datapools_3.append(path)
    else:
        datapools_3 = datapools_2

    # check head
    for path in datapools_3:
        if check_head:
            cmd = f'head -n 1 {path}'
            head_ret = get_output(cmd, **kwargs)
            if not head_ret['result']:
                continue
            if check_space:
                space_ret = get_free_space_shell(path)
                if space_ret['result']:
                    available, total = space_ret['comment']
                else:
                    available, total = 0, 0
                results.append({'path': path, 'data': head_ret['comment'], 'space': {"free": available, "size": total}})
            else:
                results.append({'path': path, 'data': head_ret['comment']})
        else:
            results.append(path)

    # check lvm
    if check_lvm:
        cmd = 'lvscan'
        cmd_ret = get_output(cmd, **kwargs)
        if not cmd_ret['result']:
            raise CommandExecutionError(f"Can not discover lvm: {cmd_ret['comment']}")
        out = cmd_ret['comment'].splitlines()
        for item in out:
            try:
                path = item.replace("'", "").replace('\\', '').split()[1]
                if re.search(r'\.meta$', path):
                    if check_head:
                        cmd = f'head -n 1 {path}'
                        head_ret = get_output(cmd)
                        if not head_ret['result']:
                            raise CommandExecutionError(f"Can not discover lvm datapools: {head_ret['comment']}")
                        results.append({'path': path, 'data': head_ret['comment']})
                    else:
                        results.append(path)
            except (ValueError, CommandExecutionError, TypeError):
                continue
    return results


@states_response
def discovery_shared_storage():
    cmd_ret = get_output('mount')
    if not cmd_ret['result']:
        comment = f"An error occurred while discovering shared storages: {cmd_ret['comment']}"
        raise CommandExecutionError(comment)
    else:
        mount_out = cmd_ret['comment'].split('\n')
        mount_dict = {}
        try:
            for mount_row in mount_out:
                if mount_row:
                    row_list = [t.strip() for s in mount_row.split(" type ") for t in s.split(" on ")]
                    row_list.extend([t.strip('()') for t in row_list[-1].split(' ')])
                    row_list.remove(row_list[2])
                    row_dict = dict(
                        connection_type=row_list[2] if row_list[2] != 'fuse.glusterfs' else 'glusterfs',
                        verbose_name='{hostname}_{mount_point}'.format(hostname=row_list[0].split(':')[0],
                                                                       mount_point=row_list[1]),
                        description='',
                        datacenter='',
                        options=dict()
                    )
                    row_dict['options'] = {
                        'volume': row_list[0].split(':')[-1],
                        'mount_point': row_list[1],
                        'hostname': row_list[0].split(':')[0],
                        'mount_options': {}
                    }
                    if row_dict['connection_type'] == 'nfs':
                        owner = get_output("stat -c '%U' {}".format(row_dict['options']['mount_point']))
                        if owner['comment'] == 'root':
                            row_dict['options']['server_use_no_root_squash'] = True
                        else:
                            row_dict['options']['server_use_no_root_squash'] = False
                    for param in row_list[3].split(','):
                        param = param.split('=')
                        if len(param) > 1:
                            try:
                                row_dict['options']['mount_options'][param[0]] = int(param[1])
                            except ValueError:
                                row_dict['options']['mount_options'][param[0]] = param[1]
                        else:
                            row_dict['options']['mount_options'][param[0]] = True
                    if row_dict['connection_type'] in mount_dict.keys():
                        mount_dict[row_dict['connection_type']].append(row_dict)
                    else:
                        mount_dict[row_dict['connection_type']] = [row_dict]
            return mount_dict
        except Exception as e:
            comment = f"An error occurred while discovering shared storages: {str(e)}. Check the 'mount' command."
            raise CommandExecutionError(comment) from None


@states_response
def discovery_cluster_storage(**kwargs):
    from .cluster import quorum_status, ocfs2_status
    from .gluster import status_cluster, status_detail_volume, check_volume_mounted

    cs_statuses = dict()

    # Gluster
    cs_statuses[GLUSTER_PATH] = {'status': False, 'volumes': []}
    res = status_cluster()
    if res['result']:
        cs_statuses[GLUSTER_PATH]['status'] = res['comment']
        volumes_res = status_detail_volume()
        if volumes_res['result']:
            volumes = volumes_res['comment']
            for volume in volumes:
                volume['mounted'] = check_volume_mounted(volume.get('name'), **kwargs)
            cs_statuses[GLUSTER_PATH]['volumes'] = volumes
    # GFS2
    res = quorum_status(suppress_err=True)
    cs_statuses[GFS2_PATH] = {'status': res['result']}  # pylint: disable=unsubscriptable-object
    # OCFS2
    res = ocfs2_status(suppress_err=True)
    cs_statuses[OCFS2_PATH] = {'status': res['result']}  # pylint: disable=unsubscriptable-object
    return cs_statuses


@states_response
def discovery_scsi_multipath(**kwargs):
    from .iscsi import transport_luns
    # SCSI
    luns_in_mpath = dict()
    ret = transport_luns(TransportProtocols.scsi, **kwargs)
    if not ret['result']:
        raise CommandExecutionError(ret['comment'])
    luns = ret['comment']
    ret = get_output('multipathd show maps json')
    if not ret['result']:
        raise CommandExecutionError(ret['comment'])
    try:
        mp_maps = json.loads(ret['comment'])
    except (json.JSONDecodeError, TypeError) as e:
        raise CommandExecutionError(f'show maps err, e={e}') from None
    for mpath_dev_id in [mp['uuid'] for mp in mp_maps['maps']]:
        if os.path.isfile(f'/etc/sys_mpath/{mpath_dev_id}'):
            # mpath with system partitions, file created during installation, do not discover
            continue
        for lun in luns:
            lun_device_id = lun.get('device_id')
            if lun_device_id == mpath_dev_id and lun_device_id not in luns_in_mpath.keys():
                luns_in_mpath[lun_device_id] = lun
    return list(luns_in_mpath.values())


@states_response
def discovery_transport_storage_multipath(as_root=True, **kwargs):
    from .iscsi import iscsi_session_rescan, transport_luns, ISCSI_TARGET_REGEX
    result = dict()

    # FC
    fc = dict()
    ret = transport_luns(TransportProtocols.fc, **kwargs)
    if not ret['result']:
        raise CommandExecutionError(ret['comment'])
    luns = ret['comment']
    ret = get_output('multipathd show maps json')
    if not ret['result']:
        raise CommandExecutionError(ret['comment'])
    try:
        mp_maps = json.loads(ret['comment'])
    except (json.JSONDecodeError, TypeError) as e:
        raise CommandExecutionError(f'show maps err, e={e}') from None

    for mpath_dev_id in [mp['uuid'] for mp in mp_maps['maps']]:
        if os.path.isfile(f'/etc/sys_mpath/{mpath_dev_id}'):
            # mpath with system partitions, file created during installation, do not discover
            continue
        # print(group_serial_number, len(list(g)))
        targets = list()
        luns_in_mpath = dict()
        for lun in luns:
            lun_device_id = lun.get('device_id')
            if lun_device_id == mpath_dev_id:
                targets.append(lun['target'])
                if lun_device_id not in luns_in_mpath.keys():
                    luns_in_mpath[lun_device_id] = lun
                else:
                    if lun['serial_number'] not in luns_in_mpath[lun_device_id]['serial_number'].split(', '):
                        luns_in_mpath[lun_device_id]['serial_number'] += ', ' + lun['serial_number']
        targets.sort()
        targets = tuple(targets)

        if targets not in fc.keys():
            fc[targets] = list(luns_in_mpath.values())
        else:
            fc[targets].extend(list(luns_in_mpath.values()))
    result['fc'] = {str(k): list({d['device_id']: d for d in v}.values()) for k, v in fc.items() if k}

    # ISCSI
    iscsi = dict()
    cmd_ret = get_output('iscsiadm -m session -P 1 2>&-', **kwargs)
    if not cmd_ret['result']:
        if cmd_ret.get('retcode', 0) != 21:
            comment = f"An error occurred while discovering transport storage on node: {cmd_ret['comment']}"
            raise CommandExecutionError(comment)
    else:
        if cmd_ret['comment']:
            iscsi_session_rescan(as_root, **kwargs)
            ret = transport_luns(TransportProtocols.iscsi, rescan=False, **kwargs)
            if not ret['result']:
                raise CommandExecutionError(ret['comment'])

            target, hostname = None, None
            iscsi_sources = {}
            for line in cmd_ret['comment'].splitlines():
                if line.strip().startswith("Target"):
                    target = re.search(ISCSI_TARGET_REGEX,
                                       line.strip()).group(0)
                    if target not in iscsi_sources.keys():
                        iscsi_sources[target] = []
                elif line.strip().startswith("Current Portal"):
                    if target:
                        _, host_string = line.split(":", 1)
                        hostname = host_string.split(',')[0].strip()
                elif line.strip().startswith("iSCSI Session State") and \
                        line.strip().split(":")[-1].strip() in ["LOGGED_IN", "ACTIVE"] and \
                        target and hostname:
                    iscsi_target_hostname = f"{target}||{hostname}"
                    iscsi_sources[target].append(iscsi_target_hostname)
                    hostname = None

            luns = ret['comment']
            iscsi_luns = groupby(sorted(luns, key=lambda x: x['serial_and_id']),
                                 lambda x: x['serial_and_id'])
            for mpath_dev_id, g in iscsi_luns:
                # print(group_serial_number, len(list(g)))
                try:
                    if os.path.isfile(f'/etc/sys_mpath/{list(g)[0]["device_id"]}'):
                        # mpath with system partitions, file created during installation, do not discover
                        continue
                except (IndexError, KeyError):
                    pass
                targets = list()
                luns_in_mpath = dict()
                for lun in luns:
                    lun_device_id = lun.get('serial_and_id')
                    if lun_device_id == mpath_dev_id:
                        targets = iscsi_sources.get(lun['target'])
                        if lun_device_id not in luns_in_mpath.keys():
                            luns_in_mpath[lun_device_id] = lun
                if targets:
                    targets = tuple(targets)
                    if targets not in iscsi.keys():
                        iscsi[targets] = list(luns_in_mpath.values())
                    else:
                        iscsi[targets].extend(list(luns_in_mpath.values()))
    result['iscsi'] = {str(k): list({d['serial_and_id']: d for d in v}.values()) for k, v in iscsi.items()}
    return result


@states_response
def get_file_mtime(path, create_as_root, owner):
    if create_as_root:
        cmd = f"stat -c %Y {path}"
    else:
        cmd = f"su {owner} -c \'stat -c %Y {path}\'"
    cmd_ret = get_output(cmd)
    if not cmd_ret['result']:
        raise CommandExecutionError(cmd_ret['comment'])
    return cmd_ret['comment']


@states_response
def get_node_usb_devices():
    """
    Возвращает usb устройства узла с любой шины.
    :return: ret ['001', '003', '05e3', '0608', 'Genesys Logic, Inc. Hub']
    "Bus 001 Device 003: ID 05e3:0608 Genesys Logic, Inc. Hub"
    """

    usb_cmd = "lsusb"
    cmd_ret = get_output(usb_cmd)
    cmd_ret_tree = get_output(usb_cmd + ' -t')
    if not cmd_ret['result']:
        comment = f"An error occurred while get node's usb devices: {cmd_ret['comment']}"
        raise CommandExecutionError(comment)
    elif not cmd_ret_tree['result']:
        comment = f"An error occurred while get node's usb tree devices: {cmd_ret_tree['comment']}"
        raise CommandExecutionError(comment)
    else:
        try:
            data = []

            def anon(line):
                bus, device, device_id, device_name = line[1], line[3][:-1], line[5], " ".join(line[6:])
                device_name = device_name or 'Unknown device'
                vendor, product = device_id.split(':')
                if int(bus):
                    data.append({"bus": bus, "device": device, "vendor": vendor,
                                 "product": product, "device_name": device_name})

            for line in cmd_ret['comment'].splitlines():
                line = line.split()
                if line:
                    if line[-1].lower() == 'hub':
                        if line[-2].lower() == 'root':
                            continue
                        else:
                            anon(line)
                    else:
                        anon(line)
            memo = dict()
            memo['devices'] = []
            result = list()
            indent = 0
            dev_dict = False
            for line in cmd_ret_tree['comment'].splitlines():
                line = line.split(' ')
                if '/:' in line:
                    memo['bus'] = format(int(line[3][:2]), '03d')
                    indent = 0
                    dev_dict = False
                elif '|__' in line:
                    memo['previous_indent'] = indent
                    memo['previous_device'] = dev_dict
                    indent = line.index('|__')
                    device = format(int(line[indent + 4][:-1]), '03d')
                    if memo['bus'] + device in memo['devices']:
                        continue
                    else:
                        memo['devices'].append(memo['bus'] + device)
                        dev_dict = {'device': device,
                                    'device_class': ' '.join(line[indent + 7:line.index(line[-2])]).split('=')[1][:-1],
                                    'bus': memo['bus']}
                        for dev in data:
                            if dev['device'] == device and dev['bus'] == dev_dict['bus']:
                                dev_dict.update(dev)
                        if 'hub' in dev_dict['device_class'].lower():
                            dev_dict['input'] = list()
                        if not memo['previous_device']:
                            result.append(dev_dict)
                        else:
                            if indent > memo['previous_indent']:
                                memo[memo['previous_indent']] = memo['previous_device']
                                memo['previous_device']['input'].append(dev_dict)
                            else:
                                if indent - 4 == 0:
                                    result.append(dev_dict)
                                else:
                                    memo[indent - 4]['input'].append(dev_dict)
            return result
        except Exception as e:
            comment = "An error occurred while get node's usb devices. " \
                      "Error: {}".format(str(e))
            raise CommandExecutionError(comment) from None


def get_mdev_name_descr(path, supp_type):
    mst_path = os.path.join(path, supp_type)
    if os.path.exists(os.path.join(mst_path, 'name')):
        # у nvidia внятное имя лежит в файлике name, а у intel сам supported_type и есть имя
        with open(os.path.join(mst_path, 'name')) as f:
            name = f.read().rstrip()
    else:
        name = supp_type
    try:
        # неясно, будет ли существовать description на каких-то иных не-NVIDIA устройствах
        with open(os.path.join(mst_path, 'description')) as f:
            descr = dict()
            dsc_raw = f.read().rstrip()
            if dsc_raw.count('\n') > dsc_raw.count(', '):
                # intel GVT-g style
                linesep, sep = '\n', ': '
            else:
                # nvidia style
                linesep, sep = ', ', '='
            for line in dsc_raw.split(linesep):
                try:
                    k, v = line.split(sep)
                    descr[k] = int(v) if str.isdecimal(v) else v
                except (IndexError, ValueError):
                    descr[line] = ''
    except FileNotFoundError:
        descr = {}
    return name, descr


@states_response
def get_node_mdev_devices(uuids=None):
    """
    Получает список uuid mdev устройств, которые должны существовать с точки зрения контроллера,
    если находит на узле лишние -- пытается удалить, ошибки игнорируются

    Возвращает доступные mdev устройства, при отсутствии -- пустой словарь в result
    :return: ret {..., 'result':
                                {
                                 name: {
                                        'md_type': <mdev_supported_types>,
                                        'avail': total_available_instances,
                                        'allocated': [uuid, ...] (full list of allocated instances of this type)
                                        'description': {'num_heads': num_heads,
                                                        'frl': frl_config,
                                                        'framebuffer': framebuffer,
                                                        'max_res': max_resolution,
                                                        'max_instances': max_instances},
                                        'modalias': modalias ('pci:'VID+PID+subven+subdev+basecode+subcode+intf_ver)
                                        'data': {
                                                <pci_id0>: {'avail': available,
                                                            'allocated': [uuid, ...] },
                                                ...
                                                }
                                        },
                                ...
                                }
                }
    пример полей:
    avail:          (int)
    pci_id:         '0000:03:00.0'
    mst:            'nvidia-155'
    name:           'GRID M10-2B'
    description:    'num_heads': 2, 'frl_config': 45, 'framebuffer': '2048M', 'max_res': '4096x2160', 'max_instance': 4
    modalias:       'pci:v000010DEd000013BDsv000010DEsd00001160bc03sc00i00'
     ^^^^ используем для проверки, что после перезагрузки под этим pci dom:bus:dev.func будет то же самое устройство
    """

    def vgpu_sort(model, dct):
        #  порядок сортировки: доступность, vRAM, подвид
        if not str.upper(model).startswith('GRID'):
            return model

        full_type = model.split('-')[-1]                        # GRID P100-12C => 12Q
        avail_key = '1' if dct[model]['avail'] == 0 else '0'    # устройства в наличии показываем первыми
        ram = re.search(r'^\d*', full_type).group(0)            # 12Q => 12
        just_type = full_type[len(ram):]                        # 12Q => Q; делим RAM и тип, чтоб 1B4 не влезало перед 6Q
        return avail_key + ram.zfill(8) + just_type             # [0/1][00000012][Q/B4]

    if isinstance(uuids, list) and os.path.exists(MDEV_DEVICES):
        for instance in [x for x in os.listdir(MDEV_DEVICES) if x not in uuids]:
            try:
                remove_mdev_instance(instance)
            except CommandExecutionError:
                pass

    ret = dict()
    if not os.path.exists(MDEV_BASE):
        return ret

    for pci_id in os.listdir(MDEV_BASE):

        mod_path = os.path.join(MDEV_BASE, pci_id)
        with open(os.path.join(mod_path, 'modalias')) as f:
            modalias = f.read()
        path = os.path.join(mod_path, 'mdev_supported_types')
        for supp_type in os.listdir(path):
            name, descr = get_mdev_name_descr(path, supp_type)

            with open(os.path.join(path, supp_type, 'available_instances')) as f:
                avail = int(f.read().rstrip())

            try:
                # список UUID созданных устройств этого типа
                alloc_list = os.listdir(os.path.join(path, supp_type, 'devices'))
            except FileNotFoundError:
                alloc_list = []

            # исходим из того, что одинаковые типы vGPU встречаются на каждом ядре единожды, уже сгруппированы в sysfs
            type_on_core = {pci_id: dict(allocated=alloc_list, avail=avail)}
            if not ret.get(name):
                ret[name] = dict(md_type=supp_type, description=descr, avail=avail, data=type_on_core,
                                 allocated=alloc_list.copy(), modalias=modalias)
            else:
                ret[name]['data'].update(type_on_core)
                ret[name]['avail'] += avail
                ret[name]['allocated'] += alloc_list

    return {k: ret[k] for k in sorted(ret, key=lambda x: vgpu_sort(x, ret))}


@states_response
def create_mdev_instance(pci_id, md_type, old_uuid=None, really_create=True):
    # если really_create == False, просто возвращаем имя и дескрипшен, ничего не создавая (добавляем mdev к шаблону)
    # old_uuid нужен при конвертации шаблона в ВМ, создаём mdev с ранее созданным uuid, лежащим в БД
    _uuid = old_uuid or uuid.uuid4().__str__()
    name, descr = get_mdev_name_descr(os.path.join(MDEV_BASE, pci_id, 'mdev_supported_types'), md_type)

    if really_create:
        try:
            with open(os.path.join(MDEV_BASE, pci_id, 'mdev_supported_types', md_type, 'available_instances'), 'r') as f:
                avail = int(f.read().rstrip())
        except OSError as e:
            raise CommandExecutionError(f'Failed to determine available MDEV instances count, e={e}') from None

        if avail == 0:
            raise CommandExecutionError('Unable to create MDEV: no available instances left on this core')

        try:
            with open(os.path.join(MDEV_BASE, pci_id, 'mdev_supported_types', md_type, 'create'), 'w') as f:
                f.write(_uuid)
        except OSError as e:
            raise CommandExecutionError(f'Unable to create MDEV: write error, e={e}') from None

        if _uuid not in os.listdir(os.path.join(MDEV_BASE, pci_id, 'mdev_supported_types', md_type, 'devices')):
            raise CommandExecutionError('Error: MDEV has not been created')
        # создаём запись для восстановления после перезагрузки
        with open(os.path.join(MDEV_DB_DIR, _uuid), 'w') as f:
            with open(os.path.join(MDEV_BASE, pci_id, 'modalias'), 'r') as ma:
                f.write(f'modalias={ma.read()}')
            f.write(f'pci_id={pci_id}\ntype={md_type}')

    return dict(uuid=_uuid, name=name, descr=descr)


@states_response
def remove_mdev_instance(_uuid):
    try:
        with open(os.path.join(MDEV_DEVICES, _uuid, 'remove'), 'w') as f:
            f.write('1')
    except OSError as e:
        # если экземпляра нет, то и ладно с ним, до нас удалили или не создали, всё равно успех
        if not isinstance(e, FileNotFoundError):
            try:
                with open(os.path.join(MDEV_DEVICES, _uuid, 'nvidia/vm_name'), 'r') as f:
                    vm = f.read().rstrip()
            except OSError:
                # если какой-то другой mdev не NVidia, например, без этой ветки
                vm = ''
            raise CommandExecutionError(f'Unable to remove MDEV (is VM {vm} still using it?), e={e}') from None

    if os.path.exists(os.path.join(MDEV_DB_DIR, _uuid)):
        # с файлом то же самое, нету и не надо, считай успешно удалили
        try:
            os.remove(os.path.join(MDEV_DB_DIR, _uuid))
        except OSError as e:
            raise CommandExecutionError(
                f'MDEV removed successfully but unable to remove DB record {_uuid}, e={e}') from None

    return 'Removal successful'


@states_response
def change_mdev_frl(_uuid, value: bool) -> None:
    # value = False -> writing 'frame_rate_limiter=0' to vgpu params and 'frl_disable=1' to mdev_recreate script
    try:
        with open(os.path.join(MDEV_DEVICES, _uuid, 'nvidia/vgpu_params'), 'r') as f_params:
            vgpu_params = [v for v in f_params.read().strip().split(', ') if v and (v != 'frame_rate_limiter=0')]
        if not value:
            vgpu_params.append('frame_rate_limiter=0')
        with open(os.path.join(MDEV_DEVICES, _uuid, 'nvidia/vgpu_params'), 'w') as f_params:
            f_params.write((', '.join(vgpu_params) + '\n') if vgpu_params else ' ')  # must be a space, not empty string
        with open(cfg_file := os.path.join(MDEV_DB_DIR, _uuid), 'r') as entry_to_recreate:
            data = [line.strip() for line in entry_to_recreate.readlines() if not line.startswith('frl_disable')] + \
                ([] if value else ['frl_disable=1'])
        with open(cfg_file, 'w') as entry_to_recreate:
            entry_to_recreate.write('\n'.join(data))
    except OSError as e:
        raise CommandExecutionError(f"Frame rate limiter for MDEV {_uuid} change failure; e={e}")


@states_response
def is_mdev_frl_enabled(_uuid) -> int:
    try:
        with open(os.path.join(MDEV_DEVICES, _uuid, 'nvidia/vgpu_params'), 'r') as params:
            data = params.read()
        return 0 if 'frame_rate_limiter=0' in data else 1
    except OSError:
        return -1


@states_response
def is_mdev_instance_available(_uuid):
    # в общем MDEV фреймворк не обязывает указывать, подключен ли экземпляр к чему-то, но nvidia позволяет
    # это узнать. Для других устройств просто проверяем, создан ли уже такой экземпляр.
    if not os.path.exists(os.path.join(MDEV_DEVICES, _uuid)):
        return False
    try:
        with open(os.path.join(MDEV_DEVICES, _uuid, 'nvidia/vm_name', 'r')) as f:
            # исходно там только \n, а если подключено, то uuid
            return len(f.read()) <= 1
    except OSError:
        return True


@states_response
def is_mdev_available_to_create(md_type, pci_id):
    path = os.path.join('/sys/class/mdev_bus', pci_id, 'mdev_supported_types', md_type, 'available_instances')
    if not os.path.exists(path):
        return False
    with open(path, 'r') as f:
        try:
            return int(f.read()) > 0
        except ValueError:
            return False


@states_response
def get_node_pci_devices():
    """
    Возвращает pci устройства узла с любой шины.
    :return: ret {..., result: [{'bus': '00',
                                 'slot': '01',
                                 'function': '01',
                                 'device_class': 'Ethernet controller',
                                 'vendor': 'Intel Corporation',
                                 'device_name': '82599ES 10-Gigabit SFI/SFP+ Network Connection'}, ...]
                 }

    01:00.1 Ethernet controller: Intel Corporation 82599ES 10-Gigabit SFI/SFP+ Network Connection (rev 01)
    01:00.1 "Ethernet controller" "Intel Corporation" "82599ES 10-Gigabit SFI/SFP+ Network Connection" -r01 "Intel Corporation" "Ethernet Server Adapter X520-2"
    """

    pci_cmd = "lspci -m"
    cmd_ret = get_output(pci_cmd)
    if not cmd_ret['result']:
        comment = f"An error occurred while get node's pci devices: {cmd_ret['comment']}"
        raise CommandExecutionError(comment)
    else:
        try:
            data = []
            for line in cmd_ret['comment'].splitlines():
                line = line.split('"')

                # убрать пробелы и пустые строки
                line = [elem.strip() for elem in line if elem.strip()]

                # распаковать 01:00.1 как 0000:01:00.1, 10000:01:00.1 как есть
                pci_split = rsplit('[:. ]', line[0])
                if len(pci_split) == 3:
                    domain, bus, slot, function_ = '0000', *pci_split
                elif len(pci_split) == 4:
                    domain, bus, slot, function_ = pci_split
                else:
                    raise CommandExecutionError(f'An error occurred while unpacking PCI address: {line[0]}')

                device_class, vendor, device_name = line[1:4]
                data.append(dict(domain=domain, bus=bus, slot=slot, function=function_,
                                 device_class=device_class, vendor=vendor, device_name=device_name))

            return data

        except Exception as e:
            comment = f"An error occurred while get node's pci devices: {str(e)}"
            raise CommandExecutionError(comment) from None


@states_response
def get_node_pci_devices_tree():
    r"""
    Используется чтобы подготовить для диалога с пользователем полное дерево
    pci устройств с названиями промежуточных и конечных узлов.
    Построчно принимаем вывод команды lspci -t (например, первые 5 строк):
        '-[0000:00]-+-00.0',
        '           +-1c.0-[05-0a]----00.0-[06-0a]--+-00.0-[07]--',
        '           |                               +-01.0-[08]--',
        '           |                               +-02.0-[09]----00.0',
        '           |                               \-04.0-[0a]--',

        Конвертим в list:
        ['0000', '00', '00', '0']
        ['1c', '0', '05', '00', '0', '06', '00', '0']
        ['|', '01', '0']
        ['|', '02', '0', '09', '00', '0']
        ['|', '04', '0']

        Возвращаем словарь:
        {
          "0000": {
            "00:00.0": {},
            "00:1c.0": {
              "05:00.0": {
                "06:00.0": {},
                "06:01.0": {},
                "06:02.0": {
                  "09:00.0": {}
                },
                "06:04.0": {}
              }
            },
            ...
            ...
    """

    def chunks(lst, n):
        return list(list(x) for x in zip_longest(*[iter(lst)] * n))

    def merge_dicts(dict1, dict2):
        """ Рекурсивно мержим dict2 и dict1 """
        if not isinstance(dict1, dict) or not isinstance(dict2, dict):
            return dict2
        for k in dict2:
            if k in dict1:
                dict1[k] = merge_dicts(dict1[k], dict2[k])
            else:
                dict1[k] = dict2[k]
        return dict1

    def parse_nested(s):
        """
        ['0000', '00', '00', '0']
        ['1c', '0', '05', '00', '0', '06', '00', '0']
        Первые 2 элемента - это всегда слот и функция шины Host bus 00,
        далее идут дочерние триплеты bus, slot, function

        Уровень вложенности определяем, отслеживая количество подстрок вида ---, -+-, \-.
            Cимвол \ является триггером уменьшения уровня вложенности (встречается перед
            последним child-элементом узла).
            Узлы --- являются мостами и при уменьшении вложенности удаляются без триггера
            если находятся на n-1 или n+1 позициях относительно удаляемого узла -+-.
            Символ | используем для ухода в ветвь кода, которая обрабатывает кейсы про вложения.
            Для обработки кейса, когда у последнего child-элемента есть вложения, используем
            переменную parent_branch_end для увеличения уровня вложенности в рамках 1 следующей итерации.

            Пример.

                         |           +-03.0-[84-89]----00.0-[85-89]--+-01.0-[86-87]--+-00.0
             |           |                               |               +-00.1
             |           |                               |               +-00.2
             |           |                               |               \-00.3
             |     последний элемент с вложением ----->  \-03.0-[88-89]--+-00.0
             |           |                                               \-00.1

             При генерации словаря используем сдвиг OFFSET - отсчет начинаем с уровня root bus:

             # -+-[0000: ff]-+-08.0 <--------- отсчет начинаем с уровня root bus
             # | +-08.2
             # | +-08.3

             При обработке последней root bus OFFSET декрементируется.
              |           -1f.2
              -[0000:00]-+-00.0 <--------- последняя root bus
        """

        tail = s[2:]
        tmp_dict = defaultdict(lambda: defaultdict(dict))
        buf_dict = {}

        # нарезаем триплеты
        sliced_tail = chunks(tail, 3)

        # будем генерить дерево, начиная с листьев
        sliced_tail.reverse()

        for i, g in enumerate(sliced_tail):
            bus, slot, function_ = g
            if len(sliced_tail) > 1 and i != 0:
                tmp_dict.pop(deleted_key)  # noqa: F821
            tmp_dict[f'{bus}:{slot}.{function_}'].update(buf_dict)
            buf_dict = deepcopy(tmp_dict)
            deleted_key = f'{bus}:{slot}.{function_}'  # noqa: F841
        return tmp_dict

    pci_tree_cmd = "lspci -t"
    cmd_ret = get_output(pci_tree_cmd)
    if not cmd_ret['result']:
        comment = f"An error occurred while get node's pci devices: {cmd_ret['comment']}"
        raise CommandExecutionError(comment)
    else:
        try:
            result_tree = defaultdict(lambda: defaultdict(dict))
            previous_line = ''
            nesting = []
            branch_end = []
            parent_branch_end = 0
            OFFSET = 2

            for line in cmd_ret['comment'].splitlines():

                nested_level = parent_branch_end
                parent_branch_end = 0

                if branch_end:
                    if nesting[-1] != '---':
                        nesting.pop()
                        branch_end.pop()
                    else:
                        while nesting[-1] == '---':
                            nesting.pop()
                        nesting.pop()
                        branch_end.pop()
                    while nesting[-1] == '---':
                        nesting.pop()

                # если уровень вложенности сброшен до root bus нужно преобразовать строку из прошлой итерации
                if not len(nesting) - OFFSET:
                    previous_line = previous_line[:2]

                # ищем в строке информацию о новых узлах или последних элементах
                nodes_info = findall(r'---|-\+-|\\', line)

                for elem in nodes_info:
                    if elem != '\\':
                        nesting.append(elem)
                    else:
                        branch_end.append(elem)
                        # у последнего child-элемента есть вложения (см. большой комментарий выше)
                        if nodes_info[-1] != elem:
                            parent_branch_end = 1
                        # если встретили последнюю root bus - декрементируем сдвиг
                        if nodes_info[0] == elem and '|' not in line:
                            OFFSET -= 1

                nested_level += len(nesting) - OFFSET - parent_branch_end

                # -\w\w\] избавляемся от -0a] в 1c.0-[05-0a], вторая шина никогда не фигурирует в выводе lspci -m
                # \[\w\w\]--$ избавляемся от шин без слотов [07]-- в +-00.0-[07]--
                # \[\w\w\-\w\w\]--$ избавляемся от range шин без слотов [04-34]--, например, в +-1c.4-[04-34]--
                # [-+\[\]:\\\.\ ] все эти символы тоже удаляем
                line = rsplit(r'-\w\w\]|\[\w\w\]--$|\[\w\w\-\w\w\]--$|[-+\[\]:\\\.\ ]', line)
                line = [elem.strip() for elem in line if elem.strip()]

                if len(line[0]) >= 4:
                    domain = line.pop(0)
                    host_bus = line.pop(0)

                if len(line) == 2:
                    slot, function_ = line
                    result_tree[domain][f'{host_bus}:{slot}.{function_}'] = {}

                if len(line) >= 5 and line[0] != '|':
                    parent_slot, parent_function = line[:2]
                    tmp_dict = parse_nested(deepcopy(line))
                    result_tree[domain][f'{host_bus}:{parent_slot}.{parent_function}'].update(tmp_dict)

                if line[0] == '|':
                    line_clone = deepcopy(line)
                    while line_clone[0] == '|':
                        line_clone.pop(0)
                    prev_line_clone = deepcopy(previous_line)

                    # восстанавливаем иерархию для случая, когда в строке нет родительских элементов
                    # ['1c', '0', '05', '00', '0', '06', '00', '0'] - previous_line
                    # ['|', '00', '1'] --> ['1c', '0', '05', '00', '0', '06', '00', '1']

                    # возможно в прошлой итерации произошло уменьшение вложенности - отбрасываем лишние триплеты
                    prev_line_clone = prev_line_clone[: 2 + 3 * nested_level]

                    prev_line_clone[-2], prev_line_clone[-1] = line_clone[:2]
                    prev_line_clone = prev_line_clone + line_clone[2:]

                    # кейс для нескольких root bus на 1 domain
                    # -+-[0000: ff]-+-08.0
                    # | +-08.2
                    # | +-08.3
                    # +-[0000: 7f] -+-08.0
                    # | +-08.0
                    if len(prev_line_clone) == 2:
                        slot, function_ = line_clone
                        result_tree[domain]['{}:{}.{}'.format(host_bus, slot, function_)] = {}
                        try:
                            tmp_dict = {f'{previous_line[0]}:{slot}.{function_}': {}}
                            result_tree[domain][f'{host_bus}:{parent_slot}.{parent_function}'].update(tmp_dict)
                        except (ValueError, TypeError, NameError):
                            pass
                    if len(prev_line_clone) >= 5:
                        parent_slot, parent_function = prev_line_clone[:2]

                        tmp_dict = parse_nested(prev_line_clone)

                        result_tree_clone = deepcopy(result_tree)
                        result_tree_clone[domain][f'{host_bus}:{parent_slot}.{parent_function}'].update(tmp_dict)

                        result_tree = merge_dicts(result_tree, result_tree_clone)

                # строки про вложенные элементы строками не считаются
                previous_line = prev_line_clone if line[0] == '|' else line

            result_tree = json.loads(json.dumps(result_tree))  # избавляемся от defaultdict
            return result_tree
        except Exception as e:
            comment = f"An error occurred while get node's pci devices tree: {e}"
            raise CommandExecutionError(comment) from None


@states_response
def get_unknown_domains(domains_ids):
    """
    Возвращает xml доменов, неизвестных на узле.
    """
    available_statuses = ['running', 'paused', 'shutdown', 'crashed']
    result = {}
    for status in available_statuses:
        result[status] = {}
    try:
        domains_cmd_ret = get_output("virsh list --all")
        if not domains_cmd_ret['result']:
            raise CommandExecutionError(domains_cmd_ret['comment'])
        domains = domains_cmd_ret['comment'].splitlines()
        active_unknown_domains = list()
        if len(domains) > 3:
            for line in domains[2:-1]:
                try:
                    _, dom_name, status = line.split()
                except ValueError:
                    continue
                if dom_name not in domains_ids:
                    active_unknown_domains.append(dom_name)
                    xml = virt_get_xml(dom_name)
                    if status in available_statuses:
                        result[status][dom_name] = xml
                    elif status == 'pmsuspended':
                        result['paused'][dom_name] = xml
        for shutoff_unknown_domain in get_shutoff_domains(domains_ids + active_unknown_domains):
            result['shutdown'][shutoff_unknown_domain['dom_id']] = shutoff_unknown_domain['dom_xml']
        return result
    except Exception as e:
        comment = f"An error occurred while get node's unknown domains: {e}"
        raise CommandExecutionError(comment) from None


@states_response
def set_ntp_servers(ntp_servers, run=True):
    local_clock = ' - 127.127.1.0\n'
    localhost = ' - 127.0.0.1\n'
    conf_template = "ntp::servers:\n{}ntp::restrict:\n{}"
    ntps = ''
    for ntp_server in ntp_servers:
        if ntp_server:
            ntps += ' - ' + ntp_server + '\n'

    ntps += local_clock

    conf = conf_template.format(ntps, localhost)
    try:
        with open(NTP_PATH, 'w') as config:
            config.write(conf)
        if run:
            cmd_ret = get_output('/usr/bin/veil-sendrole ntp')
            if not cmd_ret['result']:
                comment = f"Unable to apply ntp config: {cmd_ret['comment']}"
                raise CommandExecutionError(comment)
    except Exception as err:
        comment = f"Unable to write NTP configuration: {err}"
        raise CommandExecutionError(comment) from None
    return 'NTP configuration successfully saved.'


@states_response
def shutdown_domains(timeout):
    """
    Выключает все ВМ узла. После таймаута, если они не выключены, вырубает их жестко
    """
    try:
        domains_ids = list_domains()
        if domains_ids:
            thread_pool(virt_shutdown_by_id, domains_ids)
        else:
            return 'No active domains were found.'

        start_time = time.time()
        while True:
            domains_ids = list_domains()
            if domains_ids:
                if time.time() - start_time > timeout:
                    thread_pool(virt_destroy_by_id, domains_ids)
                time.sleep(2)
            else:
                break
        return 'All domains were shut down.'
    except Exception as e:
        comment = f"An error occurred while shutdown node's domains: {str(e)}"
        raise CommandExecutionError(comment) from None


def get_current_role() -> str:
    role = 'stale'
    cmd = f'cat {VEIL_ROLE_FILE} | grep -v "^#"'
    cmd_ret = get_output(cmd, timeout=3)
    if not cmd_ret['result']:
        return role
    roles = cmd_ret['comment'].replace('\n', ';')[:-1]
    install_status = dict(item.split('=') for item in roles.split(';'))
    if install_status['controller'] == '1' and install_status['node'] == '1':
        role = 'controller_node'
    elif install_status['controller'] == '1' and install_status['node'] == '0':
        role = 'controller'
    elif install_status['controller'] == '0' and install_status['node'] == '1':
        role = 'node'
    return role


def get_current_role_2() -> tuple:
    ctr, node = False, False
    entry_regex = r'^\s*(controller|node)=(0|1).*$'
    with open(VEIL_ROLE_FILE, 'r') as conf:
        for line in conf.readlines():
            kvpair = re.findall(entry_regex, line)
            if not len(kvpair) == 1:
                continue
            key, val = kvpair.pop()
            if key == 'controller' and int(val):
                ctr = True
            if key == 'node' and int(val):
                node = True
    return ctr, node


def install_node_complete() -> bool:
    try:
        return INSTALL_NODE_PASS in os.listdir(INSTALL_TESTS_PATH) and 'node' in get_current_role()
    except OSError:
        pass
    return False


DISTRIBUTION = distribution()
BASE_PACKAGE = f"{DISTRIBUTION}-utils"
CLI_PACKAGES = [f"{DISTRIBUTION}-cli-app", f"{DISTRIBUTION}-cli-env"]
NODE_BASE_PACKAGES = [f"{DISTRIBUTION}-node-app", f"{DISTRIBUTION}-node-env", f"{DISTRIBUTION}-node-ui"]
CONTROLLER_BASE_PACKAGES = [f"{DISTRIBUTION}-controller-app", f"{DISTRIBUTION}-controller-env",
                            f"{DISTRIBUTION}-controller-ui", f"{DISTRIBUTION}-docs"]
NODE_PACKAGES = NODE_BASE_PACKAGES + CLI_PACKAGES
CONTROLLER_PACKAGES = CONTROLLER_BASE_PACKAGES + CLI_PACKAGES
CONTROLLER_NODE_PACKAGES = CONTROLLER_BASE_PACKAGES + NODE_PACKAGES
APT_PATTERN = "apt-cache policy {} | egrep '{}' | awk '{{print $2}}'"


@states_response
def get_base_version():
    oem = get_oem()
    return {'version': oem.get('version', 'unknown')}


@states_response
def get_versions_v2(vtype: str = 'Installed|Установлен', get_role_env: bool = False, **kwargs):
    """
    vtype 'Installed' or 'Candidate'
    """
    result = {}
    role = get_current_role()
    if role == 'node':
        packages = NODE_PACKAGES
    elif role == 'controller':
        packages = CONTROLLER_PACKAGES
    elif role == 'controller_node':
        packages = CONTROLLER_NODE_PACKAGES
    else:
        raise NotImplementedError

    for package in packages:
        cmd = APT_PATTERN.format(package, vtype)
        cmd_ret = get_output(cmd, **kwargs)
        if not cmd_ret['result']:
            continue
        result[package] = dict(version='-'.join(cmd_ret['comment'].strip().split('-')[:-1]))

    cmd = APT_PATTERN.format(BASE_PACKAGE, vtype)
    cmd_ret = get_output(cmd, **kwargs)
    if not cmd_ret['result']:
        raise CommandExecutionError(cmd_ret['comment'])
    version = '-'.join(cmd_ret['comment'].strip().split('-')[:-1])
    result[BASE_PACKAGE] = dict(version=version)
    # TODO Костыль для акрониса
    result['base'] = result['veil-base'] = dict(version=version.split('-')[0])
    if 'controller' in role:
        if DISTRIBUTION == 'space':
            if f"{DISTRIBUTION}-controller-app" in result:
                result["veil-controller-app"] = result[f"{DISTRIBUTION}-controller-app"]
    if get_role_env:
        result["role"] = role
        oem = get_oem()
        result["env"] = oem.get('tag', 'unknown')
    return result


def shutdown_node(minutes=1, reboot=False, halt=False, **kwargs):
    base_cmd = "shutdown "
    if reboot:
        base_cmd += "-r"
    elif halt:
        base_cmd += "-H"
    else:
        base_cmd += "-h"
    cmd = f'{base_cmd} -h +{minutes}'
    cmd_ret = get_output(cmd, **kwargs)
    if not cmd_ret['result']:
        raise CommandExecutionError(cmd_ret['comment'])
    return f'Shutdown in {minutes} minute(s), result cmd is ({cmd}).'


@states_response
def heartbeat_update(config, shutdown=False, reboot=False, cluster_config=None, **kwargs):
    if not os.path.exists(NTP_PATH) and not config.get('built_in', True):
        set_ntp_servers([config.get('controller_ip', None)], run=False)
    if os.path.exists(FENCE_PATH):
        os.remove(FENCE_PATH)
    with open(HEARTBEAT_PATH, 'w') as f:
        json.dump(config, f, indent=4, sort_keys=True)
    if cluster_config:
        with open(CLUSTER_PATH, 'w') as f:
            json.dump(cluster_config, f, indent=4, sort_keys=True)
    redis_client = kwargs.get('redis', None)
    if redis_client:
        redis_client.set('node_reset_settings', 1)
    if shutdown or reboot:
        return shutdown_node(reboot=reboot)
    return 'Successfully update node config.'


@states_response
def set_status(status, **kwargs):
    if os.path.exists(FENCE_PATH):
        os.remove(FENCE_PATH)
    config = get_config()
    config['status'] = status
    with open(HEARTBEAT_PATH, 'w') as f:
        json.dump(config, f, indent=4, sort_keys=True)
    redis_client = kwargs.get('redis', None)
    if redis_client:
        redis_client.set('node_reset_settings', 1)
    return f'Successfully set node status to {status}.'


@states_response
def hw_info():
    cmd_ret = get_output('lshw -xml')
    if not cmd_ret['result']:
        raise CommandExecutionError(cmd_ret['comment'])
    output = cmd_ret['comment']
    try:
        d = xmltodict.parse(output)['list']
        d = json.loads(json.dumps(d).replace('@', '').replace('node', 'children'))

        known_str_fields = ['description', 'class', 'handle', 'vendor', 'product', 'serial']

        def check_fields(hw_dict: dict, new_hw: dict):
            tmp_hw = dict()
            for k, v in hw_dict.items():
                if k == 'children':
                    tmp_hw[k] = []
                    if isinstance(v, dict):
                        if 'children' in v:
                            for i in v['children']:
                                if not isinstance(i, dict):
                                    continue
                                h = check_fields(i, new_hw)
                                tmp_hw[k].append(h)
                        else:
                            h = check_fields(v, new_hw)
                            tmp_hw[k].append(h)
                    elif isinstance(v, list):
                        for i in v:
                            h = check_fields(i, new_hw)
                            tmp_hw[k].append(h)

                elif k in known_str_fields:
                    tmp_hw[k] = v
            return tmp_hw

        node_hw = check_fields(d, {})
    except Exception:
        node_hw = dict()
    return node_hw


@states_response
def ipmi_check(full=False, **kwargs):
    if not [dev for dev in ['/dev/ipmi0', '/dev/ipmi/0', '/dev/ipmidev/0']
            if os.path.exists(dev) and stat.S_ISCHR(os.lstat(dev)[stat.ST_MODE])]:
        if full:
            return 'No local IPMI devices found.'
        else:
            raise CommandExecutionError('No local IPMI devices found.')

    prefixes = ['IP Address', 'Subnet Mask', 'MAC Address', 'Default Gateway', '802.1q']
    cmd = "ipmitool lan print"
    cmd_ret = get_output(cmd, **kwargs)

    if cmd_ret['result']:
        if full:
            return '\n'.join([line for line in cmd_ret['comment'].splitlines()
                              if [xx for xx in prefixes if line.startswith(xx)]])
        else:
            ret_val = ''.join([line.split()[-1] for line in cmd_ret['comment'].splitlines()
                               if line.startswith('IP Address    ')])
            if ret_val:
                return ret_val
            else:
                raise CommandExecutionError('No valid IP address returned by IPMI device.')

    else:
        cmd = "bmc-config --checkout -S Lan_conf"
        cmd_ret = get_output(cmd, **kwargs)
        if cmd_ret['result']:
            if full:
                return '\n'.join([line.strip() for line in cmd_ret['comment'].splitlines()
                                  if not line.strip().startswith('#') and "Section" not in line.strip()])
            else:
                ret_val = ''.join([line.split()[-1] for line in cmd_ret['comment'].splitlines()
                                   if line.strip().startswith('IP_Address    ')])
                if ret_val:
                    return ret_val
                else:
                    raise CommandExecutionError('No valid IP address returned by IPMI device.')
        else:
            raise CommandExecutionError(cmd_ret['comment'])


@states_response
def ksm_stat(**kwargs):
    cmd_ret = get_output("grep . /sys/kernel/mm/ksm/*", **kwargs)
    if not cmd_ret['result']:
        raise CommandExecutionError(cmd_ret['comment'])
    return cmd_ret['comment']


@states_response
def ksm_update(ksm_sleep_time, ksm_pages_to_scan, ksm_merge_across_nodes, ksm_enable):
    cmd = f"echo {ksm_sleep_time} > '/sys/kernel/mm/ksm/sleep_millisecs'"
    cmd_ret = get_output(cmd)
    if not cmd_ret['result']:
        raise CommandExecutionError(cmd_ret['comment'])

    cmd = f"echo {ksm_pages_to_scan} > '/sys/kernel/mm/ksm/pages_to_scan'"
    cmd_ret = get_output(cmd)
    if not cmd_ret['result']:
        raise CommandExecutionError(cmd_ret['comment'])

    cmd = f"echo {ksm_merge_across_nodes} > '/sys/kernel/mm/ksm/merge_across_nodes'"
    cmd_ret = get_output(cmd)
    if not cmd_ret['result']:
        raise CommandExecutionError(cmd_ret['comment'])

    cmd = f"echo {ksm_enable} > '/sys/kernel/mm/ksm/run'"
    cmd_ret = get_output(cmd, timeout=60)
    if not cmd_ret['result']:
        raise CommandExecutionError(cmd_ret['comment'])
    return 'Successfully update ksm params.'


@states_response
def upgrade_check():
    cmd = f"{BASE_SCRIPT_PATH}/veil-upgrade -t -c"
    cmd_ret = get_output(cmd, timeout=55)
    if not cmd_ret['result']:
        raise CommandExecutionError(cmd_ret['comment'])
    return cmd_ret['comment']


@states_response
def ping(hostname, timeout):
    cmd = f'ping -c {timeout} {hostname}'
    cmd_ret = get_output(cmd, timeout=timeout + 1)
    if not cmd_ret['result']:
        raise CommandExecutionError(cmd_ret['comment'])
    return True


@states_response
def ping_extended(source, destination):
    """
    Функция пингует указанный адрес с указанного интерфейса
    """

    cmd = f'ping {destination} -I {source} -c 1 -W 1'
    cmd_ret = get_output(cmd)
    if not cmd_ret['result']:
        raise CommandExecutionError(cmd_ret['comment'])
    return True


@states_response
def node_storage_pools(flags=0):
    res = {}
    for device in list_all_storage_pools(flags=flags):
        xml = xmltodict.parse(device.XMLDesc())
        xml = json.loads(json.dumps(xml).replace('@', ''))
        res[device.name()] = xml
    return OrderedDict(sorted(res.items()))


@states_response
def node_storage_vols(flags=0):
    res = []
    for device in list_all_vols(flags=flags):
        xml = xmltodict.parse(device.XMLDesc())
        xml = json.loads(json.dumps(xml).replace('@', ''))
        res.append(xml)
    return res


@states_response
def node_storage_pool_create(name, xml):
    tmp_file_path = f'/tmp/{uuid.uuid4()}.xml'
    with open(tmp_file_path, 'w+') as f:
        f.write(xml)
    cmd_ret = states_result(get_output, cmd=f"virsh pool-define {tmp_file_path}")
    ret = cmd_ret.strip()
    os.remove(tmp_file_path)
    cmd_ret = states_result(get_output, cmd=f"virsh pool-start {name}")
    ret += f". {cmd_ret.strip()}"
    cmd_ret = states_result(get_output, cmd=f"virsh pool-autostart {name}")
    ret += f". {cmd_ret.strip()}"
    return ret


@states_response
def node_storage_pool_destroy(name):
    cmd_ret = states_result(get_output, cmd=f"virsh pool-destroy {name}")
    ret = cmd_ret.strip()
    cmd_ret = states_result(get_output, cmd=f"virsh pool-undefine {name}")
    ret += f". {cmd_ret.strip()}"
    return ret


@states_response
def node_devices(flags=0):
    res = []
    for device in list_all_devices(flags=flags):
        xml = xmltodict.parse(device.XMLDesc())
        xml = json.loads(json.dumps(xml).replace('@', ''))['device']
        if flags in [512, 1024]:
            name = device.name()
            path = name.split('_')
            if len(path) > 1:
                path = path[-1]
                if os.path.exists(f'/sys/class/fc_host/{path}/port_state'):
                    with open(f'/sys/class/fc_host/{path}/port_state') as f:
                        port_state = f.read().strip()
                        if port_state:
                            xml['port_state'] = port_state
                if os.path.exists(f'/sys/class/fc_host/{path}/speed'):
                    with open(f'/sys/class/fc_host/{path}/speed') as f:
                        speed = f.read().strip()
                        if speed:
                            xml['speed'] = speed
        res.append(xml)
    return res


@states_response
def node_device_create(xml, flags=0):
    return create_node_device(xml, flags)


def parse_lscpu(**kwargs):
    cmd_ret = json.loads(states_result(get_output, cmd="LC_ALL=en-US lscpu --json"))['lscpu']
    data = {}
    for item in cmd_ret:
        data[item['field']] = item['data']
    import psutil
    nodeinfo = {'cpu_cores': int(data['Core(s) per socket:']),
                'cpu_mhz': int(data['CPU MHz:']),
                'cpu_architecture': data['Architecture:'],
                'cpu_count': int(data['CPU(s):']),
                'cpu_threads': int(data['Thread(s) per core:']),
                'cpu_numa_nodes': int(data['NUMA node(s):']),
                'memory_count': psutil.virtual_memory().total / 1024 / 1024,
                'cpu_sockets': int(data['Socket(s):'])}
    baseline = {'model': data['Model:'],
                'vendor': data['Vendor ID:'],
                'features': data.get('Flags:', [])}
    return nodeinfo, baseline


@states_response
def hardware_info(lshw_info=True, **kwargs):
    """
    hw_info:
      module.run:
        - name: virt.node_info

    cpu_features:
      module.run:
        - name: virt.cpu_baseline
        - full: True
        - migratable: False
        - out: salt

    dmidecode_cpu:
      module.run:
        - name: smbios.records
        - rec_type: 4
        - fields: family,id,manufacturer,version,max_speed,status,thread_count

    cpu_topology:
      bazalt.get_cpu_topology

    lshw:
      cmd.run:
        - name: lshw -json
    """

    result = dict()
    try:
        result['nodeinfo'] = node_info()
        result['cpu_baseline'] = cpu_baseline(full=True, migratable=False, out='salt')
    except Exception:
        result['nodeinfo'], result['cpu_baseline'] = parse_lscpu()
    result['cpu_topology'] = get_cpu_topology()
    result['dmidecode_cpu'] = records(4, ['family', 'id', 'manufacturer', 'version',
                                          'max_speed', 'status', 'thread_count'])
    if lshw_info:
        lshw_ret = hw_info()
        if not lshw_ret['result']:
            raise CommandExecutionError(lshw_ret['comment'])
        result['lshw'] = lshw_ret['comment']
    return result


@states_response
def get_smart(brief=False, **kwargs):
    """
    если короткий вариант:
        [ {drive_description: status} ... ]
        status: 'ok', 'fail', 'na'
    иначе
    {
     drive_description:         # название, тип устройства, тип подключения по версии smartctl --scan
        {
        "DRIVE_NAME": name,     # м.б. не уникально, если за raid контроллером! напр. /dev/bus/0
        "CONN_TYPE": type,
        "IDENTIFY": [ lines ]   # секция параметров диска - пр-ль, модель, s/n, состояние смарт вкл-выкл и т.д.
        "SMART": [ lines ]      # секция данных SMART или [], если недоступно для этого устройства
        "STATUS": [ ... ],      # "OK" или сочетание алертов:
          (красный уровень)
            SYS_ERR - ошибка ком.строки smartctl, идентификации устройства, устройство в low-power
            FAIL - всё плохо, диск менять ASAP
            PREFAIL_WARN - атрибуты, критичные для состояния диска, ниже порога (значения атрибутов меньше - хуже)
          (жёлтый уровень)
            PROBLEMS_IN_THE_PAST - диск теперь OK, но какие-то атрибуты, не обяз.критичные, в прошлом падали ниже порога
            ERROR_LOG_NOT_EMPTY - диск в процессе работы выдавал ошибки, сохранил их в журнал.
            ERRS_IN_SELFTEST_LOG - самотестирование диска выявило ошибки
          (серый уровень)
            disabled - оч странно, потому что мы включаем смарт при каждом запросе
            unavailable - получаем только блок IDENTIFY по идее в таком случае
        }
     ...
    }
    """

    # TODO: 1) выбор устройства, расшир.вывод (-x), запуск в зависимости от standby/sleep, 2) запуск тестов SMART

    def check_bits(code) -> list:
        """

        :param code: код возврата smartctl
        :return:
            fail - всё плохо, диск менять ASAP
            prefail warn - атрибуты, критичные для состояния диска, ниже порога (значения атрибутов меньше - хуже)
            problems in the past - диск теперь OK, но какие-то атрибуты, не обяз.критичные, в прошлом падали ниже порога
            error log not empty - диск в процессе работы выдавал ошибки, сохранил их в журнал. смотрится с параметром -x
            errs in selftest log - самотестирование диска выявило ошибки
            если нет алертов, то ОК
        """

        ret = []

        alerts = ['FAIL', 'PREFAIL_WARN', 'PROBLEMS_IN_THE_PAST', 'ERROR_LOG_NOT_EMPTY', 'ERRS_IN_SELFTEST_LOG']
        if code & 0b00000011:
            # cmdline error, drive open err, drive identify err, drive on low-power
            return ['SYS_ERR']
        # we ignore 3 right bits as bit 2 is not important
        for index, _ in enumerate(alerts):
            if code & (0b00001000 << index):
                ret.append([alerts[index]])

        return ret

    def smart_process(input):  # pylint: disable=redefined-builtin)
        """

        :param input: выхлоп stdout smartctl
        :return: {
                    "IDENTIFY": [identify section output],
                    "SMART":    [smart data section output],
                    "STATUS":   if not empty then   {"UNAVAILABLE": 1} -- smart недоступен для устройства
                                                    {"DISABLED": 1}   -- smart отключен
                }
        """
        out = {"IDENTIFY": [], "SMART": [], "STATUS": []}
        phase = None
        for line in input.splitlines():

            if line == '=== START OF INFORMATION SECTION ===':
                phase = 'IDENTIFY'

            elif line == '=== START OF READ SMART DATA SECTION ===':
                phase = 'SMART'

            elif line:
                try:
                    out[phase].append(line)
                except KeyError:
                    pass

            if phase == 'IDENTIFY' and line.startswith('SMART support is:'):
                if 'Disabled' in line:
                    out['STATUS'].append("DISABLED")
                elif 'Unavailable' in line:
                    out['STATUS'].append('UNAVAIL')
        return out

    def brief_status(status):
        if ('UNAVAIL' in status) or ('SYS_ERR' in status) or ('DISABLED' in status):
            return 'na'
        elif ('FAIL' in status) or ('PREFAIL_WARN' in status):
            return 'fail'
        else:
            return 'ok'

    out_dict = dict()
    from .lvm import _map_blkdev_id
    dict_id = _map_blkdev_id()

    drives = get_output('/usr/sbin/smartctl --scan', **kwargs)

    # when --scan return code must be zero
    if not drives['result']:
        raise CommandExecutionError('An error occurred while getting devices list. Error: {}'.format(drives['comment']))

    for line in drives['comment'].splitlines():

        if not line:
            continue
        drv_and_type, descr = line.split('#')
        # when no available drives (as in Vagrant) string starts with '#'
        if drv_and_type:
            # descr has format like "/dev/sda, SCSI device"
            # we find device ID and change descr into '/dev/disk/by-id/DRV_ID (/dev/sda, SCSI device)'
            dev_phys_name = descr.split(', ')[0].strip() if ',' in descr else descr
            descr = f'{dict_id[dev_phys_name]} ({descr})' if dict_id.get(dev_phys_name) else descr

            drive_name, drive_type = drv_and_type.split(' -d ')
            out_dict[descr] = {"DRIVE_NAME": drive_name.strip(), "CONN_TYPE": drive_type.strip()}

    for drive in out_dict:
        retsmart, outsmart = getstatusoutput('/usr/sbin/smartctl -d {} -s on -a {}'
                                             .format(out_dict[drive]['CONN_TYPE'], out_dict[drive]['DRIVE_NAME']))

        if 'SYS_ERR' not in check_bits(retsmart):
            out_dict[drive].update(smart_process(outsmart))
            out_dict[drive]['STATUS'] += check_bits(retsmart)
        else:
            out_dict[drive]['STATUS'] = ['SYS_ERR']

        if out_dict[drive]['CONN_TYPE'] == 'scsi':
            # ata drives are detected as scsi too but are not processed via -d scsi
            # so we process such drives via -d scsi first and then via -d ata and comparing results just what's longer
            ret_ata_smart, out_ata_smart = getstatusoutput('/usr/sbin/smartctl -d ata -s on -a {}'
                                                           .format(out_dict[drive]['DRIVE_NAME']))

            if 'SYS_ERR' not in check_bits(ret_ata_smart) and len(out_ata_smart) > len(outsmart):
                out_dict[drive].update(smart_process(out_ata_smart))
                out_dict[drive]['STATUS'] += check_bits(ret_ata_smart)

        if not out_dict[drive]['STATUS']:
            out_dict[drive]['STATUS'] = ['OK']

    if brief:
        return [{k: brief_status(out_dict[k]['STATUS'])} for k in out_dict]

    return out_dict


@states_response
def sensors(brief=True, **kwargs):
    cmd = 'ipmi-sensors --no-sensor-type-output --no-header-output'
    cmd_ret = get_output(cmd, **kwargs)
    if not cmd_ret['result']:
        get_result(['ipmi-sensors', '-f'], shell=False)
        cmd_ret = get_output(cmd, **kwargs)
        if not cmd_ret['result']:
            raise CommandExecutionError(cmd_ret['comment'])
    res = cmd_ret['comment']
    if brief:
        return [{x.split('|', 5)[1].strip(): x.split('|', 5)[4].strip().replace("'", "")}
                for x in res.splitlines() if x and (len(x.split('|', 5)) > 4)]
    else:
        return [{x.split('|', 5)[1].strip(): list(map(str.strip, x.split('|', 2)[2].split('|')))}
                for x in res.splitlines() if x and (len(x.split('|', 5)) > 2)]


@states_response
def lm_sensors(**kwargs):
    res = states_result(get_output, cmd='sensors -j', **kwargs)
    try:
        return json.loads(res)
    except (json.JSONDecodeError, TypeError) as e:
        raise CommandExecutionError(f'lm_sensors err, e={e}') from None


@states_response
def virt_what(**kwargs):
    if os.path.exists('/usr/sbin/virt-what'):
        cmd_ret = get_output('virt-what', **kwargs)
        if not cmd_ret['result']:
            raise CommandExecutionError(cmd_ret['comment'])
        return cmd_ret['comment']
    else:
        return 'kvm'


@states_response
def sysctl_settings():
    result = dict(swappiness=10, vfs_cache_pressure=100, nr_hugepages=0, min_free_kbytes=0)
    for key in ('swappiness', 'vfs_cache_pressure', 'nr_hugepages', 'min_free_kbytes'):
        if os.path.exists(f"/proc/sys/vm/{key}"):
            with open(f"/proc/sys/vm/{key}", 'r') as f:
                result[key] = f.readline().strip()
    return result


@states_response
def update_sysctl_settings(swappiness=None, vfs_cache_pressure=None, nr_hugepages=None, min_free_kbytes=None):
    base_conf = "sysctl::params:\n"
    if any([swappiness is not None, vfs_cache_pressure is not None, nr_hugepages is not None, min_free_kbytes is not None]):
        if swappiness is not None:
            base_conf += f"  'vm.swappiness': '{swappiness}'\n"
        if vfs_cache_pressure:
            base_conf += f"  'vm.vfs_cache_pressure': '{vfs_cache_pressure}'\n"
        if nr_hugepages:
            base_conf += f"  'vm.nr_hugepages': '{nr_hugepages}'\n"
        if min_free_kbytes:
            base_conf += f"  'vm.min_free_kbytes': '{min_free_kbytes}'\n"
        try:
            with open(SYSCTL_PATH, 'w') as config:
                config.write(base_conf)
            cmd_ret = get_output(f'{ROLE_SCRIPT_PATH} -a sysctl')
            if not cmd_ret['result']:
                comment = f"Unable to apply sysctl config: {cmd_ret['comment']}"
                raise CommandExecutionError(comment)
        except Exception as err:
            comment = f'Unable to write sysctl configuration: {err}'
            raise CommandExecutionError(comment) from None
    else:
        comment = 'No params for update sysctl'
        raise CommandExecutionError(comment)
    return 'sysctl configuration successfully saved.'


@states_response
def grub_settings():
    with open(GRUB_FILE, 'r') as f:
        for line in f.readlines():
            if line.startswith('GRUB_CMDLINE_LINUX_DEFAULT'):
                _, value = line.split('=', maxsplit=1)
                return value
    raise CommandExecutionError('GRUB config was not found.')


@states_response
def get_render_devices():
    # return:
    # [ {'pci_id': '0000:01:00.0',
    #    'description': 'Advanced Micro Devices, Inc. [AMD/ATI] Curacao PRO [Radeon R7 370 / R9 270/370 OEM] (rev 81)'}
    #  ...
    #  ]
    ret = []
    try:
        # /dev/dri/by-path/pci-0000:01:00.0-render
        for dev in [x.split('-')[1] for x in os.listdir('/dev/dri/by-path') if x.endswith('-render')]:
            devname = get_output(f'lspci -D -s {dev}').get('comment').strip().split(None, 1)[1]
            ret.append(dict(pci_id=dev, description=devname.split(': ')[1] if ': ' in devname else devname))
    except Exception as e:
        raise CommandExecutionError('Unable to get render devices list, e={}'.format(e)) from None
    return ret


@states_response
def set_iommu(iommu_value, pcie_acs_override='downstream'):
    try:
        puppet_modify_param(VEIL_NODE_LOCAL_PATH, 'veil_node', 'grub::iommu_mode', iommu_value)
        puppet_modify_param(VEIL_NODE_LOCAL_PATH, 'veil_node', 'grub::pcie_acs_override', pcie_acs_override)
        cmd_ret = get_output(f'{ROLE_SCRIPT_PATH} -a veil_node', timeout=120)
        if not cmd_ret['result']:
            comment = f"Unable to apply node config: {cmd_ret['comment']}"
            raise CommandExecutionError(comment)
    except Exception as err:
        comment = f'Unable to write node configuration: {err}'
        raise CommandExecutionError(comment) from None

    return "Successfully changed iommu config and grub. Reboot server for applying changes."


@states_response
def check_iommu_status() -> bool:
    """
    Функция проверяет включена ли поддержка IOMMU.
    Проверки идентичны выполняемым командой virt-host-validate
    """
    return bool(os.path.isdir(iom := '/sys/kernel/iommu_groups') and len(os.listdir(iom))
                and os.path.isdir(acpi := '/sys/firmware/acpi/tables') and 'DMAR' in os.listdir(acpi))


@states_response
def remove_modules():
    cmd_ret = get_output('modprobe -r')
    if not cmd_ret['result']:
        comment = f"Unable to remove modules: {cmd_ret['comment']}"
        raise CommandExecutionError(comment)
    return cmd_ret['comment']


@states_response
def load_module(module, reload=True):
    cmd_ret = get_output(f'modprobe {module}')
    if not cmd_ret['result']:
        comment = f"Unable to load module {module}: {cmd_ret['comment']}"
        raise CommandExecutionError(comment)
    return cmd_ret['comment']


@states_response
def rm_module(module):
    cmd_ret = get_output(f'rmmod {module}')
    if not cmd_ret['result']:
        comment = f"Unable to remove module {module}: {cmd_ret['comment']}"
        raise CommandExecutionError(comment)
    return cmd_ret['comment']


@states_response
def list_watchdog_modules():
    cmd_ret = get_output('lsmod | grep -e wdt -e dog')
    if not cmd_ret['result']:
        comment = f"Unable to list modules: {cmd_ret['comment']}"
        raise CommandExecutionError(comment)
    return cmd_ret['comment']


@states_response
def check_watchdog():
    return os.path.exists('/dev/watchdog')


@states_response
def list_watchdogs():
    # pylint: disable=unsubscriptable-object
    cmd_ret = check_watchdog()
    if not cmd_ret['result']:
        raise CommandExecutionError(cmd_ret['comment'])
    if not cmd_ret['comment']:
        return ''
    cmd_ret = get_output('ls -l /dev/watchdog*')
    if not cmd_ret['result']:
        comment = f"Unable to list modules: {cmd_ret['comment']}"
        raise CommandExecutionError(comment)
    return cmd_ret['comment']


@states_response
def bmc_watchdog_action(action='get'):
    cmd_ret = get_output(f'ipmitool bmc watchdog {action}')
    if not cmd_ret['result']:
        comment = f"Unable to run bmc action: {cmd_ret['comment']}"
        raise CommandExecutionError(comment)
    return cmd_ret['comment']


@states_response
def bmc_watchdog_set(action: str, timer: int):
    cmd_ret = get_output(f' bmc-watchdog --set --timer-use=1 --timeout-action={BMC_ACTIONS[action]} '
                         f'--pre-timeout-interval=0 --clear-bios-frb2 --initial-countdown={timer}')
    if not cmd_ret['result']:
        comment = f"Unable to set bmc watchdog: {cmd_ret['comment']}"
        raise CommandExecutionError(comment)
    return cmd_ret['comment']


@states_response
def profile_create(profile, filename, path, owner):
    """
    Функция сохранения профиля узла в формате JSON в указанный датапул.
    """

    profile_path = f'{path}/{filename}'
    with open(profile_path, 'w') as f:
        json.dump(profile, f)
    try:
        chown_directory(profile_path, owner)
    except Exception as e:
        raise CommandExecutionError(f'An error occurred while changing owner of {profile_path}: {e}') from None
    return {'path': profile_path,
            'size': os.path.getsize(profile_path)}


@states_response
def profile_read(profile_path):
    """
    Функция чтения профиля.
    """

    try:
        with open(profile_path) as f:
            return json.load(f)
    except Exception as e:
        comment = 'Unable to read profile: {err}'.format(err=str(e))
        raise CommandExecutionError(comment) from None


@states_response
def autotest(**kwargs):
    cmd_ret = get_output(f"{ROLE_SCRIPT_PATH} -a full", **kwargs)
    if not cmd_ret['result']:
        comment = "Server auto test was failed. See logs in cli by command 'log autotest'"
        raise CommandExecutionError(comment)
    return "Server auto test was successfully completed."


@states_response
def grid_get_parm(verbose=True):
    """
    Функция возвращает словарь параметров шедулинга Nvidia GRID из /etc/modprobe.d/nvidia.conf
    Если задан один общий параметр на все GPU, то {"GLOB": value}
    Если заданы индивидуальные параметры для каждого, то {"DDDD:BB:dd.f": value, ... }
    value:
        "default"                   (raw == "0x00")
        "equal"                     (raw == "0x01")
        "equal with T ms timeslice" (raw == "0x00TT0001")
        "fixed"                     (raw == "0x11")
        "fixed with T ms timeslice" (raw == "0x00TT0011")
    """
    def get_sched(string):
        modes = {"00": "default", "01": "equal", "11": "fixed"}
        try:
            if not string.startswith('0x') or len(string) not in [4, 10]:
                return "INVALID"
            mode = modes[string[-2:]]
        except (IndexError, KeyError):
            return 'INVALID'
        if len(string) == 10:
            if string[:4] != '0x00' or string[6:8] != '00':
                return 'INVALID'
            return mode + f" with {int(string[4:6], 16)} ms timeslice"
        else:
            return mode

    raw = ''
    try:
        with open(GRID_CFG, 'r') as f:
            for line in f:
                line = line.split(None, 2)
                if line[:2] == ['options', 'nvidia']:
                    raw = line[2]
    except (IOError, IndexError):
        pass
    if not raw:
        return dict(GLOB='default')
    if raw.startswith('NVreg_RegistryDwords='):
        # когда одно общее значение на все GPU
        raw_value = raw.replace('"', '').split('=', 2)[2].strip()
        return {'GLOB': get_sched(raw_value) if verbose else raw_value}
    elif raw.startswith('NVreg_RegistryDwordsPerDevice='):
        # когда на каждое GPU индивидуальная настройка
        # формат вот такой:
        # "pci=DDDD:BB:dd.f;RmPVMRL=value[;pci=DDDD:BB:dd.f;RmPVMRL=value...]"
        raw_sched_list = raw.split('=', 1)[1].strip(' "').split(';')

        r = {}
        while raw_sched_list:
            if not ('pci=' in raw_sched_list[0] and 'RmPVMRL=' in raw_sched_list[1]):
                # если verbose, то это команда просмотра, там ругаемся. иначе это добавление текущих параметров
                # для объединения с новыми в grid set_parm, тогда просто считаем их дефолтными молча.
                if verbose:
                    raise CommandExecutionError(f'Invalid nvidia.conf content: {raw[:100]}')
                else:
                    return dict(GLOB='default')

            pci = raw_sched_list[0].split('=')[1]
            sched_val = raw_sched_list[1].split('=')[1]
            r[pci] = get_sched(sched_val) if verbose else sched_val
            raw_sched_list = raw_sched_list[2:]

        return r
    else:
        if verbose:
            raise CommandExecutionError(f'Invalid nvidia.conf content: {raw[:100]}')
        else:
            return dict(GLOB='default')


@states_response
def grid_set_parm(content):

    def dict2raw(inp):
        mode = {'default': '00', 'equal': '01', 'fixed': '11'}
        # timeslice не имеет смысла при задании режима Default и ограничено значениями 1-30 мс
        if (raw_mode := mode[inp.get('mode', 'default')]) != '00' and inp.get('slice', 0) != 0:
            return f'0x00{inp["slice"]:02X}00{raw_mode}'
        else:
            return f'0x{raw_mode}'

    def write_cfg(data):
        try:
            with open(GRID_CFG, 'w') as f:
                # truncate file
                f.write('')
                f.write(f'# DO NOT EDIT! This file will be overwritten by <grid set_param> CLI command\n{data}')
        except IOError as e:
            raise CommandExecutionError(f'Error writing nvidia.conf ({e})') from None

    if content.get('slice') is not None and content['slice'] not in range(1, 31):
        raise CommandExecutionError('Invalid timeslice value, must be 1-30')
    try:
        pci = content.pop('pci')
    except KeyError:
        # если не указано pci, то глобальные параметры для всех GPU
        if content['mode'] != 'default':
            write_cfg(f'options nvidia NVreg_RegistryDwords="RmPVMRL={dict2raw(content)}"')
        elif os.path.exists(GRID_CFG):
            # конфиг незачем, если всё по умолчанию
            os.remove(GRID_CFG)
    else:
        if not ((len(pci) == 12) and re.match(r'[a-fA-F0-9]{4}(:[a-fA-F0-9]{2}){2}\.[a-fA-F0-9]{1}', pci)):
            raise CommandExecutionError('Invalid PCI address, must be DDDD:BB:dd.f')
        # получаем то, что было ранее в конфиге
        cur_dict = {k: v for k, v in grid_get_parm(verbose=False)['comment'].items() if v != 'INVALID'}
        if 'GLOB' not in cur_dict:
            cur_dict.update({pci: dict2raw(content)})
        else:
            cur_dict = {pci: dict2raw(content)}
        # создаём итоговую строку. всё, что default (endswith 00) -- выкидываем для простоты
        raw_data = ';'.join([f'pci={pci};RmPVMRL={sched}'
                            for pci, sched in cur_dict.items() if not sched.endswith('00')])
        if raw_data:
            write_cfg(f'options nvidia NVreg_RegistryDwordsPerDevice="{raw_data}"')
        elif os.path.exists(GRID_CFG):
            # конфиг незачем, если всё по умолчанию
            os.remove(GRID_CFG)
    return 'Config updated successfully. Changes will take effect after node reboot.'


@states_response
def nvidia_smi(param):
    retcodes = {2: 'A supplied argument or flag is invalid',
                3: 'The requested operation is not available on target device',
                4: 'The current user does not have permission to access this device or perform this operation',
                6: 'A query to find an object was unsuccessful',
                8: "A device's external power cables are not properly attached",
                9: 'NVIDIA driver is not loaded',
                10: 'NVIDIA Kernel detected an interrupt issue with a GPU',
                12: "NVML Shared Library couldn't be found or loaded",
                13: "Local version of NVML doesn't implement this function",
                14: 'infoROM is corrupted',
                15: 'The GPU has fallen off the bus or has otherwise become inaccessible'}

    cmd_ret = get_output(f'/usr/bin/nvidia-smi {param}', suppress_err=True)
    if not cmd_ret['result']:
        err = cmd_ret.get('retcode', 255)
        if err == 127:
            raise CommandExecutionError("Unable to find nvidia-smi utility. Is NVIDIA driver installed?")
        else:
            raise CommandExecutionError(retcodes.get(err, f'Error calling nvidia-smi. Code: {err} '
                                                          f'Output:{cmd_ret.get("comment")}'))
    return cmd_ret['comment']


@states_response
def grid_sriov_mgr(action, devices):
    PCI_ADDR = r'[A-F0-9]{4}(:[A-F0-9]{2}){2}\.[A-F0-9]{1}'
    out = []
    if not (mgr := [name for name in SRIOV_MGR_LIST if os.path.isfile(name)]):
        raise CommandExecutionError('SR-IOV manager not found. Are Nvidia drivers installed?')
    SRIOV_MGR = mgr[0]

    enabled_devs = set()
    if os.path.isfile(GRID_SRIOV_ENABLE_LIST):
        with open(GRID_SRIOV_ENABLE_LIST, 'r') as f_en_list:
            enabled_devs = {dev for dev in map(str.strip, map(str.upper, f_en_list.readlines()))
                            if re.match(PCI_ADDR, dev) or dev == 'ALL'}

    if 'ALL' in (devices := set(map(str.upper, devices))):
        devices = {'ALL'}
    else:
        if not (devices and all([re.match(PCI_ADDR, pci) for pci in devices])):
            raise CommandExecutionError('All GPU PCI addresses must be in format DDDD:BB:dd.f or one word "ALL"')
    for dev in devices:
        cmd_ret = get_output(f'{SRIOV_MGR} {"-e" if action == "enable" else "-d"} {dev}')
        if not cmd_ret['result']:
            with open(GRID_SRIOV_ENABLE_LIST, 'w') as f_en_list:  # if something has been done successfully then save it
                f_en_list.write('\n'.join(enabled_devs))
            fail_dev = 'all Nvidia devices' if dev == 'ALL' else f'device {dev}'
            raise CommandExecutionError(f'Error during {"en" if action == "-e" else "dis"}abling SR-IOV on {fail_dev}.')
        out.append(cmd_ret['comment'])
        if dev == 'ALL':
            enabled_devs = {'ALL'} if action == 'enable' else set()
        else:
            if action == 'enable':
                if enabled_devs != {'ALL'}:
                    enabled_devs.add(dev)
            else:
                enabled_devs.difference_update({dev})

    with open(GRID_SRIOV_ENABLE_LIST, 'w') as f_en_list:
        f_en_list.write('\n'.join(enabled_devs))
    return '\n'.join(out)


@states_response
def backup_os(filename, datapool, description, user, task, **kwargs):
    import tempfile
    from .copy import popen_with_progress

    BACKUP_SCRIPT_PATH = f"{BASE_SCRIPT_PATH}/veil-backup-os"
    DEFAULT_PATH_TO_SAVE = f"{LOCAL_STORAGES_PATH}/default/{PATH_FILE_STORAGE}"

    progress_data = dict(type='progress', jid=task, id='', name='backup_os', result=True)
    redis = kwargs.get('redis', None)
    ssh_options = kwargs.get('ssh_options')
    if description:
        if ssh_options:
            description = re.escape(description.translate(str.maketrans({"\\": r"\\",
                                                                         "\n": "\\n",
                                                                         "\t": "\\t",
                                                                         "\b": "\\b",
                                                                         "\a": "\\a",
                                                                         "\f": "\\f",
                                                                         "\r": "\\r",
                                                                         "\v": "\\v"}))
                                    ).translate(str.maketrans({"'": r"\'\"\'\"\'",
                                                               '"': r'\"',
                                                               "`": r"\`",
                                                               "%": "%%"}))
            cmd_description = f"--description \\'{description}\\'"
        else:
            description = description.translate(str.maketrans({"\\": r"\\",
                                                               "\n": "\\n",
                                                               "\t": "\\t",
                                                               "\b": "\\b",
                                                               "\a": "\\a",
                                                               "\f": "\\f",
                                                               "\r": "\\r",
                                                               "\v": "\\v",
                                                               "'": "'\"'\"'",
                                                               '%': '%%'}))
            cmd_description = f"--description \'{description}\'"
    else:
        cmd_description = ''
    cmd = f'sudo {BACKUP_SCRIPT_PATH} {cmd_description} --progress --filename "{filename}" '
    if datapool:
        path_to_save = datapool.rstrip('/') + f'/{PATH_FILE_STORAGE}'
    else:
        path_to_save = DEFAULT_PATH_TO_SAVE
    cmd = cmd + '"' + path_to_save + '"'
    path = f'{path_to_save}/{filename}'
    if ssh_options:
        tmp_ssh_key_path = None
        try:
            tmp_fh, tmp_ssh_key_path = tempfile.mkstemp()
            with os.fdopen(tmp_fh, 'w') as tmp_file:
                tmp_file.write(str(ssh_options['private_ssh_key'].encode('utf-8').decode('unicode_escape')))
            ssh_cmd = "ssh -o \"StrictHostKeyChecking no\" -i {} {}@{} ".format(
                tmp_ssh_key_path, ssh_options['user'], ssh_options['destination_ip'])
            # bash: !\': event not found
            cmd = 'set +H; ' + ssh_cmd + cmd
            progress_data.update({'steps': {'total_steps': 2, 'step': 1}})
            p = popen_with_progress(cmd, "pv", progress_data, redis)
            if p.returncode:
                raise CommandExecutionError(
                    f'Failed to create backup os {filename}. Error: {p.stderr.read()}')
            get_size = get_output(ssh_cmd + f'\'set -o pipefail; du -b "{path}" | sed "s/\s.*//"\'')
            if not get_size['result']:
                raise CommandExecutionError(get_size['comment'])
            size = int(get_size['comment'].strip().split('\n')[-1])
            # wait until node will be alive
            progress_data['steps']['step'] += 1
            # 100 is percent, 65 is seconds
            cmd = 'for i in `seq 100`; do if [ $(($i%2)) -eq 0 ]; then echo $i; fi; sleep .65; done'
            popen_with_progress(cmd, "pv", progress_data, redis)
        except Exception as e:
            raise CommandExecutionError(f'Unknown error. Error: {e}') from None
        finally:
            if tmp_ssh_key_path:
                os.remove(tmp_ssh_key_path)
    else:
        cmd = 'set +H; ' + cmd
        p = popen_with_progress(cmd, "pv", progress_data, redis)
        if p.returncode:
            raise CommandExecutionError(
                f'Failed to create backup os {filename}. Error: {p.stdout.read()}\n{p.stderr.read()}')
        try:
            size = os.path.getsize(path)
        except Exception as e:
            raise CommandExecutionError(f'Unknown error. Error: {e}') from None
    return {'path': path, 'size': size}


@states_response
def zerofill_dd(dev, task, sz, **kwargs):
    from .copy import popen_with_progress
    from tempfile import mkstemp
    match = re.match(r'[0-9.]+', sz)
    if match is None:
        raise CommandExecutionError(f'Wrong size format: {sz}')
    multiplier = sz[match.end():]
    if multiplier:
        sz = memory_units_convertor(float(sz[:match.end()]), multiplier, 'b')
    redis = kwargs.get('redis', None)
    fh_stderr, f_stderr = mkstemp(dir=RUN_PATH, text=True)
    cmd = f'(pv -s {sz} -n /dev/zero | dd of={dev} bs=128K 2>{f_stderr}) 2>&1'
    progress_data = dict(type='progress', jid=task, id='', name='zerofill', result=True)
    # flag препятствует показу устройства в lsblk(), пока идёт стирание
    p = popen_with_progress(cmd, "pv", progress_data, redis, flag=_zerofill_flag(dev))
    with os.fdopen(fh_stderr, 'r') as f:
        f.seek(0)
        raw_lines = f.read()
    os.remove(f_stderr)
    lines = raw_lines.splitlines()
    if not lines or (p.returncode and lines[0] != f"dd: error writing '{dev}': No space left on device"):
        raise CommandExecutionError(f'Failed to fill device {dev} with zeroes. Error: {raw_lines}')
    return lines[-1]  # там статистика dd, сколько скопировано и скорость


@states_response
def set_prometheus_config(hours):
    try:
        with open(f'{LOCAL_MODULES_PATH}/prometheus.yaml', 'w+') as f:
            f.write(f"prometheus::server::storage_retention: {hours}h\n")
        cmd_ret = get_output(f'{ROLE_SCRIPT_PATH} -a prometheus', timeout=120)
        if not cmd_ret['result']:
            comment = f"Unable to apply statistics config: {cmd_ret['comment']}"
            raise CommandExecutionError(comment)
    except Exception as err:
        comment = f'Unable to write statistics config: {err}'
        raise CommandExecutionError(comment) from None

    return f"Successfully changed statistics storage retention to {hours}h."


@states_response
def set_loki_config(hours):
    try:
        with open(f'{LOCAL_MODULES_PATH}/loki.yaml', 'w+') as f:
            f.write(f"loki::retention_period: {hours}h\n")
        cmd_ret = get_output(f'{ROLE_SCRIPT_PATH} -a loki', timeout=120)
        if not cmd_ret['result']:
            comment = f"Unable to apply Loki config: {cmd_ret['comment']}"
            raise CommandExecutionError(comment)
    except Exception as err:
        comment = f'Unable to write Loki config: {err}'
        raise CommandExecutionError(comment) from None

    return f"Successfully changed Loki retention_period to {hours}h."


@states_response
def rm_rf(dirs_list: list, guard_path=r'\/storages\/(.+)\/(.+)', **kwargs):
    """
    Удаляет все данные в списке каталогов, не трогая сами каталоги.
    """
    prev_time = time.monotonic()
    for i, directory in enumerate(dirs_list):

        real_dir = os.path.realpath(directory)
        if re.match(guard_path, real_dir) is None:
            raise CommandExecutionError(f'Possibly dangerous attempt to remove all data (path={directory}).')

        dir_list = os.listdir(real_dir)
        for ii, sub_item in enumerate(dir_list):
            try:
                to_delete = os.path.join(real_dir, sub_item)
                if os.path.isdir(to_delete):
                    shutil.rmtree(os.path.join(real_dir, sub_item), ignore_errors=False)
                else:
                    os.remove(to_delete)
            except (shutil.Error, OSError) as e:
                raise CommandExecutionError(f"Failed to remove all data in directory {directory}: {e}") from None

            if kwargs.get('redis') and kwargs.get('task') and time.monotonic() - prev_time > 2:
                prev_time = time.monotonic()
                kwargs['redis'].lpush('node_results', json.dumps(
                    dict(type='progress', jid=kwargs['task'], id='', name='rm_rf', result=True,
                         comment=json.dumps({'parser': 'pv', 'progress': str(int((ii + 1) / len(dir_list) * 100)),
                                             'total_steps': len(dirs_list), 'step': i + 1}))))
        # removing extended attrs if some left after gluster
        out = get_output(f'/usr/bin/getfattr -m - "{real_dir}"', suppress_err=True, shell=True)
        if not out['result']:
            raise CommandExecutionError(out['comment'])
        for xattr in out['comment'].splitlines():
            xattr = xattr.strip()
            if xattr.startswith('#') or not xattr:
                continue
            if not get_result(['/usr/bin/setfattr', '-x', xattr, real_dir], shell=False):
                raise CommandExecutionError(f'Error removing "{xattr}" atribute in directory {real_dir}')

    return f"Successfully removed all data in directories {', '.join(dirs_list)}"


@states_response
def drop_caches(value=3):
    cmd_ret = get_output(f'sync; echo {value} > /proc/sys/vm/drop_caches', timeout=120)
    if not cmd_ret['result']:
        comment = f"Error in drop caches: {cmd_ret['comment']}"
        raise CommandExecutionError(comment)
    return f"Successfully drop caches ({value})."


@states_response
def set_hostname(name):
    cmd_ret = get_output(f"hostnamectl set-hostname '{name}'")
    if not cmd_ret['result']:
        comment = f"Error in set_hostname: {cmd_ret['comment']}"
        raise CommandExecutionError(comment)
    return f"Successfully set_hostname {name}."


def mk_authorized_keys_file(username):
    if not os.path.exists(USER_HOME.format(username=username)):
        comment = f"SSH user {username} do not exists."
        raise CommandExecutionError(comment)
    if not os.path.exists(USER_SSH_FOLDER.format(username=username)):
        os.mkdir(USER_SSH_FOLDER.format(username=username))
        open(USER_AUTHORIZED_KEYS.format(username=username), 'a').close()
    else:
        if not os.path.isfile(USER_AUTHORIZED_KEYS.format(username=username)):
            open(USER_AUTHORIZED_KEYS.format(username=username), 'a').close()


def authorized_keys_path(username='root'):
    if username == 'root':
        return ROOT_AUTHORIZED_KEYS
    else:
        mk_authorized_keys_file(username)
        return USER_AUTHORIZED_KEYS.format(username=username)


@states_response
def public_keys(username='root'):
    keys = []
    with open(authorized_keys_path(username), 'r') as f:
        for line in f.readlines():
            line = line.rstrip()
            if line.endswith(PUBLIC_KEY_IDENTIFIER):
                keys.append(line.rsplit(' ', 1)[0])
    return keys


@states_response
def public_key_attach(public_key, username='root'):
    if not public_key_check(public_key, username)['comment']:  # pylint: disable=unsubscriptable-object
        with open(authorized_keys_path(username), 'a') as f:
            f.write(f'{public_key}{PUBLIC_KEY_IDENTIFIER}\n')
    return "Public key successfully attached."


@states_response
def public_key_detach(public_key=PUBLIC_KEY_IDENTIFIER, username='root'):
    with open(authorized_keys_path(username), 'r+') as f:
        rows = f.readlines()
        f.seek(0)
        for row in rows:
            if row != '\n' and public_key not in row:
                f.write(row)
        f.truncate()
    return "Public key successfully detached."


@states_response
def public_key_check(public_key, username='root'):
    with open(authorized_keys_path(username), 'r') as f:
        rows = f.readlines()
        for row in rows:
            if public_key in row:
                return True
    return False


@states_response
def get_domains_power_state_expected() -> dict:
    result = {}
    if os.path.exists(DOMAINS_FOLDER):
        for file_name in os.listdir(DOMAINS_FOLDER):
            dom_id, _ = os.path.splitext(file_name)
            data_operator = CommonDataOperator(domain_id=dom_id, xml_operator=True, libvirt_operator=False)
            result[dom_id] = data_operator.get_field_info(data_operator.frequently_used_fields["power_state_expected"])
    return result
