"""
(c) 2023 «Dacom M» Ltd
"""
import os
from collections import OrderedDict

from veilrpc.utils.bazalt import states_response, get_output, action_service, RUN_PATH

BASE_SERVICES = [('iscsi', 'iscsid.service'),
                 ('multipath', 'multipathd.service'),
                 ('postgresql', 'postgresql@13-main.service'),
                 ('nginx', 'nginx.service'),
                 ('ntp', 'ntp.service'),
                 ('redis', 'redis-server.service'),
                 ('beanstalkd', 'beanstalkd.service'),
                 ('ttyd', 'ttyd.service'),
                 ('snmp', 'snmpd.service'),
                 ('gluster', 'glusterd.service'),
                 ('corosync', 'corosync.service'),
                 ('dlm', 'dlm.service'),
                 ('watchdog', 'watchdog.service'),
                 ('linstor-satellite', 'linstor-satellite.service'),
                 ('td-agent', 'td-agent.service'),
                 ('consul', 'consul.service')]

BASE_CONTROLLER_SERVICES = [('postgresql', 'postgresql@13-main.service'),
                            ('nginx', 'nginx.service'),
                            ('ntp', 'ntp.service'),
                            ('redis', 'redis-server.service'),
                            ('beanstalkd', 'beanstalkd.service'),
                            ('snmp', 'snmpd.service'),
                            ('watchdog', 'watchdog.service'),
                            ('loki', 'loki.service'),
                            ('linstor-controller', 'linstor-controller.service'),
                            ('prometheus', 'prometheus.service')]

BASE_CONTROLLER_SERVICES_2 = [('loki', 'loki.service'),
                              ('prometheus', 'prometheus.service'),
                              ('linstor-controller', 'linstor-controller.service')]

CONTROLLER_SERVICES = [('controller-engine', 'controller-supervisor.service'),
                       ('controller-web-api', 'gunicorn-django.service'),
                       ('controller-websockets', 'daphne-django.service'),
                       ('controller-web-proxy', 'gunicorn-proxy.service'),
                       ('controller-web-uploader', 'gunicorn-uploader.service'),
                       ('controller-log-sender', 'gunicorn-logger.service'),
                       ('controller-ws', 'gunicorn-ws.service'),
                       ('websockify', 'websockify.service')]

BASE_NODE_SERVICES = [('node-engine', 'node_supervisor.service'),
                      ('node-statistics', 'prometheus-node-exporter.service')]

NODE_SERVICES = [('node-engine', 'node_supervisor.service'),
                 ('node-web-api', 'gunicorn-django-node.service'),
                 ('node-web-proxy', 'gunicorn-proxy-node.service'),
                 ('node-web-uploader', 'gunicorn-uploader-node.service'),
                 ('node-statistics', 'prometheus-node-exporter.service'),
                 ('websockify-node', 'websockify-node.service')]

SERVICES_NODE = OrderedDict(NODE_SERVICES + BASE_SERVICES)
SERVICES_CONTROLLER = OrderedDict(CONTROLLER_SERVICES + BASE_CONTROLLER_SERVICES)
SERVICES_CONTROLLER_NODE = OrderedDict(CONTROLLER_SERVICES + BASE_CONTROLLER_SERVICES_2 +
                                       BASE_NODE_SERVICES + BASE_SERVICES)
SERVICES_FULL = OrderedDict(CONTROLLER_SERVICES + NODE_SERVICES + BASE_SERVICES)
SERVICES = {
    'node': SERVICES_NODE,
    'controller': SERVICES_CONTROLLER,
    'controller_node': SERVICES_CONTROLLER_NODE,
    'full': SERVICES_FULL
}
# Список сервисов, запуск/перезапуск/остановка которых выполняется только если они указаны явно.
explicitly_specified_only = ['postgresql@13-main.service', 'redis-server.service', 'watchdog.service']
antiviruses = [('kaspersky', 'kesl.service'),
               ('drweb', 'drweb-configd.service')]


def get_services(full=False):
    from .node import get_current_role
    if full:
        services = SERVICES['full']
    elif get_current_role() != 'stale':
        services = SERVICES[get_current_role()]
    else:
        services = None
    for antivirus in antiviruses:
        if action_service(antivirus[1], 'cat', suppress_err=True)['result']:
            services[antivirus[0]] = antivirus[1]
    return services


def get_choices(full=False):
    choices = ['all', 'controller', 'node']
    services = get_services(full=full)
    if services is not None:
        choices.extend(services.keys())
    return choices


@states_response
def services_list(formatted=True, as_root=False, **kwargs):
    results = []
    cmd = "/usr/bin/systemctl list-units --no-pager --no-legend --type=service | awk '{print $1\":\"$4}'"
    if as_root:
        cmd = "sudo " + cmd
    cmd_ret = get_output(cmd, **kwargs)
    if cmd_ret['result']:
        services = dict()
        for service in cmd_ret['comment'].splitlines():
            try:
                k, v = service.split(':')
                services[k] = v
            except ValueError:
                continue
        services_dict = get_services()
        if services_dict is not None:
            for verbose_name, service_name in services_dict.items():
                if formatted:
                    results.append('{:<25}{:>20}'.format(verbose_name,
                                                         services.get(service_name, 'stopped')))
                else:
                    results.append(dict(verbose_name=verbose_name,
                                        status=services.get(service_name, 'stopped')))
    return results


def get_services_names(choice):
    services = get_services()
    if choice == 'all':
        return [service_name for service_name in services.values()
                if service_name not in explicitly_specified_only]
    elif choice == 'controller':
        ctr_services = [s[1] for s in CONTROLLER_SERVICES]
        return [service_name for service_name in services.values()
                if service_name in ctr_services]
    elif choice == 'node':
        node_services = [s[1] for s in NODE_SERVICES]
        values = [service_name for service_name in services.values()
                  if service_name in node_services]
        return values
    else:
        return services.get(choice)


@states_response
def services_action(service, action, **kwargs):
    fpath = os.path.join(RUN_PATH, 'node_stopped')
    services_names = get_services_names(service)
    if isinstance(services_names, str):
        if services_names == 'node_supervisor.service':
            if action == 'stop':
                open(fpath, 'w').close()
            elif action in ('start', 'restart') and os.path.isfile(fpath):
                os.remove(fpath)
        action_service(services_names, action, **kwargs)
    elif isinstance(services_names, list):
        if 'node_supervisor.service' in services_names:
            if action == 'stop':
                open(fpath, 'w').close()
            elif action in ('start', 'restart') and os.path.isfile(fpath):
                os.remove(fpath)
        action_service(' '.join(services_names), action, **kwargs)
    return f'Successfully {action} {str(service)} services.'


@states_response
def services_cat(service, action, **kwargs):
    services_names = get_services_names(service)
    if isinstance(services_names, str):
        return action_service(services_names, action, **kwargs)
    elif isinstance(services_names, list):
        return action_service(' '.join(services_names), action, **kwargs)
