"""
(c) 2023 «Dacom M» Ltd
"""

import os
import time
import shutil
import uuid
import yaml

from veilrpc.utils.file_based_lock import lock
from veilrpc.utils.bazalt import states_response, CommandExecutionError, SSH_GROUP, \
    get_output, SSH_YAML, ROLE_SCRIPT_PATH, LOCAL_MODULES_PATH, BASE_PUPPET_PATH, \
    VEIL_NODE_LOCAL_PATH, action_service
from veilrpc.states.puppet import puppet_modify_param

COMMENT_TAG = "VEIL_KEY"
LIMITS_PATH = f'{LOCAL_MODULES_PATH}/limits.yaml'
DEFAULT_SSHD_CONF = f"{BASE_PUPPET_PATH}/ssh.yaml"
DEFAULT_LIMITS_CONF = f"{BASE_PUPPET_PATH}/limits.yaml"
LIMITS_ENTRIES = 'limits::entries'


def _tmp_file_path(file_path):
    path = os.path.dirname(file_path)
    file_name = os.path.basename(file_path)
    tmp_file_name = f"_{str(uuid.uuid4())}_{file_name}"
    return os.path.join(path, tmp_file_name)


def _add_authorized_key(file_path, key, key_id):
    file_path = str(file_path)
    key = str(key)
    key_id = str(key_id)
    tmp_file_path = _tmp_file_path(file_path)
    shutil.copy2(file_path, tmp_file_path)
    src_file_stat = os.stat(file_path)
    os.chown(tmp_file_path, src_file_stat.st_uid, src_file_stat.st_gid)
    with open(tmp_file_path, 'a') as tmp_file:
        tmp_file.write(f"{key} {COMMENT_TAG}:{key_id}:{time.time()}\n")
    shutil.move(tmp_file_path, file_path)


def _delete_authorized_key(file_path, key_id):
    file_path = str(file_path)
    src_file_stat = os.stat(file_path)
    key_id = str(key_id)
    tmp_file_path = _tmp_file_path(file_path)
    with open(file_path, 'r') as src_file:
        with open(tmp_file_path, 'w') as dst_file:
            for line in src_file:
                if COMMENT_TAG in line:
                    try:
                        veil_comment_position = line.find(COMMENT_TAG)
                        veil_comment_line = line[veil_comment_position:]
                        line_key_id = veil_comment_line.split(":")[1]
                        if line_key_id == key_id:
                            # didn't write to dst tmp file
                            continue
                    except IndexError:
                        # bad formated veil line? copy as-is to dst
                        pass
                dst_file.write(line)
    shutil.copystat(file_path, tmp_file_path)
    os.chown(tmp_file_path, src_file_stat.st_uid, src_file_stat.st_gid)
    shutil.move(tmp_file_path, file_path)


@states_response
def add_authorized_key(file_path, key, key_id):
    # параметр file_path используется как для конфигурации декоратора,
    # так и для передачи в функцию
    return lock(file_path)(_add_authorized_key)(
        file_path, key, key_id)


@states_response
def puppet_module_modify_param(param, value, module='ssh', **kwargs):
    _create_ssh_yaml()
    try:
        with open(SSH_YAML, 'r+') as f:
            conf = yaml.safe_load(f)
            conf[f'{module}::{param}'] = value
            f.seek(0)
            f.write(yaml.dump(conf))
            f.truncate()
    except Exception as err:
        comment = f'Unable to write configuration: {err}'
        raise CommandExecutionError(comment) from None
    if kwargs.get('apply'):
        _ssh_apply()
    return f'Param {param} successfully modified'


def _puppet_module_remove_param(param, module='ssh', **kwargs):
    _create_ssh_yaml()
    try:
        with open(SSH_YAML, 'r+') as f:
            conf = yaml.safe_load(f)
            conf.pop(f'{module}::{param}')
            f.seek(0)
            f.write(yaml.dump(conf))
            f.truncate()
    except KeyError:
        pass
    except Exception as err:
        comment = f'Unable to modify configuration: {err}'
        raise CommandExecutionError(comment) from None
    if kwargs.get('apply'):
        _ssh_apply()
    return f'Param {param} successfully removed'


@states_response
def puppet_module_get_param(param, module='ssh', **kwargs):
    _param = f'{module}::{param}'
    _create_ssh_yaml()
    try:
        with open(SSH_YAML) as f:
            conf = yaml.safe_load(f)
            return conf[_param] if _param in conf.keys() else None
    except Exception as err:
        comment = f'Unable to read {SSH_YAML}: {err}'
        raise CommandExecutionError(comment) from None


def sshd_config() -> str:
    if os.path.exists("/etc/ssh/sshd_config"):
        return "/etc/ssh/sshd_config"
    elif os.path.exists("/etc/openssh/sshd_config"):
        return "/etc/openssh/sshd_config"
    else:
        raise CommandExecutionError("no sshd config")


@states_response
def ssh_get_param(param, **kwargs):
    cmd = f"grep -r '^{param}' {sshd_config()} | awk '{{print $2}}'"
    resp = get_output(cmd, **kwargs)
    if not resp['result']:
        raise CommandExecutionError(resp['comment'])
    value = resp['comment'].strip()
    try:
        return int(value)
    except ValueError:
        return value


@states_response
def security_limits_get_param(param, **kwargs):
    cmd = f"grep -r '^{param}' /etc/security/sshd_config | awk '{{print $2}}'"
    resp = get_output(cmd, **kwargs)
    if not resp['result']:
        raise CommandExecutionError(resp['comment'])
    value = resp['comment'].strip()
    try:
        return int(value)
    except ValueError:
        return value


@states_response
def delete_authorized_key(file_path, key_id):
    return lock(file_path)(_delete_authorized_key)(
        file_path, key_id)


@states_response
def ssh_status(**kwargs):
    resp = get_output(f"cat {sshd_config()}", **kwargs)
    if not resp['result']:
        raise CommandExecutionError(resp['comment'])
    return True if f'AllowGroups {SSH_GROUP}' in resp['comment'] else False


@states_response
def ssh_root_to_group(value=True, **kwargs):
    cmd = f"usermod root -G {SSH_GROUP}" if value else "usermod root -G ''"
    resp = get_output(cmd, **kwargs)
    if not resp['result']:
        raise CommandExecutionError(resp['comment'])
    return f'The root user now has{"" if value else " no"} SSH access'


@states_response
def ssh_root_access(value=True, **kwargs):
    ret = puppet_module_modify_param(param='permit_root_login',
                                     value='yes' if value else 'without-password',
                                     **kwargs)
    if not ret['result']:
        raise CommandExecutionError(ret['comment'])
    _ssh_apply()
    return f'The root user now has{"" if value else " no"} SSH access by password'


@states_response
def ssh_get_users(info=False, **kwargs):
    resp = get_output(f"getent group {SSH_GROUP} | awk -F: '{{print $4}}'", **kwargs)
    if not resp['result']:
        raise CommandExecutionError(resp['comment'])
    users = resp['comment'].strip().split(',') if resp['comment'].strip() else []
    if info:
        users_with_info = []
        for user in users:
            if user == 'root':
                resp = ssh_get_param('PermitRootLogin', **kwargs)
                if not resp['result']:
                    raise CommandExecutionError(resp['comment'])
                user_status = False if resp['comment'].strip() == 'without-password' else True
            else:
                resp = get_output(f"passwd -S {user}", **kwargs)
                if not resp['result']:
                    raise CommandExecutionError(resp['comment'])
                user_status = False if resp['comment'].split()[1] == 'L' else True
            users_with_info.append(dict(name=user, status=user_status))
        return users_with_info
    return users


def _create_ssh_yaml():
    if not os.path.exists(SSH_YAML):
        with open(DEFAULT_SSHD_CONF, 'r') as f:
            default_sshd_conf = yaml.safe_load(f)
        for yaml_param, default_value in default_sshd_conf.items():
            for param, value in _sshd_config_settings().items():
                if param in yaml_param and value != default_value:
                    default_sshd_conf[yaml_param] = value
        with open(SSH_YAML, 'w') as f:
            f.write(yaml.dump(default_sshd_conf))


def _sshd_config_settings():
    settings = dict()
    resp = ssh_get_param(param='ClientAliveInterval')
    if resp['result']:
        settings['alive_interval'] = resp['comment']
    resp = ssh_get_param(param='ClientAliveCountMax')
    if resp['result']:
        settings['alive_count_max'] = resp['comment']
    resp = ssh_get_param(param='MaxAuthTries')
    if resp['result']:
        settings['maxauthtries'] = resp['comment']
    resp = ssh_get_param(param='MaxSessions')
    if resp['result']:
        settings['maxsessions'] = resp['comment']
    return settings


def _security_limits_settings():
    settings = dict()
    resp = get_security_limits(param='maxlogins')
    if resp['result']:
        settings['maxlogins'] = resp['comment']
    resp = get_security_limits(param='maxsyslogins')
    if resp['result']:
        settings['maxsyslogins'] = resp['comment']
    return settings


@states_response
def ssh_settings(**kwargs):
    settings = _sshd_config_settings()
    settings.update(_security_limits_settings())
    return settings


@states_response
def ssh_unlock_users(ssh_users=None, **kwargs):
    """
    Разблокирует пользователей SSH. Если список пользователей не передан, разблокирует всех.
    """
    error_list = []
    if not ssh_users:
        cmd_ret = ssh_get_users()
        if not cmd_ret['result']:
            error_list.append(cmd_ret['comment'])
            ssh_users = []
        else:
            ssh_users = cmd_ret['comment']
    for user in ssh_users:
        user = user if isinstance(user, str) else user['name']
        if user == 'root':
            cmd_ret = ssh_root_access(**kwargs)
            if not cmd_ret['result']:
                error_list.append(cmd_ret['comment'])
        else:
            cmd_ret = get_output(f'passwd -u {user}', **kwargs)
            if not cmd_ret['result']:
                error_list.append(cmd_ret['comment'])
    if error_list:
        raise CommandExecutionError(
            f'An error occurred while enabling SSH users. Errors: {", ".join(error_list)}')


@states_response
def ssh_lock_users(ssh_users, **kwargs):
    """
    Блокирует пользователей SSH.
    """
    error_list = []
    if ssh_users:
        for user in ssh_users:
            if user['name'] == 'root':
                cmd_ret = ssh_root_access(value=False, **kwargs)
                if not cmd_ret['result']:
                    error_list.append(cmd_ret['comment'])
            else:
                cmd_ret = get_output(f'passwd -l {user["name"]}', **kwargs)
                if not cmd_ret['result']:
                    error_list.append(cmd_ret['comment'])
        if error_list:
            raise CommandExecutionError(
                f'An error occurred while disabling SSH users. Errors: {", ".join(error_list)}')


@states_response
def ssh_remove_users(ssh_users=None, **kwargs):
    """
    Удаляет пользователей SSH. Если список пользователей не передан, удаляет всех.
    На этапе тестирования ни один root не пострадал.
    """
    error_list = []
    if not ssh_users:
        cmd_ret = ssh_get_users(**kwargs)
        if not cmd_ret['result']:
            error_list.append(cmd_ret['comment'])
            ssh_users = []
        else:
            ssh_users = cmd_ret['comment']
    for user in ssh_users:
        user = user if isinstance(user, str) else user['name']
        if user != 'root':
            cmd_ret = get_output(f"userdel -f -r {user}", **kwargs)
            if not cmd_ret['result']:
                error_list.append(cmd_ret['comment'])
    if error_list:
        raise CommandExecutionError(
            f'An error occurred while removing SSH users. Errors: {", ".join(error_list)}')


@states_response
def ssh_add_users(ssh_users, set_pass=True, **kwargs):
    error_list = []
    for user in ssh_users:
        cmd_ret = get_output(f"useradd {user['name']} -m -s /bin/bash -g veil -G {SSH_GROUP}", **kwargs)
        if not cmd_ret['result']:
            error_list.append(cmd_ret['comment'])
        cmd_ret = get_output(f"cp /root/.bashrc /home/{user['name']}/", **kwargs)
        if not cmd_ret['result']:
            error_list.append(cmd_ret['comment'])
    if set_pass:
        ssh_set_password(ssh_users, **kwargs)
    if error_list:
        raise CommandExecutionError(
            f'An error occurred while adding SSH users. Errors list: {", ".join(error_list)}')


@states_response
def ssh_set_password(ssh_users, **kwargs):
    error_list = []
    for user in ssh_users:
        cmd_ret = get_output("echo -e '{password}\n{password}\n' | passwd -q {user}".format(
            password=user['password'],
            user=user['name']), **kwargs)
        if not cmd_ret['result']:
            error_list.append(cmd_ret['comment'])
    if error_list:
        raise CommandExecutionError(
            f'An error occurred while changing SSH passwords. Errors list: {", ".join(error_list)}')


@states_response
def ssh_remove_entry_restrictions(**kwargs):
    _puppet_module_remove_param(param='sshd_config_allowgroups', **kwargs)
    _puppet_module_remove_param(param='sshd_config_denygroups', **kwargs)
    ret = ssh_root_access(**kwargs)
    if not ret['result']:
        raise CommandExecutionError(ret['comment'])
    return 'SSH entry restrictions have been removed'


def _ssh_apply():
    ret = get_output(f"{ROLE_SCRIPT_PATH} -a ssh")
    if not ret['result']:
        raise CommandExecutionError(ret['comment'])
    ret = get_output("systemctl restart sshd.service")
    if not ret['result']:
        raise CommandExecutionError(ret['comment'])


@states_response
def ssh_enable(**kwargs):
    _puppet_module_remove_param(param='sshd_config_denygroups', **kwargs)
    ret = puppet_module_modify_param(param='sshd_config_allowgroups', value=[SSH_GROUP, 'veil'], **kwargs)
    if not ret['result']:
        raise CommandExecutionError(ret['comment'])
    _ssh_apply()
    puppet_modify_param(VEIL_NODE_LOCAL_PATH, 'veil_node', 'ttyd_ensure', True)
    action, service = 'restart', 'ttyd'
    cmd_ret = action_service(service, action, timeout=300)
    if not cmd_ret['result']:
        raise CommandExecutionError(f"Error in {action} {service}: {cmd_ret['comment']}")
    return 'SSH successfully enabled'


@states_response
def ssh_disable(**kwargs):
    ret = puppet_module_modify_param(param='sshd_config_denygroups', value=[SSH_GROUP], **kwargs)
    if not ret['result']:
        raise CommandExecutionError(ret['comment'])
    ret = puppet_module_modify_param(param='sshd_config_allowgroups', value=['veil'], **kwargs)
    if not ret['result']:
        raise CommandExecutionError(ret['comment'])
    _ssh_apply()
    puppet_modify_param(VEIL_NODE_LOCAL_PATH, 'veil_node', 'ttyd_ensure', False)
    action, service = 'stop', 'ttyd'
    cmd_ret = action_service(service, action, timeout=300)
    if not cmd_ret['result']:
        raise CommandExecutionError(f"Error in {action} {service}: {cmd_ret['comment']}")
    return 'SSH successfully disabled'


@states_response
def ssh_alive_interval(value, **kwargs):
    ret = puppet_module_modify_param(param='sshd_client_alive_interval', value=value, **kwargs)
    if not ret['result']:
        raise CommandExecutionError(ret['comment'])
    _ssh_apply()
    return ret['comment']


@states_response
def ssh_alive_count_max(value, **kwargs):
    ret = puppet_module_modify_param(param='sshd_client_alive_count_max', value=value, **kwargs)
    if not ret['result']:
        raise CommandExecutionError(ret['comment'])
    _ssh_apply()
    return ret['comment']


@states_response
def ssh_maxauthtries(value, **kwargs):
    ret = puppet_module_modify_param(param='sshd_config_maxauthtries', value=value, **kwargs)
    if not ret['result']:
        raise CommandExecutionError(ret['comment'])
    _ssh_apply()
    return ret['comment']


@states_response
def ssh_maxsessions(value, **kwargs):
    ret = puppet_module_modify_param(param='sshd_config_maxsessions', value=value, **kwargs)
    if not ret['result']:
        raise CommandExecutionError(ret['comment'])
    _ssh_apply()
    return ret['comment']


@states_response
def ssh_sessions(**kwargs):
    ret = get_output('last | grep "still logged in"', **kwargs)
    if not ret['result']:
        raise CommandExecutionError(ret['comment'])
    return ret['comment']


@states_response
def ssh_maxlogins(value, **kwargs):
    ret = set_security_limits(param='maxlogins',
                              value=value)
    if not ret['result']:
        raise CommandExecutionError(ret['comment'])
    return ret['comment']


@states_response
def ssh_maxsyslogins(value, **kwargs):
    ret = set_security_limits(param='maxsyslogins',
                              value=value)
    if not ret['result']:
        raise CommandExecutionError(ret['comment'])
    return ret['comment']


@states_response
def set_security_limits(param, value, **kwargs):
    if os.path.exists(LIMITS_PATH):
        try:
            with open(LIMITS_PATH, 'r+') as f:
                limits_conf = yaml.safe_load(f)
                if not limits_conf.get(LIMITS_ENTRIES):
                    limits_conf[LIMITS_ENTRIES] = {}
                if not limits_conf[LIMITS_ENTRIES].get(f'@veil-admins/{param}'):
                    limits_conf[LIMITS_ENTRIES][f'@veil-admins/{param}'] = {}
                limits_conf[LIMITS_ENTRIES][f'@veil-admins/{param}']['hard'] = value
                f.seek(0)
                f.write(yaml.dump(limits_conf))
                f.truncate()
        except Exception as err:
            comment = f'Unable to write security limits configuration: {err}'
            raise CommandExecutionError(comment) from None
    else:
        with open(DEFAULT_LIMITS_CONF, 'r') as f:
            default_limits_conf = yaml.safe_load(f)
        default_limits_conf[LIMITS_ENTRIES][f'@veil-admins/{param}']['hard'] = value
        with open(LIMITS_PATH, 'w') as f:
            f.write(yaml.dump(default_limits_conf))
    cmd_ret = get_output(f'{ROLE_SCRIPT_PATH} -a limits')
    if not cmd_ret['result']:
        comment = f"Unable to apply securuty limits config: {cmd_ret['comment']}"
        raise CommandExecutionError(comment)
    return 'Security limits successfully updated.'


@states_response
def get_security_limits(param, **kwargs):
    if os.path.exists(LIMITS_PATH):
        try:
            with open(LIMITS_PATH) as f:
                limits_conf = yaml.safe_load(f)
                if limits_conf.get(LIMITS_ENTRIES) and \
                        limits_conf[LIMITS_ENTRIES].get(f'@veil-admins/{param}'):
                    return limits_conf[LIMITS_ENTRIES][f'@veil-admins/{param}']['hard']
                else:
                    return None
        except Exception as err:
            comment = f'Unable to read security limits configuration: {err}'
            raise CommandExecutionError(comment) from None
    else:
        with open(DEFAULT_LIMITS_CONF, 'r') as f:
            default_limits_conf = yaml.safe_load(f)
        with open(LIMITS_PATH, 'w') as f:
            f.write(yaml.dump(default_limits_conf))
        return default_limits_conf[LIMITS_ENTRIES][f'@veil-admins/{param}']['hard']
