"""
(c) 2023 «Dacom M» Ltd
"""

import psutil
from veilrpc.utils.bazalt import states_response, CommandExecutionError, get_output, ZPOOL_STORAGES_PATH


@states_response
def procs():
    """Return the process data"""
    cmd = 'ps'
    cmd_ret = get_output(cmd)
    if not cmd_ret['result']:
        raise CommandExecutionError(cmd_ret['comment'])
    # Get the user, pid and cmd
    ret = {}
    uind = 0
    pind = 0
    cind = 0
    plines = cmd_ret['comment'].splitlines()
    guide = plines.pop(0).split()
    if 'USER' in guide:
        uind = guide.index('USER')
    elif 'UID' in guide:
        uind = guide.index('UID')
    if 'PID' in guide:
        pind = guide.index('PID')
    if 'COMMAND' in guide:
        cind = guide.index('COMMAND')
    elif 'CMD' in guide:
        cind = guide.index('CMD')
    for line in plines:
        if not line:
            continue
        comps = line.split()
        ret[comps[pind]] = {'user': comps[uind],
                            'cmd': ' '.join(comps[cind:])}
    return ret


def _number(text):
    """
    Convert a string to a number.
    Returns an integer if the string represents an integer, a floating
    point number if the string is a real number, or the string unchanged
    otherwise.
    """
    if text.isdigit():
        return int(text)
    try:
        return float(text)
    except ValueError:
        return text


@states_response
def cpustats():
    """Return the CPU stats"""
    ret = {}
    try:
        with open('/proc/stat', 'r') as fp_:
            stats = fp_.read()
    except IOError:
        pass
    else:
        for line in stats.splitlines():
            if not line:
                continue
            comps = line.split()
            if comps[0] == 'cpu':
                ret[comps[0]] = {'idle': _number(comps[4]),
                                 'iowait': _number(comps[5]),
                                 'irq': _number(comps[6]),
                                 'nice': _number(comps[2]),
                                 'softirq': _number(comps[7]),
                                 'steal': _number(comps[8]),
                                 'system': _number(comps[3]),
                                 'user': _number(comps[1])}
            elif comps[0] == 'intr':
                ret[comps[0]] = {'total': _number(comps[1]),
                                 'irqs': [_number(x) for x in comps[2:]]}
            elif comps[0] == 'softirq':
                ret[comps[0]] = {'total': _number(comps[1]),
                                 'softirqs': [_number(x) for x in comps[2:]]}
            else:
                ret[comps[0]] = _number(comps[1])
    return ret


@states_response
def meminfo():
    ret = {}
    try:
        with open('/proc/meminfo', 'r') as fp_:
            stats = fp_.read()
    except IOError:
        pass
    else:
        for line in stats.splitlines():
            if not line:
                continue
            comps = line.split()
            comps[0] = comps[0].replace(':', '')
            ret[comps[0]] = {
                'value': comps[1],
            }
            if len(comps) > 2:
                ret[comps[0]]['unit'] = comps[2]
    return ret


@states_response
def cpuinfo():
    ret = {}
    try:
        with open('/proc/cpuinfo', 'r') as fp_:
            stats = fp_.read()
    except IOError:
        pass
    else:
        for line in stats.splitlines():
            if not line:
                continue
            comps = line.split(':')
            comps[0] = comps[0].strip()
            if comps[0] == 'flags':
                ret[comps[0]] = comps[1].split()
            else:
                ret[comps[0]] = comps[1].strip()
    return ret


@states_response
def diskstats():
    ret = {}
    try:
        with open('/proc/diskstats', 'r') as fp_:
            stats = fp_.read()
    except IOError:
        pass
    else:
        for line in stats.splitlines():
            if not line:
                continue
            comps = line.split()
            ret[comps[2]] = {
                'major': _number(comps[0]),
                'minor': _number(comps[1]),
                'device': _number(comps[2]),
                'reads_issued': _number(comps[3]),
                'reads_merged': _number(comps[4]),
                'sectors_read': _number(comps[5]),
                'ms_spent_reading': _number(comps[6]),
                'writes_completed': _number(comps[7]),
                'writes_merged': _number(comps[8]),
                'sectors_written': _number(comps[9]),
                'ms_spent_writing': _number(comps[10]),
                'io_in_progress': _number(comps[11]),
                'ms_spent_in_io': _number(comps[12]),
                'weighted_ms_spent_in_io': _number(comps[13])
            }
    return ret


@states_response
def vmstats():
    ret = {}
    try:
        with open('/proc/vmstat', 'r') as fp_:
            stats = fp_.read()
    except IOError:
        pass
    else:
        for line in stats.splitlines():
            if not line:
                continue
            comps = line.split()
            ret[comps[0]] = _number(comps[1])
    return ret


@states_response
def netstats():
    ret = {}
    try:
        with open('/proc/net/netstat', 'r') as fp_:
            stats = fp_.read()
    except IOError:
        pass
    else:
        headers = ['']
        for line in stats.splitlines():
            if not line:
                continue
            comps = line.split()
            if comps[0] == headers[0]:
                index = len(headers) - 1
                row = {}
                for field in range(index):
                    if field < 1:
                        continue
                    else:
                        row[headers[field]] = _number(comps[field])
                rowname = headers[0].replace(':', '')
                ret[rowname] = row
            else:
                headers = comps
    return ret


@states_response
def netdev():
    ret = {}
    try:
        with open('/proc/net/dev', 'r') as fp_:
            stats = fp_.read()
    except IOError:
        pass
    else:
        for line in stats.splitlines():
            if not line:
                continue
            if line.find(':') < 0:
                continue
            comps = line.split()
            # Fix lines like eth0:9999..'
            comps[0] = line.split(':')[0].strip()
            # Support lines both like eth0:999 and eth0: 9999
            comps.insert(1, line.split(':')[1].strip().split()[0])
            ret[comps[0]] = {'iface': comps[0],
                             'rx_bytes': _number(comps[1]),
                             'rx_compressed': _number(comps[7]),
                             'rx_drop': _number(comps[4]),
                             'rx_errs': _number(comps[3]),
                             'rx_fifo': _number(comps[5]),
                             'rx_frame': _number(comps[6]),
                             'rx_multicast': _number(comps[8]),
                             'rx_packets': _number(comps[2]),
                             'tx_bytes': _number(comps[9]),
                             'tx_carrier': _number(comps[15]),
                             'tx_colls': _number(comps[14]),
                             'tx_compressed': _number(comps[16]),
                             'tx_drop': _number(comps[12]),
                             'tx_errs': _number(comps[11]),
                             'tx_fifo': _number(comps[13]),
                             'tx_packets': _number(comps[10])}
    return ret


@states_response
def w():
    """Return a list of logged in users"""
    user_list = []
    cmd = 'w -h'
    cmd_ret = get_output(cmd)
    if not cmd_ret['result']:
        raise CommandExecutionError(cmd_ret['comment'])
    users = cmd_ret['comment'].splitlines()
    for row in users:
        if not row:
            continue
        comps = row.split()
        rec = {'idle': comps[3],
               'jcpu': comps[4],
               'login': comps[2],
               'pcpu': comps[5],
               'tty': comps[1],
               'user': comps[0],
               'what': ' '.join(comps[6:])}
        user_list.append(rec)
    return user_list


@states_response
def beanstalk_stats(stat='full'):
    if stat == 'full':
        stat_cmd = 'stats'
    elif stat == 'controller_herald':
        stat_cmd = f"stats-tube {stat}"
    else:
        stat_cmd = f"stats-tube {stat}_status"
    cmd = f"echo -e '{stat_cmd}\r\n' | nc localhost 11300 -N"
    cmd_ret = get_output(cmd)
    if not cmd_ret['result']:
        raise CommandExecutionError(cmd_ret['comment'])
    return cmd_ret['comment']


def get_actual_size_directory(list_pools: list) -> dict:
    ret = get_output(f'du -b {"".join([i.rstrip("/") + "/* " for i in list_pools])}')
    if not ret['result']:
        raise CommandExecutionError(ret['comment'])
    sizes = {'qcow2': {}, 'snapshot': {}}
    list_files = ret['comment'].strip().split('\n')
    if list_files[0]:
        for k, v in [tuple(x.split()) for x in list_files]:
            ext = v.split('.')[-1]
            if ext == 'qcow2':
                sizes['qcow2'].update({v: k})
            elif ext == 'snapshot':
                sizes['snapshot'].update({v: k})
    return sizes


def get_actual_size_zfs(list_pools: list) -> dict:
    l_bracket, r_bracket = ('{', '}') if len(list_pools) > 1 else ('', '')

    cmd = f'for i in {ZPOOL_STORAGES_PATH}/' \
          f'{l_bracket}{",".join([i.rstrip("/").split("/")[-1] for i in list_pools])}{r_bracket}/ ;do ' \
          f'find $i -name vdisk.qcow2 -not -path "*/\\.*" -exec du -b {{}} \\;;done'
    ret_size_disks = get_output(cmd)
    if not ret_size_disks['result']:
        raise CommandExecutionError(ret_size_disks['comment'])
    ret_size_snapshots = get_output('zfs list -o name,used -H -t snapshot -r ' +
                                    " ".join([i.rstrip("/").split("/")[-1] for i in list_pools]))
    if not ret_size_snapshots['result']:
        raise CommandExecutionError(ret_size_snapshots['comment'])
    sizes = {'qcow2': {}, 'snapshot': {}}
    vdisks = ret_size_disks['comment'].strip().split('\n')
    if vdisks[0]:
        for k, v in [tuple(x.split()) for x in vdisks]:
            sizes['qcow2'].update({v: int(k)})
    snapshots = ret_size_snapshots['comment'].strip().split('\n')
    if snapshots[0]:
        for k, v in [tuple(x.split()) for x in snapshots]:
            try:
                v = int(float(v[:-1])) * 1024 ** 'BKMGTPE'.index(v[-1])
            except ValueError:
                v = int(float(v[:-1] if v[-1].isalpha() else v))
            sizes['snapshot'].update({k.split('@')[-1]: v})
    return sizes


def get_actual_size_thinlvm(list_pools: list) -> dict:
    cmd = f"LVM_SUPPRESS_FD_WARNINGS=1 lvs --noheadings --units b -o " \
          f"lv_full_name,lv_size,data_percent {' '.join(list_pools)}"
    ret = get_output(cmd)
    if not ret['result']:
        raise CommandExecutionError(ret['comment'])
    sizes = {'qcow2': {}, 'snapshot': {}}
    list_volums = ret['comment'].strip().split('\n')
    if list_volums[0]:
        for k, v, p in [tuple(x.strip().split()) for x in list_volums]:
            mark_words = ['datapool-', 'tpool-']
            if not any(w in k for w in mark_words):
                v = int(float(v[:-1]) * float(p) / 100)
                k = f'/dev/{k}'
                if '_snapshot' in k:
                    sizes['snapshot'].update({k: v})
                else:
                    sizes['qcow2'].update({k: v})
    return sizes


@states_response
def get_actual_size(pools: dict) -> dict:
    sizes = {'qcow2': {}, 'snapshot': {}}
    if pools['directory']:
        sizes.update(get_actual_size_directory(pools['directory']))
    if pools['zfs']:
        sizes.update(get_actual_size_zfs(pools['zfs']))
    if pools['thinlvm']:
        sizes.update(get_actual_size_thinlvm(pools['thinlvm']))
    return sizes


@states_response
def user_descriptors(user, **kwargs):
    cmd = f"lsof -u {user} | grep / | sort  -k9 | wc -l"
    cmd_ret = get_output(cmd, **kwargs)
    if not cmd_ret['result']:
        raise CommandExecutionError(cmd_ret['comment'])
    return cmd_ret['comment']


def mem_stats():
    mem = psutil.virtual_memory()
    return {"memory_count": mem.total / 1024 / 1024,
            "mem_used": mem.used / 1024 / 1024,
            "mem_cached": mem.cached / 1024 / 1024,
            "mem_buffers": mem.buffers / 1024 / 1024,
            "mem_free": mem.free / 1024 / 1024}


def cpu_stats():
    cpus = psutil.cpu_percent(percpu=True)
    cpu_freq = {k: v.current for k, v in enumerate(psutil.cpu_freq(percpu=True))}
    cpu_proc_used_percent = {k: v for k, v in enumerate(cpus)}
    cpu_used_percent = sum(cpus) / len(cpus)
    load_1, load_5, load_15 = psutil.getloadavg()
    return {"cpu_proc_used_percent": cpu_proc_used_percent,
            "cpu_used_percent": cpu_used_percent,
            "cpu_used": cpu_used_percent / 100 * sum(cpu_freq.values()),
            "cpu_freq": cpu_freq,
            "cpu_count": psutil.cpu_count(),
            "cpu_load1": load_1,
            "cpu_load5": load_5,
            "cpu_load15": load_15}


def disk_used():
    disks = psutil.disk_io_counters(perdisk=True, nowrap=True)
    disk_stats = dict()
    for k, v in disks.items():
        disk_stats[k] = {"read_rate": v.read_bytes / 1024 / 1024,
                         "write_rate": v.write_bytes / 1024 / 1024}
    return {"disk_used": disk_stats}


def net_used():
    net = psutil.net_io_counters(pernic=True, nowrap=True)
    net_stats = dict()
    for k, v in net.items():
        net_stats[k] = {"transmit_rate": v.bytes_sent / 1024 / 1024 * 8,
                        "receive_rate": v.bytes_recv / 1024 / 1024 * 8}
    return {"net_used": net_stats}


def collect_data():
    stat_data = {}
    stat_data.update(mem_stats())
    stat_data.update(cpu_stats())
    stat_data.update(disk_used())
    stat_data.update(net_used())
    return stat_data
