"""
(c) 2023 «Dacom M» Ltd
"""

from collections import namedtuple
from concurrent.futures import ThreadPoolExecutor
from signal import SIGKILL
from functools import wraps, lru_cache
import json
import logging
from multiprocessing import cpu_count
import os
import shutil
import signal
import time
from typing import List, Iterable, Callable, TypeVar
from subprocess import check_output, check_call, STDOUT, DEVNULL, SubprocessError, \
    CalledProcessError, call as sub_call, TimeoutExpired, Popen, PIPE
from uuid import UUID


LOG = logging.getLogger("bazalt")
ERROR_PATTERN = 'Error in function {}: {}'

FONT_COLOR = namedtuple('FONT_COLOR', 'BLACK RED GREEN YELLOW BLUE MAGENTA CYAN WHITE')(
    BLACK='\033[1m\033[30m{}\033[0m',
    RED='\033[1m\033[31m{}\033[0m',
    GREEN='\033[1m\033[32m{}\033[0m',
    YELLOW='\033[1m\033[33m{}\033[0m',
    BLUE='\033[1m\033[34m{}\033[0m',
    MAGENTA='\033[1m\033[35m{}\033[0m',
    CYAN='\033[1m\033[36m{}\033[0m',
    WHITE='\033[1m\033[37m{}\033[0m',
)


class CommandExecutionError(Exception):
    pass


# Base
BASE_NAME = 'veil'
CONTROLLER_NAME = 'controller'
NODE_NAME = 'node'
CLI_NAME = 'cli'
REDIS_CLI_COMMANDS_QUEUE = 'node_cli_commands'
REDIS_NODE_CONFIG = 'node_config'
BASE_PATH = f'/var/lib/ecp-{BASE_NAME}'
UTILS_PATH = f'{BASE_PATH}/utils'
UTILS_OTHER_PATH = f'{UTILS_PATH}/other'
NODE_PATH = f'{BASE_PATH}/{NODE_NAME}'
NODE_OTHER_PATH = f'{NODE_PATH}/other'
NODE_PYTHON = f'{NODE_PATH}/env/bin/python'
NODE_LOCAL_DATA_PATH = f'{NODE_PATH}/local_data'
NODE_SETTINGS_PATH = f'{NODE_PATH}/bazalt'
DOMAINS_FOLDER = f'{NODE_PATH}/domains'
DOM_XML_EXT = f'{NODE_NAME}_xml'
HEARTBEAT_PATH = f'{NODE_LOCAL_DATA_PATH}/heartbeat.json'
CLUSTER_PATH = f'{NODE_LOCAL_DATA_PATH}/cluster.json'
FENCE_PATH = f'{NODE_LOCAL_DATA_PATH}/fence.json'
DATA_PATH = f'{BASE_PATH}/data'
COMMON_YAML_PATH = f'{DATA_PATH}/common.yaml'
MODULES_PATH = f'{DATA_PATH}/modules'
LOCAL_MODULES_PATH = f'{DATA_PATH}/local_modules'
VEIL_NODE_LOCAL_PATH = f'{LOCAL_MODULES_PATH}/veil_node.yaml'
VEIL_CONTROLLER_LOCAL_PATH = f'{LOCAL_MODULES_PATH}/veil_controller.yaml'
LOG_PATH = '/var/log'
BASE_LOG_PATH = f'{LOG_PATH}/{BASE_NAME}'
CONTROLLER_LOG_PATH = f'{BASE_LOG_PATH}/{CONTROLLER_NAME}'
NODE_LOG_PATH = f'{BASE_LOG_PATH}/{NODE_NAME}'
CLI_LOG_PATH = f'{BASE_LOG_PATH}/{CLI_NAME}'
WORKER_USER = f'{BASE_NAME}-{NODE_NAME}'
WORKER_USER_GID = 931
WORKER_USER_UID = 931
CONTROLLER_USER = f'{BASE_NAME}-{CONTROLLER_NAME}'
PUPPET_PATH = '/etc/puppetveil/code/environments/production'
PUPPET_MODULES_PATH = f'{PUPPET_PATH}/modules'
BASE_PUPPET_PATH = f'{PUPPET_PATH}/data/modules-defaults'
BASE_SCRIPT_PATH = "/usr/local/sbin"
ROLE_SCRIPT_PATH = f"{BASE_SCRIPT_PATH}/{BASE_NAME}-role"
NODE_CMD = f'{BASE_SCRIPT_PATH}/{BASE_NAME}-{NODE_NAME} '
CLUSTER_CMD = f'{BASE_SCRIPT_PATH}/{BASE_NAME}-cluster '
PROMETHEUS_CMD = f"{BASE_SCRIPT_PATH}/{BASE_NAME}-prometheus"
CONTROLLER_PATH = f'{BASE_PATH}/{CONTROLLER_NAME}'
CONTROLLER_APP_PATH = f'{CONTROLLER_PATH}/app'
CONTROLLER_APP_SETTINGS_PATH = f'{CONTROLLER_APP_PATH}/bazalt'
CONTROLLER_LOCAL_DATA_PATH = f'{CONTROLLER_PATH}/local_data'
CONTROLLER_OTHER_PATH = f'{CONTROLLER_PATH}/other'
CONTROLLER_PYTHON = f'{CONTROLLER_PATH}/env/bin/python'
SENDROLE_PATH = f"/usr/bin/{BASE_NAME}-sendrole"
VEIL_VAR_FILE = f"/usr/share/{BASE_NAME}/deb.subr"
RUN_PATH = f'{BASE_PATH}/run'
INSTALL_TESTS_PATH = f'{BASE_PATH}/tests/'
INSTALL_NODE_PASS = 'node_init_finalize.passed'  # nosec
INSTALL_CTRL_PASS = 'ctr_init_finalize.passed'  # nosec
INSTALL_FIRSTBOOT_PASS = 'first_boot.passed'  # nosec
UPDATE_GRUB = '/usr/sbin/update-grub'
SSH_YAML = f'{LOCAL_MODULES_PATH}/ssh.yaml'
SSH_GROUP = f'{BASE_NAME}-admins'
DEVICE_PREFIX = '/dev/'
LUN_PREFIX = f'{DEVICE_PREFIX}mapper'
CLI_PATH = f'{BASE_PATH}/{CLI_NAME}'
CLI_APP_PATH = f'{CLI_PATH}/app'
CLI_PYTHON = f'{CLI_PATH}/env/bin/python'
CLI_FILE_PATH = f'{CLI_APP_PATH}/cli'
# Путь к файлу, содержащему информацию о типе установки: node, controller, node+controller
VEIL_ROLE_FILE = f"/etc/{BASE_NAME}/media.conf"
VEIL_SE_FILE = f"/etc/{BASE_NAME}/se"
VERSION_FILE = f'/etc/{BASE_NAME}/version.build'
OEM_FILE = "/etc/oem/brand.json"
BASE_OEM = {"oem_short": "spacevm",
            "oem": "SpaceVM",
            "oem_full": "SpaceVM",
            "oem_site": "https://spacevm.ru",
            "distribution": "space",
            "company": "«Dacom M» Ltd"}
# Путь к директории, в которой хранятся файлы UEFI
NODE_UEFI_NVRAM_DIR = f"{NODE_PATH}/nvram"

# storage settings
STORAGE_ROOT_DIR = '/storages'
NFS_PATH = f"{STORAGE_ROOT_DIR}/nfs"
GLUSTERFS_PATH = f"{STORAGE_ROOT_DIR}/glusterfs"
CIFS_PATH = f"{STORAGE_ROOT_DIR}/cifs"
SHARED_STORAGES_PATHS = [NFS_PATH, GLUSTERFS_PATH, CIFS_PATH]
OCFS2_PATH = f"{STORAGE_ROOT_DIR}/ocfs2"
GFS2_PATH = f"{STORAGE_ROOT_DIR}/gfs2"
GLUSTER_PATH = f"{STORAGE_ROOT_DIR}/gluster"
CLUSTER_STORAGES_PATHS = [OCFS2_PATH, GFS2_PATH, GLUSTER_PATH]
LOCAL_STORAGES_PATH = f'{STORAGE_ROOT_DIR}/local'
ZPOOL_STORAGES_PATH = f'{STORAGE_ROOT_DIR}/zfs'
VDISK_IO_TYPES = ['io_uring', 'native', 'threads']
PATH_ISO_STORAGE = '_ISO'
PATH_FILE_STORAGE = '_LIBRARY'
PATH_NODES_HEARTBEAT = '_HEARTBEAT'
PATH_MINIO_STORAGE = '_MINIO'
PATH_TMP = '.TMP'

# Redis params names
REDIS_MASTER_STATUS = 'node_master_status'
REDIS_MASTER_IP = 'node_master_ip'
REDIS_NODE_ID = 'node_node_id'


class ListVars:
    @classmethod
    def list(cls) -> List[str]:
        """
        Метод возвращает список переменных класса.
        """

        return [value for key, value in cls.__dict__.items() if not key.startswith('__') and
                not callable(getattr(cls, key))]


class TransportProtocols(ListVars):
    """
    Класс, описывающий доступные для использования транспортные протоколы.
    """

    fc = 'fc'
    iscsi = 'iscsi'
    scsi = 'scsi'


class FENCING_TYPE(ListVars):
    virtual = "VIRTUAL"
    node = "NODE"  # watchdog

    controller_ipmi = "CONTROLLER_IPMI"  # через ipmi с контроллера
    controller_storage = "CONTROLLER_STORAGE"  # через sbd с контроллера

    distributed_ipmi = "DISTRIBUTED_IPMI"  # через ipmi из кворума
    distributed_storage = "DISTRIBUTED_STORAGE"  # через sbd из кворума
    distributed_ipmi_storage = "DISTRIBUTED_IPMI_STORAGE"  # через ipmi or sbd from quorum


class HEARTBEAT_TYPE(ListVars):
    kernel = "KERNEL_CONTROLLER"
    kernel_controller_ipmi = "KERNEL_CONTROLLER_IPMI"  # + cвязь по ipmi
    kernel_controller_storage = "KERNEL_CONTROLLER_STORAGE"  # + связь sbd

    kernel_distributed = "KERNEL_DISTRIBUTED"  # + связь через кворум
    kernel_distributed_storage = "KERNEL_DISTRIBUTED_STORAGE"  # + связь через кворум и sbd


T = TypeVar('T')


def states_response(function: Callable[..., T]) -> Callable[..., T]:
    """
    Decorator wrapper to the formation of the response for functions states.
    """

    function.wrapped = True

    @wraps(function)
    def wrapper(*args, **kwargs) -> dict:
        try:
            comment = function(*args, **kwargs)
            return {'name': function.__name__, 'result': True, 'comment': comment}
        except CalledProcessError as e:
            if e.returncode >= 0:
                desc = e.returncode
            else:
                try:
                    desc = signal.Signals(-e.returncode).name
                except ValueError:
                    desc = f'SIGNAL_{-e.returncode}'
            comment = f'retcode: {desc}, output: {e.output}'
            if 'suppress_err' not in kwargs:
                LOG.error(ERROR_PATTERN.format(function.__name__, comment))
            return {'name': function.__name__, 'result': False, 'comment': comment, 'retcode': e.returncode}
        except TimeoutExpired as e:
            if 'suppress_err' not in kwargs:
                LOG.error(ERROR_PATTERN.format(function.__name__, e))
            return {'name': function.__name__, 'result': False,
                    'comment': f'Command timed out after {e.timeout} seconds'}
        except KeyError as e:
            if 'suppress_err' not in kwargs:
                LOG.error(ERROR_PATTERN.format(function.__name__, e))
            return {'name': function.__name__, 'result': False, 'comment': f'KeyError: {str(e)}'}
        except Exception as e:
            if 'suppress_err' not in kwargs:
                LOG.error(ERROR_PATTERN.format(function.__name__, e))
            return {'name': function.__name__, 'result': False, 'comment': str(e)}

    return wrapper


def states_result(func, base_err_msg: str = 'Error', raise_exception: bool = True, get_err: bool = False, **kwargs):
    func_ret = func(**kwargs)
    if not func_ret.get('result'):
        if raise_exception:
            raise CommandExecutionError(f'{base_err_msg}: {func.__name__} {func_ret.get("comment")}')
        else:
            if get_err:
                return f'{base_err_msg}: {func.__name__} {func_ret.get("comment")}'
            return func_ret.get('result', False)
    return func_ret.get('comment') if not get_err else None


@states_response
# Таймаут 50с, чтобы в случае синхронных команд успевал поставить задачу, выполнить,
# отвалиться по таймауту, и вернуть ответ до nginx timeout 60c
def get_output(cmd, timeout: float = 50, shell: bool = True, universal_newlines: bool = True,
               suppress_err: bool = False, unix_timeout: float = 0, process_signal: int = SIGKILL, **kwargs):
    if unix_timeout > 0 and isinstance(cmd, str):
        cmd = f"timeout -s {process_signal} {unix_timeout} {cmd}"
    return check_output(cmd, stderr=DEVNULL if suppress_err else STDOUT, shell=shell, executable='/bin/bash',
                        timeout=timeout, universal_newlines=universal_newlines)


def get_result(cmd, timeout: int = 50, shell: bool = True, unix_timeout: int = 0, **kwargs):
    if unix_timeout > 0 and isinstance(cmd, str):
        cmd = f"timeout -s 9 {unix_timeout} {cmd}"
    try:
        check_output(cmd, shell=shell, timeout=timeout)
        return True
    except SubprocessError:
        return False


@states_response
def call(cmd, timeout: int = 50, shell: bool = True, unix_timeout: int = 0, **kwargs):
    if unix_timeout > 0 and isinstance(cmd, str):
        cmd = f"timeout -s 9 {unix_timeout} {cmd}"
    check_call(cmd, shell=shell, timeout=timeout)


@states_response
def call_with_args(cmd: list, timeout: int = 50):
    sub_call(cmd, timeout=timeout)


@states_response
def get_output_with_args(cmd: list, **kwargs):
    return check_output(cmd, stderr=STDOUT, universal_newlines=True, **kwargs)


def delete_file(path: str) -> bool:
    shutil.rmtree(path, ignore_errors=True)
    return True


def usr_cmd(cmd, usr):
    return 'su {} -c \'{}\''.format(usr, cmd)


def ionice_cmd(cmd: str, class_priority: int = 3) -> str:
    """
    Программа ionice задаёт класс и приоритет ввода/вывода для процесса.
    1 - real time
    2 - best-effort
    3 - idle
    """
    if class_priority not in {1, 2, 3}:
        raise CommandExecutionError(f"Invalid argument for ionice cmd. Unknown prio class {class_priority}")
    return f"ionice -c {class_priority} {cmd}"


def nice_cmd(cmd):
    return f'nice -n 19 {ionice_cmd(cmd)}'


def delete_file_with_owner(path, owner, **kwargs):
    if kwargs.get('multi_path'):
        cmd = usr_cmd('rm -f {}'.format(path), owner)
    else:
        cmd = usr_cmd('rm -f "{}"'.format(path), owner)
    return get_output(cmd, **kwargs)


def jsonize_redis_value(value) -> dict:
    data = json.loads(value.decode().replace("'", '"').replace('True', 'true').replace('False', 'false')) \
        if value else {}
    return data


def check_service(service_name, **kwargs):
    cmd_ret = get_output(['systemctl', 'status', service_name], shell=False, **kwargs)
    if not cmd_ret['result']:
        LOG.error(cmd_ret['comment'])
        return False
    return "Active: active (running)" in cmd_ret['comment']


def action_service(service_name, action='start', **kwargs):
    # слэши могут прилетать от замены дефисов (см. utils.mount)
    return get_output("sudo systemctl {} {}".format(action, service_name.replace('\\', '\\\\')), **kwargs)


def is_service_active(service_name):
    # если нужно Shell=true, надо экранировать слэши
    return get_result(["systemctl", "is-active", service_name], shell=False)


def is_service_failed(service_name):
    return get_result(["systemctl", "is-failed", service_name], shell=False)


def check_path_exists(path, **kwargs):
    return get_result(cmd=f"test -d '{path}'", **kwargs)


def check_file_exists(path, **kwargs):
    return get_result(cmd=f"test -e '{path}'", **kwargs)


def thread_pool(f, f_args: Iterable) -> list:
    with ThreadPoolExecutor(max_workers=cpu_count() // 2 or 1) as executor:
        return [r for r in executor.map(f, f_args) if r]


def retry_on_exception(ExceptionToCheck, tries=3, delay=2, backoff=2):
    def retry_decorator(f):
        @wraps(f)
        def wrapper(*args, **kwargs):
            mtries, mdelay = tries, delay
            while mtries > 1:
                try:
                    return f(*args, **kwargs)
                except ExceptionToCheck as e:
                    LOG.warning(
                        'Exception {exc} caught during execution of function {func_name} '
                        'with args {args}, kwargs {kwargs}. '
                        'Retrying...'.format(exc=e.__class__.__name__, func_name=f.__name__,
                                             args=json.dumps(args[1:]), kwargs=json.dumps(kwargs))
                    )
                    time.sleep(mdelay)
                    mtries -= 1
                    mdelay *= backoff
            return f(*args, **kwargs)

        return wrapper
    return retry_decorator


def is_uuid(uuid_str):
    """
    Функция проверяет что uuid_str - валидный uuid4
    """
    try:
        v = UUID(uuid_str, version=4)
    except ValueError:
        # non valid
        return False

    return v.hex == uuid_str.replace("-", "")


def resolve_subr_var(var):
    ret = get_output(f". {VEIL_VAR_FILE}; echo ${var}")
    if ret['result']:
        return ret['comment'].rstrip()
    return None


def check_pid(pid):
    """ Check For the existence of a unix pid. """
    try:
        os.kill(pid, 0)
    except OSError:
        return False
    else:
        return True


def create_file(file_path):
    if not os.path.exists(file_path):
        open(file_path, 'a').close()


def execution_time(func):
    @wraps(func)
    def wrapper(*args, **kwargs):
        start = time.time()
        result = func(*args, **kwargs)
        end = time.time()
        LOG.info(f'Function: {func.__name__}\nExecution time: {str(end - start)} sec')
        return result
    return wrapper


def get_config(cluster=False):
    try:
        if os.path.exists(CLUSTER_PATH if cluster else HEARTBEAT_PATH):
            with open(CLUSTER_PATH if cluster else HEARTBEAT_PATH, 'r') as f:
                return json.load(f)
        return {}
    except (IOError, json.JSONDecodeError):
        return {}


@lru_cache()
def get_oem() -> dict:
    data = BASE_OEM
    for oem_file in [VERSION_FILE, OEM_FILE]:
        try:
            if os.path.exists(oem_file):
                with open(oem_file, 'r') as f:
                    data.update(json.load(f))
        except (IOError, json.JSONDecodeError):
            pass
    return data


def distribution():
    return get_oem()['distribution']


def is_se():
    return os.path.exists(VEIL_SE_FILE)


@states_response
def clean_guaranteed(sources, clean_type, clean_count, task, **kwargs):
    """
    Clean list of files
    :param sources: list of sources
    :param clean_type: 'zero' or 'urandom'
    :param clean_count: count or cleaning tasks
    :param task: id of task
    """

    rnd_file_size = 4     # 4Mb
    max_processes = 2
    red = kwargs.get('redis', None)
    progress_data = dict(type='progress', jid=task, id='', name='clean_guaranteed', result=True)
    data = {'progress': '0', 'parser': 'dd'}
    nice_ionice_cmd = "nice -n 10 ionice -c 2 -n 0"
    rnd_path = "/tmp/rnd_file"
    os.nice(19)
    comment = 'Files were cleaned.'
    sources = [sources] if not isinstance(sources, list) else sources
    sources_len = len(sources)
    # create random or zero file
    with open(rnd_path, 'wb+') as rnd_file, open('/dev/{0}'.format(clean_type), 'rb') as dev:
        rnd_file.write(dev.read(rnd_file_size * (1024 ** 2)))
    # Проход по всем sources
    for sources_count, path in enumerate(sources):
        part_file_size = 10  # 10Gb can be change to more
        # Подготовка
        if os.path.exists(path):
            file_size = os.path.getsize(path) / (1024 ** 3)  # Gb
            files_count = int(file_size / part_file_size) if file_size > 0 else 0
            if files_count > max_processes:
                files_count = max_processes
                part_file_size = int(file_size / files_count)
            elif files_count < 1:
                files_count = 1
                part_file_size = 0

            processes = []
            results = [0] * files_count
            result, result_last, percent = 0, 0, 0
            # Общий проход по количеству циклов удаления
            for counter in range(clean_count):
                # Проход по количеству разбиений файла
                for files_counter in range(files_count):
                    offset = files_counter * part_file_size * 1024 ** 3  # bytes
                    if files_counter == files_count - 1:
                        block_clean_count = int((file_size - part_file_size * files_counter) * 1024 / rnd_file_size)
                    else:
                        block_clean_count = part_file_size * 1024 / rnd_file_size
                    if block_clean_count == 0:
                        block_clean_count = 1
                    clean_cmd = "{cmd} dd if={rnd} of={p} seek=$OFFSET bs={bs} count=1 " \
                                "conv=noerror".format(
                                    cmd=nice_ionice_cmd,
                                    rnd=rnd_path,
                                    p=path,
                                    bs='{}M'.format(rnd_file_size))
                    cycle_clean_cmd = "for i in {{0..{count}}}; do let OFFSET=$i*{size}*1024*1024+{off}; {cmd};" \
                                      " done".format(count=block_clean_count - 1,
                                                     size=rnd_file_size,
                                                     off=offset,
                                                     cmd=clean_cmd)
                    user_delete_cmd = "/bin/bash -c '{}'".format(cycle_clean_cmd)
                    processes.append(Popen(user_delete_cmd,
                                           shell=True,
                                           stdout=PIPE,
                                           stderr=PIPE,
                                           executable='/bin/bash'))
                while True:
                    for k, process in enumerate(processes):
                        if process.poll() is not None:
                            results[k] = 1
                    result = sum(results)
                    if result != result_last and result != 0:
                        percent = 100 * ((result + counter * files_count) / clean_count / files_count +
                                         sources_count / sources_len)
                        data.update({'progress': str(int(percent))})
                        progress_data['comment'] = json.dumps(progress_data)
                        if red:
                            red.lpush('node_results', json.dumps(progress_data))

                    if 0 not in results:
                        processes = []
                        results = [0] * files_count
                        result, result_last = 0, 0
                        break

                    result_last = result
            try:
                os.remove(path)
            except OSError:
                pass
            comment += f'File {path} was cleaned. '
        else:
            comment += f'File {path} does not exists. '

    return comment


def timeit(func):
    def timeit_wrapper(*args, **kwargs):
        start_time = time.perf_counter()
        result = func(*args, **kwargs)
        end_time = time.perf_counter()
        total_time = end_time - start_time
        LOG.debug(f'Function {func.__name__} took {total_time:.4f} seconds')
        return result
    return timeit_wrapper
