"""
(c) 2023 «Dacom M» Ltd
"""

from collections import namedtuple
from xml.etree import ElementTree as ET  # nosec


import os
from abc import ABC, abstractmethod
from typing import Any, Union
from veilrpc.utils.bazalt import CommandExecutionError, DOM_XML_EXT, DOMAINS_FOLDER
from veilrpc.utils.domain.user_exceptions import XMLFileError, LibvirtDomainError, SearchPathError
from veilrpc.utils.virt import virt_is_active, virt_get_xml, virt_set_metadata, virt_get_metadata, \
    register_error_handler


class DataOperator(ABC):
    """
    Абстрактный класс для работы с информацией по объекту Domain на Node
    """

    # Шаг для перемещения в xml структуре, где field_name - наименование поля, use_ns - флаг использования
    # зарегистрированных ранее пространств имен, position - позиция для выбора из списка, возвращаемого функций findall
    StepTup = namedtuple('StepTup', ['field_name', 'use_ns', 'position'], defaults=['', False, 0])

    @staticmethod
    def _convert_from_xml(xml_data: str) -> dict:
        """
        XML (str) to dict magic
        Упрощенная структура, без суррогатных ключей типа attrib и children.
        Аттрибуты и вложенные теги (детишки) находятся на одном уровне для каждого тега.
        Только повторяющиеся теги (детишки) собираются в список.
        По идее такой структурой будет удобней пользоваться - меньше итераций по спискам одинаковых тег.
        Один нюанс: если название аттрибута и вложенного тега совпадут, то название тега
        будет изменено на <имя_тега>_tag_<порядковый_номер> что бы не превратить значение аттрибута в
        список значений.
        """

        def rec(node):
            tag_dict = dict()
            node_text = node.text.strip() if isinstance(node.text, str) else ''
            if node_text:
                try:
                    tag_dict['_text_'] = int(node_text)
                except ValueError:
                    if node_text == 'True':
                        tag_dict['_text_'] = True
                    elif node_text == 'False':
                        tag_dict['_text_'] = False
                    else:
                        tag_dict['_text_'] = node_text
            if node.attrib:
                for attr, value in node.attrib.items():
                    try:
                        tag_dict[attr] = int(value)
                    except ValueError:
                        if value == 'True':
                            tag_dict[attr] = True
                        elif value == 'False':
                            tag_dict[attr] = False
                        else:
                            tag_dict[attr] = value
            num = 1
            for child in node:
                try:
                    tag_dict[child.tag].append(rec(child))
                except KeyError:
                    tag_dict[child.tag] = rec(child)
                except AttributeError:
                    if isinstance(tag_dict[child.tag], dict):
                        tag_dict[child.tag] = [tag_dict[child.tag], rec(child)]
                    else:
                        tag_dict[child.tag + '_tag_{}'.format(num)] = rec(child)
                        num += 1
            return tag_dict

        try:
            one_line_xml_str = ' '.join(xml_data.split())
            root = ET.fromstring(one_line_xml_str)
        except Exception:
            return {}

        return {root.tag: rec(root)}

    @staticmethod
    def _get_xml_root(xml_data: str, ns: dict) -> ET.Element:
        for tag in ns.items():
            ET.register_namespace(*tag)
        return ET.fromstring(xml_data)

    @staticmethod
    def _xml_element_iterator(field: StepTup, meta_element: ET.Element, ns: dict) -> Any:
        try:
            return meta_element.findall(path=field.field_name, namespaces=ns if field.use_ns else None)[field.position]
        except IndexError:
            raise SearchPathError(str(field)) from None

    @classmethod
    def _get_field(cls, step_list: list[StepTup], xml_root: ET.Element, ns: dict) -> Any:
        """
        Поиск информации по отдельному полю в элементе дерева.
        :param step_list: путь до поля в N шагов, где каждый имеет формат
        ['имя поля', True|False (поиск по именам ns), выбор элемента из группы (при единичном объекте - 0)]
        :return: искомое поле
        """
        for step in step_list:
            if not isinstance(step, cls.StepTup):
                raise SearchPathError(str(step))
        data = xml_root
        for step in step_list:
            data = cls._xml_element_iterator(step, data, ns)
        return data

    @abstractmethod
    def get_all_data(self, xml_format: bool = True, update_raw_data: bool = False):
        """ Получение всей доступной информации по объекту Domain на Node """

    @abstractmethod
    def get_field_info(self, path: list[StepTup], update_raw_data: bool = False):
        """ Получение информации по отдельному полю """

    @abstractmethod
    def update_one_field(self, path_name: str, new_value: Any, update_raw_data: bool = False):
        """ Обновление информации по одному из полей объекта Domain """

    @abstractmethod
    def create_domain_info(self, xml_data: str):
        """ Создание объекта в памяти для хранения информации по Domain """

    def update_power_state_expected(self, var: str):
        """ Изменение состояния поля power_state_expected """


class LibvirtDataOperator(DataOperator):
    """
    Реализация абстрактного класса DataOperator для работы с информаций по ВМ, хранящейся в libvirt'е.
    """
    def __init__(self, domain_id: str):
        register_error_handler()
        self._domain_uuid = domain_id
        self._libvirt_data = self._get_xml_data()
        self._ns = {'veil': 'http://controller/', 'qemu': 'http://libvirt.org/schemas/domain/qemu/1.0'}
        # обновление полей ограничивается реализованными virt_set_* функциями
        self.available_update_function = {
            'metadata': {'function': virt_set_metadata, 'params_type': str}
        }

    def _get_xml_data(self) -> str:
        xml_data = ''
        if virt_is_active(self._domain_uuid):
            xml_data = virt_get_xml(domain_name=self._domain_uuid)
        return xml_data

    def get_all_data(self, xml_format: bool = True, update_raw_data: bool = False) -> Union[dict, str]:
        """
        Получение всей доступной информации по объекту Domain на Node
        :param xml_format: выводить информацию в виде xml или конвертировать в dict
        :param update_raw_data: обновить данные, полученные при инициализации объекта.
        :return: dict или xml с доступной по Domain информацией
        """
        if not virt_is_active(self._domain_uuid):
            raise LibvirtDomainError(self._domain_uuid)
        if update_raw_data:
            self._libvirt_data = self._get_xml_data()
        return self._libvirt_data if xml_format else self._convert_from_xml(self._libvirt_data)

    def get_field_info(self, path: list[DataOperator.StepTup], update_raw_data: bool = False) -> Any:
        """
        Получение информации по отдельному полю.
        :param path: путь к полю в xml структуре.
        :param update_raw_data: обновить данные, полученные при инициализации объекта.
        :return: значение поля.
        """
        if not virt_is_active(self._domain_uuid):
            raise LibvirtDomainError(self._domain_uuid)
        if update_raw_data:
            self._libvirt_data = self._get_xml_data()
        root = self._get_xml_root(self._libvirt_data, self._ns)
        if root:
            data = self._get_field(path, root, self._ns)
            return data.text if isinstance(data, ET.Element) else None

    def update_one_field(self, path_name: str, new_value: Any, update_raw_data: bool = False) -> None:
        """
        Обновление отдельного поля. Поля, доступные для обновления ограничены реализованными функциями типа virt_set_*.
        Для контроля этого ограничения введен словарь available_update_function, содержащий пути к допустимым для
        обновления параметрам в libvirt'е.
        :param path_name: путь к полю в xml структуре.
        :param new_value: значение поля для обновления.
        :param update_raw_data: обновить данные, полученные при инициализации объекта (не актуален).
        """
        if not virt_is_active(self._domain_uuid):
            raise LibvirtDomainError(self._domain_uuid)
        if path_name in self.available_update_function:
            update_signature = self.available_update_function[path_name]
            if isinstance(new_value, update_signature.get('params_type')):
                # TODO: привести сигнатуры функций работы с параметрами libvirt к единой сигнатуре (в частности
                #  заменить metadata в ф-ии virt_set_metadata на более распространенное именование аргумента).
                update_signature['function'](self._domain_uuid, new_value)
            else:
                raise Exception('Invalid type of data being updated for libvirt data.')
        else:
            raise Exception('The command to update the field at the specified address is not implemented.')

    def create_domain_info(self, xml_data: str):
        """ Создание объекта выполняется на стороне libvirt. """
        raise NotImplementedError

    def update_power_state_expected(self, var: str) -> None:
        """
        Обновление поля power_state_expected в metadata.
        :param var: новое значение поля.
        """
        if not virt_is_active(self._domain_uuid):
            raise LibvirtDomainError(self._domain_uuid)
        # update_raw_data автоматически выполняется для libvirt
        meta_raw = virt_get_metadata(self._domain_uuid)
        metadata = ET.fromstring(meta_raw)
        item = metadata.find('power_state_expected')
        if isinstance(item, ET.Element):
            item.text = var
            self.update_one_field('metadata', ET.tostring(metadata, encoding='UTF-8').decode())

    def update_ha_enabled(self, var: str) -> None:
        """
        Изменение состояния поля ha_enabled
        :param var: новое значение поля.
        """
        raise NotImplementedError


class XMLDataOperator(DataOperator):
    """
    Реализация абстрактного класса DataOperator для работы с информаций по ВМ, хранящейся в виде xml файла.
    """
    def __init__(self, domain_id: str):
        self._domain_uuid = domain_id
        self._file_data = self._get_xml_data()
        self._ns = {'veil': 'http://controller/', 'qemu': 'http://libvirt.org/schemas/domain/qemu/1.0'}
        self.available_update_path = {
            'metadata': {
                'path': [DataOperator.StepTup('metadata', False, 0), DataOperator.StepTup('veil:instance', True, 0)],
                'function': self._update_all_metadata},
            'power_state_expected': {
                'path': [DataOperator.StepTup('metadata', False, 0), DataOperator.StepTup('veil:instance', True, 0),
                         DataOperator.StepTup('veil:power_state_expected', True, 0)],
                'function': self._update_power_state_expected_only},
            '_ha_enabled': {
                'path': [DataOperator.StepTup('metadata', False, 0), DataOperator.StepTup('veil:instance', True, 0),
                         DataOperator.StepTup('veil:_ha_enabled', True, 0)],
                'function': self._update_ha_enabled_only}
        }

    def _get_xml_data(self) -> str:
        xml_data = ''
        if self._is_domain_xml_exists():
            xml_data = self._get_local_domain_xml()
        return xml_data

    def _is_domain_xml_exists(self) -> bool:
        path_temp = f'{DOMAINS_FOLDER}/{self._domain_uuid}'
        if os.path.isfile(f'{path_temp}.xml') or os.path.isfile(f'{path_temp}.{DOM_XML_EXT}'):
            return True
        return False

    def _get_local_domain_xml(self) -> str:
        with open(self._xml_path_by_id()) as f:
            return f.read()

    def _xml_path_by_id(self) -> str:
        path_temp = f'{DOMAINS_FOLDER}/{self._domain_uuid}'
        path = f'{path_temp}.xml'
        if os.path.isfile(path):
            return path
        path = f'{path_temp}.{DOM_XML_EXT}'
        if os.path.isfile(path):
            return path
        raise CommandExecutionError(f'Domains {self._domain_uuid} XML file is not found.')

    def _update_all_metadata(self, data: ET.Element, new_value: str) -> None:
        new_metadata_instance = self._get_xml_root(new_value, self._ns)
        data_tags = {element.tag for element in data.findall('*')}
        extend_fields = []
        for element in new_metadata_instance.findall('*'):
            element.tag = f"{{{self._ns['veil']}}}{element.tag}"
            if element.tag in data_tags:
                updated_field = data.find(element.tag)
                if isinstance(updated_field, ET.Element):
                    updated_field.text = element.text
            else:
                extend_fields.append(element)
        if extend_fields:
            data.extend(extend_fields)

    def _update_power_state_expected_only(self, data: ET.Element, new_value: str) -> None:
        data.text = new_value

    def _update_ha_enabled_only(self, data: ET.Element, new_value: str) -> None:
        data.text = new_value

    def get_all_data(self, xml_format: bool = True, update_raw_data: bool = False) -> Union[dict, str]:
        """
        Получение всей доступной информации по объекту Domain на Node
        :param xml_format: выводить информацию в виде xml или конвертировать в dict
        :param update_raw_data: обновить данные, полученные при инициализации объекта.
        :return: dict или xml с доступной по Domain информацией
        """
        if update_raw_data:
            self._file_data = self._get_xml_data()
        if not self._is_domain_xml_exists():
            raise XMLFileError(self._domain_uuid)
        return self._file_data if xml_format else self._convert_from_xml(self._file_data)

    def get_field_info(self, path: list[DataOperator.StepTup], update_raw_data: bool = False) -> Any:
        """
        Получение информации по отдельному полю.
        :param path: путь к полю в xml структуре.
        :param update_raw_data: обновить данные, полученные при инициализации объекта.
        :return: значение поля.
        """
        if update_raw_data:
            self._file_data = self._get_xml_data()
        if not self._is_domain_xml_exists():
            raise XMLFileError(self._domain_uuid)
        root = self._get_xml_root(self._file_data, self._ns)
        if root:
            data = self._get_field(path, root, self._ns)
            return data.text if isinstance(data, ET.Element) else None

    def update_one_field(self, path_name: str, new_value: Any, update_raw_data: bool = False) -> None:
        """
        Обновление отдельного поля.
        :param path_name: путь к полю в xml структуре.
        :param new_value: значение поля для обновления
        :param update_raw_data: обновить данные, полученные при инициализации объекта.
        """
        if not self._is_domain_xml_exists():
            raise XMLFileError(self._domain_uuid)
        if update_raw_data:
            self._file_data = self._get_xml_data()
        if self._file_data:
            root = self._get_xml_root(self._file_data, self._ns)
            if root:
                if not self._is_domain_xml_exists():
                    raise XMLFileError(self._domain_uuid)
                data = self._get_field(self.available_update_path[path_name]['path'], root, self._ns)
                if isinstance(data, ET.Element):
                    self.available_update_path[path_name]['function'](data, new_value)
                    with open(self._xml_path_by_id(), 'w') as f:
                        f.write(ET.tostring(root, encoding='UTF-8').decode())
        else:
            raise XMLFileError(self._domain_uuid)

    def create_domain_info(self, xml_data: str) -> None:
        """
        Создание/перезапись xml файла.
        :param xml_data: xml строка.
        """
        try:
            root = self._get_xml_root(xml_data, self._ns)
            xml_path = self._xml_path_by_id() if self._is_domain_xml_exists() else \
                f'{DOMAINS_FOLDER}/{self._domain_uuid}.xml'
            with open(xml_path, 'w') as f:
                f.write(ET.tostring(root, encoding='UTF-8').decode())
        except Exception as err:
            raise err
        else:
            self._file_data = self._get_xml_data()

    def update_power_state_expected(self, var: str) -> None:
        """
        Изменение состояния поля power_state_expected
        :param var: новое значение поля.
        """
        self.update_one_field('power_state_expected', var)

    def update_ha_enabled(self, var: str) -> None:
        """
        Изменение состояния поля ha_enabled
        :param var: новое значение поля.
        """
        self.update_one_field('_ha_enabled', var)
