"""
(c) 2023 «Dacom M» Ltd
"""

from fcntl import LOCK_EX
from fcntl import LOCK_NB
from fcntl import flock
from functools import wraps
import os
from time import sleep
import secrets

NO_BLOCK = 'nb'
BLOCK = 'block'
RETRY = 'retry'


class UnableToLock(Exception):
    pass


class InvalidMode(Exception):
    pass


def lock(lock_file: str, mode: str = RETRY, retries: int = 30, max_timeout: int = 3):
    """Use flock sys-call to protect an action from multiple concurrent calls
    This decorator will attempt to get an exclusive lock on the specified file
    by using the flock system call.  As long as all callers of a protected action
    use this decorator with the same lockfile, only 1 caller will be able to
    execute at a time, all others will fail, or will be blocked.
    Usage:
        @lock('/var/run/protected.lock')
        def protected():
            # do some potentially unsafe actions
        # wait indefinitely for the lock
        @lock('/var/run/protected.lock', mode=BLOCK)
        def protected():
            # do some potentially unsafe actions

        # If the initial lock failed retry 10 more times.
        @lock('/var/run/protected.lock', mode=RETRY, retries=10)
        def protected():
            # do some potentially unsafe actions
    :param lock_file: full path to a file that will be used as a lock.
    :param mode: how should we run this? (BLOCK, NO_BLOCK, RETRY)
    :param retries: If the initial lock failed, how many more times to retry.
    :param max_timeout: How long(seconds) max to wait before retrying the lock
    """

    def decorator(target):
        @wraps(target)
        def wrapper(*args, **kwargs):
            # touch the file to create it. (not necessarily needed.)
            # will raise IOError if permission denied.
            if not (os.path.exists(lock_file) and os.path.isfile(lock_file)):
                f = open(lock_file, 'a').close()

            operation = LOCK_EX
            if mode in [NO_BLOCK, RETRY]:
                operation = operation | LOCK_NB

            f = open(lock_file, 'a')
            if mode in [BLOCK, NO_BLOCK]:
                try:
                    flock(f, operation)
                except IOError:
                    raise UnableToLock('Unable to get exclusive lock.') from None

            elif mode == RETRY:
                for i in range(0, retries + 1):
                    try:
                        flock(f, operation)
                        break
                    except IOError:
                        if i == retries:
                            raise UnableToLock('Unable to get exclusive lock.') from None
                        sleep(secrets.SystemRandom().uniform(0, max_timeout))

            else:
                raise InvalidMode(f'{mode} is not a valid mode.')

            # Execute the target
            result = target(*args, **kwargs)
            # Release the lock by closing the file
            f.close()
            return result

        return wrapper

    return decorator
