"""
(c) 2023 «Dacom M» Ltd

ChannelStatus:
  IDLE: The channel is idle.
  CONNECTING: The channel is connecting.
  READY: The channel is ready to conduct RPCs.
  TRANSIENT_FAILURE: The channel has seen a failure from which it expects to recover.
  SHUTDOWN: The channel has seen a failure from which it cannot recover.
"""

from contextlib import contextmanager
from google.protobuf.json_format import ParseError
from google.protobuf.timestamp_pb2 import Timestamp
from grpc import insecure_channel, secure_channel, ssl_channel_credentials, RpcError, FutureTimeoutError
import json
import logging
import os

from veilrpc.utils import base_pb2_grpc
from veilrpc.utils import base_pb2


LOG = logging.getLogger('bazalt')

GRPC_MAX_MESSAGE_LENGTH = 10 * 1024 * 1024  # 10m
MODULE_PATTERN = 'veilrpc.states.{}'
SERVER_PORT = 46003


class GRPCResponseType:
    heartbeat = 'heartbeat'
    task = 'task'
    progress = 'progress'
    cli = 'cli'
    event = 'event'


class GRPCJobStatus:
    success = 'SUCCESS'
    failed = 'FAILED'
    not_executed = 'NOT_EXECUTED'


class CommandExecutionError(Exception):
    pass


SERVICE_CONFIG = json.dumps(
    {
        "methodConfig": [
            {
                "name": [
                    {"service": "base.Base"},
                ],
                "retryPolicy": {
                    "maxAttempts": 4,
                    "initialBackoff": "0.1s",
                    "maxBackoff": "1s",
                    "backoffMultiplier": 2,
                    "retryableStatusCodes": ["UNAVAILABLE", "RESOURCE_EXHAUSTED", "DEADLINE_EXCEEDED"],
                },
            }
        ]
    }
)
ChannelOptions = [("grpc.service_config", SERVICE_CONFIG),
                  ("grpc.max_send_message_length", GRPC_MAX_MESSAGE_LENGTH),
                  ("grpc.max_receive_message_length", GRPC_MAX_MESSAGE_LENGTH)]
ChannelStatus = {'0': 'IDLE',
                 '1': 'CONNECTING',
                 '2': 'READY',
                 '3': 'TRANSIENT_FAILURE',
                 '4': 'SHUTDOWN'}


def parse_command_args(fargs, module, function):
    veil_args = json.dumps(fargs if fargs else None)
    return base_pb2.CommandArgs(module=module, function=function, args=veil_args)


def parse_kill_args(jid):
    return base_pb2.KillTaskArgs(jid=jid)


def parse_check_tasks_args(jids):
    msg = base_pb2.CheckArgs()
    msg.jids.extend(list(map(str, jids)))
    return msg


def parse_clear_cache_args(jids, cache_jids, ts):
    timestamp = Timestamp()
    timestamp.FromSeconds(int(ts))
    msg = base_pb2.ClearCacheArgs(ts=timestamp)
    msg.jids.extend(list(map(str, jids)))
    msg.cache_jids.extend(list(map(str, cache_jids)))
    return msg


def parse_check_request_args():
    return base_pb2.CheckRequestResponse(check=True)


def parse_reload_queues_args():
    return base_pb2.ReloadQueuesResponse(reload=True)


def parse_task_args(fargs, jid, node_id, module, function):
    veil_args = json.dumps(fargs if fargs else None)
    return base_pb2.TaskArgs(jid=jid, id=node_id, module=module, function=function, args=veil_args)


def parse_command_data(resp):
    return dict(name=resp.name, result=resp.result, comment=json.loads(resp.comment))


def parse_command_stream_data(resp):
    return dict(content=resp.content, result=resp.result, comment=resp.comment)


def parse_task_data(resp):
    return dict(jid=resp.jid, name=resp.name, result=resp.result, comment=resp.comment)


def parse_task_result(resp):
    result = dict(jid=resp.jid, node=resp.id)
    try:
        result['response'] = json.loads(resp.comment)
    except json.JSONDecodeError as e:
        LOG.error(f'Error in decoding response from node: {resp.id}: {e}')
        result['response'] = ''
    if resp.result:
        result['status'] = GRPCJobStatus.success
    else:
        result['status'] = GRPCJobStatus.failed
        result['errors'] = result['response']
    return result


def parse_progress_result(resp):
    try:
        comment = json.loads(resp.comment)
        return dict(task_id=resp.jid, progress=comment['progress'], parser=comment['parser'],
                    total_steps=comment['total_steps'], step=comment['step'])
    except (json.JSONDecodeError, KeyError) as e:
        LOG.error(f'Error in decoding progress response from node: {resp.id}: {e}')
    return None


def initial_stream_request(stub, ctr_ip):
    """
    Открывает стрим на узел с передачей своего management ip

        previous_connectivity_state = channel._channel.check_connectivity_state(True)
    # if previous_connectivity_state != 2:
    #     LOG.debug('Previous connection state: {}. Trying request to node {}.'.
    #               format(ChannelStatus[str(previous_connectivity_state)], node_uuid))
    """

    return stub.result_stream(base_pb2.NodeArgs(msg=ctr_ip), timeout=25)


def credentials(crt_path=None, crt=None):
    if crt:
        return ssl_channel_credentials(root_certificates=crt)
    elif crt_path:
        if os.path.exists(crt_path):
            with open(crt_path, 'rb') as f:
                crt = f.read()
            return ssl_channel_credentials(root_certificates=crt)
    return None


def get_channel_base(node_uuid, base_path, filename=None):
    cred = credentials(crt_path=os.path.join(base_path, filename or f'{node_uuid}.crt'))
    if cred:
        return secure_channel(f'{node_uuid}:{SERVER_PORT}', cred, options=ChannelOptions)
    else:
        return insecure_channel(f'{node_uuid}:{SERVER_PORT}', options=ChannelOptions)


@contextmanager
def get_channel(node_uuid, base_path, filename=None):
    cred = credentials(crt_path=os.path.join(base_path, filename or f'{node_uuid}.crt'))
    if cred:
        channel = secure_channel(f'{node_uuid}:{SERVER_PORT}', cred, options=ChannelOptions)
    else:
        channel = insecure_channel(f'{node_uuid}:{SERVER_PORT}', options=ChannelOptions)
    yield channel
    channel.close()
    del channel


@contextmanager
def get_channel_crt(node_uuid, crt):
    cred = credentials(crt=crt)
    if cred:
        channel = secure_channel(f'{node_uuid}:{SERVER_PORT}', cred, options=ChannelOptions)
    else:
        channel = insecure_channel(f'{node_uuid}:{SERVER_PORT}', options=ChannelOptions)
    yield channel
    channel.close()
    del channel


def get_stub_from_channel(channel):
    return base_pb2_grpc.BaseStub(channel)


class RpcCommandCrt:
    """Client for accessing the gRPC sync functionality by grpc certificate."""

    @staticmethod
    def apply(target: str, module: str, function: str, crt: str,
              fargs: dict = None, timeout: int = 55):
        with get_channel_crt(node_uuid=target, crt=crt.encode()) as channel:
            try:
                stub = get_stub_from_channel(channel)
                msg = parse_command_args(fargs, module, function)
                response = stub.apply(request=msg, timeout=timeout)
                return parse_command_data(response)
            except ParseError as e:
                msg = f"ParseError: {e}"
            except RpcError as e:
                msg = f"RpcError from function apply for target {target}. " \
                      f"status_code: {e.code().name}, details: {e.details()}"
            except Exception as e:
                msg = f"Unknown error: {e}"
            LOG.error(msg)
            return dict(result=False, comment=msg)

    @staticmethod
    def multi_apply(targets, module, function, fargs=None, timeout=55):

        results = {}
        error = False
        try:
            msg = parse_command_args(fargs, module, function)
            futures = dict()
            for target, crt in targets:
                channel = get_channel_crt(node_uuid=target, crt=crt.encode())
                stub = get_stub_from_channel(channel)
                futures[target] = stub.apply.future(request=msg), channel
            for target, s in futures.items():
                ft, ch = s
                result = ft.result(timeout=timeout)
                ch.close()
                results[target] = parse_command_data(result)
        except ParseError as e:
            msg = f"ParseError: {e}"
            error = True
        except FutureTimeoutError as e:
            msg = f"Timeout in applying task: {e}"
            error = True
        except Exception as e:
            msg = f"Unknown rpc error: {e}"
            error = True
        if error:
            LOG.error(msg)
            results = {target: dict(result=False, comment=msg) for target in targets}
        return results
