"""
(c) 2023 «Dacom M» Ltd
"""

from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import serialization, hashes
from cryptography.hazmat.primitives.asymmetric import padding
from datetime import datetime
import json
from glob import glob
import os
from redis import RedisError

from veilrpc.utils.bazalt import CONTROLLER_OTHER_PATH, CONTROLLER_PATH

SERIAL_KEY = f'{CONTROLLER_PATH}/serial.key'  # built-in serial key path
SERIAL_KEY_UUID = 'b3f2eada-1287-4f47-ad42-11ce3430839e'  # permanent in SERIAL_KEY
SERIAL_PEM = f"{CONTROLLER_OTHER_PATH}/controller.pem"
SERIAL_2_PEM = f"{CONTROLLER_OTHER_PATH}/controller_2.pem"
REDIS_LICENSE_KEY = 'CONTROLLER_LICENSE_KEY'
LICENSE_EXPTIME = 86400
UnlicenseData = {"verbose_name": "Unlicensed",
                 "nodes_limit": 0,
                 "expired_time": 0,
                 'expired': False,
                 'support_expired_time': 0,
                 'support_expired': False}


def get_license_params(unlicense_data=None, key_path: str = SERIAL_KEY, **kwargs) -> dict:
    """
    проверяется ключ из файла.
    """
    redis = kwargs.get('redis', None)
    try:
        if redis and redis.exists(f"{REDIS_LICENSE_KEY}_{key_path}"):
            data = json.loads(redis.get(f"{REDIS_LICENSE_KEY}_{key_path}").decode())
            if data:
                data['built_in'] = is_built_in(data)
                return data
    except (RedisError, json.JSONDecodeError, UnicodeError):
        pass
    if unlicense_data is None:
        unlicense_data = UnlicenseData
    filename = os.path.basename(key_path)
    unlicense_data["filename"] = filename
    try:
        private_key, private_key_2 = None, None
        if os.path.exists(SERIAL_PEM):
            with open(SERIAL_PEM, "rb") as key_file:
                private_key = serialization.load_pem_private_key(
                    key_file.read(),
                    password=None,
                    backend=default_backend())
        if os.path.exists(SERIAL_2_PEM):
            with open(SERIAL_2_PEM, "rb") as key_file_2:
                private_key_2 = serialization.load_pem_private_key(
                    key_file_2.read(),
                    password=None,
                    backend=default_backend())
        if not private_key and not private_key_2:
            return unlicense_data

        if os.path.exists(key_path):
            with open(key_path, "rb") as f:
                data = f.read()
                if private_key:
                    try:
                        decrypted_data_bytes = private_key.decrypt(
                            data,
                            padding.OAEP(
                                mgf=padding.MGF1(algorithm=hashes.SHA256()),
                                algorithm=hashes.SHA256(),
                                label=None))
                        decrypted_data = json.loads(decrypted_data_bytes)
                        decrypted_data["filename"] = filename
                        decrypted_data['built_in'] = is_built_in(decrypted_data)
                        if redis:
                            try:
                                redis.setex(name=f"{REDIS_LICENSE_KEY}_{key_path}",
                                            value=json.dumps(decrypted_data),
                                            time=LICENSE_EXPTIME)
                            except RedisError:
                                pass
                        return decrypted_data
                    except ValueError:
                        pass
                if private_key_2:
                    try:
                        decrypted_data_bytes = private_key_2.decrypt(
                            data,
                            padding.OAEP(
                                mgf=padding.MGF1(algorithm=hashes.SHA256()),
                                algorithm=hashes.SHA256(),
                                label=None))
                        decrypted_data = json.loads(decrypted_data_bytes)
                        decrypted_data["filename"] = filename
                        decrypted_data['built_in'] = is_built_in(decrypted_data)
                        if redis:
                            try:
                                redis.setex(name=f"{REDIS_LICENSE_KEY}_{key_path}",
                                            value=json.dumps(decrypted_data),
                                            time=LICENSE_EXPTIME)
                            except RedisError:
                                pass
                        return decrypted_data
                    except ValueError:
                        pass
            return unlicense_data
        else:
            return unlicense_data
    except (json.JSONDecodeError, UnicodeError, OSError):
        return unlicense_data


def get_licenses_params(unlicense_data=None, **kwargs) -> list:
    licenses = []
    for path in glob(f"{CONTROLLER_PATH}/*.key"):
        licenses.append(get_license_params(unlicense_data, path, **kwargs))
    return licenses


def get_merged_licenses_params(**kwargs) -> dict:
    redis = kwargs.get('redis', None)
    try:
        if redis and redis.exists(REDIS_LICENSE_KEY):
            data = json.loads(redis.get(REDIS_LICENSE_KEY).decode())
            if data:
                return data
    except (RedisError, json.JSONDecodeError, UnicodeError):
        pass
    merged_license = {}
    licenses = get_licenses_params(**kwargs)
    for license in licenses:  # pylint: disable=redefined-builtin
        license_uuid = license.get('uuid', '')
        merged_license_uuid = merged_license.get('uuid', '')
        if merged_license_uuid and license_uuid and license_uuid in merged_license_uuid:
            continue
        for k, v in license.items():
            if k not in merged_license:
                if k in ['expired_time', 'support_expired_time'] and v == '0':
                    v = None
                merged_license[k] = v
            else:
                if k == 'nodes_limit':
                    if isinstance(merged_license[k], int) and isinstance(v, int):
                        merged_license[k] += v
                elif k in ['expired', 'support_expired']:
                    if not v:
                        merged_license[k] = v
                elif k == 'expired_time':
                    if v and merged_license.get('expired', False):
                        if not merged_license[k] or v > merged_license[k]:
                            if v == '0':
                                v = None
                            merged_license[k] = v
                elif k == 'support_expired_time':
                    if v and merged_license.get('support_expired', False):
                        if not merged_license[k] or v > merged_license[k]:
                            if v == '0':
                                v = None
                            merged_license[k] = v
                else:
                    if not v or not merged_license[k] or not isinstance(merged_license[k], str):
                        merged_license[k] = v
                    elif v not in merged_license[k]:
                        merged_license[k] += f',{v}'
    if redis:
        try:
            redis.setex(name=REDIS_LICENSE_KEY,
                        value=json.dumps(merged_license),
                        time=LICENSE_EXPTIME)
        except (RedisError, json.JSONDecodeError):
            pass
    return merged_license


def get_nodes_count(**kwargs) -> int:
    """
    :return: количество узлов, доступных для создания
    """
    try:
        return int(get_merged_licenses_params(**kwargs).get('nodes_limit', 0))
    except ValueError:
        return 0


def get_expired_time(**kwargs):
    """
    :return: время, до которого будет работать функционал veil
    """

    return get_merged_licenses_params(**kwargs).get('expired_time', None)


def remaining_time(expired_time):
    """
    Возвращает кол-во оставшихся дней до expired_time
    """
    if expired_time:
        try:
            exp_time = datetime.strptime(expired_time, "%Y-%m-%d").date()
            now = datetime.strptime(datetime.utcnow().strftime("%Y-%m-%d"), "%Y-%m-%d").date()
            delta = exp_time - now
            days = delta.days
            return days if days >= 0 else 0
        except ValueError:
            return 0
    else:
        return 0


def get_expiration(params, support=False):
    expired = params.get('expired' if not support else 'support_expired')
    remaining_expired = remaining_time(params.get('expired_time' if not support else 'support_expired_time'))
    if not expired:
        return True
    if expired and remaining_expired > 0:
        return True
    return False


def get_activated(**kwargs) -> bool:
    """
    Проверяет nodes_limit and expired
    """
    params = get_merged_licenses_params(**kwargs)
    return True if params['nodes_limit'] and get_expiration(params) else False


def get_supported(**kwargs) -> bool:
    """
    Проверяет nodes_limit and support_expired
    """

    params = get_merged_licenses_params(**kwargs)
    return True if params['nodes_limit'] and get_expiration(params, support=True) else False


def remove_license(filename, **kwargs) -> str:
    """
    Удаляет файл лицензии, если есть
    """
    key_path = f'{CONTROLLER_PATH}/{filename}'
    if key_path != SERIAL_KEY:
        if filename == 'all.key':  # For tests only
            for path in glob(f"{CONTROLLER_PATH}/*.key"):
                os.remove(path)
        elif os.path.exists(key_path):
            os.remove(key_path)
        redis = kwargs.get('redis', None)
        if redis:
            redis.delete(REDIS_LICENSE_KEY)
        return f'Successfully remove {filename}'


def is_built_in(key_data) -> bool:
    """ Если лицензия встроенная, вернет True. """
    return key_data.get('uuid', '') == SERIAL_KEY_UUID
