"""
(c) 2023 «Dacom M» Ltd
"""

import grp
import os
import pwd
import re
import six
from time import monotonic, sleep
from veilrpc.utils.bazalt import CommandExecutionError, get_output, usr_cmd, call, action_service, is_service_active
from veilrpc.utils.file import mkdir


def _format_info(data):
    """
    Return formatted information in a pretty way.
    """
    return {'name': data.gr_name,
            'passwd': data.gr_passwd,
            'gid': data.gr_gid,
            'members': data.gr_mem}


def user_info(name):
    """
    Return user information
    """
    try:
        data = pwd.getpwnam(name)
    except KeyError:
        return {}
    else:
        return _format_info(data)


def group_info(name):
    """
    Return information about a group
    """
    try:
        grinfo = grp.getgrnam(name)
    except KeyError:
        return {}
    else:
        return _format_info(grinfo)


def _resolve_user_group_names(opts):
    """
    Resolve user and group names in related opts
    """
    name_id_opts = {'uid': user_info,
                    'gid': group_info}
    for ind, opt in enumerate(opts):
        if opt.split('=')[0] in name_id_opts:
            _givenid = opt.split('=')[1]
            _param = opt.split('=')[0]
            _id = _givenid
            if not re.match('[0-9]+$', _givenid):
                _info = name_id_opts[_param](_givenid)
                if _info and _param in _info:
                    _id = _info[_param]
            opts[ind] = _param + '=' + str(_id)
    return opts


def _list_mounts():
    ret = {}
    mounts = get_output('mount -l')

    for line in mounts['comment'].split('\n'):
        comps = re.sub(r"\s+", " ", line).split()
        if len(comps) >= 3:
            ret[comps[2]] = comps[0]
    return ret


def blkid(device=None):
    """
    Return block device attributes: UUID, LABEL, etc. This function only works
    on systems where blkid is available.
    """
    args = ""
    if device:
        args = " " + device

    ret = {}
    blkid_result = get_output('blkid' + args, timeout=300)
    if not blkid_result['result']:
        return ret

    for line in blkid_result['comment'].splitlines():
        if not line:
            continue
        comps = line.split()
        device = comps[0][:-1]
        info = {}
        device_attributes = re.split('\"*\"', line.partition(' ')[2])
        for key, value in zip(*[iter(device_attributes)] * 2):
            key = key.strip('=').strip(' ')
            info[key] = value.strip('"')
        ret[device] = info

    return ret


def _active_mountinfo(ret):
    _list = _list_mounts()
    filename = '/proc/self/mountinfo'
    if not os.access(filename, os.R_OK):
        raise CommandExecutionError(f'File not readable {filename}')

    blkid_info = blkid()

    with open(filename) as ifile:
        for line in ifile:
            comps = line.split()
            device = comps[2].split(':')
            # each line can have any number of
            # optional parameters, we use the
            # location of the separator field to
            # determine the location of the elements
            # after it.
            _sep = comps.index('-')
            device_name = comps[_sep + 2]
            device_uuid = None
            device_label = None
            if device_name:
                device_uuid = blkid_info.get(device_name, {}).get('UUID')
                device_uuid = device_uuid and device_uuid.lower()
                device_label = blkid_info.get(device_name, {}).get('LABEL')
            ret[comps[4]] = {'mountid': comps[0],
                             'parentid': comps[1],
                             'major': device[0],
                             'minor': device[1],
                             'root': comps[3],
                             'opts': _resolve_user_group_names(comps[5].split(',')),
                             'fstype': comps[_sep + 1],
                             'device': device_name,
                             'alt_device': _list.get(comps[4], None),
                             'superopts': _resolve_user_group_names(comps[_sep + 3].split(',')),
                             'device_uuid': device_uuid,
                             'device_label': device_label}
    return ret


def _active_mounts(ret):
    """
    List active mounts on Linux systems
    """
    _list = _list_mounts()
    filename = '/proc/self/mounts'
    if not os.access(filename, os.R_OK):
        raise CommandExecutionError(f'File not readable {filename}')

    with open(filename) as ifile:
        for line in ifile:
            comps = line.split()
            ret[comps[1]] = {'device': comps[0],
                             'alt_device': _list.get(comps[1], None),
                             'fstype': comps[2],
                             'opts': _resolve_user_group_names(comps[3].split(','))}
    return ret


def active(extended=False) -> dict:
    """
    List the active mounts.
    """
    ret = {}

    if extended:
        try:
            _active_mountinfo(ret)
        except CommandExecutionError:
            _active_mounts(ret)
    else:
        _active_mounts(ret)
    return ret


def umount(name, device=None, lazy=False):
    """
    Attempt to unmount a device by specifying the directory it is mounted on
    """
    if name not in active():
        return dict(name='umount', result=True, comment=f"{name} does not have anything mounted")
    # так как у нас монтирование с флагом x-systemd.automount, при простом umount ресурс может замонтироваться снова
    systemd_name = name.replace("-", r"\x2d").lstrip('/').replace('/', '-') + '.automount'
    if is_service_active(systemd_name):
        call(['systemctl', 'stop', systemd_name], shell=False)

    if name in active():
        count = 5
        while (not (ret := get_output(f'umount {device or name}', timeout=300))['result']) \
                and ((count := count - 1) > 0):
            if 'device is busy' in ret['comment'] or 'target is busy' in ret['comment']:
                if lazy:
                    ret = get_output(f'umount --lazy {device or name}', timeout=300)
                    break
                else:
                    sleep(1)
    else:
        ret = dict(name="umount", result=True, comment='')
    return ret


class _fstab_entry:
    """
    Utility class for manipulating fstab entries. Primarily we're parsing,
    formatting, and comparing lines. Parsing emits dicts expected from
    fstab() or raises a ValueError.

    Note: We'll probably want to use os.normpath and os.normcase on 'name'
    """

    class ParseError(ValueError):
        """Error raised when a line isn't parsible as an fstab entry"""

    fstab_keys = ('device', 'name', 'fstype', 'opts', 'dump', 'pass_num')

    # preserve data format
    compatibility_keys = ('device', 'name', 'fstype', 'opts', 'dump', 'pass')

    fstab_format = '{device}\t\t{name}\t{fstype}\t{opts}\t{dump} {pass_num}\n'

    @classmethod
    def dict_from_line(cls, line, keys=fstab_keys):
        if len(keys) != 6:
            raise ValueError(f'Invalid key array: {keys}')
        if line.startswith('#'):
            raise cls.ParseError("Comment!")

        comps = line.split()
        if len(comps) != 6:
            raise cls.ParseError("Invalid Entry!")

        return dict(zip(keys, comps))

    @classmethod
    def from_line(cls, *args, **kwargs):
        return cls(**cls.dict_from_line(*args, **kwargs))

    @classmethod
    def dict_to_line(cls, entry):
        return cls.fstab_format.format(**entry)

    def __str__(self):
        """string value, only works for full repr"""
        return self.dict_to_line(self.criteria)

    def __repr__(self):
        """always works"""
        return str(self.criteria)

    def pick(self, keys):
        """returns an instance with just those keys"""
        subset = dict([(key, self.criteria[key]) for key in keys])
        return self.__class__(**subset)

    def __init__(self, **criteria):
        """Store non-empty, non-null values to use as filter"""
        items = [key_value for key_value in six.iteritems(criteria) if key_value[1] is not None]
        items = [(key_value1[0], str(key_value1[1])) for key_value1 in items]
        self.criteria = dict(items)

    @staticmethod
    def norm_path(path):
        """Resolve equivalent paths equivalently"""
        return os.path.normcase(os.path.normpath(path))

    def match(self, line):
        """compare potentially partial criteria against line"""
        entry = self.dict_from_line(line)
        for key, value in six.iteritems(self.criteria):
            if entry[key] != value:
                return False
        return True


def fstab(config='/etc/fstab') -> dict:
    """
    List the contents of the fstab
    """
    ret = {}
    if not os.path.isfile(config):
        return ret
    with open(config) as ifile:
        for line in ifile:
            try:
                entry = _fstab_entry.dict_from_line(
                    line,
                    _fstab_entry.compatibility_keys)

                entry['opts'] = entry['opts'].split(',')
                while entry['name'] in ret:
                    entry['name'] += '_'

                ret[entry.pop('name')] = entry
            except _fstab_entry.ParseError:
                pass

    return ret


def rm_fstab(name, device, config='/etc/fstab'):
    """
    Remove the mount point from the fstab
    """
    modified = False
    criteria = _fstab_entry(name=name, device=device)
    lines = []

    try:
        with open(config, 'r') as ifile:
            for line in ifile:
                try:
                    if criteria.match(line):
                        modified = True
                    else:
                        lines.append(line)

                except _fstab_entry.ParseError:
                    lines.append(line)

    except (IOError, OSError) as exc:
        return f"Couldn't read from {config}: {str(exc)}"

    if modified:
        try:
            with open(config, 'w+') as ofile:
                ofile.writelines(lines)
        except (IOError, OSError) as exc:
            return f"Couldn't write to {config}: {str(exc)}"

    return True


def set_fstab(
        name,
        device,
        fstype,
        opts='defaults',
        dump=0,
        pass_num=0,
        config='/etc/fstab',
        test=False,
        match_on='auto',
        **kwargs):
    """
    Verify that this mount is represented in the fstab, change the mount
    to match the data passed, or add the mount if it is not present.

    Example^ set_fstab(name=/mnt/foo, device=/dev/sdz1, fstype=ext4)
    """

    # Fix the opts type if it is a list
    if isinstance(opts, list):
        opts = ','.join(opts)

    # preserve arguments for updating
    entry_args = {
        'name': name,
        'device': device,
        'fstype': fstype,
        'opts': opts,
        'dump': dump,
        'pass_num': pass_num,
    }

    lines = []
    ret = None

    # Transform match_on into list--items will be checked later
    if isinstance(match_on, list):
        pass
    elif not isinstance(match_on, six.string_types):
        msg = 'match_on must be a string or list of strings'
        raise CommandExecutionError(msg)
    elif match_on == 'auto':
        # Try to guess right criteria for auto....
        # NOTE: missing some special fstypes here
        specialFSes = frozenset([
            'none',
            'tmpfs',
            'sysfs',
            'proc',
            'fusectl',
            'debugfs',
            'securityfs',
            'devtmpfs',
            'cgroup',
            'btrfs'])

        if fstype in specialFSes:
            match_on = ['name']
        else:
            match_on = ['device']
    else:
        match_on = [match_on]

    # generate entry and criteria objects, handle invalid keys in match_on
    entry = _fstab_entry(**entry_args)
    try:
        criteria = entry.pick(match_on)

    except KeyError:
        def filter_fn(key):
            return key not in _fstab_entry.fstab_keys
        invalid_keys = filter(filter_fn, match_on)

        msg = f'Unrecognized keys in match_on: "{invalid_keys}"'
        raise CommandExecutionError(msg) from None

    # parse file, use ret to cache status
    if not os.path.isfile(config):
        raise CommandExecutionError(f'Bad config file "{config}"')

    try:
        with open(config, 'r') as ifile:
            for line in ifile:
                try:
                    if criteria.match(line):
                        # Note: If ret isn't None here,
                        # we've matched multiple lines
                        ret = 'present'
                        if entry.match(line):
                            lines.append(line)
                        else:
                            ret = 'change'
                            lines.append(str(entry))
                    else:
                        lines.append(line)

                except _fstab_entry.ParseError:
                    lines.append(line)

    except (IOError, OSError) as exc:
        raise CommandExecutionError(f'Couldn\'t read from {config}: {str(exc)}') from None

    # add line if not present or changed
    if ret is None:
        lines.append(str(entry))
        ret = 'new'

    if ret != 'present':  # ret in ['new', 'change']:
        try:
            with open(config, 'w+') as ofile:
                # The line was changed, commit it!
                ofile.writelines(lines)
        except (IOError, OSError) as err:
            raise CommandExecutionError(f'File not writable {config}') from err

    return ret


def mount(name, device, mkmnt=False, fstype='', opts='defaults', user=None, systemd=False):
    """
    Mounts a device as transient systemd .mount unit.
    """
    if isinstance(opts, six.string_types):
        opts = opts.split(',')
    if systemd:
        # systemd использует '-' как замену слэшей в именах юнитов, дефисы в имени маунтпоинта надо экранировать \x2d
        servicename = name.replace("-", r"\x2d").lstrip('/').replace('/', '-') + '.mount'
        with open(f'/run/systemd/transient/{servicename}', 'w') as unit:
            unit.write('[Unit]\n')
            if user is not None:
                unit.write(f'User={user}\n')
            unit.write(f'\n[Mount]\nWhat={device}\nWhere={name}\n')
            if fstype:
                unit.write(f'Type={fstype}\n')
            if opts is not None:
                unit.write(f'Options={",".join(opts)}\n')
        call(['systemctl', 'daemon-reload'], shell=False)
        action_service(servicename, 'start')
        time_0 = monotonic()
        while not is_service_active(servicename):
            sleep(1)
            if monotonic() - time_0 > 60:
                raise CommandExecutionError(f"Timeout waiting mount service {servicename} activation")
        return True

    if not os.path.exists(name) and mkmnt:
        mkdir(name, user=user)

    args = ''
    if opts is not None:
        lopts = ','.join(opts)
        args = f'-o {lopts}'
    if fstype:
        args += f' -t {fstype}'
    cmd = f'mount {args} {device} {name} '
    if user is not None:
        out = get_output(usr_cmd(cmd, user), timeout=300)
    else:
        out = get_output(cmd, timeout=300)
    if not out['result']:
        return out['comment']
    return True


def remount(name, device, mkmnt=False, fstype='', opts='defaults', user=None, force_mount=False):
    """
    Attempt to remount a device, if the device is not already mounted, mount is called
    """

    if isinstance(opts, six.string_types):
        opts = opts.split(',')
    mnts = active()
    if name in mnts:
        # The mount point is mounted, attempt to remount it with the given data
        if 'remount' not in opts:
            opts.append('remount')
        if force_mount:
            # We need to force the mount but first we should unmount
            umount(name, device)
        lopts = ','.join(opts)
        args = f'-o {lopts}'
        if fstype:
            args += f' -t {fstype}'
        cmd = f'mount {args} {device} {name} '
        if user is not None:
            out = get_output(usr_cmd(cmd, user), timeout=300)
        else:
            out = get_output(cmd, timeout=300)
        if not out['result']:
            return out['comment']
        return True
    # Mount a filesystem that isn't already
    return mount(name, device, mkmnt, fstype, opts, user=user)
