"""
(c) 2023 «Dacom M» Ltd
"""
# pylint: disable=redefined-builtin
from concurrent.futures import ThreadPoolExecutor, TimeoutError

import glob
import json
import os
import re
from abc import ABC, abstractmethod
from multiprocessing.pool import ThreadPool
from subprocess import check_call, CalledProcessError, DEVNULL, check_output
from typing import Optional, List, Union
from itertools import chain
from jinja2 import Template

from veilrpc.utils.bazalt import get_result, get_output, states_result

BASE_PATH = "/var/lib/lxc"
NET_PARAMS_DIR = '/var/lib/ecp-veil/node/bin/networks'
SRIOV_STATE_PARAMS = os.path.join(NET_PARAMS_DIR, 'sriov-state.json')
TRAFFIC_BLOCK_PARAMS = os.path.join(NET_PARAMS_DIR, 'traffic-block.json')
PORT_MIRROR_PARAMS = os.path.join(NET_PARAMS_DIR, 'port-mirrors.json')
TRAFFIC_SHAPING_PARAMS = os.path.join(NET_PARAMS_DIR, 'traffic-shaping.json')
VM_QOS_POL_MARK_PARAMS = os.path.join(NET_PARAMS_DIR, 'vmachines-qos-policy-marking.json')
SYS_CLASS_NET_DIR = '/sys/class/net'
CHECK_FIREWALL_FLAG = os.path.join(NET_PARAMS_DIR, 'check_firewall')

BASE_LXC_CONFIG = [
    'lxc.include = /usr/share/lxc/config/debian.common.conf',
    'lxc.tty.max = 4',
    'lxc.apparmor.profile = unconfined',
    'lxc.cgroup.devices.allow = a',
    'lxc.log.level = 1',
    'lxc.cgroup.memory.limit_in_bytes = 256M'
]

OVF_RULE_PATTERN = re.compile(r'.*(in_port=(?P<port>.*))(\s)(actions=(?P<action>.*)).*')


# TODO такое преобразование понадобится, если в утилите tc задавать vlan, но учитывать 16bit поле
#  (ex: tc filter add... protocol 802.1q u32 match u16 0x0064 0x0fff at -4), а не 12 bit
def convert_dec_12bit_to_hex_16bit(dec):
    """
    Функция преобразует десятичное значение для 12 bit в hex значение 16 bit
    :param dec: integer
    :return: hex
    """

    if dec not in range(0, 4096):
        raise Exception('The decimal integer should be in range between 0 to 4095')
    return f"{dec:#0{6}x}"


def convert_dscp_dec_to_tos_hex(dec):
    """
    Функция преобразует десятичное значение для 6-ти битного подполя DSCP ipv4 заголовка
    в hex значение 8 битного поля ToS
    :param dec: decimal integer
    :return: hex
    """

    if dec not in range(0, 64):
        raise Exception('The decimal integer should be in range between 0 to 63')
    return f"{dec << 2:#0{4}x}"


def create_qos_marking_rule_to_valid_str(rule: dict, vminf: str) -> str:
    rule_struct = {'direction': rule['direction'],
                   'vminf': vminf,
                   'pref_index': rule['pref_index'],
                   'dscp': rule.get('dscp'),
                   'cos': rule.get('cos'),
                   'ethertype': rule['ethertype'],
                   'vlan_id': rule.get('vlan_id'),
                   'src_mac': rule.get('src_mac'),
                   'dst_mac': rule.get('dst_mac'),
                   'ip_protocol': rule.get('ip_protocol'),
                   'src_ip': rule.get('src_ip'),
                   'dst_ip': rule.get('dst_ip'),
                   'src_port': rule.get('src_port'),
                   'dst_port': rule.get('dst_port'),
                   }

    if rule_struct['dscp']:
        rule_struct['dscp'] = convert_dscp_dec_to_tos_hex(rule_struct['dscp'])
    elif rule_struct['vlan_id']:
        rule_struct['vlan_id'] = convert_dec_12bit_to_hex_16bit(rule_struct['vlan_id'])

    temp = Template('{% if direction == "ingress" %}'
                    'tc filter add dev {{vminf}} parent ffff: pref {{pref_index}} '
                    # 'tc filter replace dev {{vminf}} parent ffff: pref {{pref_index}} '
                    '{% elif direction == "egress" %}'
                    'tc filter add dev {{vminf}} parent 1: pref {{pref_index}} '
                    # 'tc filter replace dev {{vminf}} parent 1: pref {{pref_index}} '
                    '{% endif %}'
                    '{% if ethertype == "ip" %}'
                    'protocol ip u32 '
                    '{% if ip_protocol %}'
                    '{% if ip_protocol == "tcp"%}'
                    'match ip protocol 6 0xff '
                    '{% elif ip_protocol == "udp"%}'
                    'match ip protocol 17 0xff '
                    '{% elif ip_protocol == "icmp"%}'
                    'match ip protocol 1 0xff '
                    '{% endif %}'
                    '{% endif %}'
                    '{% if dst_ip %}'
                    'match ip dst {{dst_ip}} '
                    '{% endif %}'
                    '{% if src_ip %}'
                    'match ip src {{src_ip}} '
                    '{% endif %}'
                    '{% if dst_port %}'
                    'match ip dport {{dst_port}} 0xffff '
                    '{% endif %}'
                    '{% if src_port %}'
                    'match ip sport {{src_port}} 0xffff '
                    '{% endif %}'
                    '{% if dst_mac %}'
                    'match ether dst {{dst_mac}} '
                    '{% endif %}'
                    '{% if src_mac %}'
                    'match ether src {{src_mac}} '
                    '{% endif %}'
                    '{% if not ip_protocol and not src_ip and not dst_ip '
                    'and not src_port and not dst_port and not dst_mac and not src_mac %}'
                    'match u32 0 0 '
                    '{% endif %}'
                    '{% endif %}'
                    '{% if dscp %}'
                    'action pedit munge ip tos set {{dscp}} retain 0xfc pipe csum ip'
                    '{% endif %}'
                    )
    rule_str = temp.render(rule_struct)
    return rule_str


def delete_qos_marking_rule_to_valid_str(rule: dict, vminf: str) -> str:
    rule_struct = {'direction': rule['direction'],
                   'vminf': vminf,
                   'pref_index': rule['pref_index'],
                   'ethertype': rule['ethertype'],
                   }

    temp = Template('{% if direction == "ingress" %}'
                    'tc filter delete dev {{vminf}} parent ffff: '
                    '{% elif direction == "egress" %}'
                    'tc filter delete dev {{vminf}} parent 1: '
                    '{% endif %}'
                    '{% if ethertype == "ip" %}'
                    'protocol ip '
                    '{% endif %}'
                    'pref {{pref_index}} u32'
                    )
    rule_str = temp.render(rule_struct)
    return rule_str


class VNService:
    """
    Класс описывающий контейнер LXC
    """

    __slots__ = ('name', 'config_file', 'path', 'config', 'status', 'networks')

    def __init__(self, name: str, initial: bool = False):
        self.name = name
        self.path = os.path.join(BASE_PATH, name)
        self.config_file = os.path.join(self.path, 'config')
        self.config = self._get_config(initial=initial)
        self.status = self.get_status(self.name)
        self.networks = self.get_networks()

    class STATUS:
        running = 'RUNNING'
        stopped = 'STOPPED'

    @classmethod
    def list_containers(cls, as_objects: bool = False) -> list:
        """
        Получение списка контейнеров
        :param as_objects: флаг определяющий, возвращать список контейнеров в виде списка имен или в виде списка
        экземпляров класса
        :return: возвращает список контейнеров
        """

        cmd = "lxc-ls"
        output = cls._get_shell_output(cmd)
        result = output.split() if not as_objects else [VNService(item) for item in output.split()]
        return result

    @staticmethod
    def _run_shell_cmd(cmd: str) -> bool:
        """
        Выполнение команды shell

        :param cmd: текст команды
        :return: возвращает True, если команда выполнена, и False, если выполнение команды завершилось ошибкой
        """

        try:
            check_call(cmd, shell=True, stdout=DEVNULL, stderr=DEVNULL)
            return True
        except CalledProcessError:
            return False

    @staticmethod
    def _get_shell_output(cmd: str, timeout: int = 30, universal_newlines: bool = True,
                          return_errcode: bool = False) -> str:
        """
        Выполнение команды shell и возврат вывода
        :param cmd: текст команды
        :param timeout: таймаут выполнения
        :param universal_newlines:
        :param return_errcode: флаг, определяющий возвращать текст ошибки или код возврата
        :return: возвращает вывод команды, текст ошибки или код возврата команды
        """

        try:
            output = check_output(cmd, shell=True, executable='/bin/bash', timeout=timeout,
                                  universal_newlines=universal_newlines)
            return output
        except CalledProcessError as e:
            return e.output if not return_errcode else str(e.returncode)

    def _get_config(self, initial: bool = False) -> Optional[dict]:
        """
        Преобразование текстового файла конфигурации контейнера в словарь

        :return: возвращает словарь, содержащий параметра конфигурации в качестве ключей и их значения или None, если в
        процессе чтения конфигурации возникла ошибка
        """

        if initial:
            raw_config = BASE_LXC_CONFIG
        else:
            try:
                with open(self.config_file, 'r') as f:
                    raw_config = f.read().splitlines()
            except IOError:
                return None

        config = {item.split(' = ')[0]: item.split(' = ')[1] for item in raw_config if item}
        return config

    def get_leases(self) -> list:
        headers = ('address', 'hwaddr', 'client_id', 'valid_lifetime', 'expire', 'subnet_id', 'fqdn_fwd', 'fqdn_rev',
                   'hostname', 'state')
        leases = {}
        try:
            leases_files = glob.glob(f"{self.path}/rootfs/etc/kea/dhcp4.leases*")
            for lease_file in leases_files:
                with open(lease_file, 'r') as f:
                    for raw_lease in f.read().splitlines()[1:]:
                        lease = dict(zip(headers, raw_lease.split(',')))
                        if lease['client_id']:
                            leases['client_id'] = lease
                        elif lease['hwaddr']:
                            leases['hwaddr'] = lease
        except (IOError, ValueError):
            pass
        return list(leases.values())

    def set_leases(self, leases: List[dict]):
        headers = 'address,hwaddr,client_id,valid_lifetime,expire,subnet_id,fqdn_fwd,fqdn_rev,hostname,state'
        file_to_set = f"{self.path}/rootfs/etc/kea/dhcp4.leases"
        with open(file_to_set, 'w') as f:
            f.write(headers)
            f.write('\n')
            for lease in leases:
                f.write(','.join(lease.values()))
                f.write('\n')
        return "Leases data has been set successfully.\n"

    def del_leases(self):
        leases_files = glob.glob(f"{self.path}/rootfs/etc/kea/dhcp4.leases*")
        if leases_files:
            for target_file in leases_files:
                try:
                    os.remove(target_file)
                except (OSError, Exception):
                    pass
        return "Leases data has been deleted successfully.\n"

    def get_networks(self) -> List['VNServiceNetwork']:
        """
        Получение списка интерфейсов контейнера
        """

        result = list()
        config = self._get_config()
        for idx in (0, 1):
            interface = VNServiceNetwork.get_configuration(config, idx)
            if interface:
                result.append(interface)
        return result

    def get_pid(self) -> Optional[str]:
        """
        Получние PID запущенного контейнера

        :return: вовращает PID запущенного контейнера или None, если контейнер не запущен
        """

        if not VNService.get_status(self.name) == VNService.STATUS.running:
            return None
        cmd = f"ps aux | grep {self.name}"
        raw_output = VNService._get_shell_output(cmd).splitlines()
        output = [line.split() for line in raw_output]
        for pid_info in output:
            if BASE_PATH in pid_info and self.name in pid_info:
                return pid_info[1]
        return None

    def set_config_item(self, param: str, value: str, save: bool = False) -> bool:
        """
        Установка значения параметра конфигурации контейнера

        :param param: название параметра
        :param value: значение параметра
        :param save: флаг указывает, надо ли сохранять конфигурацию после установки параметра
        :return: возвращает True, если параметр установлен, и False, если при установке значения параметра возникла
        ошибка
        """

        self.config[param] = value
        if save:
            config_str = [f"{k} = {v}" for k, v in self.config.items()]
            try:
                with open(self.config_file, 'w') as f:
                    f.write('\n'.join(config_str))
            except IOError:
                return False
        return True

    def save_config(self) -> bool:
        """
        Сохранение конфигурации контейнера в файл

        :return: возвращает True, если конфигурация сохранена, и False, если при сохранении конфигурации возникла ошибка
        """

        config_str = [f"{k} = {v}" for k, v in self.config.items()]
        try:
            with open(self.config_file, 'w') as f:
                f.write('\n'.join(config_str))
        except IOError:
            return False
        return True

    def get_config_item(self, param: str) -> Optional[str]:
        """
        Получение значения параметра конфигурации контейнера

        :param param: название параметра конфигурации
        :return: возвращает значение параметра конфигурации, если такой параметр существует, или None
        """

        if param == "lxc.rootfs.path":
            value = self.config[param][4:]
        else:
            value = self.config[param] if param in self.config.keys() else None
        return value

    def clear_config_item(self, param: str, save: bool = False) -> bool:
        """
        Удаление параметра из конфигурации контейнера

        :param param: название параметра конфигурации
        :param save: флаг указывает, надо ли сохранять конфигурацию после удаления параметра
        """

        if param in self.config.keys():
            self.config.pop(param)
        else:
            return False
        if save and not self.save_config():
            return False
        return True

    def clone(self, name: str) -> Optional['VNService']:
        """
        Клонирование базового шаблона контейнера

        :param name: название нового контейнера
        :return: возвращает клонированный контейнер в виде экземпляра класса VNService или None, если в процессе
        клонирования возникла ошибка
        """

        clone_path = f"{BASE_PATH}/{name}"
        cmd = f"mkdir {clone_path}"
        if not self._run_shell_cmd(cmd):
            return None
        cmd = f"cp -R {BASE_PATH}/vns-template/rootfs {clone_path}"
        pool = ThreadPool(processes=1)
        async_result = pool.apply_async(self._run_shell_cmd, (cmd,))
        clone_result = async_result.get(timeout=120)
        if not clone_result:
            cmd = f"rm -rf {clone_path}"
            self._run_shell_cmd(cmd)
            return None
        cmd = f"touch {clone_path}/config"
        if not self._run_shell_cmd(cmd):
            return None
        clone = VNService(name, initial=True)
        clone.set_config_item("lxc.rootfs.path", f"dir:{clone_path}/rootfs")
        clone.set_config_item("lxc.log.file", f"/var/log/veil/node/{name}.log")
        clone.set_config_item("lxc.uts.name", name)
        if not clone.save_config():
            clone.destroy()
            return None
        return clone

    def destroy(self) -> bool:
        """
        Удаление контейнера

        :return: возвращает True, если контейнер успешно удален, и False, если при удалении контейнера возникла ошибка
        """

        cmd = f"rm -rf {self.path}"
        if not self._run_shell_cmd(cmd):
            return False
        return True

    def _wait_status(self, status: str, timeout: int):
        """
        Ожидание статуса контейнера

        :param status: ожидаемый статус
        :param timeout:
        :return: возвращает True, если ожидаемый статус получен, и False, если при получении статуса возникла ошибка
        """

        cmd = f"lxc-wait -n {self.name} -s {status} -t {timeout}"
        if not self._run_shell_cmd(cmd):
            return False
        return True

    def _start(self) -> bool:
        """
        Запуск контейнера

        :return: возвращает True, если контейнер запущен, и False, если при запуске контейнера возникла ошибка
        """

        cmd = f"lxc-start -n {self.name}"
        if not self._run_shell_cmd(cmd):
            return False
        return True

    def _stop(self) -> bool:
        """
        Останов контейнера

        :return: возвращает True, если контейнер остановлен, и False, если при останове контейнера возникла ошибка
        """

        cmd = f"lxc-stop -n {self.name}"
        if not self._run_shell_cmd(cmd):
            return False
        return True

    def wait_started(self, timeout: int = 20) -> Optional[bool]:
        """
        Запуск контейнера с ожиданием статуса

        :param timeout: значение таймаута, по истечении которого задача считается не выполненной
        :return: возвращает True, если контейнер запущен, False, если до истечения таймаута контейнер не изменил статус,
        и None, если до истечения таймаута не был получен результат
        """

        executor = ThreadPoolExecutor(max_workers=2)
        wait_future = executor.submit(self._wait_status, VNService.STATUS.running, timeout)
        _ = executor.submit(self._start)
        try:
            result = wait_future.result(timeout)
            if result:
                self.status = VNService.STATUS.running
            return result
        except TimeoutError:
            return False

    def wait_stopped(self, timeout: int = 20) -> Optional[bool]:
        """
        Останов контейнера с ожиданием статуса

        :param timeout: значение таймаута, по истечении которого задача считается не выполненной
        :return: возвращает True, если контейнер остановлен, False, если до истечения таймаута контейнер не изменил
        статус, и None, если до истечения таймаута не был получен результат
        """

        executor = ThreadPoolExecutor(max_workers=2)
        wait_future = executor.submit(self._wait_status, VNService.STATUS.stopped, timeout)
        _ = executor.submit(self._stop)
        try:
            result = wait_future.result(timeout)
            if result:
                self.status = VNService.STATUS.running
                return result
            else:
                cmd = f"kill -9 {self.get_pid()}"
                return self._run_shell_cmd(cmd)
        except TimeoutError:
            cmd = f"kill -9 {self.get_pid()}"
            return self._run_shell_cmd(cmd)

    def attach_cmd(self, cmd: str, get_output: bool = False) -> Union[bool, str]:  # pylint: disable=W0621
        """
        Выполнение команды в контейнере

        :param cmd: команда, выполняемая в контейнере
        :param get_output: флаг, указывающий, надо ли получить вывод команды
        :return: возвращает True, если команда выполнена, и False, если при выполнении команды возникла ошибка; если
        передан флаг get_output, возвращает вывод команды
        """

        attach_cmd = f"lxc-attach -n {self.name} -- {cmd}"
        if get_output:
            ret = VNService._get_shell_output(attach_cmd)
            return ret
        else:
            if not self._run_shell_cmd(attach_cmd):
                return False
            return True

    @staticmethod
    def get_status(name) -> Optional[str]:
        """
        Получение текущего состояния контейнера

        :return: возвращает статус контейнера или None, в случае, если в сведениях о контейнере не содержится его
        текущего состояния
        """

        cmd = f"lxc-info -n {name}"
        ret = VNService._get_shell_output(cmd)
        raw_output = ret.splitlines()
        curr_status = next((row for row in raw_output if row.startswith("State")), None)
        if curr_status:
            state_row = curr_status.split(":")
            state = state_row[1].lstrip().rstrip()
            return state
        return None

    @classmethod
    def exists(cls, name: str) -> bool:
        """
        Проверка существования контейнера

        :param name: название контейнера
        :return: возвращает True, если контейнер существует, и False если не существует
        """

        cmd = "lxc-ls"
        output = cls._get_shell_output(cmd)
        if name in output.split():
            return True
        return False

    def create_interface(self, index: int, network_type: str, params: dict) -> bool:
        """
        Создание сетевого интерфейса контейнера

        :param index: индекс интерфейса
        :param network_type: тип интерфейса
        :param params: словарь, содержащий параметры интерфейса
        :return: возвращает True, если интерфейс создан, и False, если при создании интерфейса возникла ошибка
        """

        try:
            interface = VNServiceNetwork(index, network_type, **params)
            self.networks.append(interface)
            return True
        except (KeyError, Exception):
            return False

    def save_interface(self, index: int) -> bool:
        """
        Сохранение параметров интерфейса в конфигурацию контейнера

        :param index: индекс интерфейса
        :return: возвращает True, если сохранение успешно, и False, если при сохранении возникла ошибка
        """

        interface = next((net for net in self.networks if net.index == index), None)
        if not interface:
            return False
        self.set_config_item(f"lxc.net.{interface.index}.type", interface.network_type)
        self.set_config_item(f"lxc.net.{interface.index}.name", interface.name)
        self.set_config_item(f"lxc.net.{interface.index}.veth.pair", interface.pair)
        self.set_config_item(f"lxc.net.{interface.index}.flags", interface.flags)
        self.set_config_item(f"lxc.net.{interface.index}.hwaddr", interface.mac_address)
        self.set_config_item(f"lxc.net.{interface.index}.mtu", interface.mtu)
        self.set_config_item(f"lxc.net.{interface.index}.script.up", interface.script_up)
        self.set_config_item(f"lxc.net.{interface.index}.script.down", interface.script_down)
        if interface.ipv4_address:
            self.set_config_item(f"lxc.net.{interface.index}.ipv4.address", interface.ipv4_address)
            if interface.ipv4_gateway:
                self.set_config_item(f"lxc.net.{interface.index}.ipv4.gateway", interface.ipv4_gateway)
        if not self.save_config():
            return False
        return True

    def remove_interface(self, index: int) -> bool:
        """
        Удаление интерфейса контейнера

        :param index: индекс интерфейса
        :return: возвращает True, если удаление успешно, и False, если при удалении возникла ошибка
        """

        self.config = {k: v for k, v in self.config.items() if not k.startswith(f"lxc.net.{index}")}
        self.networks = self.get_networks()
        if not self.save_config():
            return False
        return True

    def repair_dhcp(self) -> None:
        """
        Восстановление работы сервиса dhcp посредством:
        - удаления файла /etc/kea/dhcp4.leases.pid, если сущетсвует
        """

        dhcp4_leases_pid_path = os.path.join(self.path, 'rootfs/etc/kea/dhcp4.leases.pid')
        if os.path.exists(dhcp4_leases_pid_path):
            os.remove(dhcp4_leases_pid_path)


class VNServiceNetwork:
    """
    Класс описывающий сеть (сетевой интерфейс) контейнера LXC
    """

    __slots__ = ('index', 'network_type', 'name', 'pair', 'flags', 'mac_address', 'mtu', 'script_up', 'script_down',
                 'ipv4_address', 'ipv4_gateway')

    def __init__(self, index: int, network_type: str, **kwargs):
        self.index = index
        self.network_type = network_type
        self.name = kwargs['name']
        self.pair = kwargs['pair']
        self.flags = kwargs['flags']
        self.mac_address = kwargs['mac_address']
        self.mtu = kwargs['mtu']
        self.script_up = kwargs['script_up']
        self.script_down = kwargs['script_down']
        self.ipv4_address = kwargs['ipv4_address']
        self.ipv4_gateway = kwargs['ipv4_gateway']

    @classmethod
    def get_configuration(cls, config: dict, index: int) -> Optional['VNServiceNetwork']:
        """
        Получение экземпляра класса из конфигурации контейнера
        :param config: конфигурация контейнера
        :param index: индекс интерфейса
        :return: возвращает экземпляр класса, если интерфейс есть в конфигурации контейнера или None
        """

        interface_config = {k: v for k, v in config.items() if k.startswith(f"lxc.net.{index}")}
        if interface_config:
            network_type = "veth"
            params = dict(
                name=interface_config[f"lxc.net.{index}.name"],
                pair=interface_config[f"lxc.net.{index}.veth.pair"],
                flags=interface_config[f"lxc.net.{index}.flags"],
                mac_address=interface_config[f"lxc.net.{index}.hwaddr"],
                mtu=interface_config[f"lxc.net.{index}.mtu"],
                script_up=interface_config[f"lxc.net.{index}.script.up"],
                script_down=interface_config[f"lxc.net.{index}.script.down"],
                ipv4_address=interface_config.get(f"lxc.net.{index}.ipv4.address", None),
                ipv4_gateway=interface_config.get(f"lxc.net.{index}.ipv4.gateway", None)
            )
            return VNServiceNetwork(index, network_type, **params)
        else:
            return None


class NetRuleSet(ABC):
    """
    Абстрактный класс, описывающий набор сетевых правил (блокирования трафика, зеркалирования портов и т.п.)
    """

    __slots__ = ('__config_dir',)

    def __init__(self):
        self.__config_dir = '/var/lib/ecp-veil/node/bin/networks'

    def config_dir(self):
        return self.__config_dir

    @abstractmethod
    def readable(self):
        pass

    @abstractmethod
    def find_rule(self, params: dict):
        pass

    @abstractmethod
    def remove_rule(self, rule, save: bool = False):
        pass

    @abstractmethod
    def add_rule(self, rule, save: bool = False):
        pass

    @abstractmethod
    def update_rule(self, rule, save: bool = False):
        pass

    @abstractmethod
    def save(self, apply: bool = False):
        pass

    @classmethod
    @abstractmethod
    def from_config_file(cls):
        pass

    @classmethod
    @abstractmethod
    def from_host_conf(cls):
        pass


class NetRule(ABC):
    """
    Абстрактный класс, описывающий сетевое правило (блокирования трафика, зеркалирования портов и т.п.)
    """

    __slots__ = ('applied', 'to_apply')

    def __init__(self, applied: bool = False, to_apply: bool = False):
        self.applied = applied
        self.to_apply = to_apply

    @property
    def slots(self):
        return set(chain.from_iterable(getattr(cls, "__slots__", ()) for cls in self.__class__.__mro__))

    @abstractmethod
    def readable(self):
        pass

    @abstractmethod
    def apply_rule(self):
        pass

    @abstractmethod
    def drop_rule(self):
        pass

    @abstractmethod
    def to_add_command(self):
        pass

    @abstractmethod
    def to_del_command(self):
        pass

    @classmethod
    @abstractmethod
    def from_json(cls, data: dict):
        pass


class TrafficBlockRuleSet(NetRuleSet):
    """
    Класс, описывающий набор правил блокировки трафика
    """

    __slots__ = ('__config_file', 'rules_set')

    def __init__(self):
        super(TrafficBlockRuleSet, self).__init__()
        self.__config_file = os.path.join(super().config_dir(), 'traffic-block.json')
        self.rules_set: List['TrafficBlockRule'] = []

    @property
    def config_file(self):
        return self.__config_file

    def clone(self) -> 'TrafficBlockRuleSet':
        rs = TrafficBlockRuleSet()
        for r in self.rules_set:
            rs.add_rule(r)
        return rs

    def readable(self) -> List[dict]:
        return [rule.readable() for rule in self.rules_set]

    def find_rule(self, params: dict) -> Optional['TrafficBlockRule']:
        switch = params.get('switch', None)
        port = params.get('port', None)
        if not switch or not port:
            return None
        rule = next((rule for rule in self.rules_set if rule.switch == switch and rule.port == port), None)
        return rule

    def remove_rule(self, rule: 'TrafficBlockRule', save: bool = False) -> bool:
        if not isinstance(rule, TrafficBlockRule):
            return False
        try:
            tmp_rs = self.clone()
            for r in tmp_rs.readable():
                if rule.readable() == r:
                    self.rules_set.remove(rule)
            if save and not self.save():
                return False
            return True
        except ValueError:
            return True

    def add_rule(self, rule: 'TrafficBlockRule', save: bool = False) -> bool:
        if not isinstance(rule, TrafficBlockRule):
            return False
        if rule.readable() in self.readable():
            return True
        self.rules_set.append(rule)
        if save and not self.save():
            return False
        return True

    def update_rule(self, rule: 'TrafficBlockRule', save: bool = False) -> bool:
        if not isinstance(rule, TrafficBlockRule):
            return False
        index = None
        for idx, item in enumerate(self.rules_set):
            if item.switch == rule.switch and item.port == rule.port:
                index = idx
        if index:
            self.rules_set[index] = rule
            return True
        return False

    def save(self, apply: bool = False) -> bool:
        result_rs = TrafficBlockRuleSet()
        for rule in self.rules_set:
            if rule.readable() not in result_rs.readable():
                result_rs.add_rule(rule)
        try:
            if apply:
                for rule in filter(lambda x: x.to_apply and not x.applied, result_rs.rules_set):
                    rule.apply_rule()
                    rule.applied = True
            with open(self.config_file, 'w') as f:
                f.write(json.dumps(result_rs.readable()))
                f.write('\n')
            return True
        except OSError:
            return False

    @classmethod
    def from_config_file(cls) -> 'TrafficBlockRuleSet':
        rs = TrafficBlockRuleSet()
        try:
            with open(rs.config_file, 'r') as f:
                rules = json.loads(f.read())
        except json.decoder.JSONDecodeError:
            with open(rs.config_file, 'w') as f:
                f.write("[]\n")
            rules = []
        for rule in rules:
            rs.add_rule(TrafficBlockRule.from_json(rule))
        return rs

    @classmethod
    def from_host_conf(cls) -> Optional['TrafficBlockRuleSet']:
        rs = TrafficBlockRuleSet()
        existing_switches = states_result(get_output, cmd='ovs-vsctl list-br').splitlines()
        for switch in existing_switches:
            existing_rules = states_result(get_output, cmd=f'ovs-ofctl dump-flows --names {switch}').splitlines()
            for rule in existing_rules:
                if bool(OVF_RULE_PATTERN.match(rule)):
                    port = OVF_RULE_PATTERN.search(rule).group('port').strip('"')
                    action = OVF_RULE_PATTERN.search(rule).group('action').strip('"')
                    if action == 'drop':
                        rs.rules_set.append(TrafficBlockRule(switch=switch, port=port, applied=True, to_apply=True))
        return rs


class TrafficBlockRule(NetRule):
    __slots__ = ('switch', 'port')

    def __init__(self, switch: str, port: str, applied: bool = False, to_apply: bool = False):
        super(TrafficBlockRule, self).__init__(applied, to_apply)
        self.switch = switch
        self.port = port

    def readable(self) -> dict:
        return {key: getattr(self, key) for key in self.slots}

    def apply_rule(self) -> bool:
        if not get_result(self.to_add_command()):
            return False
        return True

    def drop_rule(self):
        if not get_result(self.to_del_command()):
            return False
        return True

    def to_add_command(self) -> str:
        return f"ovs-ofctl add-flow {self.switch} table=0,in_port={self.port},actions=drop"

    def to_del_command(self) -> str:
        return f"ovs-ofctl del-flows {self.switch} table=0,in_port={self.port}"

    @classmethod
    def from_json(cls, data: dict) -> 'TrafficBlockRule':
        return TrafficBlockRule(**data)


class SriovStateRuleSet(NetRuleSet):
    """
    Класс, описывающий набор параметров для SR-IOV интерфейсов
    """

    __slots__ = ('__config_file', 'rules_set')

    def __init__(self):
        super(SriovStateRuleSet, self).__init__()
        self.__config_file = os.path.join(super().config_dir(), 'sriov-state.json')
        self.rules_set: List['SriovStateRule'] = []

    @property
    def config_file(self):
        return self.__config_file

    def readable(self) -> List[dict]:
        return [rule.readable() for rule in self.rules_set]

    def find_rule(self, params: dict) -> Optional['SriovStateRule']:
        interface = params.get('interface', None)
        if not interface:
            return None
        rule = next((rule for rule in self.rules_set if rule.interface == interface), None)
        return rule

    def remove_rule(self, rule: 'SriovStateRule', save: bool = False) -> bool:
        try:
            self.rules_set.remove(rule)
            if save and not self.save():
                return False
            return True
        except ValueError:
            return False

    def add_rule(self, rule: 'SriovStateRule', save: bool = False) -> bool:
        if not isinstance(rule, SriovStateRule):
            return False
        self.rules_set.append(rule)
        if save and not self.save():
            return False
        return True

    def update_rule(self, rule: 'SriovStateRule', save: bool = False) -> bool:
        index = None
        for idx, item in enumerate(self.rules_set):
            if item.interface == rule.interface:
                index = idx
        if index:
            self.rules_set[index] = rule
            return True
        return False

    def save(self, apply: bool = False) -> bool:
        try:
            if apply:
                for rule in filter(lambda x: x.to_apply and not x.applied, self.rules_set):
                    rule.apply_rule()
                    rule.applied = True
            with open(self.config_file, 'w') as f:
                f.write(json.dumps(self.readable()))
                f.write('\n')
            return True
        except OSError:
            return False

    @classmethod
    def from_config_file(cls) -> 'SriovStateRuleSet':
        rs = SriovStateRuleSet()
        try:
            with open(rs.config_file, 'r') as f:
                rules = json.loads(f.read())
        except json.decoder.JSONDecodeError:
            with open(rs.config_file, 'w') as f:
                f.write("[]\n")
            rules = []
        for rule in rules:
            rs.add_rule(SriovStateRule.from_json(rule))
        return rs

    @classmethod
    def from_host_conf(cls) -> 'SriovStateRuleSet':
        rs = SriovStateRuleSet()
        for device in (dev for dev in os.listdir(SYS_CLASS_NET_DIR)):
            if 'device' not in os.listdir(os.path.join(SYS_CLASS_NET_DIR, device)):
                continue
            dev_path = os.path.join(SYS_CLASS_NET_DIR, device, 'device')
            if 'sriov_totalvfs' in os.listdir(dev_path):
                cmd_ret = get_output(f'cat {os.path.join(dev_path, "sriov_numvfs")}')
                if not cmd_ret['result']:
                    continue
                if int(cmd_ret['comment']):
                    data = dict(interface=device, vfunctions=int(cmd_ret['comment']))
                    rs.add_rule(SriovStateRule.from_json(data))
        return rs


class SriovStateRule(NetRule):
    __slots__ = ('interface', 'vfunctions')

    def __init__(self, interface: str, vfunctions: int, applied: bool = False, to_apply: bool = False):
        super(SriovStateRule, self).__init__(applied, to_apply)
        self.interface = interface
        self.vfunctions = vfunctions

    def readable(self) -> dict:
        return {key: getattr(self, key) for key in self.slots}

    def apply_rule(self) -> bool:
        if not get_result(self.to_add_command()):
            return False
        return True

    def drop_rule(self) -> bool:
        if not get_result(self.to_del_command()):
            return False
        return True

    def to_add_command(self) -> str:
        return f"echo {self.vfunctions} > {os.path.join(SYS_CLASS_NET_DIR, self.interface, 'device', 'sriov_numvfs')}"

    def to_del_command(self) -> str:
        return f"echo 0 > {os.path.join(SYS_CLASS_NET_DIR, self.interface, 'device', 'sriov_numvfs')}"

    @classmethod
    def from_json(cls, data: dict) -> 'SriovStateRule':
        return SriovStateRule(**data)


class PortMirroringRuleSet(NetRuleSet):
    __slots__ = ('__config_file', 'rules_set')

    def __init__(self):
        super(PortMirroringRuleSet, self).__init__()
        self.__config_file = os.path.join(super().config_dir(), 'port-mirrors.json')
        self.rules_set: List['PortMirroringRule'] = []

    @property
    def config_file(self):
        return self.__config_file

    def readable(self) -> List[dict]:
        return [rule.readable() for rule in self.rules_set]

    def find_rule(self, params: dict) -> Optional['PortMirroringRule']:
        source = params.get('source', None)
        target = params.get('target', None)
        direction = params.get('direction', None)
        pref_index = params.get('pref_index', None)
        if not source or not target or not direction or not pref_index:
            return None
        rule = next((rule for rule in self.rules_set if rule.source == source and rule.target == target and
                     rule.direction == direction and rule.pref_index == pref_index), None)
        return rule

    def remove_rule(self, rule: 'PortMirroringRule', save: bool = False) -> bool:
        try:
            self.rules_set.remove(rule)
            if save and not self.save():
                return False
            return True
        except ValueError:
            return False

    def add_rule(self, rule: 'PortMirroringRule', save: bool = False) -> bool:
        if not isinstance(rule, PortMirroringRule):
            return False
        self.rules_set.append(rule)
        if save and not self.save():
            return False
        return True

    def update_rule(self, rule: 'PortMirroringRule', save: bool = False) -> bool:
        index = None
        for idx, item in enumerate(self.rules_set):
            if item.source == rule.source:
                index = idx
        if index:
            self.rules_set[index] = rule
            return True
        return False

    def save(self, apply: bool = False) -> bool:
        try:
            if apply:
                for rule in filter(lambda x: x.to_apply and not x.applied, self.rules_set):
                    rule.apply_rule()
                    rule.applied = True
            with open(self.config_file, 'w') as f:
                f.write(json.dumps(self.readable()))
                f.write('\n')
            return True
        except OSError:
            return False

    @classmethod
    def from_config_file(cls) -> 'PortMirroringRuleSet':
        rs = PortMirroringRuleSet()
        try:
            with open(rs.config_file, 'r') as f:
                rules = json.loads(f.read())
        except json.decoder.JSONDecodeError:
            with open(rs.config_file, 'w') as f:
                f.write("[]\n")
            rules = []
        for rule in rules:
            rs.add_rule(PortMirroringRule.from_json(rule))
        return rs

    @classmethod
    def from_host_conf(cls) -> 'PortMirroringRuleSet':
        from veilrpc.states.networking import tc_mirred_rules_exists_three_directions
        rs = PortMirroringRuleSet()
        applied_rules = tc_mirred_rules_exists_three_directions()
        for direction, source_ports in applied_rules.items():
            for port, rules in source_ports.items():
                for rule in rules:
                    data = {'name': f"Port Mirror {len(rs.rules_set)}",
                            'source': port,
                            'pref_index': rule.get('pref_index'),
                            'direction': direction,
                            'target': rule.get('dst_port')
                            }
                    rs.add_rule(PortMirroringRule.from_json(data))
        return rs


class PortMirroringRule(NetRule):
    __slots__ = ('name', 'source', 'target', 'direction', 'pref_index')

    def __init__(self, name: str, source: str, target: str, direction: str, pref_index: dict,
                 applied: bool = False, to_apply: bool = False):
        super(PortMirroringRule, self).__init__(applied, to_apply)
        self.name = name
        self.source = source
        self.target = target
        self.direction = direction
        self.pref_index = pref_index

    def readable(self) -> dict:
        return {key: getattr(self, key) for key in self.slots}

    def apply_rule(self) -> bool:
        for cmd in self.to_add_command():
            if not get_result(cmd):
                return False
        return True

    def drop_rule(self) -> bool:
        for cmd in self.to_del_command():
            if not get_result(cmd):
                return False
        return True

    def to_add_command(self) -> List[str]:
        from veilrpc.states.networking import parse_tc_qdisc_exists
        qdisc_exists = states_result(parse_tc_qdisc_exists)
        if self.direction == 'ingress':
            cmd = [f"tc filter add dev {self.source} parent ffff: "
                   f"pref {self.pref_index['ingress']} protocol all u32 match u8 0 0 "
                   f"action mirred egress mirror dev {self.target}"]
            if self.source not in qdisc_exists['ingress']:
                cmd.insert(0, f"tc qdisc replace dev {self.source} ingress")
            return cmd
        elif self.direction == 'egress':
            cmd = [f"tc filter add dev {self.source} parent 1: "
                   f"pref {self.pref_index['egress']} protocol all u32 match u8 0 0 "
                   f"action mirred egress mirror dev {self.target}"]
            if self.source not in qdisc_exists['egress']:
                cmd.insert(0, f"tc qdisc replace dev {self.source} handle 1: root prio")
            return cmd
        elif self.direction == 'ingress-and-egress':
            cmd = [f"tc filter add dev {self.source} parent ffff: "
                   f"pref {self.pref_index['ingress']} protocol all u32 match u8 0 0 "
                   f"action mirred egress mirror dev {self.target}",
                   f"tc filter add dev {self.source} parent 1: pref {self.pref_index['egress']} "
                   f"protocol all u32 match u8 0 0 action mirred egress mirror dev {self.target}"]

            if self.source not in qdisc_exists['ingress']:
                cmd.insert(0, f"tc qdisc replace dev {self.source} ingress")
            if self.source not in qdisc_exists['egress']:
                cmd.insert(0, f"tc qdisc replace dev {self.source} handle 1: root prio")
            return cmd

    def to_del_command(self) -> List[str]:
        if self.direction in ('ingress', 'ingress-and-egress'):
            result = [f"tc filter delete dev {self.source} parent ffff: protocol all "
                      f"pref {self.pref_index['ingress']} u32"]
            if self.direction == 'ingress-and-egress':
                result.append(f"tc filter delete dev {self.source} parent 1: protocol all "
                              f"pref {self.pref_index['egress']} u32")
            return result
        else:
            return [f"tc filter delete dev {self.source} parent 1: protocol all pref {self.pref_index['egress']} u32"]

    @classmethod
    def from_json(cls, data: dict) -> 'PortMirroringRule':
        return PortMirroringRule(**data)


# VMachines QoS Policy Marking to vnetwork
class VMachinesQoSPolicyMarkingRuleSet(NetRuleSet):
    __slots__ = ('__config_file', 'rules_set')

    def __init__(self):
        super(VMachinesQoSPolicyMarkingRuleSet, self).__init__()
        self.__config_file = os.path.join(super().config_dir(), 'vmachines-qos-policy-marking.json')
        self.rules_set: List['VMachinesQoSPolicyMarkingRule'] = []

    @property
    def config_file(self):
        return self.__config_file

    def readable(self) -> List[dict]:
        return [rule.readable() for rule in self.rules_set]

    def find_rule(self, params: dict) -> Optional['VMachinesQoSPolicyMarkingRule']:
        # qos_policy = params.get('qos_policy', None)
        source = params.get('source', None)
        pref_index = params.get('pref_index', None)
        direction = params.get('direction', None)
        dscp = params.get('dscp', None)
        cos = params.get('cos', None)
        ethertype = params.get('ethertype', None)
        vlan_id = params.get('vlan_id', None)
        src_mac = params.get('src_mac', None)
        dst_mac = params.get('dst_mac', None)
        ip_protocol = params.get('ip_protocol', None)
        src_ip = params.get('src_ip', None)
        dst_ip = params.get('dst_ip', None)
        src_port = params.get('src_port', None)
        dst_port = params.get('dst_port', None)
        # is_enabled = params.get('is_enabled', None)
        # rule_index = params.get('rule_index', None)
        # id = params.get('id', None)
        # vnetwork = params.get('vnetwork', None)

        if not source or not pref_index or not direction or not ethertype:
            return None

        # rule = next((rule for rule in self.rules_set if rule.qos_policy == qos_policy and rule.source == source and
        #              rule.pref_index == pref_index and rule.direction == direction and rule.dscp == dscp and
        #              rule.cos == cos and rule.ethertype == ethertype and rule.vlan_id == vlan_id and
        #              rule.src_mac == src_mac and rule.dst_mac == dst_mac and rule.ip_protocol == ip_protocol and
        #              rule.src_ip == src_ip and rule.dst_ip == dst_ip and rule.src_port == src_port and
        #              rule.dst_port == dst_port and rule.is_enabled == is_enabled and rule.rule_index == rule_index),
        #             None)

        # Правило можно однозначано опеределить по параметрам source, pref_index, direction
        # Но будем сравнивать максимально возможные параметры
        rule = next((rule for rule in self.rules_set if rule.source == source and rule.pref_index == pref_index and
                     rule.direction == direction and rule.dscp == dscp and rule.cos == cos and
                     rule.ethertype == ethertype and rule.vlan_id == vlan_id and rule.src_mac == src_mac and
                     rule.dst_mac == dst_mac and rule.ip_protocol == ip_protocol and rule.src_ip == src_ip and
                     rule.dst_ip == dst_ip and rule.src_port == src_port and rule.dst_port == dst_port),
                    None)
        return rule

    def remove_rule(self, rule: 'VMachinesQoSPolicyMarkingRule', save: bool = False) -> bool:
        try:
            self.rules_set.remove(rule)
            if save and not self.save():
                return False
            return True
        except ValueError:
            return False

    def add_rule(self, rule: 'VMachinesQoSPolicyMarkingRule', save: bool = False) -> bool:
        if not isinstance(rule, VMachinesQoSPolicyMarkingRule):
            return False
        self.rules_set.append(rule)
        if save and not self.save():
            return False
        return True

    def update_rule(self, rule: 'VMachinesQoSPolicyMarkingRule', save: bool = False) -> bool:
        index = None
        for idx, item in enumerate(self.rules_set):
            # т.к. параметр pref_index уникален в нашей реализации, то т.о. правило можно однозначано опеределить по
            # этим параметрам ниже
            if item.source == rule.source and item.pref_index == rule.pref_index and item.direction == rule.direction:
                index = idx
        if index:
            self.rules_set[index] = rule
            return True
        return False

    def save(self, apply: bool = False) -> bool:
        try:
            if apply:
                for rule in filter(lambda x: x.to_apply and not x.applied, self.rules_set):
                    rule.apply_rule()
                    rule.applied = True
            with open(self.config_file, 'w') as f:
                f.write(json.dumps(self.readable()))
                f.write('\n')
            return True
        except OSError:
            return False

    @classmethod
    def from_config_file(cls) -> 'VMachinesQoSPolicyMarkingRuleSet':
        rs = VMachinesQoSPolicyMarkingRuleSet()
        try:
            with open(rs.config_file, 'r') as f:
                rules = json.loads(f.read())
        except json.decoder.JSONDecodeError:
            with open(rs.config_file, 'w') as f:
                f.write("[]\n")
            rules = []
        for rule in rules:
            rs.add_rule(VMachinesQoSPolicyMarkingRule.from_json(rule))
        return rs

    # Важный метод, здесь дополняем data реальными параметрами: 'source': port, 'direction': direction
    # и фэйковыми: 'qos_policy': "the_true_value_is_stored_in_json_file",
    # 'is_enabled': "the_true_value_is_stored_in_json_file", 'rule_index': "the_true_value_is_stored_in_json_file",
    # 'id': "the_true_value_is_stored_in_json_file"
    # Эти параметры добавляем, т.к. мы их не можем получить путем парсинга команды tc filter show
    # (функция tc_marking_rules_exists_params_needed)
    # Добавявление параметров делается для совместимости с классом VMachinesQoSPolicyMarkingRule
    # Полученные правила будут использоваться, в том числе, в методе find_rule
    @classmethod
    def from_host_conf(cls) -> 'VMachinesQoSPolicyMarkingRuleSet':
        from veilrpc.states.networking import tc_marking_rules_exists_params_needed
        rs = VMachinesQoSPolicyMarkingRuleSet()
        applied_rules = tc_marking_rules_exists_params_needed()
        for direction, source_ports in applied_rules.items():
            for port, rules in source_ports.items():
                for rule in rules:
                    data = {'qos_policy': "the_true_value_is_stored_in_json_file",
                            'source': port,
                            'pref_index': rule.get('pref_index'),
                            'direction': direction,
                            'dscp': rule.get('dscp'),
                            'cos': rule.get('cos'),
                            'ethertype': rule.get('ethertype'),
                            'vlan_id': rule.get('vlan_id'),
                            'src_mac': rule.get('src_mac'),
                            'dst_mac': rule.get('dst_mac'),
                            'ip_protocol': rule.get('ip_protocol'),
                            'src_ip': rule.get('src_ip'),
                            'dst_ip': rule.get('dst_ip'),
                            'src_port': rule.get('src_port'),
                            'dst_port': rule.get('dst_port'),
                            'is_enabled': "the_true_value_is_stored_in_json_file",
                            'rule_index': "the_true_value_is_stored_in_json_file",
                            'id': "the_true_value_is_stored_in_json_file",
                            'vnetwork': "the_true_value_is_stored_in_json_file"
                            }
                    rs.add_rule(VMachinesQoSPolicyMarkingRule.from_json(data))
        return rs


class VMachinesQoSPolicyMarkingRule(NetRule):
    # pylint: disable=redefined-builtin
    __slots__ = ('qos_policy', 'source', 'pref_index', 'direction', 'dscp', 'cos', 'ethertype', 'vlan_id', 'src_mac',
                 'dst_mac', 'ip_protocol', 'src_ip', 'dst_ip', 'src_port', 'dst_port', 'is_enabled', 'rule_index', 'id',
                 'vnetwork')

    def __init__(self, qos_policy: str, source: str, pref_index: int, direction: str, dscp: int, cos: int,
                 ethertype: str, vlan_id: int, src_mac: str, dst_mac: str, ip_protocol: str, src_ip: str, dst_ip: str,
                 src_port: int, dst_port: int, is_enabled: bool, rule_index: int, id: str, vnetwork: str,
                 applied: bool = False, to_apply: bool = False):
        super(VMachinesQoSPolicyMarkingRule, self).__init__(applied, to_apply)
        self.qos_policy = qos_policy
        self.source = source
        self.pref_index = pref_index
        self.direction = direction
        self.dscp = dscp
        self.cos = cos
        self.ethertype = ethertype
        self.vlan_id = vlan_id
        self.src_mac = src_mac
        self.dst_mac = dst_mac
        self.ip_protocol = ip_protocol
        self.src_ip = src_ip
        self.dst_ip = dst_ip
        self.src_port = src_port
        self.dst_port = dst_port
        self.is_enabled = is_enabled
        self.rule_index = rule_index
        self.id = id
        self.vnetwork = vnetwork

    def readable(self) -> dict:
        return {key: getattr(self, key) for key in self.slots}

    def apply_rule(self) -> bool:
        for cmd in self.to_add_command():
            if not get_result(cmd):
                return False
        return True

    def drop_rule(self) -> bool:
        for cmd in self.to_del_command():
            if not get_result(cmd):
                return False
        return True

    def to_add_command(self) -> List[str]:
        from veilrpc.states.networking import parse_tc_qdisc_exists
        qdisc_exists = states_result(parse_tc_qdisc_exists)

        rule_params = dict(pref_index=self.pref_index, direction=self.direction,
                           dscp=self.dscp, cos=self.cos, ethertype=self.ethertype, vlan_id=self.vlan_id,
                           src_mac=self.src_mac, dst_mac=self.dst_mac, ip_protocol=self.ip_protocol,
                           src_ip=self.src_ip, dst_ip=self.dst_ip, src_port=self.src_port, dst_port=self.dst_port)
        vminf_name = self.source
        cmd = [create_qos_marking_rule_to_valid_str(rule_params, vminf_name)]
        if self.direction == 'ingress':
            if self.source not in qdisc_exists['ingress']:
                cmd.insert(0, f"tc qdisc replace dev {self.source} ingress")
            return cmd
        elif self.direction == 'egress':
            if self.source not in qdisc_exists['egress']:
                cmd.insert(0, f"tc qdisc replace dev {self.source} handle 1: root prio")
            return cmd

    def to_del_command(self) -> List[str]:
        rule_params = dict(pref_index=self.pref_index, direction=self.direction, ethertype=self.ethertype)
        vminf_name = self.source
        cmd = [delete_qos_marking_rule_to_valid_str(rule_params, vminf_name)]
        return cmd

    @classmethod
    def from_json(cls, data: dict) -> 'VMachinesQoSPolicyMarkingRule':
        return VMachinesQoSPolicyMarkingRule(**data)
