"""
(c) 2023 «Dacom M» Ltd
"""

from ast import literal_eval
from collections import namedtuple
from copy import deepcopy
import csv
from deepdiff import DeepDiff
from functools import reduce
import json
from ipaddress import IPv4Interface
from itertools import repeat
import os
import platform
import re
import time
from typing import List, Tuple, Union, Optional, Dict
from secrets import SystemRandom
import xmltodict
from uuid import UUID

from veilrpc.utils.bazalt import CommandExecutionError, get_output, states_response, states_result, \
    is_service_active, action_service, check_file_exists, UTILS_OTHER_PATH, get_config, \
    BASE_PATH, BASE_LOG_PATH, NODE_PATH, LOCAL_MODULES_PATH, thread_pool
from veilrpc.utils.networking import VNService, TrafficBlockRuleSet, TrafficBlockRule, SriovStateRuleSet, \
    SriovStateRule, PortMirroringRuleSet, PortMirroringRule, create_qos_marking_rule_to_valid_str, \
    delete_qos_marking_rule_to_valid_str, VMachinesQoSPolicyMarkingRuleSet, VMachinesQoSPolicyMarkingRule

# region Consts and classes
BASE_LOG_PATH = f'{BASE_LOG_PATH}/node'
BASE_HOSTS_FILE = '/etc/hosts'
HOSTS_CONF_DIR = '/etc/hosts.d'
HOSTS_UPD_UTIL = '/usr/local/sbin/veil-hosts'
BASE_FIREWALL_CONF = '/etc/nftables.conf'
BASE_DHCP_CONF = '/etc/kea/kea-dhcp4.conf'
BASE_LLDPD_CONF = '/etc/default/lldpd'
ADD_LLDPD_CONF = '/etc/lldpd.d/lldpd_config.conf'
ADD_LLDPD_DIR = '/etc/lldpd.d/'
PORT_LLDPD_CONF = '/etc/lldpd.d/lldpd_config_port_{lldpd_port}.conf'
LLDPD_PROTOCOLS = {'lldp': '',
                   'lldp_forced': '-l',
                   'lldp_disabled': '-ll',
                   'cdpv1_cdpv2': '-c',
                   'cdpv1_forced(cdpv2_enabled)': '-cc',
                   'cdpv2_forced(cdpv1_enabled)': '-ccc',
                   'cdpv2(cdpv1_disabled)': '-cccc',
                   'cdpv2_forced(cdpv1_disabled)': '-ccccc',
                   'edp': '-e',
                   'edp_forced': '-ee',
                   'fdp': '-f',
                   'fdp_forced': '-ff',
                   'sonmp': '-s',
                   'sonmp_forced': '-ss',
                   }
LLDPD_NEI_FIELDS = ['local_port', 'protocol', 'chassis_id_subtype', 'chassis_id', 'port_id_subtype', 'port_id',
                    'system_name', 'system_description', 'port_description', 'port_mfs', 'vlan_vlan_id', 'vlan_pvid',
                    'time_to_live', 'supported_system_capabilities', 'enabled_system_capabilities', 'mgmt_ip', 'age',
                    'rid']
LLDPD_CONF_ADD_FIELDS = ['chassis_id', 'mgmt_ip', 'system_name', 'system_description', 'tx_interval', 'tx_hold']
BASE_NET_DIR = '/etc/network'
BASE_NET_CONFIG = f'{BASE_NET_DIR}/interfaces'
BASE_NET_CONF_DIR = f'{BASE_NET_DIR}/interfaces.d'
BASE_NETWORK_SERVICE = 'networking.service' \
    if os.path.exists('/lib/systemd/system/networking.service') else 'network.service'
LXC_TEMPLATE = f'{BASE_PATH}/service_template/veil_template.tar.gz'
BASE_LXC_PATH = '/var/lib/lxc'
BASE_LXC_CONTAINER = 'vns-template'
FIREWALL_SERVICE = 'nftables.service'
FIREWALL_LOGGING_SERVICE = 'ulogd2.service'
DHCP_SERVICE = 'kea-dhcp4-server.service'
LLDPD_SERVICE = 'lldpd.service'
LLDPD_PATH = f'{LOCAL_MODULES_PATH}/lldpd.yaml'
BASE_LXC_CONFIG = [
    'lxc.rootfs.path = dir:/var/lib/lxc/vns-template/rootfs',
    'lxc.include = /usr/share/lxc/config/debian.common.conf',
    'lxc.tty.max = 4',
    'lxc.apparmor.profile = unconfined',
    'lxc.cgroup.devices.allow = a',
    'lxc.log.level = 1',
    'lxc.log.file = /var/log/veil/node/vns-template.log',
    'lxc.uts.name = vns-template',
    'lxc.cgroup.memory.limit_in_bytes = 256M'
]
LXC_NETWORK_KEYWORD = 'lxc.network'
LXC_UP_SCRIPTS_PATTERN = '_up.sh'
LXC_INTERNAL_INTERFACE = 'eth0'
LXC_EXTERNAL_INTERFACE = 'eth1'
TEMP_LXC_CMD_OUT = '/tmp/cmd_out'

ROUTE_INIT_PATH = f'{NODE_PATH}/bin/route.sh'
RESOLF_DIR = '/etc/resolvconf'
FILE_RESOLV_TAIL = f'{RESOLF_DIR}/resolv.conf.d/tail'
FILE_RESOLV_BASE = f'{RESOLF_DIR}/resolv.conf.d/base'
FILE_RESOLV_CONF = '/etc/resolv.conf'
FILE_RESOLV_STORED_DHCLIENT = '/run/resolvconf/interface'
FILE_DHCLIENT_CONF = '/etc/dhcp/dhclient.conf'
FILE_DHCLIENT_LEASES = '/var/lib/dhcp/dhclient.leases'
DHCLIENT_LEASES_DIR = '/var/lib/dhcp'
INTERFACE_LEASES = 'dhclient.{name}.leases'

EXCLUDED_PORTS = ('lo', 'ovs-system', 'veil-blackhole', 'veil-default', 'veil-mgmt')
SYS_CLASS_NET_DIR = '/sys/class/net'
SYS_UDEV_RULES_DIR = '/etc/udev/rules.d'
VEIL_MGMT_INTERFACE = 'veil-mgmt'
DEFAULT_SWITCH = 'veil-default'
BLACKHOLE_SWITCH = 'veil-blackhole'

OUTDATED_INTERFACES_CONF = re.compile(r'(interfaces-[0-9-]{4,5}[0-9-]{2,3}[0-9-]{2,3}[0-9-:]{2,3}[0-9-:]{2,3}[0-9]{2})')
VIRSH_HOST_NET_PATTERN = re.compile(r'(net_)(?P<port>.*)((_[a-f0-9]{2}){6})')
PROC_NET_DEV_PATTERN = re.compile(r'(.*):(.*)')
VXLAN_INF_PATTERN = re.compile(r'vxlan_sys_')
GENEV_INF_PATTERN = re.compile(r'genev_sys_')
IFCONFIG_STATE_PATTERN = re.compile(r'(.*<)(.*)(>.*)')
DHCLIENT_OPTION_PATTERN = re.compile(r'(?P<assignments>^\b\w*option\w*\s\b.*;)')
DHCLIENT_SEND_PATTERN = re.compile(r'(?P<assignments>^\b\w*send\w*\s\b.*;)')
DHCLIENT_REQUEST_PATTERN = re.compile(r'^\b\w*request\w*\s\b(?P<requests>.*);')
DHCLIENT_INTERFACE_PATTERN = re.compile(r'^(\b\w*interface\s)\"(?P<interface>.*)\"\s*'
                                        r'{\s*.*\b\w*request\s(?P<requests>.*);\s}.*')
DHCLIENT_LEASES_PATTERN = re.compile(r'dhclient.(?P<interface>[a-zA-Z0-9-_]{1,15}).leases')
LXC_INTERFACE_PATTERN = re.compile(r'^[0-9a-f]{4}eth[0-1]$')
LINK_FLIE_PATTERN = r'(\d+)-(\S+).link'
LINK_DIR = "/etc/systemd/network/"
DHCLIENT_GLOBAL_PARAMS = ['subnet-mask', 'broadcast-address', 'time-offset', 'dhcp6.fqdn', 'dhcp6.sntp-servers',
                          'netbios-name-servers', 'netbios-scope', 'interface-mtu', 'ntp-servers']
DHCLIENT_INITIAL_PARAMS = ['routers', 'domain-name-servers', 'domain-name', 'domain-search']

CONNECTION_LOSS_REPLIES = ('100% packet loss', 'Network is unreachable', 'Destination Host Unreachable')

DHCLIENT_DNS_PARAMS = ('domain-name-servers', 'domain-name', 'domain-search')
DHCLIENT6_DNS_PARAMS = ('dhcp6.name-servers', 'dhcp6.domain-search')

UPLINK_MTU = 12000
DEFAULT_MTU = 1500
VEIL_MGMT_DEFAULT_VLAN = dict(vlan_mode='access', vlan_tag=None, vlan_trunks=None)

DHCLIENT_REQ_SET = namedtuple("DhclientReqSet", "gateway, dns")(
    gateway={'routers', },
    dns={'domain-name-servers', "domain-name", 'domain-search'}
)

SYSTEM_DEVICE_NAMES = ('ovs-system', 'vxlan_sys_', 'genev_sys_')


class CommonTcpServices:
    http = ('tcp', ['80'])
    http_alt = ('tcp', ['8080'])
    https = ('tcp', ['443'])
    ftp = ('tcp', ['20', '21'])
    ssh = ('tcp', ['22'])
    telnet = ('tcp', ['23'])
    smtp = ('tcp', ['25'])
    smtp_alt = ('tcp', ['587'])
    ssl_smtp = ('tcp', ['465'])
    pop3 = ('tcp', ['110'])
    ssl_pop3 = ('tcp', ['995'])
    imap = ('tcp', ['143'])
    ssl_imap = ('tcp', ['993'])
    kerberos = ('tcp', ['88'])
    netbios = ('tcp', ['137', '138', '139'])
    iscsi = ('tcp', ['860', '3260'])
    rpcbind = ('tcp', ['111'])
    nfs = ('tcp', ['2049'])


class CommonUdpServices:
    dhcp = ('udp', ['67', '68'])
    dns = ('udp', ['53'])
    ntp = ('udp', ['123'])
    snmp = ('udp', ['161'])
    snmp_trap = ('udp', ['162'])
    netbios = ('udp', ['137', '138', '139'])
    rpcbind = ('udp', ['111'])
    nfs = ('udp', ['2049'])


class Action:
    accept = 'accept'
    drop = 'drop'


# OVS Flow parameters
FIREWALL_BACKUP_PATH = '/var/log/veil/node/net_backups/'

# OVS Flow parameters
OVS_MAC_LEARNING_AGING_TIME = 300


# Priority
class OvsPriority:
    # ACL rules
    default = 0
    user_acl = (16, 64511)
    system_acl = (64512, 65535)
    # MAC rules
    mac_learning = 100
    port_security = 200


# OVS Flow tables
class OvsTable:
    acl = 0
    mac_src = 1
    mac_select = 2
    mac_dst = 10
    output = 3


OVS_FLOW_PRESETS = dict(
    DEFAULT=f"priority={OvsPriority.default}, actions=NORMAL",
    STP_DROP=f"priority={OvsPriority.system_acl[0] + 2},dl_dst=01:80:c2:00:00:00/ff:ff:ff:ff:ff:f0 "
             f"actions={Action.drop}",
    BROADCAST_SRC_DROP=f"priority={OvsPriority.system_acl[0] + 1},dl_src=01:00:00:00:00:00/01:00:00:00:00:00 "
                       f"actions={Action.drop}",
    ARP_ALLOW=f"priority={OvsPriority.system_acl[0]},arp actions=resubmit(,{OvsTable.output})",
    MAC_LEARNING=f"table={OvsTable.mac_src},priority={OvsPriority.default + 1} actions="
                 f"learn(table={OvsTable.mac_dst},hard_timeout={OVS_MAC_LEARNING_AGING_TIME},"
                 f"NXM_OF_ETH_DST[]=NXM_OF_ETH_SRC[],load:NXM_OF_IN_PORT[]->NXM_NX_REG0[0..15]),"
                 f"resubmit(,{OvsTable.mac_select})",
    ACL_MODE=[
        f"table={OvsTable.acl},priority={OvsPriority.default} actions={Action.drop}",
        f"table={OvsTable.mac_src},priority={OvsPriority.default} actions={Action.drop}",
        f"table={OvsTable.mac_select},priority={OvsPriority.default} "
        f"actions=resubmit(,{OvsTable.mac_dst}),resubmit(,{OvsTable.output})",
        f"table={OvsTable.output},priority={OvsPriority.default} actions=NORMAL",
    ],
    PORT_SECURITY=[
        f"table={OvsTable.mac_src},priority={OvsPriority.port_security},in_port={{port}},dl_src={{mac}} "
        f"actions=resubmit(,{OvsTable.mac_dst})",
        f"table={OvsTable.mac_src},priority={OvsPriority.port_security},dl_dst={{mac}} actions=output:{{port}}"
    ]
)

# перечислить нужные протоколы, которые будут задаваться с фронта
# cat /etc/protocols
IP_PROTOCOLS = {'ip': 0,
                'hopopt': 0,
                'icmp': 1,
                'igmp': 2,
                'ggp': 3,
                'ipencap': 4,
                'st': 5,
                'tcp': 6,
                'egp': 8,
                'igp': 9,
                'pup': 12,
                'udp': 17,
                }

QOS_POLICY_MARK_RULE = ['qos_policy', 'source', 'pref_index', 'direction', 'dscp', 'cos', 'ethertype', 'vlan_id',
                        'src_mac', 'dst_mac', 'ip_protocol', 'src_ip', 'dst_ip', 'src_port', 'dst_port', 'is_enabled',
                        'rule_index', 'id', 'vnetwork']


class UplinkState:
    active = 'active'
    backup = 'backup'


class DhclientReqType:
    gateway = 'gateway'
    dns = 'dns'


class DhclientDnsRequestParams:
    nameservers = 'domain-name-servers'
    search = 'domain-search'
    domain = 'domain-name'


class DhclientRoutersRequestParams:
    gateway = 'routers'


class VnsUnexpectedErrors:
    errno1 = 'Error 0001: Unable to create folder for base container template.'
    errno2 = 'Error 0002: Unable to unpack container template root file system.'
    errno3 = 'Error 0003: Container template does not exists.'
    errno4 = 'Error 0004: Unable to set up/down script executable.'


class VnsInterfaceDirection:
    internal = 'internal'
    external = 'external'


class OvsPortType:
    system = 'system'
    internal = 'internal'
    patch = 'patch'
    vxlan = 'vxlan'
    geneve = 'geneve'
    bond = 'bond'


class AddVeilPortType:
    none = 'none'
    vm = 'vm'
    vnservices = 'vnservices'
    physical = 'physical'


class ServiceActions:
    start = 'start'
    restart = 'restart'
    stop = 'stop'


class InterfaceModes:
    dhcp = 'dhcp'
    static = 'static'
    manual = 'manual'
    loopback = 'loopback'


class IfconfigTarget:
    host = 'host'
    container = 'container'


class IpRouteAction:
    add = 'add'
    delete = 'del'


class OvsRequestParams:
    ofport = 'ofport'
    vlan_mode = 'vlan_mode'
    vlan_tag = 'tag'
    vlan_trunks = 'trunks'
    mtu = 'mtu'
    mtu_request = 'mtu_request'
    interface_type = 'type'
    name = 'name'
    interfaces = 'interfaces'
    bond_mode = 'bond_mode'
    bond_lacp = 'lacp'
    other_config = 'other_config'
    netflow = 'netflow'
    rstp = 'rstp_enable'
    peer = 'options:peer'
    remote_ip = 'options:remote_ip'
    key = 'options:key'
    local_tep = 'status:tunnel_egress_iface'
    error = 'error'
    mac_addr = 'mac'
    external_ids = 'external-ids'
    external_networks = 'external-ids:external-networks'
    virtual_networks = 'external-ids:virtual-networks'


class OvsNetflowRequestParams:
    active_timeout = 'active_timeout'
    id_to_iface = 'add_id_to_interface'
    targets = 'targets'


class EthtoolParams:
    speed = 'Speed'
    duplex = 'Duplex'
    link = 'Link detected'


class UplinkTypes:
    physical_inf = 'physical_inf'
    aggregated_inf = 'aggregated_inf'


MAC_ADDRESS_RE = "[0-9a-f]{2}([-:])[0-9a-f]{2}(\\1[0-9a-f]{2}){4}$"


def check_hwaddress(mac_address):
    if not re.match(MAC_ADDRESS_RE, mac_address.lower()):
        error = f'{mac_address} mac-address is invalid.\n'
        return False, error
    return True, None


# метка наших MAC адресов
VEIL_MAC = (0x02, 0xff, 0xf0)


def generate_mac(mac_part=None) -> str:
    """Генерирует случайный MAC-адрес"""
    _rand = SystemRandom()
    mac = VEIL_MAC + (mac_part or _rand.randint(0x00, 0x7f), _rand.randint(0x00, 0xff), _rand.randint(0x00, 0xff))
    return ':'.join(map(lambda x: "%02x" % x, mac))


def check_ip(ip: str) -> bool:
    """Проверяет валидность адреса"""
    result = True
    try:
        IPv4Interface(ip)
    except ValueError:
        result = False
    return result


# endregion


# region general entities states & functions
def gen_valid_ovs_params(ptype: str, **kwargs) -> str:
    """
    Генерация валидной строки с параметрами для передачи OVS
    :param ptype: тип интерфейса OVS
    :param kwargs: параметры
    """
    if ptype == OvsPortType.internal:
        params_string = f' mac=\\"{kwargs["mac_address"]}\\"'
    elif ptype == OvsPortType.patch:
        params_string = f' options:peer={kwargs["peer"]}'
    elif ptype in (OvsPortType.vxlan, OvsPortType.geneve):
        params_string = f' options:remote_ip={kwargs["remote_ip"]} options:key={kwargs["vni"]}'
    elif ptype == OvsPortType.bond:
        params_string = f' bond_mode={kwargs["bond_mode"]} lacp={kwargs["bond_lacp"]}'
        if kwargs["lacp_time"] and kwargs["lacp_fallback"]:
            params_string = params_string + f' other_config:lacp-time={kwargs["lacp_time"]} other_config:lacp-fallback-ab={kwargs["lacp_fallback"]}'
        if isinstance(kwargs.get("rebalance_interval"), int):
            params_string = params_string + f' other_config:bond-rebalance-interval={kwargs["rebalance_interval"]}'
    else:
        return ''
    return params_string


def get_ovs_tun_interface(remote_ip: str, key: int):
    """
    Получение интерфейса туннеля по параметрам
    :param remote_ip: ip-адрес пира
    :param key: идентификатор виртуальной сети (VNI)
    """

    cmd = f'sudo ovs-vsctl find int options:remote_ip={remote_ip} options:key={key}'
    cmd_ret = states_result(get_output, cmd=cmd)
    int_name_str = next((s for s in cmd_ret if s.startswith('name')), None)
    return int_name_str.split(' : ')[1].lstrip('"').rstrip('"') if int_name_str is not None else int_name_str


@states_response
def check_ovs_tun_interface(remote_ip: str, key: int, name: str, vswitch: str):
    """
    Проверка существования интерфейса туннеля по параметрам на вирт. коммутаторе
    :param remote_ip: ip-адрес пира
    :param key: идентификатор виртуальной сети (VNI)
    :param name: имя интерфейса
    :param vswitch: имя виртуального коммутатора
    """

    cmd = f'sudo ovs-vsctl find int options:remote_ip={remote_ip} options:key={key} name={name}'
    return True if get_bridge_of_port(name) == vswitch and get_bridge_of_interface(name) == vswitch \
                   and states_result(get_output, cmd=cmd) else False


def create_ovs_bridge(name: str):
    """
    Создание OVS-коммутатора
    :param name: название коммутатора
    """

    cmd = f'sudo ovs-vsctl --may-exist add-br {name}'
    states_result(get_output, base_err_msg=f"Failed to create switch {name}", cmd=cmd)


def remove_ovs_bridge(name: str):
    """
    Удаление OVS-коммутатора
    :param name: название коммутатора
    """

    cmd = f'sudo ovs-vsctl --if-exists del-br {name}'
    states_result(get_output, base_err_msg=f"Failed to remove switch {name}", cmd=cmd)


def enable_rstp(switch: str):
    """
    Включение RSTP на коммутаторе
    :param switch: название коммутатора
    """

    cmd = f'ovs-vsctl set br {switch} rstp_enable=true'
    states_result(get_output, base_err_msg=f"Unable to turn on RSTP on switch {switch}", cmd=cmd)


def add_port_to_switch(switch: str, port: str, ptype: str, **kwargs):
    """
    Добавление порта к OVS-коммутатору
    :param switch: название коммутатора
    :param port: название порта
    :param ptype: тип порта
    :param kwargs: параметры порта
    """

    cmd = f"sudo ovs-vsctl --may-exist add-port {switch} {port}"
    if ptype == OvsPortType.system:
        cmd += f" -- set int {port} type={ptype}"
        states_result(get_output, cmd=cmd)
    elif ptype in (OvsPortType.internal, OvsPortType.patch, OvsPortType.vxlan, OvsPortType.geneve):
        cmd += f" -- set int {port} type={ptype}{gen_valid_ovs_params(ptype, **kwargs)}"
        states_result(get_output, cmd=cmd)
    else:
        cmd = f"sudo ovs-vsctl --may-exist add-bond {switch} {port}"
        members = kwargs.pop('members')
        cmd += f" {' '.join(members)}{gen_valid_ovs_params(ptype, **kwargs)}"
        states_result(get_output, cmd=cmd)


def add_patches(obverse_switch: str, reverse_switch, obverse_patch, reverse_patch):
    """
    Соединение патчами двух OVS-коммутаторов одной командой

    :param obverse_switch:
    :param reverse_switch:
    :param obverse_patch:
    :param reverse_patch:
    """

    cmd = f"ovs-vsctl add-port {obverse_switch} {obverse_patch} -- " \
          f"set int {obverse_patch} type=patch options:peer={reverse_patch} -- " \
          f"add-port {reverse_switch} {reverse_patch} -- " \
          f"set int {reverse_patch} type=patch options:peer={obverse_patch}"
    states_result(get_output, cmd=cmd)


def del_port_from_switch(port: str, switch: str = None):
    """
    Удаление порта из OVS-коммутатора
    :param port: название порта
    :param switch: название коммутатора
    """

    cmd = 'ovs-vsctl --if-exists del-port '
    cmd += f'{port}' if switch is None else f'{switch} {port}'
    states_result(get_output, cmd=cmd)


@states_response
def del_port_from_switch_response(port, switch):
    """
    Удаление порта из OVS-коммутатора (как response)
    :param port: название порта
    :param switch: название коммутатора
    """
    del_port_from_switch(port, switch)


def set_vlan_config(interface: str, vlan_mode: Optional[str], vlan_tag: Optional[int], vlan_trunks: Optional[list]):
    """
    Установка параметров VLAN порта
    :param interface: название порта
    :param vlan_mode: режим VLAN
    :param vlan_tag: тэг VLAN
    :param vlan_trunks: список тэгов VLAN, которые будет пропускать интерфейс
    """

    mode = vlan_mode if vlan_mode else "[]"
    tag = vlan_tag if isinstance(vlan_tag, int) else "[]"
    trunks = f"[{','.join([str(v) for v in vlan_trunks])}]" if vlan_trunks else "[]"
    cmd = f'ovs-vsctl set port {interface} vlan_mode={mode} tag={tag} trunks={trunks}'
    states_result(get_output, base_err_msg=f"Failed to set vlan configuration on port {interface}", cmd=cmd)


def rollback_port_pgroup_config(interface, initial_config):
    """
    Установка исходных значений параметров VLAN и mtu на интерфейсе (откат изменений)
    """

    set_vlan_config(interface=interface, vlan_mode=initial_config['vlan_mode'], vlan_tag=initial_config.get('vlan_tag'),
                    vlan_trunks=initial_config.get('vlan_trunks'))
    if initial_config.get('mtu'):
        set_ovs_mtu(interface, initial_config['mtu'])


def check_is_ovs_bond(interface: str) -> bool:
    """
    Проверка интерфейс OVS bonded или нет
    :param interface: название интерфейса
    :return: булевое значение, определяющее результат проверки
    """
    cmd = f"ovs-appctl bond/show {interface} | grep 'bond_mode' | awk '{{print $2}}'"
    bond_mode = get_output(cmd, suppress_err=True).get('comment').strip()
    return bond_mode in ['balance-tcp', 'balance-slb', 'active-backup']


def set_ovs_bond_mtu(interface: str, mtu: int):
    """
    Устанавливает значение MTU на все интерфейсы, входящие в OVS bonded интерфейс
    :param interface: название интерфейса
    :param mtu: значение MTU
    """
    cmd = f"ovs-vsctl -d bare list port {interface} | grep interfaces"
    result = states_result(get_output, base_err_msg=f"Failed to get bond {interface} interfaces", cmd=cmd)
    if interfaces := result.split(':')[1].strip().split(' '):
        set(map(set_ovs_mtu, interfaces, [mtu] * len(interfaces)))


def set_ovs_mtu(interface: str, mtu: int):
    """
    Установка значения mtu для интерфейса OVS
    :param interface: название интерфейса
    :param mtu: значение mtu
    """

    cmd = f'ovs-vsctl set int {interface} mtu_request={mtu}'
    states_result(get_output, base_err_msg=f"Failed to set mtu configuration on port {interface}", cmd=cmd)


def set_ovs_mac_address(interface: str, mac_address: int):
    """
    Установка значения mac-адреса для интерфейса OVS
    :param interface: название интерфейса
    :param mac_address: значение mac-адреса
    """

    cmd = f'ovs-vsctl set int {interface} mac=\\"{mac_address}\\"'
    states_result(get_output, base_err_msg=f"Failed to update mac-address for interface {interface}", cmd=cmd)


def set_ovs_hwaddress(bridge: str, hwaddr: str):
    """
    Установка значения mac-адреса для коммутатора OVS
    :param bridge: название коммутатора
    :param hwaddr: значение mac-адреса
    """
    cmd = f'ovs-vsctl set bridge {bridge} other-config:hwaddr=\\"{hwaddr}\\"'
    states_result(get_output, base_err_msg=f"Failed to set hardware address for bridge {bridge}", cmd=cmd)


def set_dev_mtu(interface: str, mtu: int):
    """
    Установка значения mtu для физического интерфейса
    """

    cmd = f"ip link set dev {interface} mtu {mtu}"
    states_result(get_output, base_err_msg=f"Failed to set mtu configuration on port {interface}", cmd=cmd)


def set_arp_timeout(bridge: str, arp_timeout: int):
    """
    Установка значения ARP timeout для коммутатора OVS
    :param bridge: название коммутатора
    :param arp_timeout: значение в секундах. По умолчанию 60.
    """
    cmd = f'echo {arp_timeout} > /proc/sys/net/ipv4/neigh/{bridge}/gc_stale_time'
    states_result(get_output, base_err_msg=f"Failed to set ARP timeout for bridge {bridge}", cmd=cmd)


def get_arp_timeout(bridge: str):
    """
    Отображение значения ARP timeout для коммутатора OVS
    """
    cmd = f'cat /proc/sys/net/ipv4/neigh/{bridge}/gc_stale_time'
    return states_result(get_output, cmd=cmd).strip()


def set_port_up(interface: str):
    """
    Включение порта
    :param interface: название порта
    """

    if states_result(check_port_up_state, port=interface) is False:
        cmd = f'ifup {interface}'
        states_result(get_output, base_err_msg=f"Failed to set interface {interface} up", cmd=cmd)


def set_port_down(interface: str):
    """
    Выключение порта
    :param interface: название порта
    """

    if states_result(check_port_up_state, port=interface) is True:
        cmd = f'ifdown {interface}'
        states_result(get_output, base_err_msg=f"Failed to set interface {interface} down", cmd=cmd)


def dhclient_release(interface: str):
    """
    Освобождение адреса dhcp
    :param interface: название интерфейса
    """

    cmd = f'dhclient -r {interface}'
    states_result(get_output, base_err_msg=f"Failed to release address for interface {interface}", cmd=cmd)


def dhclient_request(interface: str):
    """
    Запрос адреса dhcp
    :param interface: название интерфейса
    """

    cmd = f'dhclient {interface}'
    states_result(get_output, base_err_msg=f"Failed to request address for interface {interface}", cmd=cmd)


def write_dhclient_conf(config: dict, timeout=20, retry=60):
    """
    Запись конфигурации dhcp запросов
    :param config: конфигурация для записи
    :param timeout: интервал времени в секундах, по истечении которого клиент считает, что dhcp сервер недоступен
    и перестает отправлять dhcp-request в данной сессии (по-умолчанию timeout=60 сек.)
    (см.timeout в функции get_output и при 'ifup имя_интерфейса' запрос dhcp клиента будет продолжаться timeout времени)
    :param retry: период времени в секундах, через который происходит повторные попытки клиента связаться
    с dhcp сервером (по-умолчанию 300 сек.)
    """

    try:
        with open(FILE_DHCLIENT_CONF, 'w') as f:
            f.write(f"timeout {timeout};\n")
            f.write('\n'.join(config['assignments']) + '\n\n')
            f.write(f"request {', '.join(config['requests'])};\n\n")
            for item in config['interfaces']:
                if item['requests']:
                    f.write(f'interface \"{item["interface"]}\" {{ also request {", ".join(item["requests"])}; }}\n')
    except OSError as e:
        raise CommandExecutionError(
            "Failed to write configuration. I/O error({0}): {1}".format(e.errno, e.strerror)) from None
    except Exception as e:
        raise CommandExecutionError("Unexpected error: {0}".format(str(e))) from None


def set_aggregation_config(interface: str, **kwargs):
    """
    Применение конфигурации агрегированного интерфейса
    """

    cmd = f'ovs-vsctl set port {interface}{gen_valid_ovs_params(OvsPortType.bond, **kwargs)}'
    states_result(get_output, base_err_msg=f"Failed to update configuration for interface {interface}", cmd=cmd)


def add_port_to_aggregation(bond: str, port: str):
    """
    Добавление порта в агрегацию
    :param bond: название агрегированного интерфейса
    :param port: название добавляемого порта
    """

    cmd = f'ovs-vsctl --may-exist add-bond-iface {bond} {port}'
    states_result(get_output, base_err_msg=f"Failed to add interface {port} to aggregated port {bond}", cmd=cmd)


def del_port_from_aggregation(bond: str, port: str):
    """
    Удаление порта из агрегации
    :param bond: название агрегированного интерфейса
    :param port: название удаляемого порта
    """

    cmd = f'ovs-vsctl --if-exists del-bond-iface {bond} {port}'
    states_result(get_output, base_err_msg=f"Failed to remove interface {port} from aggregated port {bond}", cmd=cmd)


def purge_leases(interfaces: Union[str, list]):
    """
    Удаление файлов аренды ip-адреса интерфейса
    """

    if isinstance(interfaces, str):
        del_file_by_pattern(DHCLIENT_LEASES_DIR, INTERFACE_LEASES.format(name=interfaces))
    if isinstance(interfaces, list):
        for item in interfaces:
            del_file_by_pattern(DHCLIENT_LEASES_DIR, INTERFACE_LEASES.format(name=item))


def deprecate_leases():
    """
    Удаление файлов аренды несуществующих интерфейсов
    """

    for f in os.listdir(DHCLIENT_LEASES_DIR):
        match = re.search(DHCLIENT_LEASES_PATTERN, f)
        if match:
            interface = match.group('interface')
            if not port_exists_in_ovs(interface):
                os.remove(os.path.join(DHCLIENT_LEASES_DIR, f))


def enable_promisc_mode(interface: str):
    cmd = f"ifconfig {interface} promisc"
    states_result(get_output, base_err_msg=f"Failed to enable promiscuous mode on interface {interface}", cmd=cmd)


def disable_promisc_mode(interface: str):
    cmd = f"ifconfig {interface} -promisc"
    states_result(get_output, base_err_msg=f"Failed to disable promiscuous mode on interface {interface}", cmd=cmd)


def set_vfunctions(port: str, vf_num: int) -> None:
    """
    Создает виртуальные функции у физического интерфейса
    :param port: название интерфейса
    :param vf_num: количество создаваемых виртуальных функций (всегда максимальное, т.к. для изменения их количества,
    надо сначала все удалить, потом заново создать)
    """

    cmd = f"echo {vf_num} > {os.path.join(SYS_CLASS_NET_DIR, port, 'device', 'sriov_numvfs')}"
    action = 'create' if vf_num else 'remove'
    states_result(get_output, base_err_msg=f"Failed to {action} virtual functions on interface {port}", cmd=cmd)


def set_vfunctions_persistent(port: str, vf_num: int) -> None:
    """
    Создает конфиг для того, чтобы виртуальные функции восстанавливались при перезагрузке узла
    :param port: название интерфейса
    :param vf_num: количество создаваемых виртуальных функций (всегда максимальное, т.к. для изменения их количества,
    надо сначала все удалить, потом заново создать)
    """

    rs = SriovStateRuleSet.from_config_file()
    data = dict(interface=port, vfunctions=vf_num)
    if not rs.add_rule(SriovStateRule.from_json(data), save=True):
        raise CommandExecutionError(f"Failed to set virtual functions on interface {port} persistent.")


def unset_vfunctions_persistent(port: str) -> None:
    """
    Удаляет конфиг восстановления виртуальных функций при перезагрузке узла
    :param port: название интерфейса
    """

    rs = SriovStateRuleSet.from_config_file()
    params = dict(interface=port)
    rule = rs.find_rule(params)
    if not rule:
        return
    if not rs.remove_rule(rule, save=True):
        raise CommandExecutionError(f"Failed to unset virtual functions on interface {port} persistent.")


def get_vfs_info(dev: str) -> list:
    """
    Собирает информацию о существующих виртуальных функциях физического сетевого устройства
    :param dev: название физического интерфейса
    """

    vfs_info = list()
    vfs_devices = [item for item in os.listdir(os.path.join(SYS_CLASS_NET_DIR, dev, 'device'))
                   if item.startswith('virtfn')]
    for vf_device in vfs_devices:
        vf_data = dict(name='', mac_address='', pci_domain='', pci_bus='', pci_slot='', pci_function='')
        vfs_path = os.path.join(SYS_CLASS_NET_DIR, dev, 'device', vf_device)
        if 'net' not in os.listdir(vfs_path):
            continue
        vf_data['name'] = os.listdir(os.path.join(vfs_path, 'net'))[0]
        vf_data['mac_address'] = states_result(
            get_output, cmd=f"cat {os.path.join(vfs_path, 'net', vf_data['name'], 'address')}").splitlines()[0]
        output = states_result(
            get_output, cmd=f"readlink {vfs_path} | cut -d '/' -f2").splitlines()[0].replace('.', ':').split(':')
        vf_data['pci_domain'] = int(f"0x{output[0]}", 0)
        vf_data['pci_bus'] = int(f"0x{output[1]}", 0)
        vf_data['pci_slot'] = int(f"0x{output[2]}", 0)
        vf_data['pci_function'] = int(f"0x{output[3]}", 0)
        vfs_info.append(vf_data)
    return vfs_info


@states_response
def blink(interface: str, timeout: int = 10):
    """
    Включение светодиодной индикации физического интерфейса
    :param interface: название интерфейса
    :param timeout: длительность индикации в секундах
    """

    cmd = f'/sbin/ethtool -p {interface} {timeout} &'
    states_result(get_output, base_err_msg=f"Failed to blink port {interface}", cmd=cmd)


@states_response
def create_switch(switch: str, netflow: dict = None, rstp_enable: bool = False, external_id: str = None,
                  virtual_id=None):
    """
    Создание ovs-коммутатора
    :param switch: название коммутатора
    :param netflow: параметры netflow
    :param rstp_enable: флаг, указывающий, будет ли включен rstp на коммутаторе
    :param external_id: идентификатор внешней сети
    :param virtual_id: идентификатор виртуальной стеи
    """

    create_ovs_bridge(switch)

    if netflow and netflow.get('netflow_target_ip'):
        set_netflow_params(switch, netflow)

    if rstp_enable:
        enable_rstp(switch)

    if external_id:
        cmd = f"ovs-vsctl set br {switch} external-ids:external-networks={external_id}"
        states_result(get_output, cmd=cmd)

    if virtual_id:
        cmd = f"ovs-vsctl set br {switch} external-ids:virtual-network={virtual_id}"
        states_result(get_output, cmd=cmd)

    return f'Switch {switch} created successfully.'


@states_response
def remove_switch(switch: str):
    """
    Удаление ovs-коммутатора
    :param switch: название коммутатора
    """

    remove_ovs_bridge(switch)
    return f'Switch {switch} removed successfully.'


@states_response
def remove_switches(switches: list):
    """
    Удаление ovs-коммутаторов
    :param switches: список коммутаторов
    """

    for switch in switches:
        remove_ovs_bridge(switch)
    return f"Switches {', '.join(switches)} removed successfully."


@states_response
def set_port_group_params(interfaces: list, old_config: dict, vlan_mode: Optional[str], vlan_tag: Optional[int],
                          vlan_trunks: Optional[list], mtu: Optional[int]):
    """
    Настройка vlan и mtu интерфейсов
    :param interfaces: список интерфейсов
    :param old_config: действующие параметры конфигурации
    :param vlan_mode: режим VLAN
    :param vlan_tag: тэг VLAN
    :param vlan_trunks: список тэгов VLAN, пропускаемых интерфейсом
    :param mtu: величина mtu
    """

    for interface in interfaces:
        old_vlan_config = (old_config.get('vlan_mode'), old_config.get('vlan_tag'),
                           set(old_config.get('vlan_trunks')) if old_config.get('vlan_trunks') else None)
        new_vlan_config = (vlan_mode, vlan_tag, set(vlan_trunks) if vlan_trunks else None)
        if old_vlan_config != new_vlan_config:
            set_vlan_config(interface=interface, vlan_mode=vlan_mode, vlan_tag=vlan_tag, vlan_trunks=vlan_trunks)
        if mtu != old_config.get('mtu'):
            if check_is_ovs_bond(interface=interface):
                set_ovs_bond_mtu(interface, mtu)
            else:
                set_ovs_mtu(interface, mtu)

    if not states_result(check_controller_connection, raise_exception=False):
        for interface in interfaces:
            rollback_port_pgroup_config(interface, old_config)
        raise CommandExecutionError("VLAN/MTU configuration update caused connection loss between node and "
                                    "controller. Changes cancelled.")

    return "VLAN/MTU configuration saved successfully."


@states_response
def create_internal_interface(interface: str, switch: str, ipv4_address: str, ipv4_netmask: str, mac_address: str,
                              dhcpv4: bool, config: dict):
    """
    Создание внутреннего интерфейса
    :param interface: название интерфейса
    :param switch: название коммутатора, в который подключается интерфейс
    :param ipv4_address: ip-адрес интерфейса
    :param ipv4_netmask: маска подсети интерфейса
    :param mac_address: mac-адрес интерфейса
    :param dhcpv4: флаг получения параметров по dhcp
    :param config: параметры vlan/mtu
    """

    ret = {'comment': '', 'changes': {}}

    mtu = config.pop('mtu')
    params = dict(conf_dir=BASE_NET_CONF_DIR, name=interface, hw_address=mac_address)
    if dhcpv4:
        params['mode'] = InterfaceModes.dhcp
    if not dhcpv4:
        params['mode'] = InterfaceModes.static
        params['ipv4_address'] = ipv4_address
        params['ipv4_netmask'] = ipv4_netmask
    write_interface(**params)
    add_port_to_switch(switch, interface, OvsPortType.internal, mac_address=mac_address)
    set_vlan_config(interface, **config)
    set_ovs_mtu(interface, mtu)
    if dhcpv4:
        time.sleep(5)
        ret['changes'] = states_result(get_address_info_real, interface=interface)

    ret['comment'] = f"Interface {interface} created successfully."
    return ret


@states_response
def modify_internal_interface(interface: str, old_vswitch: str, new_vswitch: str, old_group_config: dict,
                              new_group_config: dict, old_interface_config: dict, new_interface_config: dict):
    """
    Изменение параметров внутреннего интерфейса
    :param interface: название интерфейса
    :param old_vswitch: название "старого" коммутатора
    :param new_vswitch: название "нового" коммутатора
    :param old_group_config: параметры "старой" порт-группы
    :param new_group_config: параметры "новой" порт-группы
    :param old_interface_config: "старые" параметры интерфейса
    :param new_interface_config: "новые" параметры интерфейса
    """

    ret = {'comment': '', 'changes': {}}

    gateway_config = get_gateway_configuration()
    switch_modified = True if old_vswitch != new_vswitch else False
    pgroup_modified = True if bool(DeepDiff(old_group_config, new_group_config)) else False
    interface_modified = True if bool(DeepDiff(old_interface_config, new_interface_config)) else False

    dhcpv4 = new_interface_config['dhcpv4'] if interface_modified else old_interface_config['dhcpv4']
    if interface_modified:
        del_file_by_pattern(BASE_NET_CONF_DIR, interface)
        params = dict(conf_dir=BASE_NET_CONF_DIR, name=interface, hw_address=new_interface_config['mac_address'])
        if gateway_config['interface'] == interface and gateway_config['dhcpv4'] is False:
            params['ipv4_gateway'] = gateway_config['gateway']
        if new_interface_config['dhcpv4']:
            params['mode'] = InterfaceModes.dhcp
        if not new_interface_config['dhcpv4']:
            params['mode'] = InterfaceModes.static
            params['ipv4_address'] = new_interface_config['ipv4_address']
            params['ipv4_netmask'] = new_interface_config['ipv4_netmask']
            dhclient = parse_dhclient_conf()
            iface_in_dhclient = False
            iface_index = 0
            for idx, iface in enumerate(dhclient['interfaces']):
                if iface['interface'] == interface:
                    iface_in_dhclient = True
                    iface_index = idx
                    break
            if iface_in_dhclient:
                del dhclient['interfaces'][iface_index]
                write_dhclient_conf(dhclient)
        write_interface(**params)

    if switch_modified:
        del_port_from_switch(interface)
        add_port_to_switch(switch=new_vswitch,
                           port=interface,
                           ptype=OvsPortType.internal,
                           mac_address=new_interface_config['mac_address'] if interface_modified
                           else old_interface_config['mac_address'])

    mtu = new_group_config.pop('mtu') if pgroup_modified else old_group_config.pop('mtu')
    if pgroup_modified:
        set_vlan_config(interface, **new_group_config)
    elif switch_modified and not pgroup_modified:
        set_vlan_config(interface, **old_group_config)
    set_ovs_mtu(interface, mtu)

    if interface_modified:
        if not dhcpv4:
            purge_leases(interface)
        if new_interface_config['mac_address'] != old_interface_config['mac_address']:
            set_ovs_mac_address(interface, new_interface_config['mac_address'])

    if interface_modified:
        _exec = action_service(BASE_NETWORK_SERVICE, action=ServiceActions.restart)
        if not _exec['result']:
            raise CommandExecutionError(_exec['comment'])
        apply_routes_script()

    if dhcpv4:
        time.sleep(5)
        ret['changes'] = states_result(get_address_info_real, interface=interface)

    ret['comment'] = "Interface {} modified successfully.".format(interface)
    return ret


@states_response
def remove_internal_interface(interface: str):
    """
    Удаление внутреннего интерфейса
    :param interface: название интерфейса
    """

    ret = {'comment': '', 'changes': {}}

    del_port_from_switch(interface)
    del_file_by_pattern(BASE_NET_CONF_DIR, interface)
    purge_leases(interface)

    ret['comment'] = "Interface {} removed successfully.".format(interface)
    return ret


@states_response
def remove_internal_interfaces(interfaces: str):
    """
    Удаление внутреннего интерфейса
    :param interfaces: название интерфейса
    """

    ret = {'comment': '', 'changes': {}}

    for interface in interfaces:
        del_port_from_switch(interface)
        del_file_by_pattern(BASE_NET_CONF_DIR, interface)
    purge_leases(interfaces)

    ret['comment'] = f"Interfaces {', '.join(interfaces)} removed successfully."
    return ret


@states_response
def renew_interface_lease(interface: str):
    """
    Обновление адреса интерфейса
    :param interface: название интерфейса
    """

    ret = {'comment': '', 'changes': {}}

    dhclient_release(interface)
    purge_leases(interface)
    dhclient_request(interface)
    time.sleep(5)
    ret['changes'] = states_result(get_address_info_real, interface=interface)
    ret['comment'] = f"Interface {interface} address renewed successfully."
    return ret


@states_response
def set_interface_up(interface: str, check_conn: bool = True):
    """
    Включение интерфейса
    :param interface: название интерфейса
    :param check_conn:
    """

    ret = {'comment': '', 'changes': {}}

    set_port_up(interface)
    if check_conn and not states_result(check_controller_connection, raise_exception=False):
        set_port_down(interface)
        raise CommandExecutionError(f"Enabling interface {interface} caused connection loss between node "
                                    f"and controller. Changes cancelled.")

    ret['comment'] = f"Interface {interface} set to 'UP' state successfully."
    return ret


@states_response
def set_interfaces_up(interfaces: list, check_conn: bool = True):
    """
    Включение интерфейса
    :param interfaces: список интерфейсов
    :type interfaces: list
    :param check_conn:
    """

    ret = {'comment': '', 'changes': {}}

    for interface in interfaces:
        set_port_up(interface)
        if check_conn and not states_result(check_controller_connection, raise_exception=False):
            set_port_down(interface)
            raise CommandExecutionError(f"Enabling interface {interface} caused connection loss between node "
                                        f"and controller. Changes cancelled.")

    ret['comment'] = f"Interfaces {', '.join(interfaces)} set to 'UP' state successfully."
    return ret


@states_response
def set_interface_down(interface: str, check_conn: bool = True):
    """
    Включение интерфейса
    :param interface: название интерфейса
    :param check_conn:
    """

    ret = {'comment': '', 'changes': {}}

    set_port_down(interface)
    if check_conn and not states_result(check_controller_connection, raise_exception=False):
        set_port_up(interface)
        raise CommandExecutionError(f"Disabling interface {interface} caused connection loss between node "
                                    f"and controller. Changes cancelled.")

    ret['comment'] = f"Interface {interface} set to 'DOWN' state successfully."
    return ret


@states_response
def set_interfaces_down(interfaces: str, check_conn: bool = True):
    """
    Включение интерфейса
    :param interfaces: список интерфейсов
    :param check_conn:
    """

    ret = {'comment': '', 'changes': {}}

    for interface in interfaces:
        set_port_down(interface)
        if check_conn and not states_result(check_controller_connection, raise_exception=False):
            set_port_up(interface)
            raise CommandExecutionError(f"Disabling interface {interface} caused connection loss between node "
                                        f"and controller. Changes cancelled.")

    ret['comment'] = f"Interfaces {', '.join(interfaces)} set to 'DOWN' state successfully."
    return ret


@states_response
def connect_interface(interface: str, switch: str, config: dict, sys_parser: bool = False):
    """
    Подключение физического интерфейса к коммутатору
    """

    ret = {'comment': '', 'changes': {}}

    phys_infs = states_result(list_physical_infs, verbose=True, sys_parser=sys_parser)
    target_interface = next((iface for iface in phys_infs if iface['interface'] == interface), None)
    mtu = config.pop('mtu')
    write_interface(BASE_NET_CONF_DIR, interface, InterfaceModes.manual, is_connected=True,
                    promiscuous_mode=target_interface['promiscuous_mode'], mtu=target_interface['mtu'])
    add_port_to_switch(switch, interface, OvsPortType.system)
    set_vlan_config(interface, **config)
    set_ovs_mtu(interface, mtu)

    ret['comment'] = "Interface {} connected successfully.".format(interface)
    return ret


@states_response
def reconnect_interface(interface: str, old_switch: str, new_switch: str, old_config: dict, new_config: dict,
                        set_up: bool):
    """
    Переподключение физического интерфейса - изменение коммутатора и/или параметров порт-группы
    """

    ret = {'comment': '', 'changes': {}}

    switch_modified = True if old_switch != new_switch else False
    pgroup_modified = True if bool(DeepDiff(old_config, new_config)) else False

    if switch_modified:
        del_port_from_switch(interface, old_switch)
        add_port_to_switch(new_switch, interface, OvsPortType.system)

    mtu = new_config.pop('mtu') if pgroup_modified else old_config.pop('mtu')
    if pgroup_modified:
        set_vlan_config(interface, **new_config)
    elif switch_modified and not pgroup_modified:
        set_vlan_config(interface, **old_config)

    set_ovs_mtu(interface, mtu)
    # pylint: disable=expression-not-assigned
    set_port_up(interface) if set_up else set_port_down(interface)

    if not states_result(check_controller_connection, raise_exception=False):
        if switch_modified:
            del_port_from_switch(interface, new_switch)
            add_port_to_switch(old_switch, interface, OvsPortType.system)
        rollback_port_pgroup_config(interface, old_config)
        raise CommandExecutionError("Interface {} update caused connection loss between node and controller. "
                                    "Changes cancelled.".format(interface))

    ret['comment'] = "Interface {} reconnected successfully.".format(interface)
    return ret


@states_response
def disconnect_interface(interface: str, switch: str, check_conn: bool = True, sys_parser: bool = False):
    """
    Отключение физического интерфейса от коммутатора
    :param interface: название интерфейса
    :param switch: название коммутатора
    :param check_conn:
    :param sys_parser:
    """

    ret = {'comment': '', 'changes': {}}

    phys_infs = states_result(list_physical_infs, verbose=True, sys_parser=sys_parser)
    target_interface = next((iface for iface in phys_infs if iface['interface'] == interface), None)

    initial_config = states_result(get_port_pgroup_config, name=interface)
    del_port_from_switch(interface, switch)
    write_interface(BASE_NET_CONF_DIR, interface, InterfaceModes.manual, is_connected=False,
                    promiscuous_mode=target_interface['promiscuous_mode'], mtu=target_interface['mtu'])

    if check_conn and not states_result(check_controller_connection, raise_exception=False):
        add_port_to_switch(switch, interface, OvsPortType.system)
        rollback_port_pgroup_config(interface, initial_config)
        write_interface(BASE_NET_CONF_DIR, interface, InterfaceModes.manual, is_connected=True,
                        promiscuous_mode=target_interface['promiscuous_mode'], mtu=target_interface['mtu'])
        raise CommandExecutionError("Disconnecting interface {0} from switch {1} caused connection loss between "
                                    "node and controller. Changes cancelled.".format(interface, switch))

    ret['comment'] = "Interface {} disconnected successfully.".format(interface)
    return ret


@states_response
def change_mtu(interface: str, mtu: int, connected: bool, sys_parser: bool = False):
    """
    Изменение MTU физического интерфейса
    """

    ret = {'comment': '', 'changes': {}}

    phys_infs = states_result(list_physical_infs, verbose=True, sys_parser=sys_parser)
    target_interface = next((iface for iface in phys_infs if iface['interface'] == interface), None)
    # pylint: disable=expression-not-assigned
    set_ovs_mtu(interface, mtu) if connected else set_dev_mtu(interface, mtu)
    write_interface(BASE_NET_CONF_DIR, interface, InterfaceModes.manual, is_connected=connected,
                    promiscuous_mode=target_interface['promiscuous_mode'], mtu=mtu)

    ret['comment'] = f'MTU on interface {interface} changed successfully.'
    return ret


@states_response
def change_mtu_bulk(interfaces: str, mtu: int, sys_parser: bool = False):
    """
    Изменение MTU физического интерфейса

    :param interfaces: список интерфейсов
    :param mtu: значение MTU
    :param sys_parser:
    """

    ret = {'comment': '', 'changes': {}}

    phys_infs = states_result(list_physical_infs, verbose=True, sys_parser=sys_parser)
    for interface in interfaces:
        target_interface = next((iface for iface in phys_infs if iface['interface'] == interface), None)
        set_ovs_mtu(interface, mtu)
        write_interface(BASE_NET_CONF_DIR, interface, InterfaceModes.manual, is_connected=True,
                        promiscuous_mode=target_interface['promiscuous_mode'], mtu=mtu)

    ret['comment'] = f'MTU on interface {", ".join(interfaces)} changed successfully.'
    return ret


@states_response
def create_aggregation(interface: str, switch: str, members: list, connected_members: list, bond_config: dict,
                       ports_config: dict, ports_mtu: list, check_conn: bool = True, sys_parser: bool = False):
    """
    Создание агрегированного интерфейса (бондинга)
    :param interface: название интерфейса
    :param switch: название коммутатора
    :param members: список физических интерфейсов, из которых собирается бондинг
    :param connected_members: список физических интерфейсов, уже подключенных в какой-то коммутатор
    :param bond_config: параметры агрегации
    :param ports_config: параметры vlan/mtu
    :param ports_mtu:
    :param check_conn:
    :param sys_parser:
    """

    ret = {'comment': '', 'changes': {}}

    initial_config = list()

    if connected_members:
        for port in connected_members:
            port_bridge = get_bridge_of_port(port)
            port_config = states_result(get_port_pgroup_config, name=port)
            initial_config.append({'interface': port, 'switch': port_bridge, 'config': port_config})
            del_port_from_switch(port, port_bridge)
    add_port_to_switch(switch=switch, port=interface, ptype=OvsPortType.bond, members=members, **bond_config)
    set_vlan_config(interface, **ports_config)
    for entry in ports_mtu:
        states_result(change_mtu, interface=entry['port'], mtu=entry['mtu'], connected=True, sys_parser=sys_parser)

    if check_conn and not states_result(check_controller_connection, raise_exception=False):
        del_port_from_switch(interface, switch)
        if connected_members:
            for conf in initial_config:
                add_port_to_switch(conf['switch'], conf['interface'], OvsPortType.system)
                rollback_port_pgroup_config(conf['interface'], conf['config'])
        for port in members:
            write_interface(BASE_NET_CONF_DIR, port, InterfaceModes.manual, is_connected=False)
        raise CommandExecutionError("Creating interface {} caused connection loss between node and controller. "
                                    "Changes cancelled.".format(interface))

    ret['comment'] = "Interface {} created successfully.".format(interface)
    return ret


@states_response
def modify_aggregation(interface: str, bond_config: dict, check_conn: bool = True):
    """
    Изменение настроек агрегированного интерфейса (бондинга)
    :param interface: название интерфейса
    :param bond_config: параметры агрегации
    :param check_conn:
    """

    ret = {'comment': '', 'changes': {}}

    initial_config = states_result(get_bond_config, name=interface)
    set_aggregation_config(interface, **bond_config)

    if check_conn and not states_result(check_controller_connection, raise_exception=False):
        set_aggregation_config(interface, **initial_config)
        raise CommandExecutionError(f"Interface {interface} update caused connection loss between node "
                                    f"and controller. Changes cancelled.")

    ret['comment'] = f"Interface {interface} configuration updated successfully."
    return ret


@states_response
def remove_aggregation(interface: str, switch: str, initial_config: dict,
                       to_interface: Union[str, None] = None, set_up: bool = False,
                       check_conn: bool = True, sys_parser: bool = False):
    """
    Удаление агрегации
    :param interface: название агрегированного интерфейса
    :param switch: название коммутатора
    :param initial_config:
    :param to_interface:
    :param set_up:
    :param check_conn:
    :param sys_parser:
    """

    ret = {'comment': '', 'changes': {}}

    initial_bond_members = states_result(get_bond_members, name=interface)
    del_port_from_switch(interface, switch)
    for member in initial_bond_members:
        member_info = get_link_info_detailed(member)
        write_interface(BASE_NET_CONF_DIR, member, InterfaceModes.manual, is_connected=False,
                        promiscuous_mode=True if 'PROMISC' in member_info['flags'] else False,
                        mtu=member_info['mtu'])
        set_port_down(member)

    if to_interface:
        add_port_to_switch(switch, to_interface, OvsPortType.system)
        set_vlan_config(to_interface, initial_config['port_config']['vlan_mode'],
                        initial_config['port_config']['vlan_tag'], initial_config['port_config']['vlan_trunks'])
        to_inf_param = next(item for item in initial_config['ports_mtu'] if item['port'] == to_interface)
        states_result(change_mtu, interface=to_interface, mtu=to_inf_param['mtu'], connected=set_up,
                      sys_parser=sys_parser)

    if set_up:
        set_port_up(to_interface)

    if check_conn and not states_result(check_controller_connection, raise_exception=False):
        if to_interface:
            set_port_down(to_interface)
            del_port_from_switch(to_interface, switch)
        create_aggregation(interface=interface, switch=switch, members=initial_bond_members, connected_members=[],
                           bond_config=initial_config['bond_config'], ports_config=initial_config['port_config'],
                           ports_mtu=initial_config['ports_mtu'], sys_parser=sys_parser)
        for member in initial_bond_members:
            member_info = get_link_info_detailed(member)
            write_interface(BASE_NET_CONF_DIR, member, InterfaceModes.manual, is_connected=True,
                            promiscuous_mode=True if 'PROMISC' in member_info['flags'] else False,
                            mtu=member_info['mtu'])
        raise CommandExecutionError("Interface {} removal caused connection loss between node and controller. "
                                    "Changes cancelled.".format(interface))

    ret['comment'] = "Interface {} removed successfully.".format(interface)
    return ret


@states_response
def add_ports_to_aggregation(aggregation: str, interfaces: list, ports_mtu: list, sys_parser: bool = False):
    """
    Добавление физических интерфейсов в агрегацию
    :param aggregation: название агрегированного интерфейса
    :param interfaces: список интерфейсов для добавления
    :param ports_mtu:
    :param sys_parser:
    """

    ret = {'comment': '', 'changes': {}}

    phys_infs = states_result(list_physical_infs, verbose=True, sys_parser=sys_parser)
    for interface in interfaces:
        target_interface = next((iface for iface in phys_infs if iface['interface'] == interface), None)
        if target_interface is None:
            raise CommandExecutionError(f"Failed to find target interface {interface}.")
        add_port_to_aggregation(aggregation, interface)
        write_interface(BASE_NET_CONF_DIR, interface, InterfaceModes.manual, is_connected=True,
                        promiscuous_mode=target_interface['promiscuous_mode'], mtu=target_interface['mtu'])
    for entry in ports_mtu:
        states_result(change_mtu, interface=entry['port'], mtu=entry['mtu'], connected=True, sys_parser=sys_parser)

    ret['comment'] = f"Interface(s) {', '.join(interfaces)} added to aggregated port {aggregation} successfully."
    return ret


@states_response
def del_ports_to_aggregation(aggregation: str, interfaces: list, sys_parser: bool = False):
    """
    Удаление физических интерфейсов из агрегации
    :param aggregation: название агрегированного интерфейса
    :param interfaces: список интерфейсов для добавления
    :param sys_parser:
    """

    ret = {'comment': '', 'changes': {}}

    phys_infs = states_result(list_physical_infs, verbose=True, sys_parser=sys_parser)
    for interface in interfaces:
        target_interface = next((iface for iface in phys_infs if iface['interface'] == interface), None)
        if target_interface is None:
            raise CommandExecutionError(f"Failed to find target interface {interface}.")
        del_port_from_aggregation(aggregation, interface)
        write_interface(BASE_NET_CONF_DIR, interface, InterfaceModes.manual, is_connected=False,
                        promiscuous_mode=target_interface['promiscuous_mode'], mtu=target_interface['mtu'])
    ret['comment'] = f"Interface(s) {', '.join(interfaces)} removed from aggregated port {aggregation} successfully."
    return ret


@states_response
def create_vmachine_interface(xml: str, vmachine: str, vmachine_inf: str, port_config: dict,
                              live: bool = True):
    """
    Создание интерфейса виртуальной машины
    :param live: True - ВМ включена, False - выключена
    :param xml: параметры интерфейса
    :param vmachine: uuid виртуальной машины
    :param vmachine_inf: название интерфейса
    :param port_config: настройки vlan/mtu
    """

    from veilrpc.states.domain import node_attach_vmachine_interface

    ret = {'comment': '', 'changes': {}}

    mtu = port_config.pop('mtu')
    if live:
        from veilrpc.states.domain import attach_device
        states_result(attach_device, domain_id=vmachine, xml=xml)
        set_vlan_config(vmachine_inf, **port_config)
        set_ovs_mtu(vmachine_inf, mtu)
    states_result(node_attach_vmachine_interface, domain_id=vmachine, xml=xml)

    ret['comment'] = "Virtual interface created successfully."
    return ret


@states_response
def create_docker_container_interface(container_name: str, bridge: str, container_config: dict, vlan_config: dict):
    """
    Создание интерфейса Docker контейнера для подключения его к bridge
    :param container_name: имя контейнера, которой нужно присоединить к ovs bridge
    :param bridge: название ovs bridge, к которому присоединяется контейнер
    :param container_config: конфиги docker контейнера (mac, mtu, ip, gateway)
    :param vlan_config: конфиги vlan
    """
    ret = {'comment': '', 'changes': {}}

    from veilrpc.states.container import attach_docker_container, configure_docker_vlan
    states_result(attach_docker_container, container_name=container_name, bridge=bridge, **container_config)

    # TODO: придумать как достать сгенерированное название интерфейса для передачи настроек vlan через ovs-vsctl
    # set_vlan_config(container_inf, **vlan_config)

    # настройка vlan через ovs-docker
    vlan_tag = vlan_config.pop('vlan_tag')
    states_result(configure_docker_vlan, container_name=container_name, bridge=bridge, vlan_tag=vlan_tag)

    ret['comment'] = "Virtual interface created successfully."
    return ret


@states_response
def del_docker_container_interface(container_name: str, bridge: str):
    """
    Удаление интерфейса Docker контейнера для отсоединения его от bridge
    :param container_name: имя контейнера, которой нужно отсоединить от ovs bridge
    :param bridge: название ovs bridge, от которого отсоединяется контейнер
    """
    ret = {'comment': '', 'changes': {}}
    from veilrpc.states.container import detach_docker_container
    states_result(detach_docker_container, container_name=container_name, bridge=bridge)

    ret['comment'] = "Virtual interface deleted successfully."
    return ret


@states_response
def modify_vmachine_inf(new_xml: str, new_config: dict, old_config: dict, old_xml: str, vmachine: str,
                        vmachine_inf: str):
    """
    Изменение интерфейса виртуальной машины
    :param new_xml: обновленные параметры интерфейса
    :param new_config: обновленные параметры vlan интерфейса
    :param old_config: действующие параметры vlan интерфейса
    :param old_xml: действующие параметры интерфейса
    :param vmachine: uuid виртуальной машины
    :param vmachine_inf: название интерфейса
    """

    ret = {'comment': '', 'changes': {}}
    pgroup_modified = True if new_config != old_config else False
    interface_modified = True if new_xml != old_xml else False

    if interface_modified:
        from veilrpc.states.domain import attach_device, detach_device
        port_config = new_config if pgroup_modified else old_config
        mtu = port_config.pop('mtu')
        states_result(detach_device, domain_id=vmachine, xml=old_xml)
        del_port_from_switch(vmachine_inf)
        states_result(attach_device, domain_id=vmachine, xml=new_xml)
        set_vlan_config(vmachine_inf, **port_config)
        set_ovs_mtu(vmachine_inf, mtu)
    else:
        port_config = new_config if pgroup_modified else old_config
        mtu = port_config.pop('mtu')
        set_vlan_config(vmachine_inf, **port_config)
        set_ovs_mtu(vmachine_inf, mtu)

    ret['comment'] = "Virtual interface modified successfully."
    return ret


@states_response
def remove_vmachine_interface(xml: str, vmachine: str, vmachine_inf: str):
    """
    Удаление интерфейса виртуальной машины
    :param xml: параметры интерфейса
    :param vmachine: uuid виртуальной машины
    :param vmachine_inf: название интерфейса
    """

    from veilrpc.states.domain import detach_device
    states_result(detach_device, domain_id=vmachine, xml=xml)
    del_port_from_switch(vmachine_inf)
    return f"Network interface was removed from virtual machine {vmachine} successfully."


@states_response
def connect_vfunction_to_vmachine(xml: str, vmachine: str):
    """
    Подключение виртуальной функции SR-IOV к виртуальной машине
    :param xml: параметры интерфейса
    :param vmachine: uuid виртуальной машины
    """

    from veilrpc.states.domain import attach_device
    ret = states_result(attach_device, domain_id=vmachine, xml=xml)
    return ret


@states_response
def disconnect_vfunction_to_vmachine(xml: str, vmachine: str):
    """
    Отключение виртуальной функции SR-IOV от виртуальной машины
    :param xml: параметры интерфейса
    :param vmachine: uuid виртуальной машины
    """

    from veilrpc.states.domain import detach_device
    ret = states_result(detach_device, domain_id=vmachine, xml=xml)
    return ret


@states_response
def create_overlay_tunnel(name: str, protocol: str, vni: int, mtu: int, switch: str, remote_ip: str):
    """
    Создание L2 туннеля
    :param name: название туннеля
    :param protocol: протокол
    :param vni: идентификатор виртуальной сети
    :param mtu: значение mtu
    :param switch: коммутатор
    :param remote_ip: адрес удаленного энд-поинта
    """

    ret = {'comment': '', 'changes': {}}

    existing_tun_ep = get_ovs_tun_interface(remote_ip, vni)
    if existing_tun_ep:
        del_port_from_switch(existing_tun_ep)
    add_port_to_switch(switch, name, protocol, remote_ip=remote_ip, vni=vni)

    ret['comment'] = "Tunnel endpoint created successfully."
    return ret


@states_response
def create_service_tunnels(switch: str, endpoints: list):
    """
    Создание эндпоинтов L2 туннеля на указанном коммутаторе
    :param switch: название коммутатора
    :param endpoints: список создаваемых эндпоинтов
    """

    for endpoint in endpoints:
        existing_tun_ep = get_ovs_tun_interface(endpoint['remote_ip'], endpoint['vni'])
        if existing_tun_ep:
            del_port_from_switch(existing_tun_ep)
        add_port_to_switch(switch=switch, **endpoint)
    return "Tunnel endpoints created successfully."


@states_response
def remove_overlay_tunnel(name: str):
    """
    Удаление L2 туннеля
    :param name: название туннеля
    """

    del_port_from_switch(name)
    return "Tunnel endpoint removed successfully."


@states_response
def set_promisc_mode(interface: str, connected: bool, promisc_mode: bool, sys_parser: bool = False):
    """
    Включение/выключение promiscuous mode на физическом интерфейсе
    """

    ret = {'comment': '', 'changes': {}}

    phys_infs = states_result(list_physical_infs, verbose=True, sys_parser=sys_parser)
    target_interface = next((iface for iface in phys_infs if iface['interface'] == interface), None)
    mtu_needs = target_interface['mtu'] != DEFAULT_MTU

    if promisc_mode:
        enable_promisc_mode(interface)
        write_interface(BASE_NET_CONF_DIR, interface, InterfaceModes.manual, is_connected=connected,
                        promiscuous_mode=promisc_mode, mtu=target_interface['mtu'] if mtu_needs else None)
    else:
        disable_promisc_mode(interface)
        write_interface(BASE_NET_CONF_DIR, interface, InterfaceModes.manual, is_connected=connected,
                        promiscuous_mode=promisc_mode, mtu=target_interface['mtu'] if mtu_needs else None)
    result = 'enabled' if promisc_mode else 'disabled'
    ret['comment'] = f"Promiscuous mode {result} on interface {interface} successfully."
    return ret


@states_response
def enable_sriov(port: str, vf_num: int):
    """
    Стейт для функции API включения SR-IOV на интерфейсе
    :param port: название интерфейса
    :param vf_num: количество создаваемых виртуальных функций (всегда максимальное, т.к. для изменения их количества,
    надо сначала все удалить, потом заново создать)
    """

    ret = {'comment': '', 'changes': {}}

    link_info = get_link_info_detailed(port)
    promisc_mode = True if 'PROMISC' in link_info['flags'] else False
    mtu = int(link_info['mtu'])

    write_interface(BASE_NET_CONF_DIR, port, InterfaceModes.manual, is_connected=True,
                    promiscuous_mode=promisc_mode, mtu=mtu)
    states_result(set_interface_up, interface=port)
    set_vfunctions(port, vf_num)
    set_vfunctions_persistent(port, vf_num)

    time.sleep(5)
    try:
        created_vfs = get_vfs_info(port)
    except Exception as e:
        raise CommandExecutionError(f"Failed to enable SR-IOV on interface {port}: {str(e)}.") from None

    ret['comment'] = f"SR-IOV successfully enabled on interface {port}."
    ret['changes'] = created_vfs
    return ret


@states_response
def disable_sriov(port: str):
    """
    Стейт для функции API отключения SR-IOV на интерфейсе
    :param port: название интерфейса
    """

    ret = {'comment': '', 'changes': {}}

    set_vfunctions(port, 0)
    unset_vfunctions_persistent(port)
    link_info = get_link_info_detailed(port)
    promisc_mode = True if 'PROMISC' in link_info['flags'] else False
    mtu = int(link_info['mtu'])

    states_result(set_interface_down, interface=port)
    write_interface(BASE_NET_CONF_DIR, port, InterfaceModes.manual, is_connected=False,
                    promiscuous_mode=promisc_mode, mtu=mtu)

    ret['comment'] = f"SR-IOV successfully disabled on interface {port}."
    return ret


@states_response
def replace_uplink(old_uplink: str, new_uplink: str, old_members: Optional[list],
                   new_members: Optional[list]):
    """
    Стейт для замены аплинк-интерфейса подключения к внешней сети
    :param old_uplink: название старого аплинка
    :param new_uplink: название нового аплинка
    :param old_members: список интерфейсов входящих в старую агрегацию (если есть)
    :param new_members: список интерфейсов входящих в новую агрегацию (если есть)
    """

    if not old_members:
        states_result(set_interface_down, interface=old_uplink, check_conn=False)
    else:
        for member in old_members:
            states_result(set_interface_down, interface=member, check_conn=False)
    if not new_members:
        states_result(set_interface_up, interface=new_uplink, check_conn=False)
    else:
        for member in new_members:
            states_result(set_interface_up, interface=member, check_conn=False)

    if not states_result(check_controller_connection, raise_exception=False):
        if not new_members:
            states_result(set_interface_down, interface=new_uplink, check_conn=False)
        else:
            for member in new_members:
                states_result(set_interface_down, interface=member, check_conn=False)

        if not old_members:
            states_result(set_interface_up, interface=old_uplink, check_conn=False)
        else:
            for member in old_members:
                states_result(set_interface_up, interface=member, check_conn=False)
        raise CommandExecutionError(f"Replacing interface {old_uplink} with {new_uplink} caused connection loss "
                                    f"between node and controller. Changes cancelled.")


@states_response
def external_network_patch_connect(external_id: str, obverse_switch: str, obverse_patch: str, reverse_switch: str,
                                   reverse_patch: str):
    """
    Стейт для подключения внутреннего коммутатора внешней сети к внешнему коммутатору
    :param external_id: идентификатор внешней сети
    :param obverse_switch: внутренний коммутатор
    :param obverse_patch: имя патча от внутреннего коммутатора к внешнему
    :param reverse_switch: внешний коммутатор
    :param reverse_patch: имя патча от внешнего коммутатора к внутреннему
    """

    cmd = f"ovs-vsctl get br {reverse_switch} {OvsRequestParams.external_ids}"
    raw_output = states_result(get_output, raise_exception=False, cmd=cmd)
    params = ovs_param_to_dict(raw_output)
    modification_needed = True
    string_value = external_id
    if 'external-networks' in params:
        ext_networks = params['external-networks'].split(', ')
        if external_id not in ext_networks:
            ext_networks.append(external_id)
            string_value = ', '.join(ext_networks)
        else:
            modification_needed = False

    if modification_needed:
        cmd = f'ovs-vsctl set br {reverse_switch} {OvsRequestParams.external_networks}="{string_value}"'
        states_result(get_output, cmd=cmd)

    add_patches(obverse_switch, reverse_switch, obverse_patch, reverse_patch)
    return "External network connected successfully."


@states_response
def external_network_patch_disconnect(external_id: str, reverse_switch: str, reverse_patch: str):
    """
    Стейт для отключения внешнего коммутатора от внутреннего коммутатора внешней сети
    :param external_id: идентификатор внешней сети
    :param reverse_switch: внешний коммутатор
    :param reverse_patch: имя патча от внешнего коммутатора к внутреннему
    """

    cmd = f"ovs-vsctl get br {reverse_switch} {OvsRequestParams.external_ids}"
    raw_output = states_result(get_output, raise_exception=False, cmd=cmd)
    params = ovs_param_to_dict(raw_output)
    if 'external-networks' in params:
        ext_networks = params['external-networks'].split(', ')
        if external_id in ext_networks:
            ext_networks.remove(external_id)
            if len(ext_networks) == 0:
                if 'virtual-networks' not in params:
                    cmd = f"ovs-vsctl set br {reverse_switch} {OvsRequestParams.external_ids}={{}}"
                    states_result(get_output, cmd=cmd)
                else:
                    v_networks = params['virtual-networks']
                    cmd = f"ovs-vsctl set br {reverse_switch} {OvsRequestParams.external_ids}={{}}"
                    states_result(get_output, cmd=cmd)
                    cmd = f'ovs-vsctl set br {reverse_switch} {OvsRequestParams.virtual_networks}="{v_networks}"'
                    states_result(get_output, cmd=cmd)
            else:
                value = ', '.join(ext_networks)
                cmd = f'ovs-vsctl set br {reverse_switch} {OvsRequestParams.external_networks}="{value}"'
                states_result(get_output, cmd=cmd)

    del_port_from_switch(reverse_patch, reverse_switch)
    return "External network disconnected successfully."


def get_value_by_hashed_data(
        hashed_data: dict, hash_key: str, record_parameter: Tuple[str, ...] = None
) -> Union[dict, None]:
    """
    Функция возвращает (из БД ovs) значения полей (record_parameter) для записи(record) hash_key
    """
    if record_parameter:
        response_params = {}
        for param in record_parameter:
            try:
                response_params[param] = hashed_data[hash_key][param]
            except KeyError:
                response_params[param] = None
            except TypeError:
                response_params[param] = None
        return response_params
    else:
        return hashed_data.get(hash_key, None)


def hashing_ovs_db_table_by(data: dict, to_hash_by: str, table: str) -> dict:
    """
    Хэширование данных с целью последующего быстрого обращения по имени bridge
    """

    hashed_data = {}
    if not isinstance(to_hash_by, str):
        raise ValueError(f"Data can't be hashed by {to_hash_by}. Value of hashing should be string type.")
    if not data.get(table):
        raise ValueError(f"There is no table with {table} name.")
    for record in data[table]:
        if record.get(to_hash_by):
            if isinstance(record[to_hash_by], str):
                hashed_data[record[to_hash_by]] = record
            elif isinstance(record[to_hash_by], dict):
                # это костыль
                # тут учитывается только случай для передаваемого to_hash_by='_uuid'
                # значением record[to_hash_by] может быть словарь типа {'uuid': 'c37ab918-a754-4fe0-bce3-82bbf4067d0a'}
                if len(record[to_hash_by].values()) == 1 and isinstance(list(record.get('_uuid').values())[0], str):
                    hash_by = list(record.get('_uuid').values())[0]
                    hashed_data[hash_by] = record
    return hashed_data


def transform_ovs_db_data(value) -> Union[str, bool, int, list, dict]:
    """
    Функция преобразует данные из БД ovs и убирает строки обозначающие тип данных: set, map и т.д.
    Пример одного случая:
    value = ['set', [['uuid', '0805a22a-8f50-48bb-964a-c3fb9fd50049'], ['uuid', '0823f150-4293-4286-b5f8-59d366bb685d'],
    ['uuid', '4a23aaa2-6e60-4ebd-9216-bff0a3973389']]]
    return ['0805a22a-8f50-48bb-964a-c3fb9fd50049', '0823f150-4293-4286-b5f8-59d366bb685d',
    '4a23aaa2-6e60-4ebd-9216-bff0a3973389']
    """
    if isinstance(value, list):
        if isinstance(value[1], list) and value[1]:
            if value[0] == 'map':
                value = {item[0]: item[1] for item in value[1]}
            elif value[0] == 'set':
                if isinstance(value[1][0], list):
                    value = [item[1] for item in value[1]]
                else:
                    value = value[1]
            else:
                value = {value[0]: value[1]}
            return value
        else:
            # если ['uuid', '5b91ca22-85f0-46cd-a4ea-010d8a78bf0e'], а heading='_uuid', то в dump_ovs_db будет
            # {'_uuid: '5b91ca22-85f0-46cd-a4ea-010d8a78bf0e'} Т.е. подобное этому 'uuid' опускается
            return value[1]
    else:
        return value


@states_response
def dump_ovs_db() -> dict:
    """
    Получение всех данных из БД ovs в виде словаря.
    Ключ словаря это имя таблицы, значение - это данные таблицы.
    """

    dump_db = {}
    cmd = "ovsdb-client --format json dump Open_vSwitch"
    dump_db_raw = [json.loads(line) for line in states_result(get_output, cmd=cmd).splitlines()]
    for table in dump_db_raw:
        records_data = []
        data = table['data']
        headings = table['headings']
        if data:
            for record in data:
                record_data = [transform_ovs_db_data(i) for i in record]
                record_headings = dict(zip(headings, record_data))
                records_data.append(record_headings)
        else:
            records_data = dict(zip(headings, repeat(data)))
        table_name = table['caption'].split(" ")[0]
        dump_db[table_name] = records_data
    return dump_db


@states_response
def data_ovs_db() -> dict:
    result = {}
    tables_param = [('Bridge', 'name'), ('Port', '_uuid'), ('Interface', '_uuid')]
    dump_ovs = dump_ovs_db()
    if dump_ovs['result']:
        dump_ovs = dump_ovs['comment']
        for table, param in tables_param:
            hashed_data = hashing_ovs_db_table_by(data=dump_ovs, to_hash_by=param, table=table)
            result.update({f'{table}'.lower(): hashed_data})
    return result


@states_response
def get_ovs_db_bridge_ports(vswitch: str) -> List[dict]:
    bridge_params = ('ports',)
    port_params = ('name', 'trunks', 'tag', 'vlan_mode', 'bond_mode', 'other_config', 'interfaces', 'lacp',
                   'rstp_status')
    iface_params = ('name', 'ofport', 'type', 'admin_state', 'link_state', 'options', 'status', 'statistics',
                    'mac_in_use', 'mtu')

    ports_config = []
    output_ovs = states_result(data_ovs_db, raise_exception=False)
    if not output_ovs:
        return []
    bridge = get_value_by_hashed_data(hashed_data=output_ovs['bridge'], hash_key=vswitch,
                                      record_parameter=bridge_params)
    ports_bridge = bridge['ports']
    if not isinstance(ports_bridge, list):
        ports_bridge = [ports_bridge]
    for port_uuid in ports_bridge:
        port_config = {}
        port = get_value_by_hashed_data(hashed_data=output_ovs['port'], hash_key=port_uuid,
                                        record_parameter=port_params)
        if port['name'] != vswitch:
            port_config.update({param: port[param] for param in port_params})

            ifaces_config = []
            interfaces = port['interfaces']
            if not isinstance(interfaces, list):
                interfaces = [interfaces]
            for interface_uuid in interfaces:
                iface = get_value_by_hashed_data(hashed_data=output_ovs['interface'], hash_key=interface_uuid,
                                                 record_parameter=iface_params)
                if iface['name'] != vswitch:
                    ifaces_config.append({param: iface[param] for param in iface_params})
            port_config['interfaces'] = ifaces_config
            ports_config.append(port_config)
    return ports_config


@states_response
def get_ovs_bridge_fdb_table(switch: str) -> List[Dict[str, str]]:
    regex = r'(?P<port>\d+) +(?P<vlan>\d+) +(?P<mac>\S+) +(?P<age>\d+)'
    if not switch_exists(switch)['comment']:
        raise CommandExecutionError(f"Bridge with name {switch} doesn't exist.")
    cmd = f"ovs-appctl fdb/show {switch}"
    match = re.finditer(regex, states_result(get_output, cmd=cmd))
    return [m.groupdict() for m in match]


@states_response
def flush_ovs_bridge_fdb_table(switch: str) -> str:
    if not switch_exists(switch)['comment']:
        raise CommandExecutionError(f"Bridge with name {switch} doesn't exist.")
    cmd = f"ovs-appctl fdb/flush {switch}"
    states_result(get_output, cmd=cmd)
    return f"MAC-address table for virtual switch \"{switch}\" flushed successfully."


@states_response
def get_ovs_bridge_fdb_table_stats(switch: str) -> List[str]:
    if not switch_exists(switch)['comment']:
        raise CommandExecutionError(f"Bridge with name {switch} doesn't exist.")
    cmd = f"ovs-appctl fdb/stats-show {switch}"
    return [line.strip() for line in states_result(get_output, cmd=cmd).split("\n") if line]


@states_response
def clear_ovs_bridge_fdb_table_stats(switch: str) -> str:
    if not switch_exists(switch)['comment']:
        raise CommandExecutionError(f"Bridge with name {switch} doesn't exist.")
    cmd = f"ovs-appctl fdb/stats-clear {switch}"
    states_result(get_output, cmd=cmd)
    return f"Statistics for virtual switch \"{switch}\" cleared successfully."


def set_ovs_inf_type_to_veil_inf_type(iface_name: str, iface_type: str, iface_status: dict):
    """
    Не все iface_type подвергаютя замене, только те, что перечислены ниже
    """
    # в VeiL имя интерфейсов vm всегда начиначется с префикса vm_
    # интерфейс смотрит в сторону VM
    if iface_name.startswith('vm_') and iface_type == "":
        iface_type = AddVeilPortType.vm
    # интерфейс смотрит в сторону lxc контейнера, будет считаться vnservices
    elif iface_type == "" and iface_status.get('driver_name') == 'veth':
        iface_type = AddVeilPortType.vnservices
    # см. документацию к ovs-vswitchd.conf.db
    if iface_type == "" or iface_type == "system":
        iface_type = AddVeilPortType.physical
    elif iface_type is None:
        iface_type = AddVeilPortType.none
    return iface_type


@states_response
def get_ovs_bridge_fdb_table_extended(switch: str) -> List[Dict[str, str]]:
    mac_address_table_extended = []
    mac_address_table = states_result(get_ovs_bridge_fdb_table, switch=switch)
    ovs_db_ports = states_result(get_ovs_db_bridge_ports, vswitch=switch)
    for port in mac_address_table:
        iface_name, iface_type, iface_status = next(((iface['name'], iface['type'], iface['status'])
                                                     for p in ovs_db_ports for iface in p['interfaces']
                                                     if iface['ofport'] == int(port.get('port'))), None)
        n_iface_type = set_ovs_inf_type_to_veil_inf_type(iface_name=iface_name, iface_type=iface_type,
                                                         iface_status=iface_status)
        mac_address_table_extended.append({'port': port.get('port'), 'port_type': n_iface_type, 'port_name': iface_name,
                                           'vlan': port.get('vlan'), 'mac': port.get('mac'), 'age': port.get('age')})
    return mac_address_table_extended


# endregion


# region traffic control states & functions
def reload_firewall_service(vnservices_name=None):
    """
    Перезапуск службы брандмауэра
    """

    cmd = 'systemctl reload nftables.service'
    if not vnservices_name:
        states_result(get_output, base_err_msg="Failed to reload firewall service", cmd=cmd)
    else:
        cmd = f'lxc-attach -n {vnservices_name} -- {cmd}'
        states_result(get_output, base_err_msg="Failed to reload firewall service", cmd=cmd)


def write_firewall_config(ruleset: list, config_file: str):
    """
    Запись конфигурации брандмауэра
    """

    try:
        with open(config_file, 'w') as f:
            f.write(reduce(lambda a, s: a + s, map(lambda x: f"{x}\n", ruleset)))
    except OSError as e:
        raise CommandExecutionError(
            "Failed to write configuration. I/O error({0}): {1}".format(e.errno, e.strerror)) from None
    except Exception as e:
        raise CommandExecutionError("Unexpected error: {0}".format(str(e))) from None


@states_response
def parse_tc_qdisc_exists() -> dict:
    """
    Парсинг вывода утилиты tc для обнаружения дисциплин (qdisc)
    :return d:
    example: {'egress': ['vm_0d6c_0', 'vm_cc1f_0'],
             'ingress': ['vm_0d6c_0', 'vm_cc1f_0', 'vm_f83c_0']}
    """

    d = {'egress': [], 'ingress': []}
    cmd = "tc -j qdisc"
    output = states_result(get_output, cmd=cmd)
    for inf in json.loads(output):
        if inf['kind'] == 'ingress':
            d['ingress'].append(inf['dev'])
        elif inf['kind'] == 'prio':
            d['egress'].append(inf['dev'])
    return d


# section for parse action
# TODO может быть сделать более универсальную функцию
def convert_from_hex_to_bin_with_mask_invert(match: dict) -> str:
    """
    Функция возвращает значение value в bin-формате на основании кол-ва бит в выставленными в 0 в mask
    example: {'at': '0', 'value': '00580000', 'mask': 'ff03ffff'}
    """

    bin_value = bin(int(match['value'], 16))[2:].zfill(32)
    bin_mask = bin(int(match['mask'], 16))[2:].zfill(32)
    start = bin_mask.find('0')
    end = bin_mask.find('0') + bin_mask.count('0')
    return bin_value[start:end]


def parse_tc_filter_action_pedit_value(action_pedit) -> dict:
    """
    Функция вычисляет значение value на основании at, mask, value
    example: {'order': '1', 'action': 'pedit', 'at': '0', 'value': '00580000', 'mask': 'ff03ffff'}
    """

    if action_pedit['at'] == '0':  # => поле TOS => 8 bit
        m = convert_from_hex_to_bin_with_mask_invert(action_pedit)
        action_pedit['dscp'] = int(m, 2)  # здесь нужно еще поработать, я просто знаю, что здесь dscp
        del action_pedit['at']
        del action_pedit['value']
        del action_pedit['mask']
        return action_pedit  # {'order': '1', 'action': 'pedit', 'dscp': 22}


def parse_tc_filter_action_pedit(action):
    """
    Parse tc filter action pedit function
    example: "action order 1:  pedit action pipe keys 1\n \t index 5 ref 1 bind 1\n\t key #0  at 0: val 00580000 mask ff03ffff\n"
    """

    regex_action = r"action order (?P<order>\d+): *(?P<action>\S+).+?(?P<at>\d+): val (?P<value>\S+) mask (?P<mask>\S+)"
    match = re.search(regex_action, action, re.DOTALL)
    if match:
        pre_result = match.groupdict()
        result = parse_tc_filter_action_pedit_value(pre_result)
        return result


def parse_tc_filter_action_mirred(action):
    """
    Parse tc filter action mirred function
    example "action order 1: mirred (Egress Mirror to device vm_0d6c_0) pipe\n \tindex 1 ref 1 bind 1\n"
    """

    # regex_action = r"action order (?P<order>\d+): *(?P<action>\S+).+to device (?P<dst_port>\S+).+"
    regex_action = r"action order (?P<order>\d+): (?P<action>\S+) \(.+to device (?P<dst_port>\S+)\).+"
    match = re.search(regex_action, action)
    if match:
        result = match.groupdict()
    return result


def parse_tc_filter_action_csum(action: str) -> dict:
    """
    Parse tc filter action csum function
    example action: "action order 2: csum (iph) action pass\n"
    """

    regex_action = r"action order (?P<order>\d+): (?P<action>\S+) \((?P<header>\S+)\) action (?P<filter_action>\S+)"
    match = re.search(regex_action, action)
    if match:
        result = match.groupdict()
    return result


def parse_tc_filter_rule_action(actions: List[dict]) -> List[dict]:
    """
    Detected type of action
    then we'll parse it
    """

    actions_parsed = []
    regex_action = r"action order \d+: *(?P<action>\S+).+"
    for action in actions:
        match = re.search(regex_action, action)
        if match:
            if match.group('action') == 'pedit':
                action_pedit = parse_tc_filter_action_pedit(action)
                actions_parsed.append(action_pedit)
            if match.group('action') == 'mirred':
                action_mirred = parse_tc_filter_action_mirred(action)
                actions_parsed.append(action_mirred)
            if match.group('action') == 'csum':
                action_csum = parse_tc_filter_action_csum(action)
                actions_parsed.append(action_csum)
    return actions_parsed


def parse_tc_filter_action(content: List[dict]) -> List[dict]:
    action_done = []
    for rule in content:
        rule['action'] = parse_tc_filter_rule_action(rule['action'])
        action_done.append(rule)
    return action_done


# section for parse match
def bin_value_per_eight_bit(bin_value: str):
    """
    Функция разделяет строку в bin-формате на части по 8мь бит
    """
    value_per_eight_bit = []
    eight_bit = ""
    for i in bin_value:
        eight_bit += i
        if len(eight_bit) == 8:
            value_per_eight_bit.append(eight_bit)
            eight_bit = ""
    return value_per_eight_bit


# TODO может быть сделать более универсальную функцию
def convert_from_hex_to_bin_with_mask(match: Dict[str, str]) -> Union[Tuple[str, str], str]:
    """
    Функция возвращает значение value в bin-формате на основании кол-ва бит в выставленными в 1 в mask
    Исключение составляет value для 'at'= 20 - это 32 битное поле объединяет в себе src_port и dst_port.
    Value и Mask делятся на 2а по 16бит: левые 16 бит - src_port, правые 16 бит - dst_port

    match: {'at': '8', 'mask': '00ff0000', 'value': '00100000'}
    """

    if match['at'] == '20':
        # src_port
        src_value = match['value'][0:4]
        src_bin_value = bin(int(src_value, 16))[2:].zfill(16)
        src_bin_mask = match['mask'][0:4]
        src_bin_mask = bin(int(src_bin_mask, 16))[2:].zfill(16)
        src_start = src_bin_mask.find('1')
        src_end = src_bin_mask.find('1') + src_bin_mask.count('1')

        # dst_port
        dst_value = match['value'][4:]
        dst_bin_value = bin(int(dst_value, 16))[2:].zfill(16)
        dst_bin_mask = match['mask'][4:]
        dst_bin_mask = bin(int(dst_bin_mask, 16))[2:].zfill(16)
        dst_start = dst_bin_mask.find('1')
        dst_end = dst_bin_mask.find('1') + dst_bin_mask.count('1')

        return src_bin_value[src_start:src_end], dst_bin_value[dst_start:dst_end]

    bin_value = bin(int(match['value'], 16))[2:].zfill(32)
    bin_mask = bin(int(match['mask'], 16))[2:].zfill(32)
    start = bin_mask.find('1')
    end = bin_mask.find('1') + bin_mask.count('1')
    return bin_value[start:end]


def count_bit_in_mask_is_1(match):
    bin_mask = bin(int(match['mask'], 16))[2:].zfill(32)
    return bin_mask.count("1")


def pretty_match_ethertype_ip(match: dict) -> dict:
    """
    Функция находит значение по имеющимся 32битным значениям value, mask, at переданным как параметры в hex формате.
    """

    result = {}
    if isinstance(match, list):
        parts_dst_mac = []
        parts_src_mac = []

        for m in match:
            if m['at'] == "-16" or m['at'] == "-12":
                parts_dst_mac.append(m)
            if m['at'] == "-8" or m['at'] == "-4":
                parts_src_mac.append(m)

        if parts_dst_mac:
            dst_mac_bin = [convert_from_hex_to_bin_with_mask(match) for match in parts_dst_mac]
            dst_mac_bin_str = "".join([part for part in dst_mac_bin])
            dst_mac = ':'.join([''.join([hex(int(octet[0:4], 2))[2:], hex(int(octet[4:], 2))[2:]]) for octet in
                                bin_value_per_eight_bit(dst_mac_bin_str)])
            result['dst_mac'] = dst_mac

        if parts_src_mac:
            src_mac_bin = [convert_from_hex_to_bin_with_mask(match) for match in parts_src_mac]
            src_mac_bin_str = "".join([part for part in src_mac_bin])
            src_mac = ':'.join([''.join([hex(int(octet[0:4], 2))[2:], hex(int(octet[4:], 2))[2:]]) for octet in
                                bin_value_per_eight_bit(src_mac_bin_str)])
            result['src_mac'] = src_mac
        return result  # {'dst_mac': 'xxx', 'src_mac': 'xxx'} or {'dst_mac': 'xxx'} or {'src_mac': 'xxx'}

    bin_value_by_mask = convert_from_hex_to_bin_with_mask(match)
    if match['at'] == "8":  # => поле IP protocol => поле 8 бит
        m = bin_value_by_mask.ljust(8, '0')
        revert_d = dict(map(reversed, IP_PROTOCOLS.items()))
        value = list(map(lambda x: revert_d[x], [int(m, 2)]))[0]
        result = {'ip_protocol': value}
        return result  # example {'ip_protocol': 'tcp'}

    elif match['at'] == "12" or match['at'] == "16":  # => поле dst или src ip address => поле 32 бит
        m = bin_value_by_mask.ljust(32, '0')
        ip_net = '.'.join([str(int(octet, 2)) for octet in bin_value_per_eight_bit(m)]) + "/" + \
                 str(count_bit_in_mask_is_1(match))
        if match['at'] == "12":
            result = {'src_ip': ip_net}
        elif match['at'] == "16":
            result = {'dst_ip': ip_net}
        return result  # example {'dst_ip': 'xxx.xxx.xxx.xxx/xx'}

    elif match['at'] == "20":
        # => такого поля нет, но value и mask деляться пополам, т.е. 1-е -это src_port
        # => 2-е-это поле dst_port, оба поля по 16 бит
        # src_port
        if bin_value_by_mask[0]:
            src_port = int(bin_value_by_mask[0], 2)
            result['src_port'] = src_port

        # dst_port
        if bin_value_by_mask[1]:
            dst_port = int(bin_value_by_mask[1], 2)
            result['dst_port'] = dst_port

        return result  # example {'src_port': 8080, 'dst_port': 53} or {'src_port': 7777} or {'dst_port': 22}

    elif match['at'] == "0":
        if match['mask'] == '00000000' and match['value'] == '00000000':
            return {}  # возвращаю {}, если match: {'at': '0', 'mask': '00000000', 'value': '00000000'}
        else:
            return None


def parse_tc_filter_rule_match_flat(match: List[dict]) -> dict:
    """
    Функция преобразует ключ match в плоский словарь
    """

    flat_match = {}
    for m in match:
        flat_match.update(m)
    return flat_match


def parse_tc_filter_rule_match_done(content: List[dict]) -> List[dict]:
    """
    Функция окончательной обработки ключа match в каждом правиле.
    """

    for rule in content:
        pretty_match = []
        rule_match_copy = rule['match'].copy()
        for match in rule_match_copy:
            match_parsed = pretty_match_ethertype_ip(match)
            pretty_match.append(match_parsed)
            rule['match'].remove(match)
        rule['match'] = pretty_match
        rule['match'] = parse_tc_filter_rule_match_flat(pretty_match)
    return content


def parse_tc_filter_rule_match_refactor(content: List[dict]) -> List[dict]:
    """
    Функция обрабатывает в каждом правиле значения ключа match с целью вынести
    в отдельный список значения с отрицательным ключом 'at'
    """

    for rule in content:
        match_negative = []
        rule_match_copy = rule['match'].copy()
        for match in rule_match_copy:
            if int(match['at']) < 0:
                match_negative.append(match)
                rule['match'].remove(match)
        if match_negative:
            rule['match'].append(match_negative)
    return content


def parse_tc_filter_rule_match_split(content: List[dict]) -> List[dict]:
    """
    Функция разбивает в каждом правиле фильтрации строку списка ключа match на такой вид
    {'at': 'xxx', 'mask': 'xxx', 'value': 'xxx'}
    """

    rules = []
    for rule in content:
        new_matchs = []
        matchs = rule['match']
        for match in matchs:
            m = {'value': match.split(' at ')[0].split('/')[0], 'mask': match.split(' at ')[0].split('/')[1],
                 'at': match.split(' at ')[1]}
            new_matchs.append(m)
        rule['match'] = new_matchs
        rules.append(rule)
    return rules


# TODO в будущем понадобится распарсить отдельно параметры handle (hash table number, divisor, filter number)
def parse_tc_filter_rule(content: List[dict]) -> List[dict]:
    """
    Функция парсит каждое правило фильтрации
    :param content: List[dict]
    :return results: List[dict] {"protocol": "xxx", "pref": "xxx", "match": "xxx", "action": "xxx"}
    """
    results = []
    for rule in content:
        result = {}
        regex_match = r'(?P<match>\S+/\S+ at \S+)'
        # regex_prot_pref = r"\nfilter protocol (?P<ethertype>\S+) pref (?P<pref_index>\d+).+?"
        # regex_prot_pref = r"filter protocol (?P<ethertype>\S+) pref (?P<pref_index>\d+) u32 chain \d+ fh"
        regex_prot_pref = r"filter protocol (?P<ethertype>\S+) pref (?P<pref_index>\d+) u32 chain (?P<chain>\d+) " \
                          r"fh (?P<handle>\S+:\S*:\S+).+"
        # regex_action = r"(action.+\s.+\s.+\s)+"
        regex_action = r"(action.+\s.+\s.+\s.+)+"
        match = re.findall(regex_match, rule)
        if match:
            result["match"] = match

        # match = re.search(regex_prot_pref, rule, re.DOTALL)
        match = re.search(regex_prot_pref, rule)
        if match:
            # result.update(match.groupdict()) # в этом случае pref_index: str, а нужно int
            result['ethertype'] = match.group('ethertype')
            result['pref_index'] = int(match.group('pref_index'))
            result['chain'] = match.group('chain')
            result['handle'] = match.group('handle')

        match = re.findall(regex_action, rule)
        if match:
            result["action"] = match
        results.append(result)
    return results


# TODO нужно будет сделать возможность парсить каждое hash_table, когда по факту правил нет
def parse_tc_filter_output_raw(content):
    """
    Функция парсит строку, в которой все правила фильтрации и делает список правил, в котором каждый элемент
    это одно правило
    """

    # regex = r"(^filter .+\s)+(\s.+\s \s.+\s)+"
    # regex = r"(^filter .+\s)+(\s.+\s(\s \s.+\s)?)+"
    regex = r"(^filter .+\s)+(\s.+\s.+(\s \s.+\s)?)+"
    rules = [m.group() for m in re.finditer(regex, content, re.MULTILINE)]
    return rules


@states_response
def parse_tc_filter_output() -> dict:
    """
    Общая функция парсинга правил фильтрации утилиты tc, "tc filter show ..."
    :return result: {'egress': {'vm_f83c_0': [{'action': [], 'ethertype': '', 'match': {}, 'pref_index': ''},{},...,{}],
                               'vm_c44a_0': [{},{},...,{}], ...}, 'ingress': {...}}
    """

    qdisc_exists = parse_tc_qdisc_exists()['comment']
    result = {'egress': {}, 'ingress': {}}
    for direction, source_ports in qdisc_exists.items():
        if direction == 'ingress':
            for port in source_ports:
                cmd = f'tc filter show parent ffff: dev {port}'
                output = states_result(get_output, cmd=cmd)
                parced_raw = parse_tc_filter_output_raw(output)
                parced_rules = parse_tc_filter_rule(parced_raw)
                parced_rules_match_split = parse_tc_filter_rule_match_split(parced_rules)
                parced_rules_match_refactor = parse_tc_filter_rule_match_refactor(parced_rules_match_split)
                parced_rules_match_done = parse_tc_filter_rule_match_done(parced_rules_match_refactor)
                parced_rules_action_done = parse_tc_filter_action(parced_rules_match_done)
                result['ingress'][port] = parced_rules_action_done

        if direction == 'egress':
            for port in source_ports:
                cmd = f'tc filter show parent 1: dev {port}'
                output = states_result(get_output, cmd=cmd)
                parced_raw = parse_tc_filter_output_raw(output)
                parced_rules = parse_tc_filter_rule(parced_raw)
                parced_rules_match_split = parse_tc_filter_rule_match_split(parced_rules)
                parced_rules_match_refactor = parse_tc_filter_rule_match_refactor(parced_rules_match_split)
                parced_rules_match_done = parse_tc_filter_rule_match_done(parced_rules_match_refactor)
                parced_rules_action_done = parse_tc_filter_action(parced_rules_match_done)
                result['egress'][port] = parced_rules_action_done
    return result


# TODO сделать подобную функцию для правил зеркалирования портов
# TODO сейчас если трафик матчиться "любой" match 00000000/0000000, то в словаре match пустой словарь {}, это было
#  сделано для текущей совместимости правил, хранящихся в json для network_status. Но лучше, чтобы в выводе было ровно
#  такое же как и в выводе самой утилиты tc filter show ..., а именно match 00000000/00000000 at... см. вывод утилиты.
#  Тоже самое относится к функции tc_mirred_rules_exists() см. ниже

def tc_marking_rules_exists():
    """
    Функция возвращает только правила маркировки, т.е. правила в которых маркируется поле dscp
    :return marking_rules_exists_per_direction:
    {'egress': {'vm_f83c_0': [{'action': [], 'ethertype': '', 'match': {}, 'pref_index': ''},{},...,{}],
                'vm_c44a_0': [{},{},...,{}], ..., 'ingress': {...}}
    """

    pedit = 'pedit'
    dscp = 'dscp'
    marking_rules_exists_per_direction = {}
    rules_per_direction = parse_tc_filter_output()['comment']
    for direction, source_ports in rules_per_direction.items():
        rules_per_inf = {}
        for port, rules in source_ports.items():
            marking_rules = []
            if rules:
                for rule in rules:
                    for action in rule['action']:
                        if action['action'] == pedit and action.get(dscp):
                            marking_rules.append(rule)
            rules_per_inf[port] = marking_rules
        marking_rules_exists_per_direction[direction] = rules_per_inf
    return marking_rules_exists_per_direction


# TODO функция очень похожа на функцию "tc_marking_rules_exists", после тестирования можно сделать только одну функцию.
#  Комментарий по-поводу пустого словаря match см. выше
def tc_marking_rules_exists_params_needed():
    """
    Функция возвращает правила маркировки в виде удобном для сравнения с данными из БД
    (параметры правила в одном словаре)
    Добавим в словарь need_params 'direction': "значение", чтобы соответствовало данным пришедшим с БД. (см. функцию
    apply_qos_policy_marking_to_vnetwork для просмотра всех ключей приходящих с БД)
    :return marking_rules_exists_per_direction:
    {'egress': {'vm_f83c_0': [{'ethertype': '', 'pref_index': '', },{},...,{}],
                'vm_c44a_0': [{},{},...,{}], ...},
     'ingress': {...}}

    Внимание: в need_params наполняем теми значениями, к-е могут действительно быть при парсинге вывода
    tc filter show dev имя_интерфейса parent корневая_дисциплина:
    На самом деле парсятся сейчас только параметры:
    pref_index,'dscp', 'ethertype', 'src_mac', 'dst_mac', 'ip_protocol', 'src_ip', 'dst_ip', 'src_port', 'dst_port'
    Параметр 'direction': direction - берется из направления дисциплины
    Остальные параметры ставятся в None, т.к. предполагается, что они будут в дальнейшем парситься.
    """

    marking_rules_exists = tc_marking_rules_exists()
    marking_rules_exists_per_direction = {}
    dscp = 'dscp'
    for direction, source_ports in marking_rules_exists.items():
        rules_per_inf = {}
        for port, rules in source_ports.items():
            marking_rules = []
            if rules:
                for rule in rules:
                    need_params = {'pref_index': None,
                                   'direction': None,
                                   'dscp': None,
                                   'cos': None,
                                   'ethertype': None,
                                   'vlan_id': None,
                                   'src_mac': None,
                                   'dst_mac': None,
                                   'ip_protocol': None,
                                   'src_ip': None,
                                   'dst_ip': None,
                                   'src_port': None,
                                   'dst_port': None}
                    for action in rule['action']:
                        if action.get(dscp):
                            need_params.update({'dscp': action[dscp]})
                    need_params.update(rule['match'])
                    need_params.update({'ethertype': rule['ethertype']})
                    need_params.update({'pref_index': rule['pref_index']})
                    need_params.update({'direction': direction})
                    marking_rules.append(need_params)
            rules_per_inf[port] = marking_rules
        marking_rules_exists_per_direction[direction] = rules_per_inf
    return marking_rules_exists_per_direction


# сделана по аналогии с функцией tc_marking_rules_exists
def tc_mirred_rules_exists():
    """
    Функция возвращает только правила зеркалирования портов, т.е. правила, в которых присутствует слово mirred
    :return mirred_rules_exists_per_direction:
    {'egress': {'vm_cc1f_0': [{'action': [], 'ethertype': '', 'match': {}, 'pref_index': ''},{},...,{}],
                'vm_f83c_0':: [{},{},...,{}], ..., 'ingress': {'vm_cc1f_0': [],
                                                             'vm_f83c_0': [{'action': [{'action': 'mirred',
                                                                                        'dst_port': 'vm_cc1f_0)',
                                                                                        'order': '1'}],
                                                                            'chain': '0',
                                                                            'ethertype': 'all',
                                                                            'handle': '805::800',
                                                                            'match': {},
                                                                            'pref_index': 49150},
                                                                           {'action': [{'action': 'mirred',
                                                                                        'dst_port': 'vm_cc1f_0)',
                                                                                        'order': '1'}],
                                                                            'chain': '0',
                                                                            'ethertype': 'all',
                                                                            'handle': '805::801',
                                                                            'match': {},
                                                                            'pref_index': 49150}]}}
    """

    mirred = 'mirred'
    mirred_rules_exists_per_direction = {}
    rules_per_direction = parse_tc_filter_output()['comment']
    for direction, source_ports in rules_per_direction.items():
        rules_per_inf = {}
        for port, rules in source_ports.items():
            mirred_rules = []
            if rules:
                for rule in rules:
                    for action in rule['action']:
                        if action.get('action') == mirred:
                            mirred_rules.append(rule)
            rules_per_inf[port] = mirred_rules
        mirred_rules_exists_per_direction[direction] = rules_per_inf
    return mirred_rules_exists_per_direction


# TODO функция очень похожа на функцию "tc_mirred_rules_exists", после тестирования можно сделать только одну функцию.
#  Комментарий по-поводу пустого словаря match см. выше
def tc_mirred_rules_exists_params_needed():
    """
    Функция возвращает правила зеркалирования портов в виде удобном для сравнения с данными из БД
    (параметры правила в одном словаре)
    Добавим в словарь need_params 'direction': "значение", чтобы соответствовало данным пришедшим с БД. (см. функцию
    apply_qos_policy_marking_to_vnetwork для просмотра всех ключей приходящих с БД)
    :return mirred_rules_exists_per_direction:
    {'egress': {'vm_f83c_0': ['ethertype': '', 'dst_port': '', 'pref_index': '','direction': '' },{},...,{}],
                'vm_c44a_0': [{},{},...,{}], ...}, 'ingress': {...}}
    Вывод может быть таким (ниже), это не правильная ситуация для текущей реализации "зеркалирования портов" и других
    фич, использующих вывод утилиты tc filter show. Для текущей реализации pref_index - уникален во всех правилах в
    одном направлении трафика.

    {'egress': {'vm_0d6c_0': [{'direction': 'egress',
                           'dst_port': 'ens15',
                           'ethertype': 'all',
                           'pref_index': {'egress': 49152}}],
                'vm_cc1f_0': []},
    'ingress': {'vm_0d6c_0': [{'direction': 'ingress',
                            'dst_port': 'ens15',
                            'ethertype': 'all',
                            'pref_index': {'ingress': 49149}}],
                 'vm_cc1f_0': [],
                 'vm_f83c_0': [{'direction': 'ingress',
                               'dst_port': 'vm_cc1f_0',
                               'ethertype': 'all',
                               'pref_index': {'ingress': 49150}}]}}

    """

    mirred_rules_exists = tc_mirred_rules_exists()
    mirred = 'mirred'
    mirred_rules_exists_per_direction = {}
    for direction, source_ports in mirred_rules_exists.items():
        rules_per_inf = {}
        for port, rules in source_ports.items():
            mirred_rules = []
            if rules:
                for rule in rules:
                    need_params = {'pref_index': None,
                                   'direction': None,
                                   'ethertype': None,
                                   'dst_port': None}
                    for action in rule['action']:
                        if action.get('action') == mirred:
                            need_params.update({'dst_port': action.get('dst_port')})
                    need_params.update(rule['match'])
                    need_params.update({'ethertype': rule['ethertype']})
                    need_params.update({'pref_index': {direction: rule['pref_index']}})
                    need_params.update({'direction': direction})
                    mirred_rules.append(need_params)
            rules_per_inf[port] = mirred_rules
        mirred_rules_exists_per_direction[direction] = rules_per_inf
    return mirred_rules_exists_per_direction


# TODO вероятно нужно из фичи "зеркалирование портов" убрать направление 'ingress-and-egress', т.к. если будут правила
#  с одинаковыми pref_index, то как сопоставить между собой правила. Мы просто заранее ограничили в сериализаторе так,
#  что на одном порту может быть только одно правило зеркалирования. Но когда мы парсим вывод команды tc filter show
#  мы не можем знать какое правило ingress соответствует какому правилу egress в одном правиле Зеркалирования портов,
#  только если смотреть в файл json, к-й хранит правила БД
def tc_mirred_rules_exists_three_directions():
    """
    Функция проверяет наличие одинаковых имен интерфейсов в направлениях 'ingress' и 'egress' и если находит, то
    добавляет правила в направление 'ingress-and-egress и удаляет правила из направлений 'ingress' и 'egress'
    Также функция удаляет имя_порта, если на нем нет правил (если список правил пустой)
    :return result:
    {'egress': {},
    'ingress': {'vm_f83c_0': [{'direction': 'ingress',
                              'dst_port': 'vm_cc1f_0',
                              'ethertype': 'all',
                              'pref_index': {'ingress': 49150}}]},
    'ingress-and-egress': {'vm_0d6c_0': [{'direction': 'ingress-and-egress',
                                         'dst_port': 'ens15',
                                         'ethertype': 'all',
                                         'pref_index': {'egress': 49152,
                                                        'ingress': 49149}}]}}


    """
    result = {'egress': {}, 'ingress': {}, 'ingress-and-egress': {}}

    mirred_rules_exists = tc_mirred_rules_exists_params_needed()
    mirred_rules_exists_copy = deepcopy(mirred_rules_exists)
    for direction, source_ports in mirred_rules_exists_copy.items():
        for port, rules in source_ports.items():
            if not rules:
                del mirred_rules_exists[direction][port]  # удаляем имена портов, если правил нет, для intersec

    # mirred_rules_exists_copy_2 = deepcopy(mirred_rules_exists)
    # for direction, source_ports in mirred_rules_exists_copy_2.items():
    #     if not source_ports:
    #        del mirred_rules_exists[direction]

    if mirred_rules_exists.get('egress') and mirred_rules_exists.get('ingress'):
        intersec = set(port for port in mirred_rules_exists.get('egress')) & set(
            port for port in mirred_rules_exists.get('ingress'))
    else:
        result['ingress'] = mirred_rules_exists['ingress']
        result['egress'] = mirred_rules_exists['egress']
        return result

    ingress = mirred_rules_exists['ingress']
    egress = mirred_rules_exists['egress']
    copy_egress = deepcopy(egress)
    copy_ingress = deepcopy(ingress)

    for port in intersec:
        result['ingress-and-egress'][port] = []
        for rule_e in copy_egress[port]:
            same_rules = []
            for rule_i in copy_ingress[port]:
                if rule_i['dst_port'] == rule_e['dst_port'] and rule_i['ethertype'] == rule_e['ethertype']:
                    same_rule = deepcopy(rule_e)
                    same_rule['pref_index'].update({'ingress': rule_i['pref_index']['ingress']})
                    same_rule['direction'] = 'ingress-and-egress'
                    same_rules.append(same_rule)
                    for r in ingress[port]:
                        if r == rule_i:
                            ingress[port].remove(r)
                    for rr in egress[port]:
                        if rr == rule_e:
                            egress[port].remove(rr)
            if same_rules:
                for rul in same_rules:
                    if rul not in result['ingress-and-egress'][port]:
                        result['ingress-and-egress'][port].append(rul)

    result['ingress'] = ingress
    result['egress'] = egress
    result_copy = deepcopy(result)
    for direction, source_ports in result_copy.items():
        for port, rules in source_ports.items():
            if not rules:
                del result[direction][port]
    return result


# TODO решение для того, чтобы тесты "Зеркалирование портов" меньше переписывать пришлось.
#  Функция вовзращает результат, как "старая" функция parse_tc_output.
#  Впоследствии можно отказаться от этой функции и использоваться одну функцию
#  "tc_mirred_rules_exists_three_directions()"?, но в этом случае нужно править тесты
@states_response
def parse_tc_output():
    """
    Функция возвращает словарь по направлениям. Каждое направление это список портов.
    В список портов добавляется порт, если на порту настроено Зеркалирование.
    Функция выводит результат под ранее работающую функцию parse_tc_output() , чтобы в тестах не переделывать.
    :return
    {'egress': [],
    'ingress': [{'direction': 'ingress',
                'pref_index': {'ingress': 49150},
                 'source': 'vm_f83c_0',
                 'target': 'vm_cc1f_0'}],
    'ingress-and-egress': []}
    """

    result = {'egress': [], 'ingress': [], 'ingress-and-egress': []}
    mirred_rules_three_directions = tc_mirred_rules_exists_three_directions()
    for direction, source_ports in mirred_rules_three_directions.items():
        for port, rules in source_ports.items():
            if rules:
                for rule in rules:
                    data = {'source': port,
                            'pref_index': rule.get('pref_index'),
                            'direction': direction,
                            'target': rule.get('dst_port')
                            }
                    result[direction].append(data)
    return result


@states_response
def port_mirroring_create(data_config: dict) -> str:
    """
    Создание правил для зеркалирования трафика
    :param data_config:
    example data_config
    data_config = {'verbose_name': 'pm-1', 'description': '',
    'src_ports': [{'id': 'ece90184-4fd3-4aae-adfc-65def3061c68', 'name': 'vm_f83c_0',
    'entity_type': 'vmachineinf', 'connected_to': 'vnet-1', 'handeled_by': 'vm1-on-ctrl-120'},
    {'id': 'df5f7b3a-3a65-40d2-8a9b-c89a3a3a432c', 'name': 'vm_846c_0',
    'entity_type': 'vmachineinf', 'connected_to': 'vnet-1', 'handeled_by': 'vm1-on-ctrl-123'}],
    'src_ports_direction': 'ingress-and-egress',
    'dst_port': {'id': '299d1a90-0dc3-4fe3-b061-174d8c2d9387', 'name': 'ens15',
    'entity_type': 'physicalinf', 'connected_to': 'veil-default', 'handeled_by': 'ctrl-120'},
    'pref_index': {'ingress': 49150, 'egress': 49152}}

    Сохранение каждого правила зеркалирования в файл port-mirrors.json происходит по каждому интерфейсу source:
    [{"applied": true, "pref_index": {'ingress': 49150, 'egress': 49152}, "target": "ens15",
    "direction": "ingress-and-egress", "to_apply": true, "name": "pm-1", "source": "vm_f83c_0"},
    {{"applied": true, "pref_index": {'ingress': 49150, 'egress': 49152}, "target": "ens15",
    "direction": "ingress-and-egress", "to_apply": true, "name": "pm-1", "source": "vm_846c_0"}}]
    """

    qdisc_exists = parse_tc_qdisc_exists()['comment']
    rs = PortMirroringRuleSet.from_config_file()
    src_ports_list = list()
    src_ports_direction = data_config['src_ports_direction']
    dst_port = data_config['dst_port']['name']
    pref_index = data_config['pref_index']
    remove_params = dict(src_ports=[], dst_port=dst_port, src_ports_direction=src_ports_direction,
                         pref_index=pref_index)
    for src_port in data_config['src_ports']:
        rule = rs.find_rule(params=dict(source=src_port['name'], target=dst_port, direction=src_ports_direction,
                                        pref_index=pref_index))
        if rule:
            remove_params['src_ports'].append(src_port)
    if remove_params['src_ports']:
        port_mirroring_remove(remove_params, rules_set=rs)
    for src_port in data_config['src_ports']:
        src_ports_list.append(src_port["name"])
        if src_ports_direction in ('ingress', 'ingress-and-egress'):
            if src_port["name"] not in qdisc_exists['ingress']:
                cmd = f'tc qdisc replace dev {src_port["name"]} ingress'
                states_result(get_output, cmd=cmd)
            cmd = f'tc filter add dev {src_port["name"]} parent ffff: pref {pref_index["ingress"]} ' \
                  f'protocol all u32 match u8 0 0 ' \
                  f'action mirred egress mirror dev {dst_port}'
            states_result(get_output, cmd=cmd)
            if src_ports_direction == 'ingress-and-egress':
                if src_port["name"] not in qdisc_exists['egress']:
                    cmd = f'tc qdisc replace dev {src_port["name"]} handle 1: root prio'
                    states_result(get_output, cmd=cmd)
                cmd = f'tc filter add dev {src_port["name"]} parent 1: pref {pref_index["egress"]} ' \
                      f'protocol all u32 match u8 0 0 ' \
                      f'action mirred egress mirror dev {dst_port}'
                states_result(get_output, cmd=cmd)
        elif src_ports_direction == 'egress':
            if src_port["name"] not in qdisc_exists['egress']:
                cmd = f'tc qdisc replace dev {src_port["name"]} handle 1: root prio'
                states_result(get_output, cmd=cmd)
            cmd = f'tc filter add dev {src_port["name"]} parent 1: pref {pref_index["egress"]} ' \
                  f'protocol all u32 match u8 0 0 ' \
                  f'action mirred egress mirror dev {dst_port}'
            states_result(get_output, cmd=cmd)
        data = dict(name=data_config['verbose_name'], source=src_port['name'], target=dst_port,
                    direction=src_ports_direction, pref_index=pref_index, applied=True, to_apply=True)
        rs.add_rule(PortMirroringRule.from_json(data))
    rs.save()
    src_ports = ', '.join(src_ports_list)
    return f'port-mirroring configuration applied successfully: ' \
           f'source ports "{src_ports}", ' \
           f'traffic direction "{src_ports_direction}", ' \
           f'destination port "{dst_port}".'


@states_response
def port_mirroring_modify(data_db: dict, data_modify: dict) -> str:
    """
    Изменение правил для зеркалирования трафика
    :param data_db: формат параметров точно такой же, как в функции port_mirroring_create()
    :param data_modify: формат параметров точно такой же, как в функции port_mirroring_create()
    """

    qdisc_exists = parse_tc_qdisc_exists()['comment']
    port_mirroring_remove(data_db)
    rs = PortMirroringRuleSet.from_config_file()
    src_ports_list = list()
    src_ports_direction = data_modify['src_ports_direction']
    dst_port = data_modify['dst_port']['name']
    pref_index = data_modify['pref_index']
    for src_port in data_modify['src_ports']:
        src_ports_list.append(src_port['name'])
        if src_ports_direction in ('ingress', 'ingress-and-egress'):
            if src_port["name"] not in qdisc_exists['ingress']:
                cmd = f'tc qdisc replace dev {src_port["name"]} ingress'
                states_result(get_output, cmd=cmd)
            cmd = f'tc filter add dev {src_port["name"]} parent ffff: pref {pref_index["ingress"]} ' \
                  f'protocol all u32 match u8 0 0 ' \
                  f'action mirred egress mirror dev {dst_port}'
            states_result(get_output, cmd=cmd)
            if src_ports_direction == 'ingress-and-egress':
                if src_port["name"] not in qdisc_exists['egress']:
                    cmd = f'tc qdisc replace dev {src_port["name"]} handle 1: root prio'
                    states_result(get_output, cmd=cmd)
                cmd = f'tc filter add dev {src_port["name"]} parent 1: pref {pref_index["egress"]} ' \
                      f'protocol all u32 match u8 0 0 ' \
                      f'action mirred egress mirror dev {dst_port}'
                states_result(get_output, cmd=cmd)
        elif src_ports_direction == 'egress':
            if src_port["name"] not in qdisc_exists['egress']:
                cmd = f'tc qdisc replace dev {src_port["name"]} handle 1: root prio'
                states_result(get_output, cmd=cmd)
            cmd = f'tc filter add dev {src_port["name"]} parent 1: pref {pref_index["egress"]} ' \
                  f'protocol all u32 match u8 0 0 ' \
                  f'action mirred egress mirror dev {dst_port}'
            states_result(get_output, cmd=cmd)
        data = dict(name=data_modify['verbose_name'], source=src_port['name'], target=dst_port,
                    direction=src_ports_direction, pref_index=pref_index, applied=True, to_apply=True)
        rs.add_rule(PortMirroringRule.from_json(data))
    rs.save()
    src_ports = ', '.join(src_ports_list)
    return f'port-mirroring configuration modified successfully: ' \
           f'source ports "{src_ports}", ' \
           f'traffic direction "{src_ports_direction}", ' \
           f'destination port "{dst_port}".'


@states_response
def port_mirroring_remove(data_config: dict, rules_set=None) -> str:
    """
    Удаление правил для зеркалирования трафика
    """

    need_to_save, rs = (True, PortMirroringRuleSet.from_config_file()) if rules_set is None else (False, rules_set)
    src_ports = ', '.join([src_port['name'] for src_port in data_config['src_ports']])
    src_ports_direction = data_config['src_ports_direction']
    dst_port = data_config['dst_port']['name']
    pref_index = data_config['pref_index']
    for src_port in data_config['src_ports']:
        if src_ports_direction in ('ingress', 'ingress-and-egress'):
            cmd = f"tc filter delete dev {src_port['name']} parent ffff: " \
                  f"protocol all pref {pref_index['ingress']} u32"
            states_result(get_output, raise_exception=False, cmd=cmd)
            if src_ports_direction == 'ingress-and-egress':
                cmd = f"tc filter delete dev {src_port['name']} parent 1: " \
                      f"protocol all pref {pref_index['egress']} u32"
                states_result(get_output, raise_exception=False, cmd=cmd)
        elif src_ports_direction == 'egress':
            cmd = f"tc filter delete dev {src_port['name']} parent 1: " \
                  f"protocol all pref {pref_index['egress']} u32"
            states_result(get_output, raise_exception=False, cmd=cmd)
        else:
            raise CommandExecutionError(
                f"Incorrect parameter 'src_ports_direction': {data_config['src_ports_direction']}")
        rule = rs.find_rule(params=dict(source=src_port['name'], target=dst_port, direction=src_ports_direction,
                                        pref_index=pref_index))
        if rule:
            rs.remove_rule(rule)
    if need_to_save:
        rs.save()
    return f'port-mirroring configuration deleted successfully: ' \
           f'source ports "{src_ports}", ' \
           f'traffic direction "{src_ports_direction}", ' \
           f'destination port "{dst_port}".'


# endregion


# region host configuration states & functions
def start_firewall_service(get_err=False):
    """
    Запуск службы брандмауэра узла
    """

    cmd = 'systemctl start nftables.service'
    return True if states_result(
        get_output, raise_exception=False, get_err=get_err, cmd=cmd, suppress_err=not get_err) else False


def stop_firewall_service(get_err=False):
    """
    Останов службы брандмауэра узла
    """

    cmd = 'systemctl stop nftables.service'
    return True if states_result(
        get_output, raise_exception=False, get_err=get_err, cmd=cmd, suppress_err=not get_err) else False


def get_resolv_base():
    """
    Получение сохраненной основной конфигурации DNS узла
    """

    try:
        with open(FILE_RESOLV_BASE, 'r') as c:
            result = c.read().splitlines()
        return result
    except OSError as e:
        raise CommandExecutionError(
            "Unable to read resolv base configuration. I/O error({0}): {1}".format(e.errno, e.strerror)) from None
    except Exception as e:
        raise CommandExecutionError("Unexpected error: {0}".format(str(e))) from None


def split_resolv_base(source: list):
    """
    Разделение основной конфигурации DNS узла на составляющие
    """

    nameservers = list()
    search_domains = list()
    domain = None
    for line in source:
        if line.startswith('nameserver'):
            nameservers.append(line.split()[1])
        if line.startswith('search'):
            tmp = line.split()
            search_domains.extend(tmp[1:])
        if line.startswith('domain'):
            domain = line.split()[1]
    return nameservers, search_domains, domain


def purge_resolv_base():
    """
    Очистка основной конфигурации DNS узла
    """

    try:
        open(FILE_RESOLV_BASE if os.path.isfile(FILE_RESOLV_BASE) else FILE_RESOLV_CONF, 'w').close()
    except OSError as e:
        raise CommandExecutionError(f"Unable to purge configuration. I/O error({e.errno}): {e.strerror}") from None
    except Exception as e:
        raise CommandExecutionError(f"Unexpected error: {str(e)}") from None


def purge_resolv_tail():
    """
    Очистка дополнительной конфигурации DNS узла
    """

    try:
        open(FILE_RESOLV_TAIL, 'w').close()
    except OSError as e:
        raise CommandExecutionError(f"Unable to purge configuration. I/O error({e.errno}): {e.strerror}") from None
    except Exception as e:
        raise CommandExecutionError(f"Unexpected error: {str(e)}") from None


def purge_resolv():
    """
    Очистка конфигурации DNS узла
    """

    try:
        open(FILE_RESOLV_CONF, 'w').close()
    except OSError as e:
        raise CommandExecutionError(f"Unable to purge configuration. I/O error({e.errno}): {e.strerror}") from None
    except Exception as e:
        raise CommandExecutionError(f"Unexpected error: {str(e)}") from None


def write_resolv_base(nameservers: list, search: list, domain: Optional[str]):
    """
    Запись конфигурации DNS узла
    """

    try:
        with open(FILE_RESOLV_BASE if os.path.isfile(FILE_RESOLV_BASE) else FILE_RESOLV_CONF, 'w') as f:
            if nameservers:
                f.write(reduce(lambda a, s: a + s, map(lambda x: f"nameserver {x}\n", nameservers)))
            if search:
                f.write('search {}\n'.format(' '.join(search)))
            if domain:
                f.write('domain {}\n'.format(domain))
    except OSError as e:
        raise CommandExecutionError(f"Unable to write configuration. I/O error({e.errno}): {e.strerror}") from None
    except Exception as e:
        raise CommandExecutionError(f"Unexpected error: {str(e)}") from None


def update_resolv():
    """
    Обновление конфигурации DNS узла
    """
    if os.path.isfile(FILE_RESOLV_BASE):
        cmd = 'resolvconf -u'
        states_result(get_output, base_err_msg='Failed to update DNS configuration', cmd=cmd)


def get_stored_routes():
    """
    Получение сохраненных маршрутов узла
    """

    with open(ROUTE_INIT_PATH, 'r') as c:
        raw_routes_file = c.read().split('\n')
    routes = [s for s in raw_routes_file if '#' not in s and s]
    return routes


def gen_route_string(params: dict, action: str):
    """
    Генерация валидной команды для создания маршрута узла
    """

    params_list = ['ip route', action, params['destination']]
    if params.get('interface'):
        params_list.append(f"dev {params['interface']}")
    params_list.append(f"via {params['gateway']}")
    if params.get('metric'):
        params_list.append(f"metric {params['metric']}")
    return ' '.join(params_list)


def write_routes(routes: list):
    """
    Запись маршрутов узла в конфигурационный файл
    """

    try:
        with open(ROUTE_INIT_PATH, 'w') as f:
            f.write('#!/bin/sh\n')
            if routes:
                f.write(reduce(lambda a, s: a + s, map(lambda x: f"{x}\n", routes)))
    except OSError as e:
        raise CommandExecutionError(f"Unable to write configuration. I/O error({e.errno}): {e.strerror}") from None
    except Exception as e:
        raise CommandExecutionError(f"Unexpected error: {str(e)}") from None


def apply_add_del_routes(routes: list):
    """
    Применение/удаление маршрутов узла
    """

    for route in routes:
        states_result(get_output, raise_exception=False, cmd=route)


def apply_routes_script():
    """
    Применение сохраненных маршрутов
    """

    states_result(get_output, base_err_msg='Failed to update route table', cmd=ROUTE_INIT_PATH)


def get_gateway_configuration():
    """
    Получение текущей конфигурации основного шлюза узла
    """

    gateway_configuration = dict(gateway=None, dhcpv4=False, interface=None)
    cmd = "ip route | grep -E 'default' | awk '{{print $3}}'"
    gateway = states_result(get_output, cmd=cmd).rstrip('\n')
    cmd = "ip route | grep -E 'default' | awk '{{print $5}}'"
    # Может быть более 1 базового маршрута, берем первый в списке
    interface = states_result(get_output, cmd=cmd).rstrip('\n').split('\n')
    if gateway:
        gateway_configuration['gateway'] = gateway
    if interface and not interface[0] == '':
        gateway_configuration['interface'] = interface[0]
        if states_result(check_interface_mode, interface=interface[0]):
            # check dhclient config
            dhclient = parse_dhclient_conf()
            iface = next((inf for inf in dhclient['interfaces'] if inf['interface'] == interface[0]), None)
            if iface and 'routers' in iface['requests']:
                gateway_configuration['dhcpv4'] = True
    return gateway_configuration


def get_dns_configuration():
    """
    Получение текущей активной конфигурации DNS узла
    """

    dns_servers_configuration = dict(nameservers=[], search=[], domain=None, dhcpv4=False, interface=None)
    nameservers, search, domain = [], [], None
    if os.path.isfile(FILE_RESOLV_BASE):
        resolv_base = get_resolv_base()
        nameservers, search, domain = split_resolv_base(resolv_base)
        if any([nameservers, search, domain]):
            resolv_conf = parse_resolv_base()
            dns_servers_configuration['nameservers'] = list(dict.fromkeys(resolv_conf['nameservers']))
            dns_servers_configuration['search'] = list(dict.fromkeys(resolv_conf['search']))
            dns_servers_configuration['domain'] = resolv_conf['domain']
            return dns_servers_configuration

    elif os.path.isfile(FILE_RESOLV_CONF):
        p_r_c = parse_resolv_conf()
        nameservers = p_r_c['nameservers']
        search = p_r_c['search']
        domain = p_r_c['domain']
        if any([nameservers, search, domain]):
            dns_servers_configuration['nameservers'] = nameservers
            dns_servers_configuration['search'] = search
            dns_servers_configuration['domain'] = domain
            return dns_servers_configuration

    dhclient = parse_dhclient_conf()
    for interface in dhclient['interfaces']:
        if any(req in DHCLIENT_DNS_PARAMS for req in interface['requests']) and states_result(
                check_interface_mode, interface=interface['interface']):
            dns_servers_configuration['dhcpv4'] = True
            dns_servers_configuration['interface'] = interface['interface']
            iface_options = get_last_interface_lease(interface['interface'])
            dns_servers_configuration['nameservers'] = iface_options['nameservers'] \
                if iface_options['nameservers'] is not None else []
            dns_servers_configuration['search'] = iface_options['search'] \
                if iface_options['search'] is not None else []
            dns_servers_configuration['domain'] = iface_options['domain']
            break
    return dns_servers_configuration


def remove_dhclient_requests(req_type: str) -> None:
    """
    Удаление параметра из конфигурации dhclient
    """

    dhclient = parse_dhclient_conf()
    req_set = DHCLIENT_REQ_SET.__getattribute__(req_type)
    for item in dhclient['interfaces']:
        item['requests'] = list(set(item['requests']) - req_set)
    write_dhclient_conf(dhclient)


def update_dhclient_requests(interface: str, req_type: str) -> dict:
    """
    Обновление параметров dhclient для указанного интерфейса
    """

    dhclient = parse_dhclient_conf()
    req_set = DHCLIENT_REQ_SET.__getattribute__(req_type)
    if any(map(lambda x: x['interface'] == interface, dhclient['interfaces'])):
        for item in dhclient['interfaces']:
            if item['interface'] == interface:
                item['requests'] = list(set(item['requests']) | req_set)
                break
    else:
        dhclient['interfaces'].append(dict(interface=interface, requests=list(req_set)))
    return dhclient


def lldpd_puppet_config(add=True):
    """
    Добавление службы lldp в конфигурацию puppet
    """

    if add:
        if not os.path.exists(LLDPD_PATH):
            with open(LLDPD_PATH, 'w') as config:
                config.write("lldpd::manage_service: true\n")
    else:
        if os.path.exists(LLDPD_PATH):
            os.remove(LLDPD_PATH)


def write_lldpd_config(config_lines: Union[list, str], config_file: str):
    """
    Запись конфигурации службы lldp
    """

    try:
        with open(config_file, 'w') as f:
            if not isinstance(config_lines, list):
                config_lines = [config_lines]
            f.write(reduce(lambda a, s: a + s, map(lambda x: f"{x}\n", config_lines)))
    except OSError as e:
        raise CommandExecutionError(
            "Failed to write configuration. I/O error({0}): {1}".format(e.errno, e.strerror)) from None
    except Exception as e:
        raise CommandExecutionError("Unexpected error: {0}".format(str(e))) from None


def transform(data):
    def flat_json(y):
        out = {}

        def flatten(x, name=''):
            if type(x) is dict:
                for a in x:
                    flatten(x[a], name + a + '_')
            elif type(x) is list:
                i = 0
                for a in x:
                    flatten(a, name + str(i) + '_')
                    i += 1
            else:
                out[name[:-1]] = x

        flatten(y)
        return out

    def parse_flat_json(data_flat):
        d_ed = dict.fromkeys(LLDPD_NEI_FIELDS, '')
        l_mgmt_ip = []
        l_cap_mix_en = []
        l_cap_en = []
        r_local_port = r'(?P<local_port>.+)_via'
        r_chassis_id_subtype = r'chassis_((?P<system_name>.+)?_)?id_type'
        r_chassis_id_value = r'chassis_((.+)?_)?id_value'
        r_chassis_descr = r'chassis_((.+)?_)?descr'
        r_chassis_mgmt_ip = r'chassis_((.+)?_)?mgmt-ip'
        r_chassis_capability_type = r'chassis_((.+)?_)?capability_(\d+_)?type'
        r_chassis_capability_enabled = r'chassis_((.+)?_)?capability_(\d+_)?enabled'

        for k, v in data_flat.items():
            if 'via' in k:
                match = re.search(r_local_port, k)
                if match:
                    d_ed['local_port'] = match.group('local_port')
                d_ed['protocol'] = v
            elif 'rid' in k:
                d_ed['rid'] = v
            elif 'age' in k:
                d_ed['age'] = v
            elif 'chassis' in k:

                match = re.search(r_chassis_id_subtype, k)
                if match:
                    d_ed['chassis_id_subtype'] = v
                    d_ed['system_name'] = match.group('system_name')
                    continue

                match = re.search(r_chassis_id_value, k)
                if match:
                    d_ed['chassis_id'] = v
                    continue

                match = re.search(r_chassis_descr, k)
                if match:
                    d_ed['system_description'] = v
                    continue

                match = re.search(r_chassis_mgmt_ip, k)
                if match:
                    l_mgmt_ip.append(v)
                    continue

                match = re.search(r_chassis_capability_type, k)
                if match:
                    l_cap_mix_en.append(v)
                    continue

                match = re.search(r_chassis_capability_enabled, k)
                if match:
                    l_cap_mix_en.append(v)

            elif 'port_id_type' in k:
                d_ed['port_id_subtype'] = v
            elif 'port_id_value' in k:
                d_ed['port_id'] = v
            elif 'port_descr' in k:
                d_ed['port_description'] = v
            elif 'port_ttl' in k:
                d_ed['time_to_live'] = v
            elif 'port_mfs' in k:
                d_ed['port_mfs'] = v
            elif 'vlan_vlan-id' in k:
                d_ed['vlan_vlan_id'] = v
            elif 'vlan_pvid' in k:
                d_ed['vlan_pvid'] = v

        for item in zip(l_cap_mix_en[::2], l_cap_mix_en[1::2]):
            if item[1]:
                l_cap_en.append(item[0])

        d_ed['supported_system_capabilities'] = l_cap_mix_en[::2]
        d_ed['enabled_system_capabilities'] = l_cap_en
        d_ed['mgmt_ip'] = l_mgmt_ip

        return d_ed

    items = [parse_flat_json(flat_json(item)) for item in data]
    return items


@states_response
def check_firewall_service():
    """
    Получение состояния службы брандмауэра
    """

    cmd = 'systemctl status nftables.service'
    return True if states_result(get_output, raise_exception=False, cmd=cmd, suppress_err=True) else False


@states_response
def start_firewall():
    """
    Запуск службы брандмауэра
    """

    err = start_firewall_service(get_err=True)
    if err:
        raise CommandExecutionError(err)
    return "Firewall started successfully."


@states_response
def stop_firewall():
    """
    Останов службы брандмауэра
    """

    err = stop_firewall_service(get_err=True)
    if err:
        raise CommandExecutionError(err)
    return "Firewall stopped successfully."


@states_response
def restart_firewall():
    """
    Перезапуск службы брандмауэра
    """

    err = stop_firewall_service(get_err=True)
    if err:
        raise CommandExecutionError(err)
    err = start_firewall_service(get_err=True)
    if err:
        raise CommandExecutionError(err)

    return "Firewall restarted successfully."


@states_response
def apply_node_policy(ruleset: list, config_file: str):
    """
    Запись правил брандмауэра узла в файл конфигурации и последующий перезапуск службы для применения конфигурации
    :param ruleset: правила брандмауэра
    :param config_file: путь к файлу конфигурации
    """

    write_firewall_config(ruleset, config_file)
    firewall_state = states_result(check_firewall_service)
    if firewall_state:
        reload_firewall_service()
    return 'Firewall configuration applied successfully.'


@states_response
def update_hosts(old_address: str, new_address: str):
    """
    Обновление файла hosts
    """

    if not new_address:
        raise CommandExecutionError("Failed to update hosts config file. IP-address can't be empty.")
    hostname = states_result(get_output, cmd='hostname').rstrip('\n')
    conf_file = f"{os.path.join(HOSTS_CONF_DIR, hostname)}.conf"
    try:
        UUID(hostname)
        hostname_is_uuid = True
    except ValueError:
        hostname_is_uuid = False
    cmd = f"echo '{new_address:16}{hostname}' > {conf_file}"
    states_result(get_output, base_err_msg='Failed to save hosts config file', cmd=cmd)
    states_result(get_output, base_err_msg='Failed to update hosts configuration', cmd=f"{HOSTS_UPD_UTIL} -c")
    if not hostname_is_uuid:
        states_result(get_output, cmd=f"rm {conf_file}")


@states_response
def set_static_routes(static_routes: list):
    """
    Добавление статических маршрутов
    :param static_routes: список маршрутов
    """

    ret = {'comment': '', 'changes': {}}

    existing_routes = get_stored_routes()
    routes_list = [gen_route_string(route, IpRouteAction.add) for route in static_routes]
    existing_routes.extend(routes_list)
    write_routes(existing_routes)
    apply_add_del_routes(routes_list)

    ret['comment'] = f"Static routes {', '.join([dest.get('destination', None) for dest in static_routes])} added successfully."
    return ret


@states_response
def del_static_routes(static_routes: list):
    """
    Удаление статических маршрутов
    :param static_routes: список маршрутов
    """

    ret = {'comment': '', 'changes': {}}

    existing_routes = get_stored_routes()
    routes_to_store = [route for route in existing_routes if
                       route not in [gen_route_string(route, IpRouteAction.add) for route in static_routes]]
    routes_to_delete = [gen_route_string(route, IpRouteAction.delete) for route in static_routes]
    write_routes(routes_to_store)
    apply_add_del_routes(routes_to_delete)

    ret['comment'] = f"Static routes {', '.join([dest.get('destination', None) for dest in static_routes])} removed successfully."
    return ret


@states_response
def set_veil_mgmt_dhcp(old_interface_params: dict, new_interface_params: dict, old_gateway_config: dict,
                       new_gateway_config: Union[dict, None], restart_engine=True):
    """
    Изменение параметров veil-mgmt
    """

    ret = {'comment': '', 'changes': {}}
    mac_address_to_set = new_interface_params.get('mac_address', old_interface_params['mac_address'])
    result_interface_config = dict(
        ipv4_address='',
        ipv4_netmask='',
        mac_address=mac_address_to_set,
        dhcpv4=True)

    # Выключаем порт и перезаписываем файл конфигурации /etc/network/interfaces.d/veil-mgmt
    set_port_down(VEIL_MGMT_INTERFACE)
    del_file_by_pattern(BASE_NET_CONF_DIR, VEIL_MGMT_INTERFACE)
    params = dict(conf_dir=BASE_NET_CONF_DIR,
                  name=VEIL_MGMT_INTERFACE,
                  hw_address=mac_address_to_set,
                  mode=InterfaceModes.dhcp)
    write_interface(**params)

    # Изменяем mac адрес интерфейса на ходу
    set_ovs_mac_address(VEIL_MGMT_INTERFACE, mac_address_to_set)

    # Добавляем запрос дефолтного маршрута по dhcp
    if new_gateway_config:
        dhclient = parse_dhclient_conf()
        iface_in_dhclient = False
        iface_index = 0
        for idx, iface in enumerate(dhclient['interfaces']):
            if iface['interface'] == VEIL_MGMT_INTERFACE:
                iface_in_dhclient = True
                iface_index = idx
                break
        if iface_in_dhclient:
            if DhclientRoutersRequestParams.gateway not in dhclient['interfaces'][iface_index]['requests']:
                dhclient['interfaces'][iface_index]['requests'].append(DhclientRoutersRequestParams.gateway)
        else:
            dhclient['interfaces'].append(dict(interface=VEIL_MGMT_INTERFACE,
                                               requests=[DhclientRoutersRequestParams.gateway]))
        write_dhclient_conf(dhclient)

    # Включаем порт, ждем 5 сек (чтобы успел получить адрес) и проверяем связь с контроллером.
    # Если связи нет - откатываемся на первоначальные настройки
    set_port_up(VEIL_MGMT_INTERFACE)
    time.sleep(5)

    # Парсим новые адрес/маску, шлюз
    real_params = states_result(get_address_info_real, interface=VEIL_MGMT_INTERFACE)
    # case: на veil-mgmt не было ip-адреса и затем снова выполнили set-dhcp
    if not real_params['ipv4_address'] and not old_interface_params['ipv4_address']:
        raise CommandExecutionError(
            f"Interface {VEIL_MGMT_INTERFACE} didn't have ip-address and ip-address "
            f"has not been assigned to interface {VEIL_MGMT_INTERFACE} by dhcp server. "
            f"Check your network connection, network settings, dchp server.")
    result_interface_config['ipv4_address'] = real_params['ipv4_address']
    result_interface_config['ipv4_netmask'] = real_params['ipv4_netmask']

    conf = get_config()
    controller_ip = conf.get('controller_ip')
    if controller_ip and not states_result(check_controller_connection,
                                           raise_exception=False, controller_ip=controller_ip):
        set_port_down(VEIL_MGMT_INTERFACE)
        del_file_by_pattern(BASE_NET_CONF_DIR, VEIL_MGMT_INTERFACE)
        params = dict(conf_dir=BASE_NET_CONF_DIR,
                      name=VEIL_MGMT_INTERFACE,
                      hw_address=old_interface_params['mac_address'],
                      mode=InterfaceModes.static,
                      ipv4_address=old_interface_params['ipv4_address'],
                      ipv4_netmask=old_interface_params['ipv4_netmask'],
                      ipv4_gateway=old_gateway_config['gateway'])
        write_interface(**params)

        set_ovs_mac_address(VEIL_MGMT_INTERFACE, old_interface_params['mac_address'])
        set_port_up(VEIL_MGMT_INTERFACE)
        raise CommandExecutionError("Changing kernel interface configuration caused connection loss between node "
                                    "and controller. Changes cancelled.")

    result_gateway_config = get_gateway_configuration()
    # обновляем /etc/hosts
    states_result(update_hosts, old_address=old_interface_params['ipv4_address'],
                  new_address=result_interface_config['ipv4_address'])
    if restart_engine:
        cmd = 'systemctl restart node_supervisor.service'
        states_result(get_output, base_err_msg='Failed to restart node supervisor service', cmd=cmd, timeout=180)

    ret['comment'] = 'Management interface configuration updated successfully.'
    ret['changes'] = dict(gateway_config=result_gateway_config,
                          interface_params=result_interface_config)
    return ret


@states_response
def set_veil_mgmt_static(old_interface_params: dict, new_interface_params: dict, old_gateway_config: dict,
                         new_gateway_config: Union[dict, None], restart_engine=True):
    """
    Изменение параметров veil-mgmt
    """

    ret = {'comment': '', 'changes': {}}
    mac_address_to_set = new_interface_params.get('mac_address', old_interface_params['mac_address'])
    result_interface_config = dict(
        ipv4_address=new_interface_params['ipv4_address'],
        ipv4_netmask=new_interface_params['ipv4_netmask'],
        mac_address=mac_address_to_set,
        dhcpv4=False)

    # Выключаем порт и перезаписываем файл конфигурации /etc/network/interfaces.d/veil-mgmt
    set_port_down(VEIL_MGMT_INTERFACE)
    del_file_by_pattern(BASE_NET_CONF_DIR, VEIL_MGMT_INTERFACE)
    params = dict(conf_dir=BASE_NET_CONF_DIR,
                  name=VEIL_MGMT_INTERFACE,
                  hw_address=mac_address_to_set,
                  mode=InterfaceModes.static,
                  ipv4_address=new_interface_params['ipv4_address'],
                  ipv4_netmask=new_interface_params['ipv4_netmask'])
    if old_gateway_config['interface'] == VEIL_MGMT_INTERFACE or old_gateway_config['interface'] is None:
        params['ipv4_gateway'] = new_gateway_config.get(
            'gateway', old_gateway_config['gateway']) if new_gateway_config else None
    write_interface(**params)

    # Изменяем mac адрес интерфейса на ходу
    set_ovs_mac_address(VEIL_MGMT_INTERFACE, mac_address_to_set)

    # Чистим dhclient.conf
    if old_interface_params['dhcpv4']:
        purge_leases(VEIL_MGMT_INTERFACE)
        dhclient = parse_dhclient_conf()
        iface_entry = next((e for e in dhclient['interfaces'] if e['interface'] == VEIL_MGMT_INTERFACE), None)
        if iface_entry is not None:
            dhclient['interfaces'].remove(iface_entry)
            write_dhclient_conf(dhclient)

    # Включаем порт и проверяем связь с контроллером. Если связи нет - откатываемся на первоначальные настройки
    set_port_up(VEIL_MGMT_INTERFACE)
    conf = get_config()
    controller_ip = conf.get('controller_ip')
    if old_interface_params['ipv4_address'] == controller_ip:
        old_net = IPv4Interface(f"{old_interface_params['ipv4_address']}/{old_interface_params['ipv4_netmask']}")
        new_net = IPv4Interface(f"{new_interface_params['ipv4_address']}/{new_interface_params['ipv4_netmask']}")
        if new_net != old_net:
            controller_ip = new_interface_params['ipv4_address']
    if controller_ip and not states_result(check_controller_connection,
                                           raise_exception=False, controller_ip=controller_ip):
        # case: на veil-mgmt не было ip-адреса, например, dhcp клиент не получил ip-адрес
        if not old_interface_params['ipv4_address']:
            raise CommandExecutionError(
                f"Changing kernel interface configuration caused connection loss between node and controller. "
                f"Changes cancelled. Interface {VEIL_MGMT_INTERFACE} didn't have ip-address. "
                f"Check your network connection, network settings.")
        set_port_down(VEIL_MGMT_INTERFACE)
        del_file_by_pattern(BASE_NET_CONF_DIR, VEIL_MGMT_INTERFACE)
        params = dict(
            conf_dir=BASE_NET_CONF_DIR,
            name=VEIL_MGMT_INTERFACE,
            hw_address=old_interface_params['mac_address'],
            mode=InterfaceModes.static if old_interface_params['dhcpv4'] is False else InterfaceModes.dhcp)
        if old_interface_params['dhcpv4'] is False:
            params['ipv4_address'] = old_interface_params['ipv4_address']
            params['ipv4_netmask'] = old_interface_params['ipv4_netmask']
            if old_gateway_config['dhcpv4'] is False:
                params['ipv4_gateway'] = old_gateway_config['gateway']
        write_interface(**params)

        set_ovs_mac_address(VEIL_MGMT_INTERFACE, old_interface_params['mac_address'])

        if old_gateway_config['dhcpv4']:
            dhclient = parse_dhclient_conf()
            dhclient['interfaces'].append(dict(interface=VEIL_MGMT_INTERFACE,
                                               requests=[DhclientRoutersRequestParams.gateway]))
            write_dhclient_conf(dhclient)

        states_result(update_hosts, old_address=result_interface_config['ipv4_address'],
                      new_address=old_interface_params['ipv4_address'])

        set_port_up(VEIL_MGMT_INTERFACE)
        raise CommandExecutionError("Changing kernel interface configuration caused connection loss between node "
                                    "and controller. Changes cancelled.")

    result_gateway_config = get_gateway_configuration()
    # обновляем /etc/hosts
    states_result(update_hosts, old_address=old_interface_params['ipv4_address'],
                  new_address=result_interface_config['ipv4_address'])
    if restart_engine:
        cmd = 'systemctl restart node_supervisor.service'
        states_result(get_output, base_err_msg='Failed to restart node supervisor service', cmd=cmd, timeout=180)

    ret['comment'] = 'Management interface configuration updated successfully.'
    ret['changes'] = dict(gateway_config=result_gateway_config,
                          interface_params=result_interface_config)
    return ret


@states_response
def rewrite_dhclient_conf(is_dhcp: bool):
    dhclient = parse_dhclient_conf()
    dhclient['requests'] = DHCLIENT_GLOBAL_PARAMS
    if is_dhcp:
        dhclient['interfaces'] = [dict(interface=VEIL_MGMT_INTERFACE, requests=DHCLIENT_INITIAL_PARAMS)]
    write_dhclient_conf(dhclient)


@states_response
def get_node_network_config():
    return dict(dns_config=get_dns_configuration(),
                gateway_config=get_gateway_configuration(),
                static_routes=parse_routes(),
                firewall_active=states_result(check_firewall_service))


@states_response
def dump_node_networks(as_root: bool = True, sys_parser: bool = False) -> dict:
    """
    Стейт для сбора структурированой информации о сетевых устройствах узла
    """

    result = dict(switches=[], vnetworks=[], free_ports=[], sriov_ports=[])

    if sys_parser:
        raw_hardware_devices, raw_other_devices, raw_ovs_bridges = parse_sys_class_net(as_root)
    else:
        network_devices = dump_network_devices(as_root)
        raw_hardware_devices, raw_other_devices, raw_ovs_bridges = parse_network_devices(network_devices, as_root)
    used_hardware_devices = list()
    states_result(transform_network_config, as_root=as_root)

    # Расчленяем список сетевого железа и распихиваем по категориям:
    # - свободные (не подключенные к OpenVswitch'н) сетевые карточки без поддержки SR-IOV
    # - сетевые карточки с включенным SR-IOV
    # - оставшиеся сетевые карточки
    while len(raw_hardware_devices):
        dev = raw_hardware_devices.pop()
        dev.update(get_physical_inf_params(dev['name']))
        if not port_exists_in_ovs(dev['name'], as_root) and not iface_exists_in_ovs(dev['name'], as_root):
            # pylint: disable=expression-not-assigned
            result['free_ports'].append(dev) if dev['sriov_enabled'] is False else result['sriov_ports'].append(dev)
        else:
            if device_in_aggregation(dev['name'], as_root):
                dev['bonded'] = True
            used_hardware_devices.append(dev)

    physical_devices_mapping = dict()
    for index, value in enumerate(used_hardware_devices):
        key = value['name']
        physical_devices_mapping[key] = index

    virtual_devices_mapping = dict()
    for index, value in enumerate(raw_other_devices):
        key = value['name']
        virtual_devices_mapping[key] = index

    while len(raw_ovs_bridges):
        bridge = raw_ovs_bridges.pop()
        br = bridge['name']
        rstp_enabled = True if get_ovs_bridge_param(br, OvsRequestParams.rstp, as_root) == 'true' else False
        bridge_ports = get_ovs_bridge_ports(br, as_root)
        if rstp_enabled:
            vnetwork_id = states_result(get_vnetwork_id, switch=br)
            lswitch_ret = dict(switch=br, vmachines_infs=[], tunnel_endpoints=[], vnservices_infs=[],
                               uplink_connections=[], netflow_params=None, vnetwork_id=vnetwork_id)
            lswitch_ret['netflow_params'] = states_result(check_netflow_params, switch=br)
            lswitch_ret['vmachines_infs'] = get_ovs_bridge_vm_ifaces(br, as_root)
            while len(bridge_ports):
                port = bridge_ports.pop()
                if port in virtual_devices_mapping.keys():
                    pass
                else:
                    port_type = get_ovs_interface_param(port, OvsRequestParams.interface_type, as_root=as_root)
                    if port_type == OvsPortType.patch:
                        dev = dict(name=port, peer=get_ovs_interface_param(port, OvsRequestParams.peer,
                                                                           as_root=as_root).strip('"'))
                        lswitch_ret['uplink_connections'].append(dev)
                    elif port_type in (OvsPortType.vxlan, OvsPortType.geneve):
                        tep = dict(endpoint_port=port, remote_ip='', key='', local_iface='')
                        endpoint_remote_ip = get_ovs_interface_param(port, OvsRequestParams.remote_ip,
                                                                     as_root=as_root).strip('"')
                        endpoint_key = get_ovs_interface_param(port, OvsRequestParams.key,
                                                               as_root=as_root).strip('"')
                        endpoint_local = get_ovs_interface_param(port, OvsRequestParams.local_tep,
                                                                 as_root=as_root).strip('"')
                        tep.update(dict(remote_ip=endpoint_remote_ip, key=endpoint_key, local_iface=endpoint_local))
                        lswitch_ret['tunnel_endpoints'].append(tep)
            result['vnetworks'].append(lswitch_ret)
        else:
            vswitch_ret = dict(switch=br, internal_infs=list(), physical_infs=list(), aggregated_infs=list(),
                               vnetwork_connections=list(), port_mirroring=None)
            while len(bridge_ports):
                port = bridge_ports.pop()
                if port in physical_devices_mapping.keys():
                    entry_index = physical_devices_mapping[port]
                    device = used_hardware_devices[entry_index]
                    device.update(get_physical_inf_params(port))
                    device.update(states_result(get_port_pgroup_config, name=port, as_root=as_root))
                    device['up_state'] = states_result(check_port_up_state, port=port)
                    device['connected'] = True
                    vswitch_ret['physical_infs'].append(device)
                else:
                    port_type = get_ovs_interface_param(port, OvsRequestParams.interface_type, as_root=as_root)
                    if port in virtual_devices_mapping.keys():
                        if port_type == OvsPortType.internal:
                            port_ret = states_result(get_internal_inf, name=port, as_root=as_root)
                            port_data = dict(name=port, **port_ret)
                            vswitch_ret['internal_infs'].append(port_data)
                        elif bool(re.match(LXC_INTERFACE_PATTERN, port)):
                            pass
                    else:
                        if port_type == OvsPortType.patch:
                            dev = dict(name=port, peer=get_ovs_interface_param(port, OvsRequestParams.peer,
                                                                               as_root=as_root).strip('"'))
                            vswitch_ret['vnetwork_connections'].append(dev)
                            continue
                        aggregation = states_result(get_bond_config, name=port, as_root=as_root)
                        aggregation.update(states_result(get_port_pgroup_config, name=port, get_mtu=False))
                        aggregation['name'] = port
                        aggregation['members'] = list()
                        members_ids = get_ovs_port_param(port,
                                                         OvsRequestParams.interfaces,
                                                         as_root).lstrip('[').rstrip(']').split(', ')
                        members = [get_ovs_interface_param(i, OvsRequestParams.name,
                                                           as_root=as_root).strip('"') for i in members_ids]
                        for member in members:
                            member_index = physical_devices_mapping[member]
                            member_data = used_hardware_devices[member_index]
                            member_data['connected'] = True
                            member_data['up_state'] = states_result(check_port_up_state, port=member)
                            member_data.update(get_physical_inf_params(member))
                            aggregation['members'].append(member_data)
                        vswitch_ret['aggregated_infs'].append(aggregation)
            result['switches'].append(vswitch_ret)
    return result


@states_response
def dump_free_ports(as_root: bool = True, sys_parser: bool = False):
    """
    Получение списка неподключенных к OpenVswitch физических интерфейсов
    """

    result = list()
    if sys_parser:
        hardware_devices, _, _ = parse_sys_class_net(as_root)
    else:
        net_devices = dump_network_devices()
        hardware_devices, _, _ = parse_network_devices(net_devices)
    for dev in hardware_devices:
        if not port_exists_in_ovs(dev['name']) and not iface_exists_in_ovs(dev['name'], as_root):
            dev['up_state'] = states_result(check_port_up_state, port=dev['name'])
            dev.update(get_physical_inf_params(dev['name']))
            result.append(dev)
    return result


@states_response
def get_node_networks(sys_parser: bool = False):
    """
    Сбор информации для синхронизации
    """

    return dict(comment="Node's network configuration gathered successfully.",
                changes=dict(general_config=states_result(get_node_network_config),
                             network_entities=states_result(dump_node_networks, sys_parser=sys_parser)))


@states_response
def flush_non_management_entities():
    """
    Удаление "пользовательских" сетевых сущностей
    """

    ret = {'comment': '', 'changes': {}}

    for bridge in get_ovs_bridges():
        if bridge not in (DEFAULT_SWITCH, BLACKHOLE_SWITCH):
            for port in get_ovs_bridge_ports(bridge):
                port_type = get_ovs_interface_param(port, OvsRequestParams.interface_type)
                port_interfaces = get_ovs_port_param(port,
                                                     OvsRequestParams.interfaces).lstrip('[').rstrip(']').split(', ')
                if port_type == OvsPortType.internal:
                    del_port_from_switch(port)
                    del_file_by_pattern(BASE_NET_CONF_DIR, port)
                elif port_type == OvsPortType.system:
                    del_port_from_switch(port, bridge)
                    set_port_down(port)
                elif port_type in (OvsPortType.geneve, OvsPortType.patch, OvsPortType.vxlan):
                    del_port_from_switch(port, bridge)
                else:
                    bond_members = list()
                    for interface_id in port_interfaces:
                        bond_members.append(get_ovs_interface_param(interface_id, OvsRequestParams.name))
                    for member in bond_members:
                        set_port_down(member)
                    del_port_from_switch(port, bridge)
            remove_ovs_bridge(bridge)
        else:
            for port in get_ovs_bridge_ports(bridge):
                port_type = get_ovs_interface_param(port, OvsRequestParams.interface_type)
                if port_type == OvsPortType.internal and port != VEIL_MGMT_INTERFACE:
                    del_port_from_switch(port)
                    del_file_by_pattern(BASE_NET_CONF_DIR, port)
                elif port_type in (OvsPortType.geneve, OvsPortType.patch, OvsPortType.vxlan):
                    del_port_from_switch(port, bridge)
    dhclient = parse_dhclient_conf()
    dhclient['interfaces'] = [iface for iface in dhclient['interfaces'] if iface['interface'] == VEIL_MGMT_INTERFACE]
    write_dhclient_conf(dhclient)

    ret['comment'] = 'Network configuration flushed successfully.'
    return ret


@states_response
def flush_routing_configuration(gateway: bool, static_routes: bool):
    if gateway:
        gateway_conf = get_gateway_configuration()
        inf_name = gateway_conf['interface']
        if gateway_conf['dhcpv4'] is False and inf_name:
            with open(os.path.join(BASE_NET_CONF_DIR, inf_name), 'r') as f:
                interface_config = f.read().splitlines()
            interface_config = [s for s in interface_config if 'gateway' not in s]
            states_result(set_interface_down, interface=inf_name, check_conn=False)
            with open(os.path.join(BASE_NET_CONF_DIR, inf_name), 'w') as f:
                f.write(reduce(lambda a, s: a + s, map(lambda x: f"{x}\n", interface_config)))
            states_result(set_interface_up, interface=inf_name, check_conn=False)
        interfaces = list()
        dhclient = parse_dhclient_conf()
        for iface_conf in dhclient['interfaces']:
            interfaces.append(iface_conf['interface'])
        remove_dhclient_requests(DhclientReqType.gateway)
        for interface in interfaces:
            dhclient_release(interface)
            dhclient_request(interface)
    if static_routes:
        existing_routes = parse_routes()
        if existing_routes:
            states_result(del_static_routes, static_routes=existing_routes)


@states_response
def flush_network_config(uplink: str, management_params: dict, sys_parser: bool = False):
    """
    Функция сброса сетевой конфигурации
    :param uplink:
    :type uplink: str
    :param management_params:
    :type management_params: dict
    :param sys_parser:
    """

    current_config = states_result(dump_node_networks, sys_parser=sys_parser)
    states_result(flush_routing_configuration, gateway=True, static_routes=True)
    states_result(set_dns_configuration, dns_config=dict(nameservers=[], search=[], domain='', dhcpv4=False))
    # Step 3: disable all SR-IOVs
    for port in current_config['sriov_ports']:
        states_result(disable_sriov, port=port['name'])
    # Step 4: set all interfaces DOWN, remove all ports from OVS, remove all port mirrors etc., remove all bridges
    # except veil-default and veil-blackhole
    for bridge in current_config['switches']:
        for internal_inf in bridge['internal_infs']:
            states_result(remove_internal_interface, interface=internal_inf['name'])
        for physical_inf in bridge['physical_infs']:
            states_result(disconnect_interface, interface=physical_inf['name'], switch=bridge['switch'],
                          check_conn=False, sys_parser=sys_parser)
        for aggregated_inf in bridge['aggregated_infs']:
            states_result(remove_aggregation, interface=aggregated_inf['name'], switch=bridge['switch'],
                          initial_config={}, check_conn=False, sys_parser=sys_parser)
        for patch in bridge['vnetwork_connections']:
            del_port_from_switch(patch['name'])
        if bridge['switch'] not in (DEFAULT_SWITCH, BLACKHOLE_SWITCH):
            states_result(remove_switch, switch=bridge['switch'])
    for bridge in current_config['vnetworks']:
        states_result(remove_switch, switch=bridge['switch'])

    # Configuration steps
    uplink_mtu = DEFAULT_MTU
    if not management_params['mac_address']:
        for port in states_result(list_physical_infs, verbose=True, sys_parser=sys_parser):
            if port['interface'] == uplink:
                management_params['mac_address'] = port['mac_address']
                uplink_mtu = port['mtu']
    # Step 5: add uplink port to veil-default
    uplink_vlan_config = dict(vlan_mode='trunk', vlan_tag=None, vlan_trunks=None, mtu=uplink_mtu)
    states_result(connect_interface, interface=uplink, switch=DEFAULT_SWITCH, config=uplink_vlan_config,
                  sys_parser=sys_parser)
    states_result(set_interface_up, interface=uplink, check_conn=False)
    # Step 6: add veil-mgmt interface with specified params
    mgmt_vlan_config = VEIL_MGMT_DEFAULT_VLAN.copy()
    mgmt_vlan_config['mtu'] = DEFAULT_MTU
    if management_params['vlan']:
        mgmt_vlan_config.update(vlan_tag=management_params['vlan'])
    states_result(create_internal_interface,
                  interface=VEIL_MGMT_INTERFACE,
                  switch=DEFAULT_SWITCH,
                  ipv4_address=management_params.get('ipv4_address'),
                  ipv4_netmask=management_params.get('ipv4_netmask'),
                  mac_address=management_params['mac_address'],
                  dhcpv4=management_params['dhcpv4'],
                  config=mgmt_vlan_config)

    # Step 7: set gateway
    if management_params['dhcpv4']:
        dhclient_release(VEIL_MGMT_INTERFACE)
        dhclient_config = update_dhclient_requests(VEIL_MGMT_INTERFACE, DhclientReqType.gateway)
        write_dhclient_conf(dhclient_config)
        dhclient_request(VEIL_MGMT_INTERFACE)
    else:
        gateway = management_params['ipv4_gateway']
        states_result(set_interface_down, interface=VEIL_MGMT_INTERFACE, check_conn=False)
        interface_config_path = os.path.join(BASE_NET_CONF_DIR, VEIL_MGMT_INTERFACE)
        with open(interface_config_path, 'r') as f:
            raw_interface_config = f.read().splitlines()
        interface_config = [s for s in raw_interface_config if 'gateway' not in s]
        interface_config.append(f"    gateway {gateway}")
        with open(interface_config_path, 'w') as f:
            f.write(reduce(lambda a, s: a + s, map(lambda x: f"{x}\n", interface_config)))
        states_result(set_interface_up, interface=VEIL_MGMT_INTERFACE, check_conn=False)

    return "Network configuration flushed successfully."


@states_response
def set_dns_configuration(dns_config: dict):
    """
    Установка конфигурации DNS узла
    :param dns_config: словарь, содержащий параметры DNS
    """

    ret = {'comment': '', 'changes': None}

    initial_config = get_dns_configuration()
    # 1. если начальный конфиг статика и новый конфиг статика: очищаем конфиг днс и записываем новый
    if initial_config['dhcpv4'] is False and dns_config['dhcpv4'] is False:
        purge_resolv_base()
        write_resolv_base(nameservers=dns_config['nameservers'], search=dns_config['search'],
                          domain=dns_config['domain'])
        ret['changes'] = dict(nameservers=dns_config['nameservers'], search=dns_config['search'],
                              domain=dns_config['domain'], dhcpv4=False, interface=None)
    # 2. если начальный конфиг статика, а новый - динамический: очищаем конфиг днс, добавляем днс-запросы в dhcp-конфиг
    #    и делаем dhcp-запрос с новыми параметрами
    elif initial_config['dhcpv4'] is False and dns_config['dhcpv4'] is True:
        purge_resolv_base()
        dhclient_config = update_dhclient_requests(dns_config['interface'], DhclientReqType.dns)
        write_dhclient_conf(dhclient_config)
        dhclient_request(dns_config['interface'])
        time.sleep(3)
        ret['changes'] = get_dns_configuration()
    # 3. если начальный конфиг динамический, а новый статический: удаляем днс-запросы из dhcp-конфига, записываем
    #    статический конфиг и делаем dhcp-запрос с новыми параметрами
    elif initial_config['dhcpv4'] is True and dns_config['dhcpv4'] is False:
        remove_dhclient_requests(DhclientReqType.dns)
        del_file_by_pattern(FILE_RESOLV_STORED_DHCLIENT, initial_config['interface'])
        write_resolv_base(nameservers=dns_config['nameservers'], search=dns_config['search'],
                          domain=dns_config['domain'])
        dhclient_request(initial_config['interface'])
        time.sleep(3)
        ret['changes'] = dict(nameservers=dns_config['nameservers'], search=dns_config['search'],
                              domain=dns_config['domain'], dhcpv4=False, interface=None)
    # 4. если и начальный и новый конфиги - динамические
    else:
        remove_dhclient_requests(DhclientReqType.dns)
        del_file_by_pattern(FILE_RESOLV_STORED_DHCLIENT, initial_config['interface'])
        dhclient_config = update_dhclient_requests(dns_config['interface'], DhclientReqType.dns)
        # dhclient_release(dns_config['interface'])
        write_dhclient_conf(dhclient_config)
        dhclient_request(dns_config['interface'])
        ret['changes'] = get_dns_configuration()

    update_resolv()
    ret['comment'] = 'DNS configuration applied successfully.'
    return ret


@states_response
def set_gateway_configuration(gateway_conf: dict):
    """
    Установка конфигурации основного шлюза узла
    :param gateway_conf: словарь, содержащий параметры основного шлюза
    """

    ret = {'comment': '', 'changes': {}}

    initial_config = get_gateway_configuration()
    updated_iface = states_result(get_internal_inf, name=gateway_conf['interface'])
    interface_changed = True if gateway_conf['interface'] != initial_config['interface'] else False
    conf = get_config()
    controller_ip = conf.get('controller_ip')
    if gateway_conf['interface'] == VEIL_MGMT_INTERFACE and controller_ip:
        existing_routes = parse_routes()
        route = next((r for r in existing_routes if r['destination'] == controller_ip and
                      r['interface'] == VEIL_MGMT_INTERFACE), None)
        if route is not None:
            states_result(del_static_routes, raise_exception=False, static_routes=[route])

    if initial_config['interface'] and initial_config['gateway'] is not None:
        interface_config_path = os.path.join(BASE_NET_CONF_DIR, initial_config['interface'])
        with open(interface_config_path, 'r') as f:
            interface_config = f.read().splitlines()
        result_config = [s for s in interface_config if 'gateway' not in s and 'ip route' not in s]
        with open(interface_config_path, 'w') as f:
            f.write(reduce(lambda a, s: a + s, map(lambda x: f"{x}\n", result_config)))
        if initial_config['dhcpv4']:
            remove_dhclient_requests(DhclientReqType.gateway)
            purge_leases(initial_config['interface'])

    if gateway_conf['dhcpv4'] is False:
        interface_config_path = os.path.join(BASE_NET_CONF_DIR, gateway_conf['interface'])
        with open(interface_config_path, 'r') as f:
            raw_interface_config = f.read().splitlines()
        interface_config = [s for s in raw_interface_config if 'gateway' not in s]
        if updated_iface['dhcpv4']:
            interface_config.append(f"    post-up ip route add default via {gateway_conf['gateway']} "
                                    f"dev {gateway_conf['interface']}")
        else:
            interface_config.append(f"    gateway {gateway_conf['gateway']}")
        with open(interface_config_path, 'w') as f:
            f.write(reduce(lambda a, s: a + s, map(lambda x: f"{x}\n", interface_config)))
    else:
        purge_leases(gateway_conf['interface'])
        dhclient_config = update_dhclient_requests(gateway_conf['interface'], DhclientReqType.gateway)
        write_dhclient_conf(dhclient_config)

    if controller_ip:
        if initial_config['interface'] == VEIL_MGMT_INTERFACE and interface_changed:
            route = dict(destination=controller_ip, gateway=initial_config['gateway'], interface=VEIL_MGMT_INTERFACE)
            states_result(set_static_routes, static_routes=[route])
        if not states_result(check_controller_connection, raise_exception=False):
            pass

    _exec = action_service(BASE_NETWORK_SERVICE, action=ServiceActions.restart)
    if not _exec['result']:
        raise CommandExecutionError(_exec['comment'])
    apply_routes_script()
    default_route = get_gateway_configuration()
    static_routes = parse_routes()
    ret['comment'] = 'Gateway configuration applied successfully.'
    ret['changes'] = dict(default_route=default_route, static_routes=static_routes)
    return ret


@states_response
def check_add_lldpd_exists(path: str = ADD_LLDPD_CONF):
    """
    Проверка наличия конфигурационного файла службы lldp
    """

    return check_file_exists(path)


@states_response
def check_lldpd_service():
    """
    Проверка состояния службы lldp
    """

    return is_service_active(LLDPD_SERVICE)


@states_response
def parse_intfs_lldpd(sys_parser: bool = False) -> list:
    """
    Получение списка интерфейсов, задействованных службой lldp
    """

    return [entry['interface'] for entry in states_result(list_physical_infs, verbose=True, sys_parser=sys_parser) if
            entry['connected'] and entry['state'] == 'UP']


@states_response
def start_lldpd_service():
    """
    Запуск службы lldp
    """

    lldpd_puppet_config()
    if not states_result(check_lldpd_service):
        _exec = action_service(LLDPD_SERVICE, action=ServiceActions.start)
        if not _exec['result']:
            raise CommandExecutionError(_exec['comment'])
        return 'lldpd has been changed to active state.'
    else:
        return 'lldpd was already in active state.'


@states_response
def stop_lldpd_service() -> str:
    """
    Останов службы lldp
    """

    lldpd_puppet_config(add=False)
    if check_lldpd_service()['comment']:
        _exec = action_service(LLDPD_SERVICE, action=ServiceActions.stop)
        if not _exec['result']:
            raise CommandExecutionError(_exec['comment'])
        return 'lldpd has been changed to inactive state.'
    else:
        return 'lldpd was already in inactive state.'


@states_response
def restart_lldpd_service() -> str:
    """
    Перезапуск службы lldp
    """

    lldpd_puppet_config()
    _exec = action_service(LLDPD_SERVICE, action=ServiceActions.restart)
    if not _exec['result']:
        raise CommandExecutionError(_exec['comment'])
    return 'lldpd has been restarted successfully.'


@states_response
def get_lldpd_neighbors(sys_parser: bool = False) -> Union[list, str]:
    """
    Получение информации о соседях на узле в формате json
    """

    if not states_result(check_lldpd_service):
        return 'lldpd service is not active. You should activate it.'
    cmd = f"lldpcli show neighbors details ports \"{','.join(states_result(parse_intfs_lldpd, sys_parser=sys_parser))}\" -f json"
    neighbors = json.loads(states_result(get_output, cmd=cmd))
    if not neighbors['lldp']:
        return []
    elif type(neighbors['lldp']['interface']) == dict:
        return transform([neighbors['lldp']['interface']])
    elif type(neighbors['lldp']['interface']) == list:
        return transform(neighbors['lldp']['interface'])
    else:
        return []


@states_response
def parse_lldpcli() -> dict:
    """
    Функция парсит вывод команды утилиты lldpcli
    """

    d = dict()
    r_chassis = (r' *ChassisID: +(?P<chassis_id_subtype>\D+) (?P<chassis_id>.+)\n'
                 r' *SysName: +(?P<system_name>.+)\n'
                 r' *SysDescr: +(?P<system_description>.+)\n'
                 r'(?P<mgmt_ip>(?: *MgmtIP: +.+\n)+)'
                 )
    r_conf_run = (r' *Transmit delay: +(?P<tx_interval>\d+)\n'
                  r'.+'
                  r' *Transmit hold: +(?P<tx_hold>\d+)\n')
    cmd = 'lldpcli show chassis details'
    output = states_result(get_output, cmd=cmd)
    match = re.search(r_chassis, output)
    if match:
        d.update(match.groupdict())
        d['mgmt_ip'] = ', '.join([i.split()[1] for i in match.groupdict().get('mgmt_ip').split('\n') if i])
        del d['chassis_id_subtype']

    cmd = 'lldpcli show running-configuration'
    output = states_result(get_output, cmd=cmd)
    match = re.search(r_conf_run, output, re.DOTALL)
    if match:
        d.update(match.groupdict())
    return d


@states_response
def parse_lldpd_config_base(config_file: str = BASE_LLDPD_CONF):
    lldpd_config_base = dict()
    lldpd_config_base['protocols'] = []
    regex = r'DAEMON_ARGS="(.*)"'
    if os.path.exists(config_file):
        try:
            with open(config_file) as f:
                for line in f:
                    if line.startswith('DAEMON_ARGS'):
                        match = re.search(regex, line)
                        if match:
                            config_base = match.group(1).split()
                            if '-ll' not in config_base:
                                lldpd_config_base['protocols'].append('lldp')
                            revert_d = dict(map(reversed, LLDPD_PROTOCOLS.items()))
                            # pylint: disable=cell-var-from-loop
                            lldpd_config_base['protocols'].extend(list(map(lambda x: revert_d[x], config_base)))
        except Exception as e:
            raise CommandExecutionError("Unexpected error: {0}".format(str(e))) from None
    if not lldpd_config_base['protocols']:
        lldpd_config_base['protocols'].append('lldp')
    return lldpd_config_base


@states_response
def parse_lldpd_config_add(config_file: str = ADD_LLDPD_CONF) -> dict:
    lldpd_config_add = dict.fromkeys(LLDPD_CONF_ADD_FIELDS, '')
    regex = r'\"(.*)\"'
    if os.path.exists(config_file):
        try:
            with open(config_file, 'r') as f:
                read_file_lines = f.read().splitlines()
                for line in read_file_lines:
                    if 'chassisid' in line:
                        match = re.search(regex, line)
                        lldpd_config_add['chassis_id'] = match.group(1)
                    elif 'ip management pattern' in line:
                        match = re.search(regex, line)
                        lldpd_config_add['mgmt_ip'] = match.group(1)
                    elif 'system hostname' in line:
                        match = re.search(regex, line)
                        lldpd_config_add['system_name'] = match.group(1)
                    elif 'system description' in line:
                        match = re.search(regex, line)
                        lldpd_config_add['system_description'] = match.group(1)
                    elif 'tx-interval' in line:
                        match = re.search(regex, line)
                        lldpd_config_add['tx_interval'] = int(match.group(1))
                    elif 'tx-hold' in line:
                        match = re.search(regex, line)
                        lldpd_config_add['tx_hold'] = int(match.group(1))
        except Exception as e:
            raise CommandExecutionError("Unexpected error: {0}".format(str(e))) from None
    else:
        lldpd_config_add.update(states_result(parse_lldpcli))
    return lldpd_config_add


@states_response
def parse_lldpd_config_ports(config_dir: str = ADD_LLDPD_DIR, sys_parser: bool = False) -> dict:
    conf_ports = {}
    ports = []
    r_conf_file = r'lldpd_config_port_\S+.conf'
    r_status = r'configure ports "(\S+)" lldp status "(\D+)"'
    r_descr = r'configure ports "(\S+)" lldp portdescription "(.+)"'
    files = [f for f in os.listdir(config_dir) if os.path.isfile(config_dir + f)]
    for filename in files:
        conf_port = {}
        match = re.search(r_conf_file, filename)
        if match:
            try:
                with open(ADD_LLDPD_DIR + filename, 'r') as f:
                    for line in f:
                        match = re.search(r_status, line.strip())
                        if match:
                            conf_port['port'] = match.group(1)
                            conf_port['status'] = match.group(2)

                        match = re.search(r_descr, line.strip())
                        if match:
                            conf_port['portdescription'] = match.group(2)
            except Exception as e:
                raise CommandExecutionError("Unexpected error: {0}".format(str(e))) from None

            ports.append(conf_port)
    if not ports:
        infs = [p for p in states_result(list_physical_infs, verbose=True, sys_parser=sys_parser)
                if p['connected'] and p['state'] == 'UP']
        for port in infs:
            conf_port = dict()
            conf_port['port'] = port['interface']
            conf_port['status'] = 'rx-and-tx'
            conf_port['portdescription'] = port['interface']
            ports.append(conf_port)
    conf_ports['ports'] = ports
    return conf_ports


@states_response
def change_lldpd_config_base(data: dict, config_file: str = BASE_LLDPD_CONF) -> str:
    """
    Запись конфигурации демона lldpd узла в базовый файл конфигурации
    :param data: конфигурация демона lldpd
    :param config_file: путь к файлу конфигурации
    :return: str "lldpd to base configuration file applied successfully." or raise Exception "CommandExecutionError"
    """

    config_str = ' '.join(list(map(lambda x: LLDPD_PROTOCOLS[x], data.get('protocols')))).strip(' ')
    config_line = f'DAEMON_ARGS="{config_str}"'
    write_lldpd_config(config_line, config_file)
    return 'lldpd to base configuration file applied successfully.'


@states_response
def change_lldpd_config_add(data: dict, config_file: str = ADD_LLDPD_CONF) -> str:
    """
    Запись конфигурации демона lldpd в дополнительный файл конфигурации и последующий перезапуск службы для
    применения конфигурации.
    При включении службы lldpd выполняются строки из файла config_file
    :param data: словарь конфигурации
    :param config_file: путь к дополнительному файлу конфигурации демона lldpd
    имя файла любая строка, а расширение должно быть 'conf'
    по-умолчанию используется значение '/etc/lldpd.d/lldpd_config.conf'
    :return: str "lldpd to add configuration file applied successfully." or raise Exception "CommandExecutionError"
    """

    try:
        with open(config_file, 'w') as f:
            for k, v in data.items():
                if k == 'chassis_id':
                    f.write(f'configure system chassisid "{v}"\n')
                elif k == 'mgmt_ip':
                    f.write(f'configure system ip management pattern "{v}"\n')
                elif k == 'system_name':
                    f.write(f'configure system hostname "{v}"\n')
                elif k == 'system_description':
                    f.write(f'configure system description "{v}"\n')
                elif k == 'tx_interval':
                    f.write(f'configure lldp tx-interval "{v}"\n')
                elif k == 'tx_hold':
                    f.write(f'configure lldp tx-hold "{v}"\n')
    except OSError as e:
        raise CommandExecutionError(
            "Failed to write configuration. I/O error({0}): {1}".format(e.errno, e.strerror)) from None
    except Exception as e:
        raise CommandExecutionError("Unexpected error: {0}".format(str(e))) from None
    return 'lldpd to add configuration file applied successfully.'


@states_response
def change_lldpd_config_ports(data: dict) -> str:
    """
    Запись конфигурации демона lldpd в дополнительные файлы конфигурации для порта и последующий перезапуск
    службы для применения конфигурации.
    При включении службы lldpd выполняются строки из файла config_file
    :param data: вложенный словарь конфигурации
    data должен приходить в таком виде
    Пример:
    "ports": [{"port": "имя_порта", "status": "", "portdescription": ""},..., {...}}]
    "port" имя_интерфейса выбирается из списка на front-end
    "status" может принимать такие значения: rx-and-tx, rx-only, tx-only, disabled. ВЫбирается из списка на front-end
    "portdescription" любая строка
    :return: str "lldpd to port configuration files applied successfully." or raise Exception "CommandExecutionError"
    """

    for d in data.get('ports'):
        conf_lines = []
        port = d.get('port')
        config_file = PORT_LLDPD_CONF.format(lldpd_port=port)
        status = d.get('status')
        descr = d.get('portdescription')
        conf_lines.append(f'configure ports "{port}" lldp status "{status}"')
        conf_lines.append(f'configure ports "{port}" lldp portdescription "{descr}"')
        write_lldpd_config(conf_lines, config_file)

    return 'lldpd to ports configuration files applied successfully.'


@states_response
def change_lldpd_config(config_base: dict, config_add: dict, config_ports: dict) -> str:
    try:
        states_result(change_lldpd_config_add, data=config_add)
        states_result(change_lldpd_config_ports, data=config_ports)
        states_result(change_lldpd_config_base, data=config_base)
    except Exception as e:
        raise CommandExecutionError(str(e)) from None
    if states_result(check_lldpd_service):
        restart_lldpd_service()
    return 'lldpd configuration applied successfully.'


@states_response
def set_firewall_logging(firewall_logging):
    """
    Запуск/остановка службы логирования брандмауэра
    :param firewall_logging статус логирования брандмауэра
    """

    if firewall_logging:
        err = start_firewall_logging_service(get_err=True)
        if err:
            raise CommandExecutionError(err)
        return "Firewall logging started successfully."

    else:
        err = stop_firewall_logging_service(get_err=True)
        if err:
            raise CommandExecutionError(err)
        return "Firewall logging stopped successfully."


def start_firewall_logging_service(get_err=False):
    """
    Запуск службы логирования брандмауэра узла
    """

    cmd = 'systemctl start ulogd2.service'
    return True if states_result(
        get_output, raise_exception=False, get_err=get_err, cmd=cmd, suppress_err=not get_err) else False


def stop_firewall_logging_service(get_err=False):
    """
    Остановка службы логирования брандмауэра узла
    """

    cmd = 'systemctl stop ulogd2.service'
    return True if states_result(
        get_output, raise_exception=False, get_err=get_err, cmd=cmd, suppress_err=not get_err) else False


def _get_folder_files(path_to_dir):
    try:
        content_dir = os.listdir(path_to_dir)
    except OSError as e:
        raise CommandExecutionError(
            f"Failed to get directory content {path_to_dir}. Error({e.errno}): {e.strerror}") from None
    return [f for f in content_dir if os.path.isfile(path_to_dir + f)]


@states_response
def link_iface_name_to_mac(iface_name: str, sys_parser: bool = False) -> str:
    """
    Запись в специальный файл соответствия между именем физического интерфейса и mac-адресом.
    Таким образом имя физического интерфейса не будет меняться в ОС.
    """

    unlink_iface_name_to_mac(iface_name, sys_parser=sys_parser)
    files = _get_folder_files(LINK_DIR)
    if not files:
        filename = f"1-{iface_name}.link"
    else:
        if len(files) == 98:  # https://manpages.debian.org/testing/udev/systemd.link.5.en.html
            raise CommandExecutionError("Can't link interface name to mac-address. "
                                        "You should change file name /lib/systemd/network/99-default.link")
        exist_number = [re.search(LINK_FLIE_PATTERN, f).group(1) for f in files]
        number = next((number for number in range(1, 99) if str(number) not in exist_number), None)
        filename = f"{str(number)}-{iface_name}.link"
    target_file = os.path.join(LINK_DIR, filename)
    mac_address = next((i["mac_address"] for i in states_result(list_physical_infs, verbose=True, sys_parser=sys_parser)
                        if i['interface'] == iface_name), None)
    if not mac_address:
        return f"There is no mac-address {mac_address} to interface name {iface_name}.\n"
    temp = f"[Match]\nMACAddress={mac_address}\n\n[Link]\nName={iface_name}\n"
    try:
        with open(target_file, "w") as f:
            f.write(temp)
    except OSError as e:
        raise CommandExecutionError(f"Failed to write configuration into file {target_file}. "
                                    f"I/O error({e.errno}): {e.strerror}") from None
    except Exception as e:
        raise CommandExecutionError(f"Unexpected error: {str(e)}") from None
    return f"Link interface name {iface_name} to mac-address {mac_address} successfully.\n"


@states_response
def unlink_iface_name_to_mac(iface_name: str, sys_parser: bool = False) -> str:
    """
    Удаление специального файла соответствия между именем физического интерфейса и mac-адресом.
    Таким образом имя физического интерфейса может меняться в ОС.
    """

    files = _get_folder_files(LINK_DIR)
    filename = next((f for f in files if files and re.search(LINK_FLIE_PATTERN, f).group(2) == iface_name), None)
    if not filename:
        return f"There is no interface name {iface_name}.\n"
    target_file = os.path.join(LINK_DIR, filename)
    mac_address = next((i["mac_address"] for i in states_result(list_physical_infs, verbose=True, sys_parser=sys_parser)
                        if i['interface'] == iface_name), None)
    if not mac_address:
        return f"There is no mac-address {mac_address} to interface name {iface_name}.\n"
    try:
        os.remove(target_file)
    except OSError as e:
        raise CommandExecutionError(f"Failed to delete file {target_file}. Error({e.errno}): {e.strerror}") from None
    except Exception as e:
        raise CommandExecutionError(f"Unexpected error: {str(e)}") from None
    return f"Unlink interface name {iface_name} to mac-address {mac_address} successfully.\n"


# endregion


# region virtual networks states & functions
def prepare_base_template(template_extracted=False):
    target_path = os.path.join(BASE_LXC_PATH, BASE_LXC_CONTAINER)
    if not template_extracted:
        states_result(get_output, cmd=f'mkdir {os.path.join(BASE_LXC_PATH, BASE_LXC_CONTAINER)}')
        states_result(get_output, cmd=f'tar --numeric-owner -xzf {LXC_TEMPLATE} -C {target_path}', timeout=180)

    with open(os.path.join(target_path, 'config'), 'w') as f:
        f.write(reduce(lambda a, s: a + s, map(lambda x: f"{x}\n", BASE_LXC_CONFIG)))

    if not VNService.exists(BASE_LXC_CONTAINER):
        raise CommandExecutionError(VnsUnexpectedErrors.errno3)

    base = VNService(BASE_LXC_CONTAINER)
    net_config = os.path.join(base.get_config_item('lxc.rootfs.path'), BASE_NET_CONFIG.lstrip('/'))
    write_interfaces_config(net_config)
    write_interface(os.path.join(base.get_config_item('lxc.rootfs.path'), BASE_NET_CONF_DIR.lstrip('/')),
                    'lo', InterfaceModes.loopback)


def modify_lxc_interface(vns: 'VNService', params: dict, net_index: int):
    """
    Изменение параметров интерфейса в конфигурации контейнера
    """

    if params.get('mac_address'):
        vns.set_config_item(f"lxc.net.{net_index}.hwaddr", f"{params['mac_address']}")
    if params.get('dhcpv4'):
        vns.clear_config_item(f"lxc.net.{net_index}.ipv4.address")
        vns.clear_config_item(f"lxc.net.{net_index}.ipv4.gateway")
    else:
        if params.get('ipv4_address'):
            vns.set_config_item(f"lxc.net.{net_index}.ipv4.address", params['ipv4_address'])
        if params.get('ipv4_gateway'):
            vns.set_config_item(f"lxc.net.{net_index}.ipv4.gateway", params['ipv4_gateway'])
    vns.save_config()


def remove_lxc_interface_files(vns_path: str, iface: str):
    """
    Удаления файла конфигурации интерфейса контейнера по указанному пути
    """

    try:
        del_file_by_pattern(vns_path, iface)
    except Exception as e:
        raise CommandExecutionError(str(e)) from None


def write_lxc_interface_scripts(vns_path: str, params: dict):
    """
    Запись конфигурационных скриптов для сетевых интерфейсов контейнера
    """

    script_up = os.path.join(vns_path, params['name'] + '_up.sh')
    script_down = os.path.join(vns_path, params['name'] + '_down.sh')

    try:
        with open(script_up, 'w') as c:
            c.write('#!/bin/bash\n\n')
            c.write(f'BRIDGE="{params["switch"]}"\n')
            c.write('ovs-vsctl --may-exist add-br $BRIDGE\n')
            c.write('ovs-vsctl --if-exists del-port $BRIDGE $5\n')
            c.write('ovs-vsctl --may-exist add-port $BRIDGE $5\n')
            if params.get('vlan_id'):
                c.write(f'ovs-vsctl set port $5 vlan_mode=access tag={params["vlan_id"]}\n')
            c.write(f'ovs-vsctl set int $5 mtu_request={UPLINK_MTU}\n')
            c.write('ip link set $5 down\n')

        with open(script_down, 'w') as c:
            c.write('#!/bin/bash\n\n')
            c.write(f'BRIDGE="{params["switch"]}"\n')
            c.write('ovs-vsctl --if-exists del-port $BRIDGE $5\n')
    except OSError as e:
        raise CommandExecutionError(f"Failed to write configuration. I/O error({e.errno}): {e.strerror}") from None
    except Exception as e:
        raise CommandExecutionError(f"Unexpected error: {str(e)}") from None

    states_result(get_output, cmd=f'chmod +x {script_up}')
    states_result(get_output, cmd=f'chmod +x {script_down}')


def get_lxc_ifconfig(vns: 'VNService', interface: str) -> list:
    """
    Получение текущей конфигурации сетевого интерфейса контейнера
    """

    output = vns.attach_cmd(f"ifconfig {interface}", get_output=True)
    return output.split('\n')


def get_lxc_default_gateway(vns: 'VNService') -> str:
    """
    Получение текущего основного шлюза контейнера
    """

    raw_output = vns.attach_cmd("ip route", get_output=True)
    output = raw_output.split('\n')
    return output[0].split()[2] if output else ''


def parse_lxc_address(source: str) -> dict:
    """
    Получение ip-адреса из текущей конфигурации сетевого интерфейса контейнера
    """

    try:
        info = source.lstrip().split()
        if 'inet' in info and 'netmask' in info:
            params = dict(ipv4_address=info[1], ipv4_netmask=info[3])
        else:
            params = dict(ipv4_address='', ipv4_netmask='')
        return params
    except Exception as e:
        raise CommandExecutionError(f"Failed to parse container ip-address: {str(e)}") from None


def set_lxc_links_up(vns_path: str):
    """
    Включение интерфейсов контейнера
    """

    try:
        for script in os.listdir(vns_path):
            if re.search(LXC_UP_SCRIPTS_PATTERN, script):
                with open(os.path.join(vns_path, script), 'r') as c:
                    config = c.read().splitlines()
                while 'ip link set $5 down' in config:
                    config.remove('ip link set $5 down')
                with open(os.path.join(vns_path, script), 'w') as f:
                    f.write(reduce(lambda a, s: a + s, map(lambda x: f"{x}\n", config)))
    except OSError as e:
        raise CommandExecutionError(
            "Unable to update configuration. I/O error({0}): {1}".format(e.errno, e.strerror)) from None
    except Exception as e:
        raise CommandExecutionError("Unexpected error: {0}".format(str(e))) from None


def set_lxc_links_down(vns_path: str):
    """
    Выключение интерфейсов контейнера
    """

    try:
        for script in os.listdir(vns_path):
            if re.search(LXC_UP_SCRIPTS_PATTERN, script):
                with open(os.path.join(vns_path, script), 'r') as c:
                    config = c.read().splitlines()
                if 'ip link set $5 down' not in config:
                    with open(os.path.join(vns_path, script), 'a') as c:
                        c.write('ip link set $5 down\n')
    except OSError as e:
        raise CommandExecutionError(
            "Unable to update configuration. I/O error({0}): {1}".format(e.errno, e.strerror)) from None
    except Exception as e:
        raise CommandExecutionError("Unexpected error: {0}".format(str(e))) from None


def vn_service_change_enabled(vns_name: str, service_name: str, command: str) -> bool:
    """
    Изменение автозагрузки сетевой службы
    :param vns_name: название контейнера сетевых служб
    :param service_name: название сетевой службы
    :param command: состояние автозагрузки (enable/disable)
    """

    c = VNService(vns_name)
    _exec = c.attach_cmd(f"systemctl {command} {service_name}")
    if not _exec:
        return False
    return True


def vn_service_start(vns_name: str, service_name: str) -> bool:
    """
    Запуск сетевой службы
    :param vns_name: название контейнера сетевых служб
    :param service_name: название сетевой службы
    """

    c = VNService(vns_name)
    _exec = c.attach_cmd(f"systemctl start {service_name}")
    if not _exec:
        return False
    return True


def vn_service_stop(vns_name: str, service_name: str) -> bool:
    """
    Останов сетевой службы
    :param vns_name: название контейнера сетевых служб
    :param service_name: название сетевой службы
    """

    c = VNService(vns_name)
    _exec = c.attach_cmd(f"systemctl stop {service_name}")
    if not _exec:
        return False
    return True


def write_dhcp_config(config: dict, config_file: str):
    """
    Запись конфигурации dhcp-сервера сетевых служб
    :param config: конфигурация для записи
    :param config_file: файл конфигурации
    """

    try:
        with open(config_file, 'w') as c:
            json.dump(config, c, indent=4)
    except OSError as e:
        raise CommandExecutionError(
            "Failed to write configuration. I/O error({0}): {1}".format(e.errno, e.strerror)) from None
    except Exception as e:
        raise CommandExecutionError("Unexpected error: {0}".format(str(e))) from None


def update_dhcp_config(conf: dict, data: dict):
    """
    Изменение конфигурации dhcp-сервера сетевых служб
    """

    try:
        if LXC_INTERNAL_INTERFACE not in conf['Dhcp4']['interfaces-config']['interfaces']:
            conf['Dhcp4']['interfaces-config']['interfaces'].append(LXC_INTERNAL_INTERFACE)
        subnet_config = dict()
        subnet_config['subnet'] = data['subnet']
        subnet_config['interface'] = LXC_INTERNAL_INTERFACE
        if len(data['dhcp_pools']):
            subnet_config['pools'] = [dict(
                pool=f"{pool['first_address']} - {pool['last_address']}") for pool in data['dhcp_pools']]
        if len(data['dhcp_options']):
            subnet_config['option-data'] = data['dhcp_options']
        if len(data['dhcp_reservations']):
            subnet_config['reservations'] = data['dhcp_reservations']
        if data['dhcp_lease_lifetime']:
            subnet_config['valid-lifetime'] = data['dhcp_lease_lifetime']

        if not len(conf['Dhcp4']['subnet4']):
            conf['Dhcp4']['subnet4'].append(subnet_config)
        else:
            conf['Dhcp4']['subnet4'][0] = subnet_config
        return conf
    except Exception as e:
        raise CommandExecutionError(str(e)) from None


def get_netflow_id(switch: str):
    """
    Получение id конфигурации netflow в OVS
    """

    cmd = f'ovs-vsctl get br {switch} netflow'
    cmd_ret = states_result(get_output, cmd=cmd)
    return cmd_ret.splitlines()[0]


def get_netflow_param(netflow, param):
    """
    Получение параметров конфигурации netflow
    :param netflow: id конфигурации в OVS
    :param param: название параметра
    """

    cmd = f'ovs-vsctl get netflow {netflow} {param}'
    cmd_ret = states_result(get_output, cmd=cmd)
    return cmd_ret.splitlines()[0]


def set_netflow_params(switch: str, params: dict):
    """
    Установка параметров netflow коммутатора
    :param switch: название коммутатора
    :param params: параметры
    """

    cmd = f'ovs-vsctl set br {switch} netflow=@nf_{switch} -- --id=@nf_{switch} create NetFlow ' \
          f'targets=\\"{params["netflow_target_ip"]}:{params["netflow_target_port"]}\\" ' \
          f'add_id_to_interface={"true" if params["netflow_id_to_iface"] else "false"}'
    states_result(get_output, base_err_msg=f'Failed to set netflow parameters for switch {switch}', cmd=cmd)

    cmd = f'ovs-vsctl set NetFlow {switch} active_timeout={params["netflow_active_timeout"]}'
    states_result(get_output, base_err_msg=f'Failed to set netflow timeout for switch {switch}', cmd=cmd)


def flush_netflow(switch: str):
    """
    Сброс параметров netflow коммутатора
    :param switch: название коммутатора
    """

    cmd = f'ovs-vsctl clear br {switch} netflow'
    states_result(get_output, base_err_msg=f'Failed to reset netflow parameters for switch {switch}', cmd=cmd)


@states_response
def ping_from_node(targets, cmd_str=None):
    """
    Пингование адресов с нескольких нод
    :param targets: цели для пинга
    :param cmd_str: команда пинга
    """
    if not cmd_str:
        cmd_str = "nping -c 2 {target} || true"

    def ping_target(target):
        return {target: states_result(get_output, raise_exception=False, cmd=cmd_str.format(target=target))}

    return thread_pool(ping_target, targets)


@states_response
def ping_from_node_df(target, data_size, interface):
    """
    Пингование адреса с ноды, с выключенной фрагментацией (df = don't fragment)
    :param target: цель для пинга
    :param data_size: длина пакета
    :param interface: source интерфейс
    """
    cmd_str = "nping --icmp --interface {interface} --data-length {data_size} -df -c 2 {target} || true"
    return states_result(get_output, raise_exception=False, cmd=cmd_str.format(target=target,
                                                                               interface=interface,
                                                                               data_size=data_size))


@states_response
def vn_service_get_leases(vns_name: str):
    """
    Получение dhcp4 leases
    """

    if not VNService.exists(vns_name):
        raise CommandExecutionError("Virtual network services container does not exist.")

    return VNService(vns_name).get_leases()


@states_response
def vn_service_set_leases(vnservices_name: str, leases: List[dict]):
    """
    Установка dhcp4 leases
    """

    if not VNService.exists(vnservices_name):
        raise CommandExecutionError("Virtual network services container does not exist.")

    vnservices = VNService(vnservices_name)
    vnservices.del_leases()
    vnservices.set_leases(leases)
    if VNService.get_status(vnservices_name) == VNService.STATUS.running:
        state = states_result(vn_service_state, vns_name=vnservices_name, service_name=DHCP_SERVICE)
        if state['active']:
            if not vn_service_stop(vnservices_name, DHCP_SERVICE):
                raise CommandExecutionError(f"Failed to stop {DHCP_SERVICE}.")
            time.sleep(1)
            state = states_result(vn_service_state, vns_name=vnservices_name, service_name=DHCP_SERVICE)
            if state['active']:
                raise CommandExecutionError(f"Failed to stop {DHCP_SERVICE}.")
            if not vn_service_start(vnservices_name, DHCP_SERVICE):
                raise CommandExecutionError(f"Failed to start {DHCP_SERVICE}.")
            time.sleep(1)
            state = states_result(vn_service_state, vns_name=vnservices_name, service_name=DHCP_SERVICE)
            if not state['active']:
                raise CommandExecutionError(f"Failed to start {DHCP_SERVICE}.")
    return 'Leases data has been set successfully.'


@states_response
def vn_service_del_leases(vnservices_name: str):
    """
    Удаление dhcp4 leases
    """

    if not VNService.exists(vnservices_name):
        raise CommandExecutionError("Virtual network services container does not exist.")

    VNService(vnservices_name).del_leases()
    if VNService.get_status(vnservices_name) == VNService.STATUS.running:
        state = states_result(vn_service_state, vns_name=vnservices_name, service_name=DHCP_SERVICE)
        if state['active']:
            if not vn_service_stop(vnservices_name, DHCP_SERVICE):
                raise CommandExecutionError(f"Failed to stop {DHCP_SERVICE}.")
            time.sleep(1)
            state = states_result(vn_service_state, vns_name=vnservices_name, service_name=DHCP_SERVICE)
            if state['active']:
                raise CommandExecutionError(f"Failed to stop {DHCP_SERVICE}.")
            if not vn_service_start(vnservices_name, DHCP_SERVICE):
                raise CommandExecutionError(f"Failed to start {DHCP_SERVICE}.")
            time.sleep(1)
            state = states_result(vn_service_state, vns_name=vnservices_name, service_name=DHCP_SERVICE)
            if not state['active']:
                raise CommandExecutionError(f"Failed to start {DHCP_SERVICE}.")
    return 'Leases data has been deleted successfully.'


@states_response
def create_vns_container(vnservices_name: str, dhcp_conf: dict, fw_conf: list, memory_limit: int, cpu_arch: str):
    """
    Создание контейнера сетевых служб из шаблона
    :param vnservices_name: название контейнера
    :param dhcp_conf: базовая конфигурация службы dhcp
    :param fw_conf: базовая конфигурация службы брандмауэра
    :param memory_limit: лимит оперативной памяти контейнера
    :param cpu_arch: архитектура CPU узла, на котором разворачивается контейнер
    """

    if VNService.exists(vnservices_name):
        old_vns = VNService(vnservices_name)
        if not old_vns.destroy():
            raise CommandExecutionError("Unable to remove old container with the same name.")

    prepare_base_template(template_extracted=True if VNService.exists(BASE_LXC_CONTAINER) else False)
    base = VNService(BASE_LXC_CONTAINER)

    vnservices = base.clone(vnservices_name)
    if not vnservices:
        raise CommandExecutionError("Cloning of base template failed.")
    vnservices.set_config_item("lxc.cgroup.memory.limit_in_bytes", f"{memory_limit}M")
    vnservices.set_config_item("lxc.arch", cpu_arch)
    vnservices.save_config()

    vns_root = vnservices.get_config_item('lxc.rootfs.path')
    try:
        with open(os.path.join(vns_root, 'etc/hostname'), 'w') as c:
            c.write(f'{vnservices_name}')
    except OSError as e:
        raise CommandExecutionError(
            "Failed to write configuration. I/O error({0}): {1}".format(e.errno, e.strerror)) from None
    except Exception as e:
        raise CommandExecutionError("Unexpected error: {0}".format(str(e))) from None

    vns_net_conf_file = os.path.join(vns_root, BASE_NET_CONFIG.lstrip('/'))
    vns_net_conf_dir = os.path.join(vns_root, BASE_NET_CONF_DIR.lstrip('/'))
    write_interfaces_config(vns_net_conf_file)
    write_interface(vns_net_conf_dir, 'lo', InterfaceModes.loopback)
    write_firewall_config(fw_conf, os.path.join(vns_root, BASE_FIREWALL_CONF.lstrip('/')))
    write_dhcp_config(dhcp_conf, os.path.join(vns_root, BASE_DHCP_CONF.lstrip('/')))

    return 'Virtual network services container created successfully.'


@states_response
def create_vns_interface(vnservices_name: str, ifaces_params: dict):
    """
    Запись конфигурации сетевого интерфейса контейнера
    :param vnservices_name: название контейнера
    :param ifaces_params: параметры интерфейса
    """

    ret = {'comment': '', 'changes': {}}

    if not VNService.exists(vnservices_name):
        raise CommandExecutionError("Virtual network services container does not exist.")

    vns_container = VNService(vnservices_name)
    for iface_params in ifaces_params:
        index = 0
        network_type = "veth"
        params = dict()
        if len(vns_container.networks) == 0:
            index = 0
        if len(vns_container.networks) == 1 and iface_params['direction'] == VnsInterfaceDirection.external:
            index = 1
        params['name'] = iface_params['name'][4:]
        params['pair'] = iface_params['name']
        params['flags'] = "up"
        params['mac_address'] = iface_params['mac_address']
        params['mtu'] = iface_params['mtu']
        params['ipv4_address'] = iface_params['ipv4_address'] if iface_params['ipv4_address'] else None
        params['ipv4_gateway'] = iface_params['ipv4_gateway'] if iface_params['ipv4_gateway'] else None
        params['script_up'] = os.path.join(vns_container.path, params['pair'] + '_up.sh')
        params['script_down'] = os.path.join(vns_container.path, params['pair'] + '_down.sh')
        vns_container.create_interface(index, network_type, params)
        vns_container.save_interface(index)

        vns_root = vns_container.get_config_item('lxc.rootfs.path')
        net_conf_dir = os.path.join(vns_root, BASE_NET_CONF_DIR.lstrip('/'))
        vns_path = vns_container.path
        if not iface_params.get('switch'):
            iface_params['switch'] = BLACKHOLE_SWITCH
        write_lxc_interface_scripts(vns_path, iface_params)
        write_interface(conf_dir=net_conf_dir, name=iface_params['name'][4:],
                        mode=InterfaceModes.dhcp if iface_params['dhcpv4'] else InterfaceModes.manual)

    ret['comment'] = "Configuration changed successfully."
    return ret


@states_response
def remove_vns_interface(vnservices_name: str, params: dict):
    """
    Удаление конфигурации сетевого интерфейса контейнера
    :param vnservices_name: название контейнера
    :param params: параметры интерфейса
    """

    ret = {'comment': '', 'changes': {}}

    if not VNService.exists(vnservices_name):
        raise CommandExecutionError("Virtual network services container does not exist.")

    vns = VNService(vnservices_name)
    vns_root = vns.get_config_item('lxc.rootfs.path')
    vns_path = os.path.join(BASE_LXC_PATH, vnservices_name)
    vns_interfaces_conf = os.path.join(vns_root, BASE_NET_CONF_DIR.lstrip('/'))
    net_index = 0 if params['direction'] == VnsInterfaceDirection.internal else 1
    if not vns.remove_interface(net_index):
        raise CommandExecutionError("Failed to remove network services interface.")
    remove_lxc_interface_files(vns_path, params['name'])
    remove_lxc_interface_files(vns_interfaces_conf, params['name'][4:])

    if vns.status == VNService.STATUS.running:
        if not vns.wait_stopped():
            raise CommandExecutionError("Failed to restart network services.")
        for net in vns.get_networks():
            del_port_from_switch(net.pair)
        if not vns.wait_started():
            raise CommandExecutionError("Failed to restart network services.")

    ret['comment'] = "Configuration changed successfully."
    return ret


@states_response
def modify_vns_interface(vnservices_name: str, interface: str, params: dict):
    """
    Изменение параметров сетевого интерфейса контейнера
    :param vnservices_name: название контейнера
    :param interface: название интерфейса
    :param params: словарь с обновленными параметрами
    """

    ret = {'comment': '', 'changes': {}}

    if not VNService.exists(vnservices_name):
        raise CommandExecutionError("Virtual network services container does not exist.")

    vns = VNService(vnservices_name)
    vns_root = vns.get_config_item('lxc.rootfs.path')
    vns_interfaces_conf = os.path.join(vns_root, BASE_NET_CONF_DIR.lstrip('/'))
    net_index = 0 if params['direction'] == VnsInterfaceDirection.internal else 1
    modify_lxc_interface(vns, params, net_index)
    remove_lxc_interface_files(vns_interfaces_conf, interface[4:])
    write_interface(conf_dir=vns_interfaces_conf, name=interface[4:],
                    mode=InterfaceModes.dhcp if params.get('dhcpv4', False) else InterfaceModes.manual)

    if vns.status == VNService.STATUS.running:
        if not vns.wait_stopped():
            raise CommandExecutionError("Failed to restart network services.")
        for net in vns.get_networks():
            del_port_from_switch(net.pair)
        if not vns.wait_started():
            raise CommandExecutionError("Failed to restart network services.")

    ret['comment'] = "Configuration changed successfully."
    return ret


@states_response
def create_vns_interface_alias(vnservices_name: str, alias_id: str, ipv4_address: str):
    """ Создаёт алиас на внешнем интерфейсе для floating ip. """

    ret = {'comment': '', 'changes': {}}
    if not VNService.exists(vnservices_name):
        raise CommandExecutionError("Virtual network services container does not exist.")

    vns_container = VNService(vnservices_name)
    vns_root = vns_container.get_config_item('lxc.rootfs.path')
    net_conf_dir = os.path.join(vns_root, BASE_NET_CONF_DIR.lstrip('/'))
    inf_index = 0

    if os.path.exists(f"{net_conf_dir}/{LXC_EXTERNAL_INTERFACE}-{alias_id}"):
        ret['comment'] = "Alias configuration already exists. "
        ret['comment'] += states_result(modify_vns_interface_alias, vnservices_name=vnservices_name,
                                        alias_id=alias_id, ipv4_address=ipv4_address).get('comment')
    else:
        for inf_conf in os.listdir(net_conf_dir):
            if inf_conf.startswith(f"{LXC_EXTERNAL_INTERFACE}-"):
                with open(f"{net_conf_dir}/{inf_conf}", "r") as conf:
                    for line in conf.readlines():
                        line = line.strip()
                        if line.startswith("iface"):
                            iface_index = int(line.split().pop(1).split(':').pop(1))
                            inf_index = iface_index if iface_index > inf_index else inf_index
        inf_index += 1
        write_interface_alias(conf_dir=net_conf_dir, mode=InterfaceModes.static,
                              inf_name=f"{LXC_EXTERNAL_INTERFACE}:{inf_index}",
                              conf_name=f"{LXC_EXTERNAL_INTERFACE}-{alias_id}",
                              ipv4_address=ipv4_address, ipv4_netmask="255.255.255.255")
        if vns_container.status == VNService.STATUS.running:
            cmd = f"ip address add {ipv4_address} brd {ipv4_address} dev {LXC_EXTERNAL_INTERFACE} " \
                  f"label {LXC_EXTERNAL_INTERFACE}:{inf_index}"
            if not vns_container.attach_cmd(cmd=cmd):
                raise CommandExecutionError("Failed to add interface alias.")
        ret['comment'] = "Alias configuration create successfully."
    return ret


@states_response
def remove_vns_interface_alias(vnservices_name: str, alias_id: str):
    """ Удаление конфигурации алиаса внешнего интерфейса для floating ip """

    ret = {'comment': '', 'changes': {}}

    if not VNService.exists(vnservices_name):
        raise CommandExecutionError("Virtual network services container does not exist.")

    vns_container = VNService(vnservices_name)
    vns_root = vns_container.get_config_item('lxc.rootfs.path')
    vns_interfaces_conf = os.path.join(vns_root, BASE_NET_CONF_DIR.lstrip('/'))
    if not os.path.exists(f"{vns_interfaces_conf}/{LXC_EXTERNAL_INTERFACE}-{alias_id}"):
        raise CommandExecutionError("Alias configuration not found.")
    with open(f"{vns_interfaces_conf}/{LXC_EXTERNAL_INTERFACE}-{alias_id}", "r") as conf:
        for line in conf.readlines():
            line = line.strip()
            if line.startswith("address"):
                ipv4_address = line.split().pop(1)
    remove_lxc_interface_files(vns_interfaces_conf, f"{LXC_EXTERNAL_INTERFACE}-{alias_id}")
    if vns_container.status == VNService.STATUS.running:
        cmd = f"ip address del {ipv4_address} dev {LXC_EXTERNAL_INTERFACE}"
        if not vns_container.attach_cmd(cmd=cmd):
            raise CommandExecutionError("Failed to remove interface alias.")
    ret['comment'] = "Alias configuration remove successfully."
    return ret


@states_response
def modify_vns_interface_alias(vnservices_name: str, alias_id: str, ipv4_address: str):
    """ Изменение параметров алиаса сетевого интерфейса контейнера для floating ip """

    ret = {'comment': '', 'changes': {}}

    if not VNService.exists(vnservices_name):
        raise CommandExecutionError("Virtual network services container does not exist.")

    vns_container = VNService(vnservices_name)
    vns_root = vns_container.get_config_item('lxc.rootfs.path')
    net_conf_dir = os.path.join(vns_root, BASE_NET_CONF_DIR.lstrip('/'))
    if not os.path.exists(f"{net_conf_dir}/{LXC_EXTERNAL_INTERFACE}-{alias_id}"):
        raise CommandExecutionError("Alias configuration not found.")
    with open(f"{net_conf_dir}/{LXC_EXTERNAL_INTERFACE}-{alias_id}", "r") as conf:
        for line in conf.readlines():
            line = line.strip()
            if line.startswith("iface"):
                inf_index = int(line.split().pop(1).split(':').pop(1))
            if line.startswith("address"):
                old_ipv4_address = line.split().pop(1)
    remove_lxc_interface_files(net_conf_dir, f"{LXC_EXTERNAL_INTERFACE}-{alias_id}")
    write_interface_alias(conf_dir=net_conf_dir, mode=InterfaceModes.static,
                          inf_name=f"{LXC_EXTERNAL_INTERFACE}:{inf_index}",
                          conf_name=f"{LXC_EXTERNAL_INTERFACE}-{alias_id}",
                          ipv4_address=ipv4_address, ipv4_netmask="255.255.255.255")
    if vns_container.status == VNService.STATUS.running:
        cmd = f"ip address del {old_ipv4_address} dev {LXC_EXTERNAL_INTERFACE}"
        if not vns_container.attach_cmd(cmd=cmd):
            raise CommandExecutionError("Failed to modify interface alias.")
        cmd = f"ip address add {ipv4_address} brd {ipv4_address} dev {LXC_EXTERNAL_INTERFACE} " \
              f"label {LXC_EXTERNAL_INTERFACE}:{inf_index}"
        if not vns_container.attach_cmd(cmd=cmd):
            raise CommandExecutionError("Failed to modify interface alias.")
    ret['comment'] = "Alias configuration modify successfully."
    return ret


@states_response
def get_vns_interface_alias(vnservices_name: str) -> dict:
    """ Возвращает список алиасов сетевого интерфейса контейнера для floating ip """

    if not VNService.exists(vnservices_name):
        raise CommandExecutionError("Virtual network services container does not exist.")

    vns_container = VNService(vnservices_name)
    vns_root = vns_container.get_config_item('lxc.rootfs.path')
    net_conf_dir = os.path.join(vns_root, BASE_NET_CONF_DIR.lstrip('/'))
    aliases = dict()
    for inf in os.listdir(net_conf_dir):
        if inf.startswith(f"{LXC_EXTERNAL_INTERFACE}-"):
            with open(f"{net_conf_dir}/{inf}", "r") as conf:
                for line in conf.readlines():
                    line = line.strip()
                    if line.startswith("address"):
                        ipv4_address = line.split().pop(1)
                        aliases.update({inf: ipv4_address})
    return aliases


@states_response
def vn_service_state(vns_name: str, service_name: str):
    """
    Получение состояния сетевой службы виртуальной сети
    :param vns_name: название контейнера сетевых служб
    :param service_name: название сетевой службы
    """

    c = VNService(vns_name)
    try:
        is_active_output = c.attach_cmd(f"systemctl is-active {service_name}", get_output=True)
        is_enabled_output = c.attach_cmd(f"systemctl is-enabled {service_name}", get_output=True)
        return {'active': True if is_active_output == "active\n" else False,
                'enabled': True if is_enabled_output == "enabled\n" else False}
    except Exception as err:
        raise CommandExecutionError("Unexpected error. Unable to get service state.") from err


@states_response
def get_vnservices_states():
    """
    Получение состояния всех сетевых служб виртуальных сетей
    """

    ret = dict()
    containers = [c for c in VNService.list_containers() if c != BASE_LXC_CONTAINER]
    for container in containers:
        c = VNService(container)
        ret[c.name] = dict()
        data = dict()
        data['state'] = True if c.status == VNService.STATUS.running else False
        data['services'] = dict()
        if not data['state']:
            data['services'].update({'dhcp': {'active': False, 'enabled': False},
                                     'firewall': {'active': False, 'enabled': False}})
        else:
            dhcp_state = states_result(vn_service_state, vns_name=c.name, service_name=DHCP_SERVICE)
            firewall_state = states_result(vn_service_state, vns_name=c.name, service_name=FIREWALL_SERVICE)
            data['services'].update({'dhcp': dhcp_state, 'firewall': firewall_state})
        ret[c.name].update(data)
    return ret


@states_response
def set_uplink(lswitch: str, vswitch: str, obverse_patch: str, reverse_patch: str, uplink_state: str):
    """
    Соединение логического коммутатора с аплинк-коммутатором
    :param lswitch: название логического коммутатора
    :param vswitch: название аплинк-коммутатора
    :param obverse_patch: название патча от логического коммутатора к аплинку
    :param reverse_patch: название обратного патча (от аплинка к логическому коммутатору)
    :param uplink_state: состояние аплинк-соедиенения (active/backup)
    """

    if uplink_state == UplinkState.backup:
        patch_params = ((lswitch, obverse_patch, reverse_patch), (vswitch, reverse_patch, obverse_patch))
        block_rule_set = TrafficBlockRuleSet.from_config_file()
        rule_to_add = TrafficBlockRule(switch=lswitch, port=obverse_patch, applied=True, to_apply=True)
        if not block_rule_set.add_rule(rule_to_add):
            raise CommandExecutionError(f"Failed to save blocking rule for '{lswitch}'.")
        rule_to_add = TrafficBlockRule(switch=vswitch, port=reverse_patch, applied=True, to_apply=True)
        if not block_rule_set.add_rule(rule_to_add, save=True):
            raise CommandExecutionError(f"Failed to save blocking rule for '{vswitch}'.")
        for param_set in patch_params:
            switch, obverse_patch, reverse_patch = param_set
            add_port_to_switch(switch, obverse_patch, OvsPortType.patch, peer=reverse_patch)
            cmd = f"ovs-ofctl add-flow {switch} table=0,in_port={obverse_patch},actions=drop"
            states_result(get_output, base_err_msg=f'Failed to add blocking rule for switch {switch}', cmd=cmd)
    else:
        add_port_to_switch(lswitch, obverse_patch, OvsPortType.patch, peer=reverse_patch)
        add_port_to_switch(vswitch, reverse_patch, OvsPortType.patch, peer=obverse_patch)

    return f"Logical switch {lswitch} connected to uplink switch {vswitch} successfully."


@states_response
def del_uplink(lswitch: str, vswitch: str, obverse_patch: str, reverse_patch: str, uplink_state: str):
    """
    Отсоединение логического коммутатора от аплинк-коммутатора
    :param lswitch: название логического коммутатора
    :param vswitch: название аплинк-коммутатора
    :param obverse_patch: название патча от логического коммутатора к аплинку
    :param reverse_patch: название обратного патча (от аплинка к логическому коммутатору)
    :param uplink_state: состояние аплинк-соедиенения (active/backup)
    """

    block_rule_set = TrafficBlockRuleSet.from_config_file()

    del_port_from_switch(obverse_patch, switch=lswitch)
    del_port_from_switch(reverse_patch, switch=vswitch)
    rule_to_delete = block_rule_set.find_rule(dict(switch=lswitch, port=obverse_patch))
    if rule_to_delete and not block_rule_set.remove_rule(rule_to_delete):
        raise CommandExecutionError(f"Failed to remove uplink for virtual network switch '{lswitch}'.")
    rule_to_delete = block_rule_set.find_rule(dict(switch=vswitch, port=reverse_patch))
    if rule_to_delete and not block_rule_set.remove_rule(rule_to_delete, save=True):
        raise CommandExecutionError(f"Failed to remove uplink for virtual network switch '{lswitch}'.")

    return f"Virtual network switch {lswitch} disconnected from uplink switch {vswitch} successfully."


@states_response
def set_uplink_state(lswitch: str, vswitch: str, obverse_patch: str, reverse_patch: str, uplink_state: str):
    """
    Изменение статуса соединения логического коммутатора с аплинк-коммутатором - активное/резервное
    :param lswitch: название логического коммутатора
    :param vswitch: название аплинк-коммутатора
    :param obverse_patch: название патча от логического коммутатора к аплинку
    :param reverse_patch: название обратного патча (от аплинка к логическому коммутатору)
    :param uplink_state: состояние аплинк-соедиенения (active/backup)
    """

    block_rule_set = TrafficBlockRuleSet.from_config_file()
    ok_msg = f"Uplink connection for {lswitch} through switch {vswitch} is set to {uplink_state} state successfully."

    if uplink_state == UplinkState.backup:
        commands = [f"ovs-ofctl add-flow {lswitch} table=0,in_port={obverse_patch},actions=drop",
                    f"ovs-ofctl add-flow {vswitch} table=0,in_port={reverse_patch},actions=drop"]
        for cmd in commands:
            states_result(get_output,
                          base_err_msg=f"Failed to set uplink state to 'backup' for logical switch {lswitch}.",
                          cmd=cmd)
        block_rule_set = TrafficBlockRuleSet.from_config_file()
        rule_to_add = TrafficBlockRule(switch=lswitch, port=obverse_patch, applied=True, to_apply=True)
        if not block_rule_set.add_rule(rule_to_add):
            raise CommandExecutionError(f"Failed to save blocking rule for '{lswitch}'.")
        rule_to_add = TrafficBlockRule(switch=vswitch, port=reverse_patch, applied=True, to_apply=True)
        if not block_rule_set.add_rule(rule_to_add, save=True):
            raise CommandExecutionError(f"Failed to save blocking rule for '{vswitch}'.")
    else:
        commands = [f"ovs-ofctl del-flows {lswitch} table=0,in_port={obverse_patch}",
                    f"ovs-ofctl del-flows {vswitch} table=0,in_port={reverse_patch}"]
        for cmd in commands:
            states_result(get_output,
                          base_err_msg=f"Failed to set uplink state to 'active' for logical switch {lswitch}.",
                          cmd=cmd)
        rule_to_delete = block_rule_set.find_rule(dict(switch=lswitch, port=obverse_patch))
        if not rule_to_delete:
            return ok_msg
        else:
            if not block_rule_set.remove_rule(rule_to_delete):
                raise CommandExecutionError(f"Failed to remove uplink for logical switch '{lswitch}'.")
        rule_to_delete = block_rule_set.find_rule(dict(switch=vswitch, port=reverse_patch))
        if not rule_to_delete:
            return ok_msg
        else:
            if not block_rule_set.remove_rule(rule_to_delete, save=True):
                raise CommandExecutionError(f"Failed to remove uplink for logical switch '{lswitch}'.")

    return ok_msg


@states_response
def vns_connect_uplink(vnservices_name: str, interface: str, switch: str, vlan_id: int):
    """
    Создание скриптов подключения внешнего интерфейса контейнера к указанному аплинк-коммутатору
    :param vnservices_name: название контейнера
    :param interface: название интерфейса
    :param switch: название коммутатора
    :param vlan_id:
    """

    ret = {'comment': '', 'changes': {}}

    if not VNService.exists(vnservices_name):
        raise CommandExecutionError("Virtual network services container does not exist.")

    vns = VNService(vnservices_name)
    vns_path = vns.path
    prev_state = VNService.get_status(vnservices_name)

    if prev_state == VNService.STATUS.running:
        if not vns.wait_stopped():
            raise CommandExecutionError("Failed to stop network services.")
        for net in vns.get_networks():
            del_port_from_switch(net.pair)

    remove_lxc_interface_files(vns_path, interface)
    iface_params = dict(name=interface, switch=switch, vlan_id=vlan_id)
    write_lxc_interface_scripts(vns_path, iface_params)
    cur_state = VNService.get_status(vnservices_name)
    if cur_state != prev_state and not vns.wait_started():
        raise CommandExecutionError("Failed to start network services.")

    ret['comment'] = "Virtual network services connected to uplink switch successfully."
    return ret


@states_response
def apply_vnservices_policy(ruleset: list, vnservices_params):
    """
    Запись правил брандмауэра в файл конфигурации и последующий перезапуск служб в контейнерах, если они запущены
    :param ruleset: правила брандмауэра
    :param vnservices_params: список словарей с именами контейнеров и их конфигурационных файлов
    """

    for vns in vnservices_params:
        write_firewall_config(ruleset, vns['config_file'])
        if VNService.exists(vns['name']) and VNService.get_status(vns['name']) == VNService.STATUS.running:
            firewall_state = states_result(vn_service_state, vns_name=vns['name'], service_name=FIREWALL_SERVICE)
            if firewall_state['active']:
                reload_firewall_service(vns['name'])
                time.sleep(1)
                state = states_result(vn_service_state, vns_name=vns['name'], service_name=FIREWALL_SERVICE)
                if not state['active']:
                    raise CommandExecutionError(f"Failed to start {FIREWALL_SERVICE}.")
    return 'Firewall configuration applied successfully.'


@states_response
def get_dhcp_config(config_path: str) -> dict:
    """
    Получение конфигурации DHCP-сервера
    """

    try:
        with open(config_path, 'r') as f:
            conf = json.load(f)
        return conf
    except OSError as e:
        raise CommandExecutionError(
            "Unable to read DHCP configuration. I/O error({0}): {1}".format(e.errno, e.strerror)) from None
    except Exception as e:
        raise CommandExecutionError("Unexpected error while reading DHCP conf: {0}".format(str(e))) from None


@states_response
def set_dhcp_config(vnservices_name: str, dhcp_config: dict):
    """
    Запись конфигурации службы dhcp
    :param vnservices_name: название контейнера
    :param dhcp_config: конфигурация службы
    """

    if not VNService.exists(vnservices_name):
        raise CommandExecutionError("Virtual network services container does not exist.")

    vns_container = VNService(vnservices_name)
    vns_root = vns_container.get_config_item('lxc.rootfs.path')
    dhcp_conf_file = os.path.join(vns_root, BASE_DHCP_CONF.lstrip('/'))
    existing_config = states_result(get_dhcp_config, config_path=dhcp_conf_file)
    result_config = update_dhcp_config(existing_config, dhcp_config)
    write_dhcp_config(result_config, dhcp_conf_file)

    if VNService.get_status(vnservices_name) == VNService.STATUS.running:
        state = states_result(vn_service_state, vns_name=vnservices_name, service_name=DHCP_SERVICE)
        if state['active']:
            if not vn_service_stop(vnservices_name, DHCP_SERVICE):
                raise CommandExecutionError(f"Failed to stop {DHCP_SERVICE}.")
            if not vn_service_start(vnservices_name, DHCP_SERVICE):
                raise CommandExecutionError(f"Failed to start {DHCP_SERVICE}.")
            time.sleep(1)
            state = states_result(vn_service_state, vns_name=vnservices_name, service_name=DHCP_SERVICE)
            if not state['active']:
                raise CommandExecutionError(f"Failed to start {DHCP_SERVICE}.")
    time.sleep(1)
    return "Configuration applied successfully."


@states_response
def modify_vnetwork_netflow(switch: str, netflow: dict):
    """
    Изменение параметров netflow
    :param switch: название коммутатора
    :param netflow: параметры netflow
    """
    ret = {'comment': '', 'changes': {}}

    if all(map(lambda x: x is None, list(netflow.values()))):
        flush_netflow(switch)
    else:
        set_netflow_params(switch, netflow)

    ret['comment'] = 'Netflow parameters changed successfully.'
    return ret


@states_response
def check_netflow_params(switch: str):
    """
    Получение параметров netflow коммутатора
    """

    result = dict(netflow_target_ip=None,
                  netflow_target_port=None,
                  netflow_active_timeout=None,
                  netflow_id_to_iface=None)
    netflow = get_netflow_id(switch)
    if netflow == '[]':
        return result

    raw_target = get_netflow_param(netflow, OvsNetflowRequestParams.targets).lstrip('["').rstrip('"]').split(':')
    result['netflow_target_ip'] = raw_target[0]
    result['netflow_target_port'] = int(raw_target[1])
    result['netflow_active_timeout'] = int(get_netflow_param(netflow, OvsNetflowRequestParams.active_timeout))
    result['netflow_id_to_iface'] = True if get_netflow_param(
        netflow, OvsNetflowRequestParams.id_to_iface) == 'true' else False
    return result


@states_response
def get_vnetwork_id(switch: str):
    """
    Получение параметра id виртуальной сети, к которой относится коммутатор
    """

    cmd = f"ovs-vsctl get br {switch} {OvsRequestParams.external_ids}"
    raw_output = states_result(get_output, raise_exception=False, cmd=cmd)
    params = ovs_param_to_dict(raw_output)
    vnetwork_id = params['virtual-network'] if 'virtual-network' in params else None
    return vnetwork_id


@states_response
def remove_vns_container(vnservices_name: str):
    """
    Удаление контейнера сетевых служб
    :param vnservices_name: название контейнера
    """

    ret = {'comment': '', 'changes': {}}

    if VNService.exists(vnservices_name):
        vns_container = VNService(vnservices_name)
        if vns_container.status == VNService.STATUS.running:
            stopped = vns_container.wait_stopped()
            if stopped is None or stopped is False:
                raise CommandExecutionError("Unable to stop network services.")
            vns_networks = vns_container.get_networks()
            for net in vns_networks:
                del_port_from_switch(net.pair)
        vns_container.destroy()

    ret['comment'] = 'Virtual network services container removed successfully.'
    return ret


@states_response
def flush_vns_containers():
    """
    Удаление всех существующих контейнеров сетевых служб
    """

    ret = {'comment': '', 'changes': {}}

    containers = [c for c in VNService.list_containers(as_objects=True) if c.name != BASE_LXC_CONTAINER]
    for container in containers:
        container.destroy()

    ret['comment'] = "Virtual network service containers removed successfully."
    return ret


@states_response
def set_vns_memory_limit(vnservices_name: str, memory_limit: int):
    """
    Установка лимита используемой памяти контейнера сетевых служб
    :param vnservices_name: название контейнера
    :param memory_limit: значение лимита в мегабайтах
    """

    ret = {'comment': '', 'changes': {}}

    if not VNService.exists(vnservices_name):
        raise CommandExecutionError("Virtual network services container does not exist.")

    vns_container = VNService(vnservices_name)
    vns_container.set_config_item("lxc.cgroup.memory.limit_in_bytes", f"{memory_limit}M")
    if not vns_container.save_config():
        raise CommandExecutionError("Unexpected error. Saving configuration failed.")
    if vns_container.status == VNService.STATUS.running:
        expected_dhcp_status = states_result(vn_service_state, vns_name=vnservices_name, service_name=DHCP_SERVICE)
        expected_firewall_status = states_result(vn_service_state,
                                                 vns_name=vnservices_name, service_name=FIREWALL_SERVICE)
        if not vns_container.wait_stopped():
            raise CommandExecutionError("Failed to stop network services.")
        vns_networks = vns_container.get_networks()
        for net in vns_networks:
            del_port_from_switch(net.pair)
        if not vns_container.wait_started():
            raise CommandExecutionError("Failed to start network services.")
        current_dhcp_status = states_result(vn_service_state, vns_name=vnservices_name, service_name=DHCP_SERVICE)
        current_firewall_status = states_result(vn_service_state,
                                                vns_name=vnservices_name, service_name=FIREWALL_SERVICE)
        if expected_dhcp_status['active'] and not current_dhcp_status['active']:
            if not vn_service_start(vnservices_name, DHCP_SERVICE):
                raise CommandExecutionError(f"Failed to start {FIREWALL_SERVICE}.")
        elif not expected_dhcp_status['active'] and current_dhcp_status['active']:
            if not vn_service_stop(vnservices_name, DHCP_SERVICE):
                raise CommandExecutionError(f"Failed to stop {DHCP_SERVICE}.")
        elif expected_firewall_status['active'] and not current_firewall_status['active']:
            if not vn_service_start(vnservices_name, FIREWALL_SERVICE):
                raise CommandExecutionError(f"Failed to start {FIREWALL_SERVICE}.")
        elif not expected_firewall_status['active'] and current_firewall_status['active']:
            if not vn_service_stop(vnservices_name, FIREWALL_SERVICE):
                raise CommandExecutionError(f"Failed to stop {FIREWALL_SERVICE}.")

    ret['comment'] = "Configuration changed successfully."
    return ret


@states_response
def vns_container_action(vnservices_name: str, action: str):
    """
    Запуск/перезапуск/останов контейнера сетевых служб
    :param vnservices_name: название контейнера
    :param action: выполняемое действие
    """

    ret = {'comment': '', 'changes': {}}

    if not VNService.exists(vnservices_name):
        raise CommandExecutionError("Virtual network services container does not exist.")
    vns_container = VNService(vnservices_name)
    vns_root = vns_container.get_config_item('lxc.rootfs.path')
    vns_path = vns_container.path
    vns_net_conf = os.path.join(vns_root, BASE_NET_CONF_DIR.lstrip('/'))

    if action == ServiceActions.start:
        if vns_container.status == VNService.STATUS.running:
            return ret
        set_lxc_links_up(vns_path)
        if not vns_container.wait_started():
            set_lxc_links_down(vns_path)
            raise CommandExecutionError("Unable to start network services.")
        ret['comment'] = 'Virtual network services container started successfully.'
    elif action == ServiceActions.stop:
        if vns_container.status == VNService.STATUS.stopped:
            return ret
        if not vns_container.wait_stopped():
            raise CommandExecutionError("Unable to stop network services.")
        set_lxc_links_down(vns_path)
        vns_networks = vns_container.get_networks()
        for net in vns_networks:
            del_port_from_switch(net.pair)
        ret['comment'] = 'Virtual network services container stopped successfully.'
    else:
        if vns_container.status == VNService.STATUS.stopped:
            raise CommandExecutionError('Virtual network services container is not running.')
        if not vns_container.wait_stopped():
            raise CommandExecutionError("Unable to stop network services.")
        vns_networks = vns_container.get_networks()
        for net in vns_networks:
            del_port_from_switch(net.pair)
        if not vns_container.wait_started():
            raise CommandExecutionError("Unable to start network services.")
        ret['comment'] = 'Virtual network services container restarted successfully.'

    if action in (ServiceActions.start, ServiceActions.restart):
        if check_iface_conf_exists(vns_net_conf, 'eth1'):
            if states_result(check_interface_mode, conf_dir=vns_net_conf, interface='eth1'):
                time.sleep(7)
                ext_iface_config = dict(dhcpv4=True,
                                        ipv4_address='',
                                        ipv4_netmask='',
                                        ipv4_gateway='')
                vns_interface_params = get_lxc_ifconfig(vns_container, 'eth1')
                if vns_interface_params:
                    ext_iface_config.update(**parse_lxc_address(vns_interface_params[1]))
                ext_iface_config['ipv4_gateway'] = get_lxc_default_gateway(vns_container)
                ret['changes'] = ext_iface_config
        time.sleep(1)
        state = states_result(vn_service_state, vns_name=vnservices_name, service_name=DHCP_SERVICE)
        if state['enabled'] and not state['active']:
            raise CommandExecutionError(f"Failed to start {DHCP_SERVICE}.")
        state = states_result(vn_service_state, vns_name=vnservices_name, service_name=FIREWALL_SERVICE)
        if state['enabled'] and not state['active']:
            raise CommandExecutionError(f"Failed to start {FIREWALL_SERVICE}.")
    return ret


@states_response
def dhcp_service_action(vnservices_name: str, action: str):
    """
    Запуск/перезапуск/останов службы dhcp
    :param vnservices_name: название контейнера
    :type vnservices_name: str
    :param action: выполняемое действие
    :type action: str
    :return
    """

    ret = {'comment': '', 'changes': {}}
    dhcp_running = False

    if not VNService.exists(vnservices_name):
        raise CommandExecutionError("Virtual network services container does not exist.")

    vns_container = VNService(vnservices_name)
    vns_path = vns_container.path

    if not VNService.get_status(vnservices_name) == VNService.STATUS.running:
        firewall_was_running = False
        if action == ServiceActions.restart:
            raise CommandExecutionError("Unable to restart service while service container is not running. "
                                        "Use 'START' function instead.")
        if action == ServiceActions.stop:
            raise CommandExecutionError("Unable to stop service while service container is not running.")

        set_lxc_links_up(vns_path)
        if not vns_container.wait_started():
            set_lxc_links_down(vns_path)
            raise CommandExecutionError("Failed to start network services.")
    else:
        _firewall_state = states_result(vn_service_state, vns_name=vnservices_name, service_name=FIREWALL_SERVICE)
        firewall_was_running = _firewall_state['active']

    dhcp_state = states_result(vn_service_state, vns_name=vnservices_name, service_name=DHCP_SERVICE)
    if action == ServiceActions.start and not dhcp_state['active']:
        if not vn_service_start(vnservices_name, DHCP_SERVICE):
            raise CommandExecutionError(f"Failed to start {DHCP_SERVICE}.")
        time.sleep(1)
        state = states_result(vn_service_state, vns_name=vnservices_name, service_name=DHCP_SERVICE)
        if not state['active']:
            raise CommandExecutionError(f"Failed to start {DHCP_SERVICE}.")
        ret['comment'] = 'DHCP service started successfully.'
        dhcp_running = True
    elif action == ServiceActions.start and dhcp_state['active']:
        ret['comment'] = 'DHCP service is already running.'
    elif action == ServiceActions.stop and dhcp_state['active']:
        if not vn_service_stop(vnservices_name, DHCP_SERVICE):
            raise CommandExecutionError(f"Failed to stop {DHCP_SERVICE}.")
        time.sleep(1)
        state = states_result(vn_service_state, vns_name=vnservices_name, service_name=DHCP_SERVICE)
        if state['active']:
            raise CommandExecutionError(f"Failed to stop {DHCP_SERVICE}.")
        ret['comment'] = 'DHCP service stopped successfully.'
        if not firewall_was_running:
            vns_container.wait_stopped()
            set_lxc_links_down(vns_path)
            vns_networks = vns_container.get_networks()
            for net in vns_networks:
                del_port_from_switch(net.pair)
    elif action == ServiceActions.stop and not dhcp_state['active']:
        raise CommandExecutionError("DHCP service is not running.")
    elif action == ServiceActions.restart and dhcp_state['active']:
        if not vn_service_stop(vnservices_name, DHCP_SERVICE):
            raise CommandExecutionError(f"Failed to stop {DHCP_SERVICE}.")
        time.sleep(1)
        state = states_result(vn_service_state, vns_name=vnservices_name, service_name=DHCP_SERVICE)
        if state['active']:
            raise CommandExecutionError(f"Failed to stop {DHCP_SERVICE}.")
        if not vn_service_start(vnservices_name, DHCP_SERVICE):
            raise CommandExecutionError(f"Failed to start {DHCP_SERVICE}.")
        time.sleep(1)
        state = states_result(vn_service_state, vns_name=vnservices_name, service_name=DHCP_SERVICE)
        if not state['active']:
            raise CommandExecutionError(f"Failed to start {DHCP_SERVICE}.")
        ret['comment'] = 'DHCP service restarted successfully.'
        dhcp_running = True
    elif action == ServiceActions.restart and not dhcp_state['active']:
        raise CommandExecutionError("Unable to restart service while it is not running. Use 'START' function instead.")

    firewall_state = states_result(vn_service_state, vns_name=vnservices_name, service_name=FIREWALL_SERVICE)
    ret['changes'] = dict(dhcp_running=dhcp_running, firewall_running=firewall_state['active'])
    return ret


@states_response
def firewall_service_action(vnservices_name: str, action: str):
    """
    Запуск/перезапуск/останов службы брандмауэра
    :param vnservices_name: название контейнера
    :param action: выполняемое действие
    """

    ret = {'comment': '', 'changes': {}}
    firewall_running = False

    if not VNService.exists(vnservices_name):
        raise CommandExecutionError("Virtual network services container does not exist.")

    vns_container = VNService(vnservices_name)
    vns_path = vns_container.path

    if not VNService.get_status(vnservices_name) == VNService.STATUS.running:
        dhcp_was_running = False
        if action == ServiceActions.restart:
            raise CommandExecutionError("Unable to restart service while service container is not running. "
                                        "Use 'START' function instead.")
        if action == ServiceActions.stop:
            raise CommandExecutionError("Unable to stop service while service container is not running.")

        set_lxc_links_up(vns_path)
        if not vns_container.wait_started():
            set_lxc_links_down(vns_path)
            raise CommandExecutionError("Failed to start network services.")
    else:
        _dhcp_state = states_result(vn_service_state, vns_name=vnservices_name, service_name=DHCP_SERVICE)
        dhcp_was_running = _dhcp_state['active']

    firewall_state = states_result(vn_service_state, vns_name=vnservices_name, service_name=FIREWALL_SERVICE)
    if action == ServiceActions.start and not firewall_state['active']:
        if not vn_service_start(vnservices_name, FIREWALL_SERVICE):
            raise CommandExecutionError(f"Failed to start {FIREWALL_SERVICE}.")
        time.sleep(1)
        state = states_result(vn_service_state, vns_name=vnservices_name, service_name=FIREWALL_SERVICE)
        if not state['active']:
            raise CommandExecutionError(f"Failed to start {FIREWALL_SERVICE}.")
        ret['comment'] = 'Firewall service started successfully.'
        firewall_running = True
    elif action == ServiceActions.start and firewall_state['active']:
        ret['comment'] = "Firewall service is already running."
    elif action == ServiceActions.stop and firewall_state['active']:
        if not vn_service_stop(vnservices_name, FIREWALL_SERVICE):
            raise CommandExecutionError(f"Failed to stop {FIREWALL_SERVICE}.")
        time.sleep(1)
        state = states_result(vn_service_state, vns_name=vnservices_name, service_name=FIREWALL_SERVICE)
        if state['active']:
            raise CommandExecutionError(f"Failed to stop {FIREWALL_SERVICE}.")
        ret['comment'] = 'Firewall service stopped successfully.'
        if not dhcp_was_running:
            vns_container.wait_stopped()
            set_lxc_links_down(vns_path)
            vns_networks = vns_container.get_networks()
            for net in vns_networks:
                del_port_from_switch(net.pair)
    elif action == ServiceActions.stop and not firewall_state['active']:
        raise CommandExecutionError("Firewall service is not running.")
    elif action == ServiceActions.restart and firewall_state['active']:
        if not vn_service_stop(vnservices_name, FIREWALL_SERVICE):
            raise CommandExecutionError(f"Failed to stop {FIREWALL_SERVICE}.")
        time.sleep(1)
        state = states_result(vn_service_state, vns_name=vnservices_name, service_name=FIREWALL_SERVICE)
        if state['active']:
            raise CommandExecutionError(f"Failed to stop {FIREWALL_SERVICE}.")
        if not vn_service_start(vnservices_name, FIREWALL_SERVICE):
            raise CommandExecutionError(f"Failed to start {FIREWALL_SERVICE}.")
        time.sleep(1)
        state = states_result(vn_service_state, vns_name=vnservices_name, service_name=FIREWALL_SERVICE)
        if not state['active']:
            raise CommandExecutionError(f"Failed to start {FIREWALL_SERVICE}.")
        ret['comment'] = 'Firewall service restarted successfully.'
        firewall_running = True
    elif action == ServiceActions.restart and not firewall_state['active']:
        raise CommandExecutionError("Unable to restart service while it is not running. Use 'START' function instead.")

    dhcp_state = states_result(vn_service_state, vns_name=vnservices_name, service_name=DHCP_SERVICE)
    ret['changes'] = dict(dhcp_running=dhcp_state['active'], firewall_running=firewall_running)
    return ret


@states_response
def set_dhcp_autostart(vnservices_name: str, autostart: bool):
    """
    Включение/отключение автозапуска службы dhcp
    :param vnservices_name: название контейнера
    :param autostart: флаг автозапуска
    """

    command = 'enable' if autostart else 'disable'

    if not VNService.exists(vnservices_name):
        raise CommandExecutionError("Virtual network services container does not exist.")

    vns = VNService(vnservices_name)
    if not VNService.get_status(vnservices_name) == VNService.STATUS.running:
        if not vns.wait_started():
            raise CommandExecutionError("Failed to start network services.")
        if not vn_service_change_enabled(vnservices_name, DHCP_SERVICE, command):
            if not vns.wait_stopped():
                pass
            vns_networks = vns.get_networks()
            for net in vns_networks:
                del_port_from_switch(net.pair)
            raise CommandExecutionError(f"Failed to {command} {DHCP_SERVICE}.")
        if not vns.wait_stopped():
            raise CommandExecutionError("Failed to stop network services.")
        vns_networks = vns.get_networks()
        for net in vns_networks:
            del_port_from_switch(net.pair)
    else:
        if not vn_service_change_enabled(vnservices_name, DHCP_SERVICE, command):
            raise CommandExecutionError(f"Failed to {command} {DHCP_SERVICE}.")

    return f"DHCP-service {command}d successfully."


@states_response
def set_firewall_autostart(vnservices_name: str, autostart: bool):
    """
    Включение/отключение автозапуска службы брандмауэра
    :param vnservices_name: название контейнера
    :param autostart: флаг автозапуска
    """

    command = 'enable' if autostart else 'disable'

    if not VNService.exists(vnservices_name):
        raise CommandExecutionError("Virtual network services container does not exist.")

    vns = VNService(vnservices_name)
    if not VNService.get_status(vnservices_name) == VNService.STATUS.running:
        if not vns.wait_started():
            raise CommandExecutionError("Failed to start network services.")
        if not vn_service_change_enabled(vnservices_name, FIREWALL_SERVICE, command):
            if not vns.wait_stopped():
                pass
            vns_networks = vns.get_networks()
            for net in vns_networks:
                del_port_from_switch(net.pair)
            raise CommandExecutionError(f"Failed to {command} {FIREWALL_SERVICE}.")
        if not vns.wait_stopped():
            raise CommandExecutionError("Failed to stop network services.")
        vns_networks = vns.get_networks()
        for net in vns_networks:
            del_port_from_switch(net.pair)
    else:
        if not vn_service_change_enabled(vnservices_name, FIREWALL_SERVICE, command):
            raise CommandExecutionError(f"Failed to {command} {FIREWALL_SERVICE}.")

    return f"Firewall service {command}d successfully."


@states_response
def service_repair_dhcp(vnservices_name: str) -> str:
    """
    Восстановление работы dhcp сервиса в контейнере lxc
    :param vnservices_name: название контейнера lxc
    """

    vns_container = VNService(vnservices_name)
    if not VNService.exists(vnservices_name):
        raise CommandExecutionError("Virtual network services container does not exist.")

    if not VNService.get_status(vnservices_name) == VNService.STATUS.running:
        raise CommandExecutionError("Virtual network services container is not running.")

    dhcp_state = states_result(vn_service_state, vns_name=vnservices_name, service_name=DHCP_SERVICE)
    if not dhcp_state['active']:
        vns_container.repair_dhcp()
        vn_service_start(vnservices_name, DHCP_SERVICE)
        dhcp_state = states_result(vn_service_state, vns_name=vnservices_name, service_name=DHCP_SERVICE)
        if not dhcp_state['active']:
            return "Restoring of dhcp service for virtual network failed."
        return "Restoring of dhcp service for virtual network successfully."
    else:
        return "Restoring of dhcp service for virtual network does not need."


@states_response
def set_firewall_logging_service(vnservices_name: str, firewall_logging: bool) -> str:
    """
    Запуск/остановка службы логирования брандмауэра в контейнере
    :param vnservices_name: название контейнера
    :param firewall_logging: статус логирования брандмауэра
    """

    if not VNService.exists(vnservices_name):
        raise CommandExecutionError("Virtual network services container does not exist.")
    if firewall_logging:
        if not vn_service_start(vnservices_name, FIREWALL_LOGGING_SERVICE):
            raise CommandExecutionError(f"Failed to start {FIREWALL_LOGGING_SERVICE}.")
        return "Firewall logging for service started successfully."

    else:
        if not vn_service_stop(vnservices_name, FIREWALL_LOGGING_SERVICE):
            raise CommandExecutionError(f"Failed to start {FIREWALL_LOGGING_SERVICE}.")
        return "Firewall logging for service stopped successfully."
# endregion


# region OVS Flow rules (ACL)
def is_valid_bridge(bridge: str) -> bool:
    """Проверяет существование OVS свитча и что он не является системным"""
    if bridges := get_ovs_bridges():
        return bridge in bridges and bridge not in (DEFAULT_SWITCH, BLACKHOLE_SWITCH)
    return False


def ovs_del_flows(bridge: str) -> bool:
    """Удаляет все правила OVS Flow на коммутаторе."""
    if not is_valid_bridge(bridge):
        raise CommandExecutionError("Bridge is system or does not exists.")
    cmd = f"ovs-ofctl del-flows {bridge}"
    result = get_output(cmd, suppress_err=True).get('result')
    return result


def ovs_del_flow(bridge: str, flow: str) -> bool:
    """Удаляет правило OVS Flow на коммутаторе."""
    if not is_valid_bridge(bridge):
        raise CommandExecutionError("Bridge is system or does not exists.")
    cmd = f"ovs-ofctl --strict del-flows {bridge} '{flow.split('actions').pop(0)}'"
    result = get_output(cmd, suppress_err=True).get('result')
    return result


def ovs_add_flow(bridge: str, rule: str) -> bool:
    """Добавляет правило OVS Flow на коммутатор."""
    if not is_valid_bridge(bridge):
        raise CommandExecutionError("Bridge is system or does not exists.")
    cmd = f"ovs-ofctl add-flow {bridge} '{rule}'"
    result = get_output(cmd, suppress_err=True).get('result')
    return result


def ovs_add_flows(bridge: str, rules: list) -> bool:
    """Добавление правил OVS Flow на коммутатор."""
    result = all(map(ovs_add_flow, [bridge] * len(rules), rules))
    return result


def ovs_get_flows(bridge: str, table: int = None) -> list:
    """Получает список правил OVS Flow на коммутаторе."""
    cmd = f"ovs-ofctl --names --no-stats dump-flows {bridge}"
    if table is not None:
        cmd = f"{cmd} table={table}"
    response = get_output(cmd, suppress_err=True)
    if not response['result']:
        raise CommandExecutionError(f"Failed to get flows from {bridge}.")
    return list(map(str.strip, response['comment'].replace('=', '>').strip().split('\n')))


@states_response
def set_ovs_flow_default(bridge: str):
    """Возврат OVS Flow коммутатора в исходное состояние."""
    if ovs_del_flows(bridge):
        if not ovs_add_flow(bridge, OVS_FLOW_PRESETS['DEFAULT']):
            raise CommandExecutionError(f"Failed to set default flows at {bridge}.")
    else:
        raise CommandExecutionError(f"Failed to clean flows at {bridge}.")


@states_response
def ovs_flow_acl_mode(bridge: str, status: bool) -> str:
    """Установка правил OVS Flow для возможности задания ACL или возврат к состоянию по-умолчанию."""
    if ovs_del_flows(bridge):
        if status:
            if not ovs_add_flows(bridge, OVS_FLOW_PRESETS['ACL_MODE']):
                raise CommandExecutionError(f"Failed to set flows for ACL at {bridge}.")
            return f"Successfully enable ACL mode at {bridge}."
        else:
            if not ovs_add_flow(bridge, OVS_FLOW_PRESETS['DEFAULT']):
                raise CommandExecutionError(f"Failed to set default flows at {bridge}.")
            return f"Successfully disable ACL mode at {bridge}."
    else:
        raise CommandExecutionError(f"Failed to clean flows at {bridge}.")


@states_response
def ovs_flow_mac_learning(bridge: str, status: bool):
    """Установка правила MAC Learning."""
    if status:
        if not ovs_add_flow(bridge, OVS_FLOW_PRESETS['MAC_LEARNING']):
            raise CommandExecutionError(f"Failed to set MAC learning flow at {bridge}.")
    else:
        if not ovs_del_flow(bridge, OVS_FLOW_PRESETS['MAC_LEARNING']):
            raise CommandExecutionError(f"Failed to remove MAC learning flow at {bridge}.")


@states_response
def ovs_flow_port_security(bridge: str, mac: str, port: str):
    """Статическая привязка MAC-адреса к порту коммутатора."""
    rules = OVS_FLOW_PRESETS['PORT_SECURITY'].copy()
    rules = list(map(lambda rule: rule.format(port=port, mac=mac), rules))
    if not ovs_add_flows(bridge, rules):
        raise CommandExecutionError(f"Failed to set static MAC-address to port at {bridge}.")


@states_response
def ovs_flow_arp(bridge: str, status: bool):
    """Разрешить/запретить ARP - запросы."""
    if status:
        if not ovs_add_flow(bridge, OVS_FLOW_PRESETS['ARP_ALLOW']):
            raise CommandExecutionError(f"Failed to set flow to allow arp at {bridge}")
    else:
        if not ovs_del_flow(bridge, OVS_FLOW_PRESETS['ARP_ALLOW']):
            raise CommandExecutionError(f"Failed to remove flow to allow arp at {bridge}")


@states_response
def ovs_flow_stp_drop(bridge: str, status: bool):
    """Добавит/убрать STP drop правило."""
    if status:
        if not ovs_add_flow(bridge, OVS_FLOW_PRESETS['STP_DROP']):
            raise CommandExecutionError(f"Failed to set flow to STP drop at {bridge}")
    else:
        if not ovs_del_flow(bridge, OVS_FLOW_PRESETS['STP_DROP']):
            raise CommandExecutionError(f"Failed to remove flow to STP drop at {bridge}")


@states_response
def ovs_flow_broadcast_src_drop(bridge: str, status: bool):
    """Добавит/убрать broadcast src drop правило."""
    if status:
        if not ovs_add_flow(bridge, OVS_FLOW_PRESETS['BROADCAST_SRC_DROP']):
            raise CommandExecutionError(f"Failed to set flow to broadcast src drop at {bridge}")
    else:
        if not ovs_del_flow(bridge, OVS_FLOW_PRESETS['BROADCAST_SRC_DROP']):
            raise CommandExecutionError(f"Failed to remove flow to broadcast src drop at {bridge}")


def ovs_rule_parser(rule: str) -> dict:
    """Парсит OVS Flow правило в словарь"""
    match, actions = rule.split('actions')
    params = dict(action=Action.drop if Action.drop in actions else Action.accept)
    for param in match.split(','):
        if '>' not in param:
            params['protocol'] = param.strip()
        else:
            k, v = param.strip().split('>')
            params[k] = v.strip()
    if 'table' not in params.keys():
        params['table'] = 0
    return params


def get_ovs_acl_indexes(bridge: str) -> list:
    """Поиск индекса (priority) для вставки OVS Flow правила"""
    rules = ovs_get_flows(bridge=bridge, table=OvsTable.acl)
    parsed_rules = map(ovs_rule_parser, rules)
    return [int(p['priority']) for p in filter(lambda p: int(p['priority']), parsed_rules)]


def ovs_flow_rule_to_str(bridge: str, rule: dict) -> str:
    """Формирует из словаря OVS Flow правило"""
    acl_params = dict(protocol=set())
    indexes = get_ovs_acl_indexes(bridge)
    if (rule_index := rule.get('rule_index')) is not None:
        rule_index = max(OvsPriority.user_acl) - rule_index
        acl_params['priority'] = rule_index
    else:
        acl_params['priority'] = min(indexes) - 1 if indexes else max(OvsPriority.user_acl)
    if src_address := rule.get('src_address'):
        address = src_address.pop()
        if re.match(MAC_ADDRESS_RE, address.lower()):
            acl_params['dl_src'] = address
        elif check_ip(address):
            acl_params['protocol'].add('ip')
            acl_params['nw_src'] = address
        else:
            raise CommandExecutionError(f"Invalid parameter src_address = {src_address}")
    if dst_address := rule.get('dst_address'):
        address = dst_address.pop()
        if re.match(MAC_ADDRESS_RE, address.lower()):
            acl_params['dl_dst'] = address
        elif check_ip(address):
            acl_params['protocol'].add('ip')
            acl_params['nw_dst'] = address
        else:
            raise CommandExecutionError(f"Invalid parameter dst_address = {dst_address}")
    if protocol := rule.get('protocol'):
        if protocol != 'none':
            acl_params['protocol'].add(protocol)
        if protocol == 'icmp':
            if icmp_type := rule.get('icmp_type'):
                acl_params['icmp_type'] = int(icmp_type.pop())
            if icmp_code := rule.get('icmp_code'):
                acl_params['icmp_code'] = int(icmp_code.pop())
    if in_interface := rule.get('in_interface'):
        bridges = get_ovs_bridge_ports(bridge)
        if in_interface not in bridges:
            raise CommandExecutionError(f"Invalid interface {in_interface}")
        acl_params['in_port'] = in_interface
    if src_port := rule.get('src_port'):
        try:
            port = int(src_port.pop())
        except ValueError:
            raise CommandExecutionError("Invalid type of src_port.") from None
        else:
            acl_params['tp_src'] = port
    if dst_port := rule.get('dst_port'):
        try:
            port = int(dst_port.pop())
        except ValueError:
            raise CommandExecutionError("Invalid type of dst_port.") from None
        else:
            acl_params['tp_dst'] = port
    acl = list()
    for key, value in acl_params.items():
        if value is not None:
            if key == 'protocol':
                acl.append(','.join(value))
            else:
                acl.append(f"{key}={value}")
    acl = ','.join(acl)
    if out_interface := rule.get('out_interface'):
        acl = f"{acl} actions=output:{out_interface}"
    else:
        if action := rule.get('action'):
            if action == Action.drop:
                acl = f"{acl} actions=drop"
            else:
                acl = f"{acl} actions=resubmit(,{OvsTable.mac_src})"
        else:
            acl = f"{acl} actions=drop"
    return acl


@states_response
def ovs_flow_add_acl(bridge: str, rule: dict) -> str:
    """Добавляет ACL правило на коммутаторе"""
    acl = ovs_flow_rule_to_str(bridge, rule)
    if not ovs_add_flow(bridge, acl):
        raise CommandExecutionError(f"Failed to add acl flow at {bridge}")
    return f"OVS ACL at vswitch {bridge} added successfully."


@states_response
def ovs_flow_remove_acl(bridge: str, rule: dict) -> str:
    """Удаляет ACL правило на коммутаторе"""
    acl = ovs_flow_rule_to_str(bridge, rule)
    if not ovs_del_flow(bridge, acl):
        raise CommandExecutionError(f"Failed to add acl flow at {bridge}")
    return f"OVS ACL at vswitch {bridge} removed successfully."


@states_response
def ovs_flow_get_acl(bridge: str, table: int = OvsTable.acl) -> list:
    """ Возвращает список правил в формате для VMachinesAccessRule """

    key_remap = dict(priority='rule_index',
                     nw_src='src_address', nw_dst='dst_address',
                     dl_src='src_address', dl_dst='dst_address',
                     tp_src='src_port', tp_dst='dst_port',
                     icmp_type='icmp_type', icmp_code='icmp_code')
    rules = list(map(ovs_rule_parser, ovs_get_flows(bridge=bridge, table=table)))
    result = list()
    for rule in rules:
        tmp = dict()
        if 'priority' in rule:
            rule['priority'] = int(rule['priority'])
            if min(OvsPriority.system_acl) <= rule['priority'] <= max(OvsPriority.system_acl):
                rule['priority'] = max(OvsPriority.system_acl) - rule['priority'] + max(OvsPriority.user_acl)
            elif OvsPriority.default <= rule['priority'] < min(OvsPriority.user_acl):
                rule['priority'] = max(OvsPriority.system_acl) - rule['priority']
            else:
                rule['priority'] = max(OvsPriority.user_acl) - rule['priority']
        for key, value in rule.items():
            if key in key_remap:
                if key != 'priority':
                    tmp[key_remap[key]] = [value]
                else:
                    tmp[key_remap[key]] = value
            else:
                tmp[key] = value
        result.append(tmp)

    return result

# endregion


# region auxiliary and service states & functions
def ovs_param_to_clean_string(value: str) -> str:
    """
    Очистка строки от лишних кавычек, символов переноса строки и т.п.
    """

    def remove_escape_sec(string: str) -> (bool, str):
        mod = False
        if string.endswith('\r'):
            string = string.strip('\r')
            mod = True
            return mod, string
        if string.endswith('\n'):
            string = string.strip('\n')
            mod = True
            return mod, string
        return mod, string

    modified = True
    while modified:
        modified, value = remove_escape_sec(value)

    result = value.lstrip('"').rstrip('"')
    return result


def ovs_param_to_dict(value: str) -> dict:
    """
    Преобразование строки вида {key=value, key=value} в словарь
    """

    value = ovs_param_to_clean_string(value)
    if value == '{}':
        result = literal_eval(value)
    else:
        key_pattern = re.compile(r'(?P<key>[a-z-_]*)(=)')
        for match in key_pattern.finditer(value):
            value = value.replace(match.group("key"), f'"{match.group("key")}"')
        value = value.replace('=', ':')
        result = literal_eval(value)
    return result


def write_interfaces_config(config_file: str) -> None:
    """
    Запись базовой конфигурации
    :param config_file: путь к файлу конфигурации
    """

    try:
        with open(config_file, 'w') as c:
            c.write(f'source {BASE_NET_DIR}/interfaces.d/*\n')
    except OSError as e:
        raise CommandExecutionError(
            "Failed to write configuration. I/O error({0}): {1}".format(e.errno, e.strerror)) from None
    except Exception as e:
        raise CommandExecutionError("Unexpected error: {0}".format(str(e))) from None


def write_interface(conf_dir: str, name: str, mode: str, ipv4_address: str = None, ipv4_netmask: str = None,
                    ipv4_gateway: str = None, hw_address: str = None, is_connected: bool = True,
                    promiscuous_mode: bool = False, mtu: int = None) -> None:
    """
    Запись параметров интерфейса в конфигурационный файл
    :param conf_dir: директория, содержащая конфигурационные файлы сетевых интерфейсов
    :param name: название интерфейса
    :param mode: режим работы интерфейса
    :param ipv4_address: ip-адрес
    :param ipv4_netmask: маска подсети
    :param ipv4_gateway: основной шлюз
    :param hw_address: физический адрес
    :param is_connected: флаг, указывающий, подключен ли интерфейс
    :param promiscuous_mode: флаг, указывающий, включен ли неразборчивый режим
    :param mtu: величина параметра mtu
    """

    try:
        with open(os.path.join(conf_dir, name), 'w') as c:
            if mode in (InterfaceModes.loopback, InterfaceModes.static, InterfaceModes.dhcp) or \
                    (mode == InterfaceModes.manual and is_connected):
                c.write(f"auto {name}\n")
            if mode in (InterfaceModes.static, InterfaceModes.dhcp):
                c.write(f"allow-hotplug {name}\n")
            c.write(f"iface {name} inet {mode}\n")
            if hw_address:
                c.write(f"{' ':4}hwaddress ether {hw_address}\n")
            if mode == InterfaceModes.static:
                c.write(f"{' ':4}address {ipv4_address}\n")
                c.write(f"{' ':4}netmask {ipv4_netmask}\n")
            if ipv4_gateway:
                if mode == InterfaceModes.static:
                    c.write(f"{' ':4}gateway {ipv4_gateway}\n")
                if mode == InterfaceModes.dhcp:
                    c.write(f"{' ':4}post-up ip route add default via {ipv4_gateway} dev {name}\n")
            if mode == InterfaceModes.manual:
                if promiscuous_mode:
                    c.write(f"{' ':4}post-up ifconfig {name} promisc\n")
                if mtu is not None:
                    c.write(f"{' ':4}post-up ip link set dev {name} mtu {mtu}\n")
    except OSError as e:
        raise CommandExecutionError(
            "Failed to write configuration. I/O error({0}): {1}".format(e.errno, e.strerror)) from None
    except Exception as e:
        raise CommandExecutionError("Unexpected error: {0}".format(str(e))) from None


def write_interface_alias(conf_dir: str, mode: str, inf_name: str, conf_name: str,
                          ipv4_address: str = None, ipv4_netmask: str = None) -> None:
    """ Запись параметров алиаса в конфигурационный файл. """

    try:
        with open(os.path.join(conf_dir, conf_name), 'w') as c:
            c.write(f"auto {inf_name}\n")
            c.write(f"iface {inf_name} inet {mode}\n")
            if mode == InterfaceModes.static:
                c.write(f"{' ':4}address {ipv4_address}\n")
                c.write(f"{' ':4}netmask {ipv4_netmask}\n")
    except OSError as e:
        raise CommandExecutionError(
            "Failed to write configuration. I/O error({0}): {1}".format(e.errno, e.strerror)) from None
    except Exception as e:
        raise CommandExecutionError("Unexpected error: {0}".format(str(e))) from None


def split_net_config(raw_content: List[str]) -> Tuple[List[dict], bool]:
    """
    Разбиение содержимого файла конфигурации сетевых интерфейсов /etc/network/interfaces на конфигурации отдельных
    интерфейсов
    :param raw_content: содержимое файла /etc/network/interfaces в виде списка строк
    :return: список словарей, содержащих имя интерфейса и его конфигурация; флаг, требуется ли дальнейшая трансформация
    конфигурации
    """

    if raw_content == [f'source {BASE_NET_DIR}/interfaces.d/*']:
        return [], False

    ifaces_configs = []
    transformation_needed = True
    raw_config = (s for s in raw_content if not s.startswith('#') and not s.startswith('source') and not s.isspace()
                  and bool(s.strip()) is True)
    iface_conf = []
    iface_name = ''
    for s in raw_config:
        if s.startswith('auto'):
            if iface_conf:
                ifaces_configs.append(dict(name=iface_name, config=iface_conf))
                iface_conf = []
            iface_name = s.split()[1]
        iface_conf.append(s)
    if iface_conf:
        ifaces_configs.append(dict(name=iface_name, config=iface_conf))

    return ifaces_configs, transformation_needed


def get_iface_config_as_list(config: str) -> list:
    """
    Получение конфигурации интерфейса
    :param config: путь к файлу настроек сетевого интерфейса
    :return: список строк, очищенный от комментариев, пустых строк и т.п., содержащих параметры конфигурации интерфейса
    """

    with open(config, 'r') as f:
        raw_config = f.read().splitlines()
    return [s.lstrip() for s in raw_config if not s.startswith('#') and not s.isspace() and bool(s) is True]


def get_iface_config_as_dict(config: list) -> dict:
    """
    Преобразовывание конфигурации интерфейса в словарь
    :param config: список строк, содержащих параметры конфигурации сетевого интерфейса
    :return: словарь, содержащий параметры конфигурации интерфейса
    """

    result = dict(name='', mode='')
    for s in config:
        if s.startswith('auto'):
            result.update(name=s.split()[1])
        if s.startswith('iface'):
            result.update(mode=s.split()[3])
        if s.startswith('address'):
            result.update(ipv4_address=s.split()[1])
        if s.startswith('netmask'):
            result.update(ipv4_netmask=s.split()[1])
        if s.startswith('gateway'):
            result.update(ipv4_gateway=s.split()[1])
        if s.startswith('hwaddress'):
            result.update(mac_address=s.split()[2])
    return result


def check_iface_conf_exists(conf_dir: str, interface: str) -> bool:
    """
    Проверка существования файла конфигурации сетевого интерфейса
    :param conf_dir: путь к директории, хранящей файлы конфигурации
    :param interface: название интерфейса
    :return: булевое значение, определяющее, была ли проверка успешной
    """

    conf_file = os.path.join(conf_dir, interface)
    return True if os.path.exists(conf_file) and os.path.isfile(conf_file) else False


def del_file_by_pattern(search_path: str, pattern: str) -> None:
    """
    Удаление файла по указанному пути, если имя файла соответствует переданному паттерну
    """

    for f in os.listdir(search_path):
        if re.search(pattern, f):
            os.remove(os.path.join(search_path, f))
            break


def get_vmachines_interfaces(as_root=True) -> list:
    """
    Получение списка интерфейсов виртуальных машин, подключенных к OVS-коммутаторам
    """

    return [port for bridge in get_ovs_bridges(as_root) for port in get_ovs_bridge_ports(bridge, as_root) if port.startswith('vm_')]


def get_all_interfaces(as_root=True) -> list:
    """
    Получение списка всех интерфейсов, подключенных к OVS-коммутаторам
    """

    return [port for bridge in get_ovs_bridges(as_root) for port in get_ovs_bridge_ports(bridge, as_root)]


def get_patches() -> list:
    """
    Возвращает список словарей вида {switch: name, port: name} для последующей блокировки/разблокировки трафика
    """

    result = []
    for bridge in get_ovs_bridges():
        rstp_enabled = True if get_ovs_bridge_param(bridge, OvsRequestParams.rstp) == 'true' else False
        if rstp_enabled:
            tunnel_infs_exist = [iface for iface in get_ovs_bridge_ifaces(bridge)  # noqa
                                 if get_ovs_interface_param(iface, OvsRequestParams.interface_type) in
                                 (OvsPortType.vxlan, OvsPortType.geneve)]
            if tunnel_infs_exist:
                for iface in get_ovs_bridge_ifaces(bridge):
                    port_type = get_ovs_interface_param(iface, OvsRequestParams.interface_type)
                    if port_type == OvsPortType.patch:
                        result.append(dict(switch=bridge, port=iface))
                        peer = get_ovs_interface_param(iface, OvsRequestParams.peer).strip('"')
                        peer_bridge = get_bridge_of_port(peer)
                        result.append(dict(switch=peer_bridge, port=peer))
    return result


def get_tunnels() -> list:
    """
    Возвращает список словарей вида {switch: name, port: name} для последующей блокировки/разблокировки трафика
    """

    return [dict(switch=bridge, port=iface) for bridge in get_ovs_bridges() for iface in get_ovs_bridge_ifaces(bridge)
            if get_ovs_interface_param(iface, OvsRequestParams.interface_type) in (OvsPortType.vxlan,
                                                                                   OvsPortType.geneve)]


def port_exists_in_ovs(name: str, as_root=True) -> bool:
    """
    Проверка, существует ли порт в базе данных OpenVswitch
    """
    if as_root:
        cmd = f"sudo ovs-vsctl find port name={name}"
    else:
        cmd = f"ovs-vsctl find port name={name}"
    return True if states_result(get_output, cmd=cmd) else False


def iface_exists_in_ovs(name: str, as_root=True) -> bool:
    """
    Проверка, существует ли интерфейс в базе данных OpenVswitch
    """
    if as_root:
        cmd = f'sudo ovs-vsctl find interface name={name}'
    else:
        cmd = f'ovs-vsctl find interface name={name}'
    return True if states_result(get_output, cmd=cmd) else False


def device_in_aggregation(name: str, as_root=True) -> bool:
    """
    Проверка, входит ли физический интерфейс в состав агрегации
    """

    return True if iface_exists_in_ovs(name, as_root) and not port_exists_in_ovs(name, as_root) else False


def get_link_info_detailed(name: str) -> dict:
    """
    Получение данных об интерфейсе утилитой ip
    """

    cmd = f"ip -j -p -d link show dev {name}"
    output = states_result(get_output, cmd=cmd)
    return next((conf for conf in json.loads(output) if conf and conf['ifname'] == name), None)


def parse_routes() -> List[dict]:
    """
    Получение информации о текущих сохраненных статических маршрутах узла
    """

    with open(ROUTE_INIT_PATH, 'r') as f:
        content = f.read().splitlines()

    result = list()
    routes = [s for s in content if not s.startswith('#') and s != '\n' and bool(s) is True]
    for index_route, row in enumerate(routes):
        route = dict(destination='', interface=None, gateway='', metric=None, index=index_route)
        params = row.replace('ip route add', 'net').split()
        for i, _ in enumerate(params):
            if params[i] == 'net':
                route['destination'] = params[i + 1]
            if params[i] == 'dev':
                route['interface'] = params[i + 1]
            if params[i] == 'via':
                route['gateway'] = params[i + 1]
            if params[i] == 'metric':
                route['metric'] = int(params[i + 1])
        result.append(route)
    return result


def parse_resolv_conf() -> dict:
    """
    Получение информации о текущей общей конфигурации DNS узла
    """

    result = dict(search=[], nameservers=[], domain=None)
    try:
        with open(FILE_RESOLV_CONF, 'r') as f:
            raw_file = f.read().splitlines()
            raw_conf = [s for s in raw_file if s and not s.startswith('#')]
        result['nameservers'] = [s.split()[1] for s in raw_conf if s.startswith('nameserver')]
        for s in raw_conf:
            if s.startswith('search'):
                result['search'] = s.split()[1:]
                break
        for s in raw_conf:
            if s.startswith('domain'):
                result['domain'] = s.split()[1]
                break
    except OSError as e:
        raise CommandExecutionError(
            f"Failed to read resolv.conf configuration. I/O error({e.errno}): {e.strerror}") from None
    except Exception as e:
        raise CommandExecutionError(f"Unexpected error during parsing resolv.conf: {str(e)}") from None
    return result


def parse_resolv_base() -> dict:
    """
    Получение информации о текущей основной конфигурации DNS узла
    """

    result = dict(search=[], nameservers=[], domain=None)
    try:
        with open(FILE_RESOLV_BASE, 'r') as f:
            raw_file = f.read().splitlines()
            raw_conf = [s for s in raw_file if s and not s.startswith('#')]
        result['nameservers'] = [s.split()[1] for s in raw_conf if s.startswith('nameserver')]
        for s in raw_conf:
            if s.startswith('search'):
                result['search'] = s.split()[1:]
                break
        for s in raw_conf:
            if s.startswith('domain'):
                result['domain'] = s.split()[1]
                break
    except OSError as e:
        raise CommandExecutionError(
            f"Failed to read resolv base configuration. I/O error({e.errno}): {e.strerror}") from None
    except Exception as e:
        raise CommandExecutionError(f"Unexpected error: {str(e)}") from None
    return result


def parse_dhclient_conf() -> dict:
    """
    Получение конфигурации DHCP-клиента узла
    """

    result = dict(assignments=[], requests=[], interfaces=[])

    try:
        with open(FILE_DHCLIENT_CONF, 'r') as f:
            raw_file = f.read().splitlines()
            raw_conf = [s for s in raw_file if s and not s.startswith('#')]
    except OSError as e:
        raise CommandExecutionError(f"Failed to read DHCP configuration. I/O error({e.errno}): {e.strerror}") from None
    except Exception as e:
        raise CommandExecutionError(f"Unexpected error: {str(e)}") from None

    for s in raw_conf:
        if bool(DHCLIENT_OPTION_PATTERN.match(s)):
            result['assignments'].append(DHCLIENT_OPTION_PATTERN.search(s).string)
        elif bool(DHCLIENT_SEND_PATTERN.match(s)):
            result['assignments'].append(DHCLIENT_SEND_PATTERN.search(s).string)
        elif bool(DHCLIENT_REQUEST_PATTERN.match(s)):
            result['requests'].extend(DHCLIENT_REQUEST_PATTERN.search(s).groupdict()['requests'].split(', '))
        elif bool(DHCLIENT_INTERFACE_PATTERN.match(s)):
            ret = DHCLIENT_INTERFACE_PATTERN.search(s).groupdict()
            ret['requests'] = ret['requests'].split(', ')
            result['interfaces'].append(ret)
    return result


def parse_dhclient_leases(iface: str) -> List[dict]:
    """
    Функция парсит файл /var/lib/dhcp/dhclient.{iface}.leases и возвращает список структурированных данных
    """

    leases_file = f"/var/lib/dhcp/dhclient.{iface}.leases"
    structed_leases = list()
    if not os.path.exists(leases_file):
        return structed_leases
    with open(leases_file, 'r') as f:
        raw_leases = f.read().splitlines()
    raw_leases = [s.rstrip(';').lstrip() for s in raw_leases]
    stack = []
    lease = []
    leases_list = []
    for s in raw_leases:
        if s == 'lease {':
            stack.append('{')
        elif s == '}':
            stack.pop()
            leases_list.append(lease)
            lease = []
        elif s.startswith('default-duid'):
            continue
        else:
            if '{' in stack:
                lease.append(s)

    for lease_info in leases_list:
        lease_dict = dict(interface='', options=[])
        options = []
        for s in lease_info:
            if 'interface' in s:
                lease_dict['interface'] = s.split()[1].strip('"')
            if 'expire' in s:
                lease_dict['expire'] = ' '.join(s.split()[2:])
            if 'fixed-address' in s:
                lease_dict['ipv4_address'] = s.split()[1]
            if 'option' in s:
                raw_option = s.split()[1:]
                option_key = raw_option[0]
                option_value = raw_option[1] if ',' not in raw_option[1] else raw_option[1].split(',')
                options.append({option_key: option_value})
        lease_dict['options'] = options
        structed_leases.append(lease_dict)
    return structed_leases


def get_last_interface_lease(interface: str) -> dict:
    """
    Функция получает список структурированных данных о dhcp-арендах, ищет актуальные данные о переданном интерфейсе и
    возвращает словарь с данными о полученных интерфейсом опциях
    """

    result = dict(interface=interface, gateway=None, nameservers=None, domain=None, search=None, ipv4_address=None,
                  ipv4_netmask=None)
    leases = parse_dhclient_leases(interface)
    if not leases:
        return result
    last_inf_lease = sorted([item for item in leases if item['interface'] == interface], key=lambda k: k['expire'])[-1]
    if not last_inf_lease:
        return result
    for option in last_inf_lease['options']:
        if option.get('routers'):
            result['gateway'] = option['routers']
        elif option.get('domain-name-servers'):
            result['nameservers'] = option['domain-name-servers'] if isinstance(option['domain-name-servers'], list) \
                else [option['domain-name-servers']]
        elif option.get('domain-search'):
            result['search'] = option['domain-search'] if isinstance(option['domain-search'], list) \
                else [option['domain-search']]
        elif option.get('domain-name'):
            result['domain'] = option['domain-name']
        elif option.get('subnet-mask'):
            result['ipv4_netmask'] = option['subnet-mask']
        else:
            continue
    result['ipv4_address'] = last_inf_lease.get('ipv4_address')
    return result


def get_bridge_of_port(port: str) -> Optional[str]:
    """
    Получения названия коммутатора, которому принадлежит порт
    """

    cmd = f'sudo ovs-vsctl port-to-br {port}'
    ret = get_output(cmd, suppress_err=True)
    if ret['result']:
        return ret['comment'].rstrip('\n')
    else:
        return None


def get_bridge_of_interface(interface: str) -> Optional[str]:
    """
    Получение названия коммутатора, которому принадлежит интерфейс
    """
    cmd = f'sudo ovs-vsctl iface-to-br {interface}'
    ret = get_output(cmd, suppress_err=True)
    if ret['result']:
        return ret['comment'].rstrip('\n')
    else:
        return None


def get_ovs_bridges(as_root=True) -> list:
    """
    Получение списка коммутаторов
    """
    if as_root:
        cmd = 'sudo ovs-vsctl list-br'
    else:
        cmd = 'ovs-vsctl list-br'
    return states_result(get_output, cmd=cmd).splitlines()


def get_ovs_bridge_param(bridge: str, param: str, as_root=True) -> str:
    """
    Получение параметра коммутатора
    """
    if as_root:
        cmd = f'sudo ovs-vsctl get br {bridge} {param}'
    else:
        cmd = f'ovs-vsctl get br {bridge} {param}'
    return states_result(get_output, cmd=cmd).rstrip('\n')


def get_ovs_bridge_ports(bridge: str, as_root=True) -> list:
    """
    Получение списка портов коммутатора
    """
    if as_root:
        cmd = f'sudo ovs-vsctl list-ports {bridge}'
    else:
        cmd = f'ovs-vsctl list-ports {bridge}'
    return states_result(get_output, cmd=cmd).splitlines()


def get_ovs_bridge_ifaces(bridge: str, as_root=True) -> list:
    """
    Получение списка интерфейсов коммутатора
    """
    if as_root:
        cmd = f'sudo ovs-vsctl list-ifaces {bridge}'
    else:
        cmd = f'ovs-vsctl list-ifaces {bridge}'
    return states_result(get_output, cmd=cmd).splitlines()


def get_ovs_bridge_vm_ifaces(bridge: str, as_root=True) -> list:
    """
    Получение списка вм интерфейсов коммутатора
    """

    if as_root:
        cmd = f'sudo ovs-vsctl list-ifaces {bridge}'
    else:
        cmd = f'ovs-vsctl list-ifaces {bridge}'
    vm_infs = states_result(get_output, cmd=cmd).splitlines()
    return [inf for inf in vm_infs if inf.startswith('vm_')]


def get_ovs_port_param(port: str, param: str, as_root=True) -> str:
    """
    Получение параметра порта коммутатора
    """
    if as_root:
        cmd = f'sudo ovs-vsctl get port {port} {param}'
    else:
        cmd = f'ovs-vsctl get port {port} {param}'
    return states_result(get_output, cmd=cmd).rstrip('\n')


@states_response
def get_ovs_bridges_configs():
    """
    Получение портов/интерфейсов коммутаторов
    """
    result = dict()
    bridges = get_ovs_bridges()
    for bridge in bridges:
        result[bridge] = dict(ifaces=get_ovs_bridge_ifaces(bridge),
                              ports=get_ovs_bridge_ports(bridge))
    return result


@states_response
def get_ovs_bridge_configs(bridge: str):
    """
    Получение портов/интерфейсов коммутатора
    """
    return dict(ifaces=get_ovs_bridge_ifaces(bridge), ports=get_ovs_bridge_ports(bridge))


def get_ovs_bond_params(port: str) -> list:
    """
    Получение параметра бонда
    """

    cmd = f'sudo ovs-appctl bond/show {port}'
    return states_result(get_output, cmd=cmd).split('\n')


def get_ovs_interface_param(port: str, param: str, as_root=True) -> str:
    """
    Получение параметра интерфейса коммутатора
    """
    if as_root:
        cmd = f'sudo ovs-vsctl --if-exists get int {port} {param}'
    else:
        cmd = f'ovs-vsctl --if-exists get int {port} {param}'
    return states_result(get_output, cmd=cmd).rstrip('\n')


def get_physical_inf_params(name: str) -> dict:
    """
    Получение параметров физического интерфейса
    """

    settings = dict(link_speed='', duplex_mode='', link_state=False)
    keys = ['Speed', 'Duplex', 'Link detected']
    cmd = "ethtool {port} | grep -E '{settings}' | awk '{{print $1$2$3}}'".format(port=name,
                                                                                  settings='|'.join(keys))
    result = states_result(get_output, cmd=cmd)
    for line in result.splitlines():
        key = line.split(':')[0]
        value = line.split(':')[1]
        if key == 'Speed':
            value = '-' if value == 'Unknown!' else value
            settings['link_speed'] = value
        if key == 'Duplex':
            value = '-' if value == 'Unknown!(255)' else value
            settings['duplex_mode'] = value
        if key == 'Linkdetected':
            settings['link_state'] = True if value == 'yes' else False
    return settings


def dump_network_devices(as_root=True) -> list:
    """
    Получение информации об устройствах класса network от утилиты lshw в виде словаря
    *из-за того, что используемая версия lshw некорректно формирует json, приходится получать данные в xml и
    преобразовывать их в json
    :return: список словарей, описывающих сетевые устройства узла
    """
    if as_root:
        cmd = 'sudo lshw -c network -disable SCSI -xml 2>/dev/null'
    else:
        cmd = 'lshw -c network -disable SCSI -xml 2>/dev/null'
    output = states_result(get_output, cmd=cmd)
    dev_xml = xmltodict.parse(output)['list']
    dev_json = json.loads(json.dumps(dev_xml).replace('@', ''))
    dev_json['netdev'] = dev_json.pop('node')
    for dev in dev_json['netdev']:
        if isinstance(dev, dict) and 'node' in dev.keys():
            dev['netdev'] = dev.pop('node')

    # Проверим наличие в выводе lshw устройств с драйвером openvswitch для e2k
    # Если устройств нет, формируем необходимый json для каждого OVS бриджа
    if platform.machine() == 'e2k' and 'openvswitch' not in output:
        ovs_netdev = {
            'id': '',
            'claimed': 'true',
            'class': 'network',
            'description': 'Ethernet interface',
            'physid': '',
            'logicalname': '',
            'serial': '',
            'configuration': {'setting': [
                {'id': 'broadcast', 'value': 'yes'},
                {'id': 'driver', 'value': 'openvswitch'},
                {'id': 'driverversion', 'value': ''},
                {'id': 'multicast', 'value': 'yes'}]
            },
            'capabilities': {'capability': [{'id': 'ethernet'}, {'id': 'physical', '#text': 'Physical interface'}]},
            'hints': {'hint': {'name': 'icon', 'value': 'network'}},
        }
        # Список портов в формате csv, чтобы легче разделить было
        cmd = 'ovs-vsctl -f csv list port'
        if as_root:
            cmd = f'sudo {cmd}'
        output = get_output(cmd, suppress_err=True).get('comment').strip().splitlines()
        # Взять имена бриджей
        bridges = list(map(lambda br: br['name'], csv.DictReader(output)))
        for idx, bridge in enumerate(bridges):
            # Пропустить устройства, которые есть в выводе lshw
            # pylint: disable=cell-var-from-loop
            if any(map(lambda dev: bridge in dev.values(), dev_json['netdev'])):
                continue
            netdev = ovs_netdev.copy()
            cmd = f'ovs-vsctl -f csv list Interface {bridge}'
            if as_root:
                cmd = f'sudo {cmd}'
            bridge_params = get_output(cmd, suppress_err=True).get('comment').strip().splitlines()
            bridge_params = next(csv.DictReader(bridge_params))
            if bridge_params.get('admin_state', 'down') == 'down':
                netdev['disabled'] = 'true'
                netdev['configuration']['setting'].append({'id': 'link', 'value': 'no'})
            elif bridge_params.get('link_state', 'down') == 'up':
                netdev['configuration']['setting'].append({'id': 'link', 'value': 'yes'})
            else:
                netdev['configuration']['setting'].append({'id': 'link', 'value': 'no'})
            netdev['serial'] = bridge_params.get('mac_in_use', 'Unknown!').replace('"', '')
            netdev['id'] = f'network:{idx}'
            netdev['physid'] = f'{idx + 1}'
            netdev['logicalname'] = bridge
            dev_json['netdev'].append(netdev)

    return dev_json['netdev']


def ethtool(name: str, as_root=True) -> dict:
    """
    Вывод ethtool -i name в dict.
    С --json и без него выводит одинаковый результат.
    Как заработает вывод json, можно указать параметр --json и не парсить строки.
    """
    cmd = f"ethtool -i {name}"
    if not as_root:
        cmd = f"sudo {cmd}"
    result = get_output(cmd, suppress_err=True)
    if result.get('result'):
        info = dict(map(lambda i: i.split(": "), result.get('comment').strip().splitlines()))
    else:
        info = {'driver': 'Unknown', 'version': 'Unknown', 'bus-info': 'Unknown'}
    return info


def get_physical_inf_device_info(name: str, as_root=True) -> dict:
    """ Собирает необходимую информацию по устройству """
    info = dict(pci_info='', vendor='Unknown', model='Unknown', driver='', driver_version='')
    ethtool_info = ethtool(name, as_root)
    cmd = f"udevadm info -q property {SYS_CLASS_NET_DIR}/{name}"
    if not as_root:
        cmd = f"sudo {cmd}"
    result = get_output(cmd, suppress_err=True)
    if result.get('result'):
        device = dict(map(lambda p: p.split('='), result.get('comment').splitlines()))
        info['pci_info'] = device.get('ID_PATH', ethtool_info['bus-info'])
        info['vendor'] = device.get('ID_VENDOR_FROM_DATABASE', info['vendor'])
        info['model'] = device.get('ID_MODEL_FROM_DATABASE', info['model'])
        info['driver'] = device.get('ID_NET_DRIVER', ethtool_info['driver'])
        info['driver_version'] = ethtool_info['version']

    return info


def parse_sys_class_net(as_root=True) -> Tuple[List[dict], List[dict], List[dict]]:
    """
    Собирает данные об сетевых интерфейсах и выбирает физические интерфейсы.
    Использует данные ethtool, udev, ip
    """
    hardware_devices = list()
    other_devices = list()
    ovs_bridges = list()
    existing_bridges = get_ovs_bridges(as_root)
    dev_names = [dev_name for dev_name in os.listdir(SYS_CLASS_NET_DIR) if dev_name != 'lo']
    for name in dev_names:
        link_info = get_link_info_detailed(name)
        if 'device' in os.listdir(os.path.join(SYS_CLASS_NET_DIR, name)):
            device = get_physical_inf_device_info(name, as_root)
            dev_path = os.path.join(SYS_CLASS_NET_DIR, name, 'device')
            dev_info = dict(
                name=name,
                mac_address=link_info['address'],
                pci_info=device['pci_info'],
                vendor=device['vendor'],
                model=device['model'],
                driver=device['driver'],
                driver_version=device['driver_version'],
                up_state="UP" in link_info['flags'],
                sriov_ready=False,
                sriov_enabled=False,
                vf_count=0,
                bonded=False,
                connected=False,
                promiscuous_mode=True if 'PROMISC' in link_info['flags'] else False,
                max_mtu=int(link_info['max_mtu']),
                mtu=int(link_info['mtu']))
            if 'sriov_totalvfs' in os.listdir(dev_path):
                dev_info['sriov_ready'] = True
                dev_info['vf_count'] = int(
                    states_result(get_output, cmd=f'cat {os.path.join(dev_path, "sriov_totalvfs")}'))
                if int(states_result(get_output, cmd=f'cat {os.path.join(dev_path, "sriov_numvfs")}')):
                    dev_info['sriov_enabled'] = True
                    dev_info['virtual_functions'] = get_vfs_info(dev_info['name'])
                else:
                    dev_info['sriov_enabled'] = False
            hardware_devices.append(dev_info)
        else:
            ethtool_info = ethtool(name, as_root)
            if any(map(lambda x: x in name, SYSTEM_DEVICE_NAMES)):  # pylint: disable=cell-var-from-loop
                continue
            dev_info = dict(name=name)
            dev_info['mac_address'] = link_info['address']
            dev_info['driver'] = ethtool_info['driver']
            # pylint: disable=expression-not-assigned
            other_devices.append(dev_info) if name not in existing_bridges else ovs_bridges.append(dev_info)

    return hardware_devices, other_devices, ovs_bridges


def parse_network_devices(data: list, as_root=True) -> Tuple[List[dict], List[dict], List[dict]]:
    """
    Выбирает из полученного списка устройства, которые можно отнести к физическим интерфейсам
    :param data: данные о сетевых устройствах в виде списка словарей
    :param as_root:
    """

    hardware_devices = list()
    other_devices = list()
    ovs_bridges = list()
    existing_bridges = get_ovs_bridges(as_root)
    while len(data):
        dev = data.pop()
        if 'handle' in dev.keys() and dev['handle'].split(':')[0] == 'PCI' and 'Virtual Function' not in dev['product']:
            if dev.get('netdev'):
                if isinstance(dev['netdev']['capabilities']['capability'], dict) \
                        and dev['netdev']['capabilities']['capability'].get('id') != 'physical':
                    continue
                if not any(map(lambda x: x['id'] == 'physical', dev['netdev']['capabilities']['capability'])):
                    continue
                if isinstance(dev['netdev']['logicalname'], str):
                    logical_name = dev['netdev']['logicalname']
                elif isinstance(dev['netdev']['logicalname'], list):
                    logical_name = next((i for i in dev['netdev']['logicalname'] if not i.startswith("/")), None)
                    if not logical_name:
                        continue
                else:
                    continue
                dev_path = os.path.join(SYS_CLASS_NET_DIR, logical_name, 'device')
                vagrant_instance = True
                up_state = True if 'disabled' not in dev['netdev'].keys() else False
                name = logical_name
            else:
                if isinstance(dev['capabilities']['capability'], dict) and dev['capabilities']['capability'].get(
                        'id') != 'physical':
                    continue
                if not any(map(lambda x: x['id'] == 'physical', dev['capabilities']['capability'])):
                    continue
                if isinstance(dev['logicalname'], str):
                    logical_name = dev['logicalname']
                elif isinstance(dev['logicalname'], list):
                    logical_name = next((i for i in dev['logicalname'] if not i.startswith("/")), None)
                    if not logical_name:
                        continue
                else:
                    continue
                dev_path = os.path.join(SYS_CLASS_NET_DIR, logical_name, 'device')
                vagrant_instance = False
                up_state = True if 'disabled' not in dev.keys() else False
                name = logical_name
            driver_info = dict(driver='', version='')
            for entry in dev['configuration']['setting']:
                if entry['id'] == 'driver':
                    driver_info['driver'] = entry['value']
                    if all(map(lambda x: bool(driver_info[x]) is True, driver_info.keys())):
                        break
                if entry['id'] == 'driverversion':
                    driver_info['version'] = entry['value']
                    if all(map(lambda x: bool(driver_info[x]) is True, driver_info.keys())):
                        break
            link_info = get_link_info_detailed(name)
            dev_info = dict(name=name,
                            mac_address=dev['serial'] if not vagrant_instance else dev['netdev']['serial'],
                            pci_info=dev['handle'],
                            vendor=dev['vendor'],
                            model=dev['product'],
                            driver=driver_info['driver'],
                            driver_version=driver_info['version'],
                            up_state=up_state,
                            sriov_ready=False,
                            sriov_enabled=False,
                            vf_count=0,
                            bonded=False,
                            connected=False,
                            promiscuous_mode=True if 'PROMISC' in link_info['flags'] else False,
                            max_mtu=int(link_info['max_mtu']),
                            mtu=int(link_info['mtu']))
            if 'sriov_totalvfs' in os.listdir(dev_path):
                dev_info['sriov_ready'] = True
                dev_info['vf_count'] = int(
                    states_result(get_output, cmd=f'cat {os.path.join(dev_path, "sriov_totalvfs")}'))
                if int(states_result(get_output, cmd=f'cat {os.path.join(dev_path, "sriov_numvfs")}')):
                    dev_info['sriov_enabled'] = True
                    dev_info['virtual_functions'] = get_vfs_info(dev_info['name'])
                else:
                    dev_info['sriov_enabled'] = False
            hardware_devices.append(dev_info)
        else:
            if dev.get('logicalname'):
                if any(map(lambda x: x in dev['logicalname'], SYSTEM_DEVICE_NAMES)):
                    continue
                dev_info = dict(name=dev['logicalname'])
                if dev.get('serial'):
                    dev_info['mac_address'] = dev['serial']
                for entry in dev['configuration']['setting']:
                    if entry['id'] == 'driver':
                        dev_info['driver'] = entry['value']
                        break
                # pylint: disable=expression-not-assigned
                other_devices.append(dev_info) if dev_info['name'] not in existing_bridges \
                    else ovs_bridges.append(dev_info)
    return hardware_devices, other_devices, ovs_bridges


def rescan_hardware(as_root: bool = True, sys_parser: bool = False) -> list:
    """
    Сканирование физических сетевых устройств узла
    """

    if sys_parser:
        hardware_devices, other_devices, _ = parse_sys_class_net(as_root)
    else:
        hardware_devices, other_devices, _ = parse_network_devices(dump_network_devices(as_root), as_root)
    hw_dev_names = [dev['name'] for dev in hardware_devices]
    other_dev_names = [dev['name'] for dev in other_devices]
    stored_configs = [dev for dev in os.listdir(BASE_NET_CONF_DIR)
                      if dev not in ('lo', 'veil-blackhole', 'veil-mgmt') and dev not in other_dev_names]
    if all(dev in stored_configs for dev in hw_dev_names) and all(dev in hw_dev_names for dev in stored_configs):
        return hw_dev_names
    else:
        removed_devices = list(set(stored_configs).difference(set(other_dev_names + hw_dev_names)))
        added_devices = list(set(hw_dev_names).difference(set(stored_configs)))
        for dev in removed_devices:
            del_port_from_switch(dev)
            del_file_by_pattern(BASE_NET_CONF_DIR, dev)
            # hw_dev_names.remove(dev)  # здесь была ошибка list.remove(x): x not in list
        for dev in added_devices:
            write_interface(BASE_NET_CONF_DIR, dev, InterfaceModes.manual, is_connected=False)
            hw_dev_names.append(dev)
    return hw_dev_names


def dump_device_params(name: str, data: list) -> dict:
    """
    Возвращает информацию о физическом интерфейсе
    :param name: данные о сетевых устройствах в виде списка словарей
    """

    device = next((dev for dev in data if dev['name'] == name), None)
    if device is None:
        raise CommandExecutionError()
    return device


@states_response
def check_controller_connection(controller_ip: Optional[str] = None):
    """
    Проверка связи узла с контроллером
    :param controller_ip: ip-адрес контроллера
    :return: булевое значение, определяющее, была ли проверка успешной
    """

    try:
        if controller_ip is None:
            conf = get_config()
            controller_ip = conf.get('controller_ip')
        cmd = f'ping {controller_ip} -c 5 -i 0.2'
        cmd_ret = states_result(get_output, cmd=cmd)
        output = cmd_ret.split('\n')
        return False if any(failed_reply in s for s in output for failed_reply in CONNECTION_LOSS_REPLIES) else True
    except (IOError, json.JSONDecodeError) as e:
        raise CommandExecutionError(f'Error opening configuration file: {str(e)}') from None
    except Exception as e:
        raise CommandExecutionError(str(e)) from None


@states_response
def check_interface_mode(conf_dir: str = BASE_NET_CONF_DIR, interface: str = VEIL_MGMT_INTERFACE,
                         mode: str = InterfaceModes.dhcp):
    """
    Проверка режима работы сетевого интерфейса
    :param conf_dir: директория, содержащая файлы конфигурации сетевых интерфейсов
    :param interface: название интерфейса
    :param mode: проверяемый режим работы сетевого интерфейса
    :return: булевое значение, определяеющее, была ли проверка успешной
    """

    conf_file = os.path.join(conf_dir, interface)
    config = get_iface_config_as_dict(get_iface_config_as_list(conf_file))
    return mode == config['mode']


@states_response
def get_address_info_real(interface: str):
    """
    Получение реальной ip-конфигурации сетевого интерфейса
    :param interface: название интерфейса
    :return: словарь, содержащий параметры конфигурации интерфейса, или None
    """

    cmd = f"ip -j -p address show {interface}"
    output = states_result(get_output, cmd=cmd)
    config = next((conf for conf in json.loads(output) if conf and conf['ifname'] == interface), None)
    if config is None:
        return None
    address_info = next((info for info in config.get('addr_info') if info['label'] == interface), None)
    return dict(mac_address=config['address'],
                ipv4_address=address_info.get('local') if address_info and address_info.get('local') else '',
                ipv4_netmask=str(IPv4Interface(f"{address_info['local']}/{address_info['prefixlen']}").netmask) if
                address_info and address_info.get('local') and address_info.get('prefixlen') else '',
                dhcpv4=states_result(check_interface_mode, interface=interface))


@states_response
def get_link_info_real(interface: str):
    """
    Получени реальной конфигурации сетевого устройства
    :param interface: название интерфейса
    :return: словарь, содержащий параметры конфигурации интерфейса, или None
    """

    cmd = f"ip -j -p link show {interface}"
    output = states_result(get_output, cmd=cmd)
    config = next((conf for conf in json.loads(output) if conf and conf['ifname'] == interface), None)
    if config is None:
        return None
    return dict(up_state=True if 'UP' in config['flags'] else False,
                flags=config['flags'],
                mac_address=config['address'],
                mtu=config['mtu'])


@states_response
def block_port_traffic(switch_port_pairs: List[dict]):
    """
    Добавляет правило, дропающее трафик через указанный порт, в таблицу потоков (flows) указанного свитча
    :param switch_port_pairs: список пар свитч-порт
    """

    block_rule_set = TrafficBlockRuleSet.from_config_file()
    errors = []
    for pair in switch_port_pairs:
        port = pair['port']
        switch = pair['switch']
        if port_exists_in_ovs(port):
            cmd = f"ovs-ofctl add-flow {switch} table=0,in_port={port},actions=drop"
            err = states_result(get_output, base_err_msg=f"Failed to block port {port} on switch {switch}.",
                                raise_exception=False, get_err=True, cmd=cmd)
            if err:
                errors.append(err)
            rule_to_add = TrafficBlockRule(switch=switch, port=port, applied=True, to_apply=True)
            if not block_rule_set.add_rule(rule_to_add):
                errors.append(f"Failed to save blocking rule for port '{port}' on switch '{switch}'.")
    if errors:
        raise CommandExecutionError('\n'.join(errors))
    block_rule_set.save()


@states_response
def unblock_port_traffic(switch_port_pairs: List[dict]):
    """
    Удаляет правило, дропающее трафик через указанный порт, в таблицу потоков (flows) указанного свитча
    :param switch_port_pairs: список пар свитч-порт
    """

    block_rule_set = TrafficBlockRuleSet.from_config_file()
    errors = []
    for pair in switch_port_pairs:
        port = pair['port']
        switch = pair['switch']
        if port_exists_in_ovs(port):
            cmd = f"ovs-ofctl del-flows {switch} table=0,in_port={port}"
            if not states_result(get_output, raise_exception=False, cmd=cmd):
                pass
            rule_to_delete = block_rule_set.find_rule(dict(switch=switch, port=port))
            if rule_to_delete and not block_rule_set.remove_rule(rule_to_delete):
                errors.append(f"Failed to remove blocking rule for port '{port}' on switch '{switch}'.")
    if errors:
        raise CommandExecutionError('\n'.join(errors))
    block_rule_set.save()


@states_response
def list_physical_infs(verbose: bool = False, sys_parser: bool = False):
    """
    Получение списка физических интерфейсов узла
    :param verbose: флаг, определяющий возвращается ли детализированная информация
    :param sys_parser: флаг, определяющий метод определения сетевых устройств
    """

    if sys_parser:
        hardware_devices, _, _ = parse_sys_class_net()
    else:
        net_devices = dump_network_devices()
        hardware_devices, _, _ = parse_network_devices(net_devices)

    if not verbose:
        return [dev['name'] for dev in hardware_devices]
    else:
        result = list()
        for dev in hardware_devices:
            port_data = dict(interface=dev['name'], mac_address=dev['mac_address'], connected=False, switch=None,
                             bonded=False, sriov_ready=dev['sriov_ready'], vf_count=dev['vf_count'],
                             sriov_enabled=dev['sriov_enabled'], promiscuous_mode=dev['promiscuous_mode'],
                             max_mtu=dev['max_mtu'], mtu=dev['mtu'],
                             state='UP' if states_result(check_port_up_state, port=dev['name']) else 'DOWN',
                             pci_info=dev['pci_info'], vendor=dev['vendor'], model=dev['model'],
                             driver=dev['driver'], driver_version=dev['driver_version'],
                             link_state=get_physical_inf_params(dev['name'])['link_state'])
            if port_exists_in_ovs(dev['name']):
                port_data['connected'] = True
                port_data['switch'] = get_bridge_of_port(dev['name'])
            elif device_in_aggregation(dev['name']):
                port_data['connected'] = True
                port_data['switch'] = get_bridge_of_port(dev['name'])
                port_data['bonded'] = True
                port_data['switch'] = get_bridge_of_interface(dev['name'])
            result.append(port_data)
        return result


@states_response
def get_port_pgroup_config(name: str, get_mtu: bool = True, as_root=True):
    """
    Получение конфигурации VLAN порта
    """

    result = dict(vlan_mode=None, vlan_tag=None, vlan_trunks=None)

    vlan_mode = get_ovs_port_param(name, OvsRequestParams.vlan_mode, as_root).strip('"')
    vlan_tag = get_ovs_port_param(name, OvsRequestParams.vlan_tag, as_root).strip('"')
    vlan_trunks = get_ovs_port_param(name, OvsRequestParams.vlan_trunks, as_root).lstrip('[').rstrip(']')

    result['vlan_mode'] = vlan_mode if vlan_mode != '[]' else None
    result['vlan_tag'] = int(vlan_tag) if vlan_tag != '[]' else None
    result['vlan_trunks'] = [int(vlan) for vlan in vlan_trunks.split(',')] if vlan_trunks else None
    if get_mtu:
        mtu = get_ovs_interface_param(name, OvsRequestParams.mtu_request, as_root)
        if mtu == '[]':
            mtu = get_ovs_interface_param(name, OvsRequestParams.mtu, as_root)
            set_ovs_mtu(name, int(mtu))
        result['mtu'] = int(mtu)

    return result


@states_response
def get_bond_config(name: str, as_root=True) -> dict:
    raw_bond_mode = get_ovs_port_param(name, OvsRequestParams.bond_mode, as_root).strip('"')
    bond_mode = raw_bond_mode if raw_bond_mode != '[]' else None
    raw_bond_lacp = get_ovs_port_param(name, OvsRequestParams.bond_lacp, as_root).strip('"')
    bond_lacp = raw_bond_lacp if raw_bond_lacp != '[]' else None
    raw_other_config = get_ovs_port_param(name, OvsRequestParams.other_config, as_root).strip('"')

    other_config = dict()
    if raw_other_config != '{}':
        other_params = [param.strip() for param in raw_other_config.lstrip('{').rstrip('}').split(',')]
        for item in other_params:
            k = item.split('=')[0].replace('-', '_')
            v = item.split('=')[1].strip('"')
            if k == 'lacp_fallback_ab':
                k = 'lacp_fallback'
            if k == 'bond_rebalance_interval':
                k = 'rebalance_interval'
                v = int(v)
            other_config[k] = v
        if 'lacp_fallback' not in other_config:
            other_config['lacp_fallback'] = None
        if 'lacp_time' not in other_config:
            other_config['lacp_time'] = None
        if 'rebalance_interval' not in other_config:
            other_config['rebalance_interval'] = None
    else:
        other_config['lacp_fallback'] = None
        other_config['lacp_time'] = None
        other_config['rebalance_interval'] = None

    result = dict(bond_mode=bond_mode, bond_lacp=bond_lacp, lacp_fallback=other_config['lacp_fallback'],
                  lacp_time=other_config['lacp_time'], rebalance_interval=other_config['rebalance_interval'])
    return result


@states_response
def get_bond_additional_info(name: str) -> dict:
    bond_params = dict()
    bond_params_output = get_ovs_bond_params(name)

    for index, bond_param in enumerate(bond_params_output):
        if 'updelay' in bond_param:
            bond_params["updelay"] = bond_param.split()[1]
        elif 'downdelay' in bond_param:
            bond_params["downdelay"] = bond_param.split()[1]
        elif 'rebalance' in bond_param:
            bond_params["rebalance_interval"] = bond_param.split()[2]
        elif 'lacp_status' in bond_param:
            bond_params["lacp_status"] = bond_param.split()[1]
        elif 'active member mac' in bond_param or 'active slave mac' in bond_param:
            bond_params["mac"] = bond_param.split()[3]
        elif 'member' in bond_param or 'slave' in bond_param:
            bond_params["members"] = list()
            for line in bond_params_output[index:-1]:
                if line.startswith("member") or line.startswith('slave'):
                    bond_data = dict()
                    _, name, state = line.split()
                    name = name.rstrip(':')
                    bond_data["name"] = name
                    bond_data["state"] = state
                    # bond_data["status"] = "not active"
                # elif 'active' in line:
                #     status, _ = line.split()
                #     bond_data["status"] = status
                elif 'may_enable' in line:
                    _, may_enable = line.split()
                    bond_data["may_enable"] = may_enable
                    bond_params["members"].append(bond_data)
            break
    return bond_params


@states_response
def get_bond_members(name: str) -> list:
    """
    Получение списка интерфейсов - участников агрегации
    """

    result = list()
    members_ids = get_ovs_port_param(name, OvsRequestParams.interfaces).lstrip('[').rstrip(']').split(', ')
    for member in members_ids:
        result.append(get_ovs_interface_param(member, OvsRequestParams.name).strip('"'))
    return result


@states_response
def get_internal_inf(name: str, as_root=True) -> dict:
    """
    Получение параметров конфигурации внутреннего интерфейса
    """

    params = dict(ipv4_address='', ipv4_netmask='', mac_address='', dhcpv4=False, vlan_mode=None, vlan_tag=None,
                  vlan_trunks=None)

    params.update(states_result(get_address_info_real, interface=name))
    params.update(states_result(get_port_pgroup_config, name=name, as_root=as_root))
    return params


@states_response
def get_physical_link(switch: str, sys_parser: bool = False) -> dict:
    """
    Получение параметров дефолтного аплинка узла (при добавлении узла к контроллеру)
    """

    excluded_port_types = ('internal', 'patch', 'vxlan', 'geneve')
    if sys_parser:
        hardware_devices, _, _ = parse_sys_class_net()
    else:
        network_devices = dump_network_devices()
        hardware_devices, _, _ = parse_network_devices(network_devices)
    params = dict()
    ports = get_ovs_bridge_ports(switch)
    ports = [port for port in ports if get_ovs_interface_param(
        port, OvsRequestParams.interface_type) not in excluded_port_types]
    if len(ports) > 1:
        raise CommandExecutionError(f"Initial configuration should be with only one uplink connection on default "
                                    f"switch {switch}. Connected ports: {', '.join(ports)}")
    name = get_ovs_port_param(ports[0], OvsRequestParams.name).strip('"')
    raw_interfaces = get_ovs_port_param(name, OvsRequestParams.interfaces)
    related_interfaces = [iface_id.strip() for iface_id in raw_interfaces.lstrip('[').rstrip(']').split(',')]
    if len(related_interfaces) == 1:
        iface_name = get_ovs_interface_param(related_interfaces[0], OvsRequestParams.name).strip('"')
        if name == iface_name:
            uplink_type = UplinkTypes.physical_inf
        else:
            uplink_type = UplinkTypes.aggregated_inf
    else:
        uplink_type = UplinkTypes.aggregated_inf

    params['name'] = name
    params['type'] = uplink_type

    if uplink_type == UplinkTypes.physical_inf:
        params.update(dump_device_params(name, hardware_devices))
        params.update(get_physical_inf_params(name))
        params.update(states_result(get_port_pgroup_config, name=name))
    else:
        members_params = list()
        for iface_id in related_interfaces:
            member_info = dict(name=get_ovs_interface_param(iface_id, OvsRequestParams.name).strip('"'))
            member_info.update(dump_device_params(member_info['name'], hardware_devices))
            member_info.update(get_physical_inf_params(member_info['name']))
            members_params.append(member_info)
        params['members'] = members_params
        params.update(states_result(get_port_pgroup_config, name=name, get_mtu=False))
        params.update(states_result(get_bond_config, name=name))
    return params


@states_response
def get_node_firewall_config():
    """
    Получение правил брандмауэра узла в виде построчной конфигурации nftables
    """

    try:
        with open(BASE_FIREWALL_CONF, 'r') as f:
            config = f.read().splitlines()
    except OSError as e:
        raise CommandExecutionError(
            "Failed to write configuration. I/O error({0}): {1}".format(e.errno, e.strerror)) from None
    except Exception as e:
        raise CommandExecutionError("Unexpected error: {0}".format(str(e))) from None
    return config


@states_response
def transform_network_config(as_root=True, sys_parser: bool = False):
    """
    Стейт для "трансформации" конфигурации интерфейсов узла - разбиение конфигурационного файла /etc/network/interfaces
    на отдельные конфигурационные файлы для каждого интерфейса
    """

    with open(BASE_NET_CONFIG, 'r') as f:
        raw_config = f.read().splitlines()
    existing_config, processing_needed = split_net_config(raw_config)
    if processing_needed:
        for entry in existing_config:
            iface = entry['name']
            with open(os.path.join(BASE_NET_CONF_DIR, iface), 'w') as f:
                f.write(reduce(lambda a, s: a + s, map(lambda x: f"{x}\n", entry['config'])))

        write_interfaces_config(BASE_NET_CONFIG)
        free_ports = states_result(dump_free_ports, as_root=as_root)
        for port in free_ports:
            write_interface(BASE_NET_CONF_DIR, port['name'], InterfaceModes.manual,
                            is_connected=False if not port['sriov_enabled'] else True,
                            promiscuous_mode=port['promiscuous_mode'], mtu=port['mtu'])
            if not port['sriov_enabled']:
                set_port_down(port['name'])
    hw_ports = rescan_hardware(as_root, sys_parser=sys_parser)
    purge_leases([port for port in hw_ports])
    deprecate_leases()
    if not is_service_active(BASE_NETWORK_SERVICE):
        action_service(BASE_NETWORK_SERVICE, action=ServiceActions.restart)


@states_response
def purge_ports_with_errors(as_root=True) -> str:
    """
    Удаление портов с ошибкой из коммутатора
    """

    ports_to_delete = []
    for bridge in get_ovs_bridges(as_root):
        for port in get_ovs_bridge_ports(bridge, as_root):
            ptype = get_ovs_interface_param(port, OvsRequestParams.interface_type, as_root)
            error = get_ovs_interface_param(port, OvsRequestParams.error, as_root).rstrip('"').lstrip('"')
            if error in (f"could not open network device {port} (No such device)",
                         f"could not add network device {port} to ofproto (File exists)"):
                ports_to_delete.append(port)
            if ptype == OvsPortType.patch:
                peer = '_'.join([i for i in reversed(port.split('_'))])
                if error == f"No usable peer '{peer}' exists in 'system' datapath.":
                    ports_to_delete.append(port)
    _ = [r for r in map(del_port_from_switch, ports_to_delete)]
    return "Ports with errors purged successfully."


@states_response
def remove_unplugged_ports(ports: list):
    """
    Удаляет конфиги интерфейсов, физически отсутствующих в системе
    """

    ret = {'comment': '', 'changes': {}}
    for port in ports:
        del_file_by_pattern(BASE_NET_CONF_DIR, port)

    ret['comment'] = 'Outdated configs removed successfully.'
    return ret


@states_response
def add_discovered_ports(ports: list):
    """
    Добавляет конфиги интерфейсов, добавленных в систему
    """

    ret = {'comment': '', 'changes': {}}
    for port in ports:
        write_interface(BASE_NET_CONF_DIR, port['name'], mode=InterfaceModes.manual, is_connected=port['is_connected'],
                        promiscuous_mode=port['promisc_mode'], mtu=port['mtu'])
        if not port['up_state']:
            set_port_down(port['name'])
    ret['comment'] = "Discovered ports' configs saved successfully."
    return ret


@states_response
def configure_sriov_ports(ports: list):
    """
    Перезаписывает конфиги SR-IOV интерфейсов
    """

    ret = {'comment': '', 'changes': {}}
    for port in ports:
        write_interface(BASE_NET_CONF_DIR, port['name'], mode=InterfaceModes.manual, is_connected=True,
                        promiscuous_mode=port['promisc_mode'], mtu=port['mtu'])
        set_port_up(port['name'])
    ret['comment'] = "SR-IOV ports' configs saved successfully."
    return ret


@states_response
def remove_outdated_interfaces_conf():
    """
    Удаляет устаревшие записи о конфигурации интерфейсов
    """
    if os.path.isdir(BASE_NET_DIR):
        for f in os.listdir(f'{BASE_NET_DIR}/'):
            if re.search(OUTDATED_INTERFACES_CONF, f):
                os.remove(os.path.join(f'{BASE_NET_DIR}/', f))


@states_response
def switch_exists(name):
    """
    Проверка существования коммутатора
    """

    return name in get_ovs_bridges()


@states_response
def check_port_in_switch(switch, port):
    """
    Проверка существования порта на коммутаторе
    """

    return port in get_ovs_bridge_ports(switch)


@states_response
def check_patches_exist(obverse_switch, reverse_switch, obverse_patch, reverse_patch):
    """
    Проверка существования патча
    """

    return (obverse_patch in get_ovs_bridge_ports(obverse_switch)) and \
           (reverse_patch in get_ovs_bridge_ports(reverse_switch))


@states_response
def check_patches_not_exist(vswitch, lswitch, obverse_patch, reverse_patch):
    """
    Проверка отсутствия патча
    """

    return (reverse_patch not in get_ovs_bridge_ports(vswitch)) and (obverse_patch not in get_ovs_bridge_ports(lswitch))


@states_response
def check_port_up_state(port: str) -> bool:
    """
    Проверка состояния порта
    """

    port_config = states_result(get_link_info_real, interface=port)
    return port_config['up_state']


@states_response
def check_promisc_mode(interface: str, mode: bool) -> bool:
    """
    Проверка соответствия режима работы физического интерфейса
    """

    link_info = get_link_info_detailed(interface)
    real_mode = True if 'PROMISC' in link_info['flags'] else False
    return real_mode == mode


@states_response
def check_mtu(interface: str, mtu: int) -> bool:
    """
    Проверка соответствия значения MTU интерфейса
    """

    link_info = get_link_info_detailed(interface)
    real_mtu = int(link_info['mtu'])
    return real_mtu == mtu


@states_response
def get_full_mtu_info(interface: str) -> dict:
    """
    Получение минимального, текущего и максимального значения MTU интерфейса
    """

    link_info = get_link_info_detailed(interface)
    real_mtu = int(link_info['mtu'])
    min_mtu = int(link_info['min_mtu'])
    max_mtu = int(link_info['max_mtu'])
    return dict(real_mtu=real_mtu, min_mtu=min_mtu, max_mtu=max_mtu)


@states_response
def check_traffic_block_saved(switch: str, port: str, saved: bool):
    rs = TrafficBlockRuleSet.from_config_file()
    rule_exists = True if next(
        (rule for rule in rs.readable() if rule['switch'] == switch and rule['port'].endswith(port)), None) else False
    return True if (saved and rule_exists) or (not saved and not rule_exists) else False


@states_response
def check_traffic_block_applied(switch: str, port: str, applied: bool):
    rs = TrafficBlockRuleSet.from_host_conf()
    rule_exists = True if next(
        (rule for rule in rs.readable() if rule['switch'] == switch and rule['port'].endswith(port)), None) else False
    return True if (applied and rule_exists) or (not applied and not rule_exists) else False


@states_response
def auto_configure_nic(data):
    map_params_keys = {'action': '-a',
                       'mode': '-m',
                       'timeout': '-t',
                       'uplink_name': '-u',
                       'ipv4_address': '-i',
                       'ipv4_netmask': '-s',
                       'mac_address': '-h',
                       'ipv4_gateway': '-g',
                       'dns_servers': '-d'}
    map_values_keys = []
    for k, v in data.items():
        if v:
            if isinstance(v, list) and k == 'dns_servers':
                map_values_keys.append(f"{map_params_keys[k]} {','.join(v)}")
            else:
                map_values_keys.append(f"{map_params_keys[k]} {v}")
    cmd = f"{UTILS_OTHER_PATH}/auto_configure_nic.sh -f {' '.join(map_values_keys)}"
    states_result(get_output, cmd=cmd)
    return "Auto configure nic applied successfully."


@states_response
def get_blocks_map():
    rs = TrafficBlockRuleSet.from_config_file()
    return rs.readable()


@states_response
def remove_deprecated_traffic_blocks(rules_to_delete: list):
    rs = TrafficBlockRuleSet.from_config_file()
    for rule in rules_to_delete:
        r = rs.find_rule(rule)
        if r:
            rs.remove_rule(r)
    rs.save()


# endregion


@states_response
def apply_qos_policy_marking_to_vnetwork(qos_policy_ruleset: dict) -> str:
    """
    Функция применяет правила QoS Маркировки на узле. Применяется на каждый интерфейс. Итерфейсы принадлежат указанной
    виртуальной сети.
    :param qos_policy_ruleset

    qos_policy_ruleset = {'qos_policy': uuid, 'vnetwork': uuid, 'config: {'egress': {'rules': [], 'vminfs_name': []},
                                                                         {'ingress': {'rules': [], 'vminfs_name': []}

    Example qos_policy_ruleset:
    {'qos_policy': '7b3a01e4-c93d-4727-b95c-b6743d373e43', 'vnetwork': 'a5203dba-a686-4d66-8560-c2502441439a',
    'config: {
    {'egress': {'rules': [], 'vminfs_name': []},
    'ingress': {'rules': [{'id': '52ea5453-85ae-4915-92d2-a433f1651621',
    'rule_index': 0,
    'pref_index': 49152,
    'is_enabled': True,
    'direction': 'ingress',
    'dscp': 21,
    'cos': None,
    'ethertype': 'ip',
    'vlan_id': None,
    'src_mac': None,
    'dst_mac': None,
    'ip_protocol': 'tcp',
    'src_ip': None,
    'dst_ip': '192.168.20.120/32',
    'src_port': None,
    'dst_port': None}], 'vminfs_name': ['vm_f83c_0', 'vm_0d6c_0', 'vm_cc1f_0']}}
    }}

    Сохранение каждого правила маркировки в файл vmachines-qos-policy-marking.json происходит
    по каждому интерфейсу source:
    [{"ip_protocol": "tcp", "src_mac": null, "dst_mac": null, "vlan_id": null, "src_port": null,
    "direction": "ingress", "pref_index": 49152, "cos": null, "dst_port": null, "ethertype": "ip", "to_apply": true,
    "src_ip": null, "applied": true, "dscp": 21, "qos_policy": "7b3a01e4-c93d-4727-b95c-b6743d373e43",
    "dst_ip": "192.168.20.120/32", "source": "vm_f83c_0", 'rule_index': 0, 'id': '52ea5453-85ae-4915-92d2-a433f1651621',
    'is_enabled': True, "vnetwork": "a5203dba-a686-4d66-8560-c2502441439a"},
    {...,"source": "vm_0d6c_0"},
    {...,"source": "vm_cc1f_0"}]
    """

    rs = VMachinesQoSPolicyMarkingRuleSet.from_config_file()
    qos_pocily = qos_policy_ruleset['qos_policy']
    vnetwork = qos_policy_ruleset['vnetwork']
    ruleset_by_dir = qos_policy_ruleset['config']
    qdisc_exists = parse_tc_qdisc_exists()['comment']
    existing_vm_interfaces = get_vmachines_interfaces()

    # удаляем все правила для указанной qos_policy и vnetwork
    rules_to_remove = []
    for rule in rs.rules_set:
        if rule.qos_policy == qos_pocily and rule.vnetwork == vnetwork:
            data = {k: getattr(rule, k) for k in QOS_POLICY_MARK_RULE}
            rules_to_remove.append(data)

    if rules_to_remove:
        copy_rules_to_remove = rules_to_remove.copy()
        vinf_ruls = {}
        for rule in rules_to_remove:
            ruls = []
            for r in copy_rules_to_remove:
                if r['source'] == rule['source']:
                    ruls.append(r)
            vinf_ruls[rule['source']] = ruls
        for vminf, rules in vinf_ruls.items():
            qos_marking_rules_remove(rules=rules, inf_name=vminf, rules_set=rs)

    # Создание правил

    # TODO добавить общее условие, если правила есть, то создать дисциплину и создаем правила
    # создаем входящую дисциплину,
    # если интерфейс ВМ существует на узле и на этом интерфейсе нет входящей дисциплины
    for vminf_name in ruleset_by_dir['ingress']['vminfs_name']:
        if vminf_name in existing_vm_interfaces and vminf_name not in qdisc_exists['ingress']:
            cmd = f'tc qdisc replace dev {vminf_name} handle ffff: ingress'
            states_result(get_output, cmd=cmd)

    # создаем только те правила, к-е пришли для направления ingress и пришедших интерфейсов
    # создаем на всех интерфейсах, к-е существуют.
    # Добавляем в это правило указанную qos_policy и source.
    for vminf_name in ruleset_by_dir['ingress']['vminfs_name']:
        for rule in ruleset_by_dir['ingress']['rules']:
            if rule['is_enabled']:
                if vminf_name in existing_vm_interfaces:
                    cmd = create_qos_marking_rule_to_valid_str(rule, vminf_name)
                    states_result(get_output, cmd=cmd)
                rule.update({'qos_policy': qos_pocily})
                rule.update({'vnetwork': vnetwork})
                rule.update({'source': vminf_name})
                data = {k: rule.get(k) for k in QOS_POLICY_MARK_RULE}
                data.update({'applied': True, 'to_apply': True})
                rs.add_rule(VMachinesQoSPolicyMarkingRule.from_json(data))

    # TODO добавить общее условие, если правила есть, то создать дисциплину и создаем правила
    # создаем исходящую дисциплину,
    # если интерфейс ВМ существует на узле и на этом интерфейсе нет исходящей дисциплины
    for vminf_name in ruleset_by_dir['egress']['vminfs_name']:
        if vminf_name in existing_vm_interfaces and vminf_name not in qdisc_exists['egress']:
            cmd = f'tc qdisc replace dev {vminf_name} handle 1: root prio'
            states_result(get_output, cmd=cmd)

    # создаем только те правила, к-е пришли для направления egress и пришедших интерфейсов
    # создаем на всех интерфейсах, к-е существуют.
    # Добавляем в это правило указанную qos_policy и source.
    for vminf_name in ruleset_by_dir['egress']['vminfs_name']:
        for rule in ruleset_by_dir['egress']['rules']:
            if rule['is_enabled']:
                if vminf_name in existing_vm_interfaces:
                    cmd = create_qos_marking_rule_to_valid_str(rule, vminf_name)
                    states_result(get_output, cmd=cmd)
                rule.update({'qos_policy': qos_pocily})
                rule.update({'vnetwork': vnetwork})
                rule.update({'source': vminf_name})
                data = {k: rule.get(k) for k in QOS_POLICY_MARK_RULE}
                data.update({'applied': True, 'to_apply': True})
                rs.add_rule(VMachinesQoSPolicyMarkingRule.from_json(data))
    rs.save()
    return 'qos policy marking apply to vnetwork successfully'


@states_response
def undo_apply_qos_policy_marking_to_vnetwork(qos_policy_ruleset: dict) -> str:
    """
    Функция удаления всех правил для указзаной политики QoS Virtual Machine на указанных интерфейсах и vnetwork
    """

    rs = VMachinesQoSPolicyMarkingRuleSet.from_config_file()
    qos_pocily = qos_policy_ruleset['qos_policy']
    vnetwork = qos_policy_ruleset['vnetwork']

    # удаляем все правила для указанной qos_policy
    rules_to_remove = []
    for rule in rs.rules_set:
        if rule.qos_policy == qos_pocily and rule.vnetwork == vnetwork:
            data = {k: getattr(rule, k) for k in QOS_POLICY_MARK_RULE}
            rules_to_remove.append(data)

    if rules_to_remove:
        copy_rules_to_remove = rules_to_remove.copy()
        vinf_ruls = {}
        for rule in rules_to_remove:
            ruls = []
            for r in copy_rules_to_remove:
                if r['source'] == rule['source']:
                    ruls.append(r)
            vinf_ruls[rule['source']] = ruls
        for vminf, rules in vinf_ruls.items():
            qos_marking_rules_remove(rules=rules, inf_name=vminf, rules_set=rs)

    rs.save()
    return 'qos policy marking undo apply to vnetwork successfully'


# TODO
#  сделать так же как в функции port_mirroring_remove.
#  А именно убрать строки:
#  exists_marking_rules = tc_marking_rules_exists_params_needed()
#  infs_name_marking_rules = [inf for inf, rules in exists_marking_rules[direction].items() if rules]
#  if inf_name in infs_name_marking_rules:
#  И функцию states_result() представить в виде states_result(get_output, raise_exception=False, cmd=cmd).
#  Сделать это лучше после тестов.
@states_response
def qos_marking_rules_remove(rules, inf_name, rules_set=None):
    """
    Функция удаления правил маркировки VMachine QoS Policy
    :param rules:
    :param inf_name:
    :param rules_set:
    :param direction:
    """

    infs_name_marking_rules = []
    exists_marking_rules = tc_marking_rules_exists_params_needed()
    for _, config in exists_marking_rules.items():
        infs_name_marking_rules += [inf for inf, rules in config.items() if rules]
    if infs_name_marking_rules:
        infs_name_marking_rules = set(infs_name_marking_rules)

    if not rules_set:
        rules_set = VMachinesQoSPolicyMarkingRuleSet.from_config_file()
    for rule in rules:
        if inf_name in infs_name_marking_rules:
            cmd = delete_qos_marking_rule_to_valid_str(rule, inf_name)
            states_result(get_output, cmd=cmd)
        # т.е. в правилах rules_set(хранится в json) ищем правило rule, иначе не сможем выполнить remove_rule
        find_rule = rules_set.find_rule(rule)
        if find_rule:
            rules_set.remove_rule(find_rule)
    return "qos policy marking rule has deleted successfully"


def ping_rtt_extractor(ping_output):
    """
    Получение rtt и кода

    code 0: 0 потерянных пакетов
    code 1: потерянные пакеты, но часть получена
    code 2: ошибка, либо 100% потерянных пакетов
    """
    result_lines = ping_output.splitlines()
    rtt_str, rcvd_lost_str, code = None, None, 2

    for line in result_lines:
        if rtt_str and rcvd_lost_str:
            break
        if not rtt_str and 'avg rtt:' in line.lower():
            rtt_str = line
        if not rcvd_lost_str \
                and 'Rcvd: ' in line \
                and 'Lost: ' in line:
            rcvd_lost_str = line
    if rcvd_lost_str:
        if 'Lost: 0' in rcvd_lost_str:
            code = 0  # без потерянных пакетов
        elif 'Rcvd: 0' in rcvd_lost_str:
            code = 2  # все пакеты потеряны
        else:
            code = 1  # часть пакетов потеряна
    return rtt_str, code


def ping_is_successful_with_df(target, interface, data_size):
    """
    Проверка, возможно ли пингануть цель пингом с выключенной фрагментацией (при определённой длине пакета)
    """
    response = states_result(ping_from_node_df, target=target, data_size=data_size, interface=interface)
    _, code = ping_rtt_extractor(response)

    if code == 2:
        # если все пакеты потеряны, значит data_size (размер пакета) не подошёл - возвращаем False
        return False
    return True


def optimal_mtu_binary_search(target, interface, min_mtu, max_mtu, accuracy):
    """
    Поиск оптимального mtu, используя бинарный поиск
    """

    # генерируем список mtu с шагом accuracy
    mtu_range = [x for x in range(min_mtu, max_mtu + accuracy, accuracy)]

    left = 0
    right = len(mtu_range) - 1
    icmp_extra_bytes = 28  # тестовый пинг добавляет 28 байтов (ip + icmp header)

    while left < right:
        mid = (left + right) // 2
        result = ping_is_successful_with_df(target=target,
                                            interface=interface,
                                            data_size=mtu_range[mid])
        if result and not ping_is_successful_with_df(target=target,
                                                     interface=interface,
                                                     data_size=mtu_range[mid + 1]):
            return mtu_range[mid] + icmp_extra_bytes
        elif ping_is_successful_with_df(target=target,
                                        interface=interface,
                                        data_size=mtu_range[mid]):
            left = mid + 1
        else:
            right = mid

    # если подходящий mtu, оказался максимальным значением из списка
    if left != len(mtu_range) and ping_is_successful_with_df(target=target,
                                                             interface=interface,
                                                             data_size=mtu_range[left]):
        return mtu_range[left] + icmp_extra_bytes

    # если оптимальное mtu не найдено - возвращаем 0
    return 0


@states_response
def find_optimal_mtu(target, interface, accuracy):
    """
    Поиск оптимального MTU, с изменением текущего MTU на максимальный и обратно
    """
    optimal_mtu = 0
    mtu_info = states_result(get_full_mtu_info, interface=interface)
    current_mtu = mtu_info['real_mtu']
    try:
        min_mtu = mtu_info['min_mtu']
        max_mtu = mtu_info['max_mtu']
        # изменяем значение MTU на максимальное
        set_ovs_mtu(interface=interface, mtu=max_mtu)
        # запускаем поиск оптимального MTU
        optimal_mtu = optimal_mtu_binary_search(target, interface, min_mtu, max_mtu, accuracy)

        # изменяем значение MTU на прежнее
        set_ovs_mtu(interface=interface, mtu=current_mtu)
    except Exception as error:
        # в случае ошибки, откатываем изменения
        set_ovs_mtu(interface=interface, mtu=current_mtu)
        raise error
    return optimal_mtu


# функции вызываемые из тестов
@states_response
def vmachines_qos_policy_marking_stored_rules():
    rs = VMachinesQoSPolicyMarkingRuleSet()
    try:
        with open(rs.config_file, 'r') as f:
            stored_rules = json.load(f)
    except Exception as e:
        raise CommandExecutionError("Unexpected error: {0}".format(str(e))) from None
    return stored_rules


@states_response
def vmachines_qos_policy_marking_applied_rules():
    return tc_marking_rules_exists_params_needed()
# endregion
